/*LINTLIBRARY*/
/*  miriad_read.c

    This code provides a Miriad read facility.

    Copyright (C) 1995-2003  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*

    This file contains the various utility routines for reading data files in
  Miriad Image format.


    Written by      Richard Gooch   27-SEP-1995

    Updated by      Richard Gooch   28-SEP-1995

    Updated by      Richard Gooch   16-OCT-1995: Fixed bug in <make_desc> where
  axis count + 1 was improperly used.

    Updated by      Richard Gooch   13-DEC-1995: Fixed bug in <make_desc> where
  Miriad units were not translated into FITS units.

    Updated by      Richard Gooch   12-APR-1996: Changed to new documentation
  format.

    Updated by      Richard Gooch   8-MAY-1996: Added support for reading mask
  file in Miriad Image datasets.

    Updated by      Richard Gooch   17-MAY-1996: Check for "MOSTABLE" keyword
  and ignore.

    Updated by      Richard Gooch   30-MAY-1996: Cleaned code to keep
  gcc -Wall -pedantic-errors happy.

    Updated by      Richard Gooch   3-JUN-1996: No longer fiddle reversed axes.

    Updated by      Richard Gooch   17-JUN-1996: Subtracted 1.0 from CRPIXn
  values.

    Updated by      Richard Gooch   25-JUN-1996: Trap "ANGLE" axis name and
  convert to "Axis n (deg)" otherwise name is multiply used. Improved
  efficiency of data reader several times.

    Updated by      Richard Gooch   28-JUN-1996: No longer use FITS reader to
  read data.

    Updated by      Richard Gooch   6-AUG-1996: Copied "CRVALn" value to
  reference field of dimension descriptor.

    Updated by      Richard Gooch   15-AUG-1996: No longer update <<reference>>
  field of dimension descriptor.

    Updated by      Richard Gooch   29-AUG-1996: Allowed some axes to not have
  names but still use as many supplied names as possible.

    Updated by      Richard Gooch   5-SEP-1996: Made use of <pio_read_floats>.

    Updated by      Richard Gooch   25-SEP-1996: Changed to implement the new
  FITS-style co-ordinate handling, where dimension co-ordinates range from 0
  to length - 1. Fixed scale conversion of rest frequency.

    Updated by      Richard Gooch   26-SEP-1996: "RESTFREQ" keyword now scaled.

    Updated by      Richard Gooch   29-OCT-1996: Tidied up macros to keep
  Solaris 2 compiler happy.

    Updated by      Richard Gooch   7-JAN-1997: os.h no longer defines
  <network_type_bytes>. Removed os.h

    Updated by      Richard Gooch   20-JAN-1997: Trap "Radius (R\dJ\u)" axis
  name and convert to "Radius n (R\dJ\u)" otherwise name is multiply used.

    Updated by      Richard Gooch   21-FEB-1997: Accomodate integer data type.

    Updated by      Richard Gooch   14-APR-1997: Fixed bug in header reader
  when binary items (i.e. "mostable") are not at the end of the file: was not
  skipping to next padding boundary.

    Updated by      Richard Gooch   23-APR-1997: Added skip attribute.

    Updated by      Richard Gooch   3-JUL-1997: Added default "1/F" value for
  "CELLSCAL" keyword for spectral cubes.

    Updated by      Richard Gooch   23-JUL-1997: Fixed bug when skipping mask
  values.

    Updated by      Richard Gooch   8-AUG-1997: Added memory-mapping support.

    Updated by      Richard Gooch   25-SEP-1997: No longer clear array when
  allocating.

    Updated by      Richard Gooch   24-FEB-1998: Switch to new array_pointer
  which has free function.

    Updated by      Richard Gooch   11-MAR-1998: Made experimental use of
  madvise(2) and MADV_SEQUENTIAL. MADV_WILLNEED makes things worse if the data
  is not actually used.

    Updated by      Richard Gooch   18-MAR-1998: Do not abort if array alloc
  fails in <make_desc>.

    Updated by      Richard Gooch   24-MAR-1998: Fixed bug introducted in
  <miriad_read> on 11-MAR-1998: mmapped pointer was not written into context
  hence not writted into arrayp.

    Updated by      Richard Gooch   13-MAY-1998: Added fallback to using anon.
  mmapped array.

    Updated by      Richard Gooch   17-AUG-1998: No longer fail if items have
  extra bytes: just drain the extra data.

    Updated by      Richard Gooch   23-APR-2000: Fixed hyperlink.

    Updated by      Richard Gooch   13-MAR-2001: Fixed bogus Miriad "CAR" proj.

    Updated by      Richard Gooch   27-FEB-2002: Documented header file format
  and fixed check for item name length (was checking for MIN_ITEM_SIZE+2 but
  should be MIN_ITEM_SIZE-2).

    Updated by      Richard Gooch   28-FEB-2002: Fixed documentation. Tolerate
  unknown header item data types by skipping data.

    Updated by      Richard Gooch   8-MAR-2002: Supported reading headers for
  visibility data.

    Updated by      Richard Gooch   25-MAY-2002: First working version of
  visibility data reading code. Auxilary array (BL -> Ant pair) not yet filled
  in.

    Updated by      Richard Gooch   28-MAY-2002: Fixed UV reader to support
  missing spectra (i.e. some baselines for a particular timestamp missing).
  Requires two passes over the visdata file. Result is cached in "pol+time+bl"
  file. Auxilary array (BL -> Ant pair) now filled in.

    Updated by      Richard Gooch   29-MAY-2002: Added reading of "flags" file
  in UV reader. Skip reading UV correlation data on first pass (optimisation).
  Fill main UVW array with blanks for missing time/BL records.

    Updated by      Richard Gooch   30-MAY-2002: Sanitise polarisation
  dimension when reading UV data, if appropriate.

    Updated by      Richard Gooch   5-JUN-2002: Fixed bug adding keywords when
  reading UV data and not sanitising: items with no data caused crash.

    Updated by      Richard Gooch   20-JUN-2002: Close and re-open data context
  in <foreign_miriad_read> to avoid VM exhaustion.

    Updated by      Richard Gooch   24-JUN-2002: Fixed bug in UV flags reading:
  bit orders were reversed.

    Updated by      Richard Gooch   3-JUL-2002: Supported compressed image
  data.

    Updated by      Richard Gooch   16-AUG-2002: Renamed struct "_header" to
  "miriad_header" to avoid conflict with broken OSF1 system headers.

    Updated by      Richard Gooch   23-DEC-2002: Supported spherical axes
  without projections (i.e. default to flat aka/ cartesian).

    Updated by      Richard Gooch   24-DEC-2002: Switched from <ch_drain*> to
  <ch_skip*>.

    Updated by      Richard Gooch   29-DEC-2002: Fixed bugs introduced by last
  change.

    Updated by      Richard Gooch   27-JAN-2003: Removed unused variable.

    Updated by      Richard Gooch   28-MAY-2003: Removed assumption that a
  vartable entry implies a visdata entry, otherwise can SEGV.

    Last updated by Richard Gooch   18-SEP-2003: Fixed bug in function
  <foreign_miriad_read>: copy old arrayp before changing. Recent addition of
  <<change_callbacks>> exposed this problem.


*/
#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <stdarg.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <os.h>
#ifdef HAS_MADVISE
#  include <sys/mman.h>
#endif
#include <karma.h>
#include <k_list.h>
#include <karma_foreign.h>
#include <karma_dsrw.h>
#include <karma_pio.h>
#include <karma_dmp.h>
#include <karma_ch.h>
#include <karma_ds.h>
#include <karma_ex.h>
#include <karma_st.h>
#include <karma_m.h>
#include <karma_a.h>
#include <karma_s.h>
#include <karma_r.h>
#include <karma_p.h>


#define MIN_ITEM_SIZE                         16
#define BLOCK_LENGTH                          1048576
#define BUF_SIZE                              4096
#define MAGIC_NUMBER                          486734598
#define NUM_VISDATA_ARRAYS                    3

#define VISDATA_DIM_INDEX_POLARISATION        0
#define VISDATA_DIM_INDEX_TIME                1
#define VISDATA_DIM_INDEX_BASELINE            2
#define VISDATA_DIM_INDEX_CHANNEL             3

#define VISUVW_DIM_INDEX_TIME                 0
#define VISUVW_DIM_INDEX_BASELINE             1

#define VISDATA_ENTRY_TYPE_DESCRIPTOR         0
#define VISDATA_ENTRY_TYPE_DATA               1
#define VISDATA_ENTRY_TYPE_END_OF_RECORD      2


#define VERIFY_CONTEXT(ctx) if (ctx == NULL) \
{fprintf (stderr, "NULL context passed\n"); \
 a_prog_bug (function_name); } \
if (ctx->magic_number != MAGIC_NUMBER) \
{fprintf (stderr, "Invalid context object\n"); \
 a_prog_bug (function_name); }

#define TIMEDIFF_US(start, stop) \
  (((stop).tv_sec - (start).tv_sec)*1000000 + (stop).tv_usec - (start).tv_usec)

/*  Description of the Miriad header format
    The header is written in big-endian format. The header contains a number
    of "items", which roughly correspond to FITS keyword-value pairs. The items
    are written one after the other in the header file. The first item is
    placed at the start of the file. There is no indication of how many items
    are present, other than reaching EndOfFile. The items are placed in what
    appears to be reverse order (i.e. the "naxis" keyword usually comes last).
    The structure of each item is:

    16 bytes containing the null-terminated keyword name with a single byte
    at the end indicating the number of data bytes (D) in the item. Thus, the
    maximum keyword name length is 14 bytes.

    D bytes of data, containing the type of the data and the data value, in the
    following format:

	4 byte unsigned integer indicating the type of the data. The following
	values are defined:
	    0: unstructured, application-specific binary data
	    1: signed 1 byte integer (appears to be a variable-length string)
	    2: signed 4 byte integer
	    3: signed 2 byte integer
	    4: IEEE 4 byte float
	    5: IEEE 8 byte float
	    6: variable-length string, length is (item_length - 4) bytes
	    7: single precision complex (4 bytes for real and imaginary)

	N bytes of padding, to align the data that follows to a data-sized
	boundary. There is no padding for variable-length string data or
	unstructured binary data.

	The data.

    M bytes of padding, to align the item that follows to a 16 byte boundary.
*/

/*  Structure declarations  */

typedef struct _header_item
{
    char *name;
    char *data;
    unsigned int type;
    unsigned int num_values;     /*  Number of data values    */
    unsigned int buf_len;        /*  Size of buffer in bytes  */
    struct _header_item *prev;
    struct _header_item *next;
    double d_data;               /*  Used for axis scaling    */
} *HeaderItem;

struct miriad_data_context_type
{
    unsigned int magic_number;
    Channel image_channel;
    CONST float *mmapped_ptr;
    flag first_time;
    Channel mask_channel;
    int last_mask_bit_in_word;
    unsigned long mask_word;
};

typedef struct
{
    unsigned int length;      /*  Number of bytes to read            */
    HeaderItem header_item;
    unsigned int count;       /*  Number of times this has been set  */
    char type;
    char name[1];             /*  This must come last                */
} *VartableItem;

struct index_struct
{
    int polarisation;
    int corr;
    int time;
    int baseline;
    int coord;
};

struct array_struct
{
    char *data;
    array_desc *desc;
};

typedef struct
{
    unsigned int num_items;
    struct index_struct indices;
    struct array_struct vis;      /*  The visibility data                 */
    struct array_struct ant;      /*  The baseline -> antenna pair table  */
    struct array_struct uvw;      /*  The BL,Time -> U,V,W table          */
    VartableItem *items;
} *VarTable;

typedef struct sized_array
{
    unsigned long num_values;
    unsigned long buflen;
    double *data;
} *SizedArray;

typedef struct miriad_header
{
    flag have_frequency_axis;
    flag have_cellscal;
    int naxis;
    unsigned int num_dim;
    VarTable vartable;
    struct sized_array polarisations;
    struct sized_array times;
    struct sized_array baselines;
    HeaderItem first;
    HeaderItem last;
} *Header;


/*  Declarations of private functions follow  */
STATIC_FUNCTION (multi_array *_foreign_miriad_read_header,
		 (CONST char *dirname, flag data_alloc, flag sanitise,
		  Header header, ...) );
STATIC_FUNCTION (multi_array *_foreign_miriad_read_header_argp,
		 (CONST char *dirname, flag data_alloc, flag sanitise,
		  Header header, va_list argp) );
STATIC_FUNCTION (flag _foreign_miriad_read_header_items,
		 (Channel channel, Header header, va_list argp) );
STATIC_FUNCTION (multi_array *_foreign_miriad_process_header,
		 (Header header, flag data_alloc, flag sanitise,
		  va_list argp) );
STATIC_FUNCTION (flag _foreign_miriad_read_miriad_type,
		 (Channel channel, unsigned int *type) );
STATIC_FUNCTION (flag _foreign_miriad_read_header_value,
		 (Channel channel, HeaderItem item, unsigned int length) );
STATIC_FUNCTION (multi_array *_foreign_miriad_read_make_desc,
		 (Header header, flag data_alloc, flag sanitise) );
STATIC_FUNCTION (array_desc *_foreign_miriad_read_make_image_array_desc,
		 (HeaderItem first_item, signed int naxis, flag sanitise) );
STATIC_FUNCTION (flag _foreign_miriad_read_make_visdata_array_descs,
		 (Header header,
		  array_desc **arr_descs, unsigned int *num_arrays) );
STATIC_FUNCTION (HeaderItem _foreign_miriad_read_find_item,
		 (HeaderItem first_item, CONST char *name) );
STATIC_FUNCTION (void _foreign_miriad_read_free_header,
		 (HeaderItem first_item) );
STATIC_FUNCTION (flag _foreign_miriad_read_get_next_mask_value,
		 (KMiriadDataContext context, flag *mask) );
STATIC_FUNCTION (flag _foreign_miriad_read_convert_units,
		 (HeaderItem first_item) );
STATIC_FUNCTION (flag _foreign_miriad_read_scale_item,
		 (HeaderItem first_item, CONST char *name, double scale) );
STATIC_FUNCTION (flag _foreign_miriad_read_read_data,
		 (KMiriadDataContext context, char *data, uaddr num_values,
		  unsigned long *blank_count) );
STATIC_FUNCTION (HeaderItem _foreign_miriad_read_split_binary_item,
		 (HeaderItem item) );
STATIC_FUNCTION (HeaderItem _foreign_miriad_read_alloc_items,
		 (unsigned int num, CONST char **names,
		  CONST unsigned int *types) );
STATIC_FUNCTION (VarTable _foreign_miriad_read_vartable, (CONST char *dir) );
STATIC_FUNCTION (void _foreign_miriad_read_free_vartable, (VarTable vt) );
STATIC_FUNCTION (flag _foreign_miriad_read_visdata_header,
		 (CONST char *dirname, Header header) );
STATIC_FUNCTION (flag _foreign_miriad_read_sized_array,
		 (Channel channel, SizedArray array) );
STATIC_FUNCTION (flag _foreign_miriad_read_write_sized_array,
		 (Channel channel, SizedArray array) );
STATIC_FUNCTION (flag _foreign_miriad_read_visdata_chunk,
		 (Channel channel, VarTable vartable, flag *end_of_file,
		  unsigned int *entry_type, int *index, int skip_index) );
STATIC_FUNCTION (flag _foreign_miriad_read_visdata_data,
		 (CONST char *dirname, Header header,
		  multi_array *multi_desc) );
STATIC_FUNCTION (void blank_plane,
		 (CONST array_desc *arr_desc, char *data,
		  unsigned int y_dim_index, unsigned int x_dim_index) );


/*  Private data  */
static flag debug = FALSE;
static flag first_time = TRUE;


/*  Public functions follow  */

/*EXPERIMENTAL_FUNCTION*/
flag foreign_miriad_test2 (CONST char *dirname, flag test_dir)
/*  [SUMMARY] Test if a directory is a Miriad Image file.
    <dirname> The directory name of the Miriad dataset name.
    <test_dir> If TRUE will first test if the dataset name is a directory. Set
    this to FALSE only if you know the dataset name is a directory.
    [RETURNS] TRUE if the directory is a Miriad Image file, else FALSE.
*/
{
    struct stat statbuf;
    char filename[STRING_LENGTH];
    static CONST char *extensions[] = {"", ".gz", ".bz2", NULL};

    if (test_dir)
    {
	if (stat (dirname, &statbuf) != 0)
	{
	    if (errno == ENOENT) return (FALSE);
	    fprintf (stderr, "Error statting file: \"%s\"\t%s\n",
		     dirname, ERRSTRING);
	    return (FALSE);
	}
	if ( !S_ISDIR (statbuf.st_mode) ) return (FALSE);
    }
    sprintf (filename, "%s/header", dirname);
    if ( !r_check_file_variants (filename, extensions) ) return (FALSE);
    sprintf (filename, "%s/visdata", dirname);
    if ( r_check_file_variants (filename, extensions) ) return (TRUE);
    sprintf (filename, "%s/image", dirname);
    if ( r_check_file_variants (filename, extensions) ) return (TRUE);
    return (FALSE);
}   /*  End Function foreign_miriad_test2  */

/*OBSOLETE_FUNCTION*/
flag foreign_miriad_test (CONST char *dirname)
/*  [SUMMARY] Test if a directory is a Miriad Image file.
    <dirname> The directory name of the Miriad dataset name.
    [RETURNS] TRUE if the directory is a Miriad Image file, else FALSE.
*/
{
    return foreign_miriad_test2 (dirname, TRUE);
}   /*  End Function foreign_miriad_test  */

/*OBSOLETE_FUNCTION*/
multi_array *foreign_miriad_read_header (Channel channel, flag data_alloc,
					 flag sanitise, ...)
/*  [SUMMARY] Read a Miriad Image file header.
    [PURPOSE] This routine will read the header of a Miriad Image file from a
    channel. The data section is NOT read.
    <channel> The channel to read from.
    <data_alloc> If TRUE, the data space is allocated.
    <sanitise> If TRUE, Miriad axes with length 1 are ignored. This is highly
    recommended.
    [VARARGS] The optional attributes are given as pairs of attribute-key
    attribute-value pairs. This list must be terminated with
    FA_MIRIAD_READ_HEADER_END. See [<foreign_ATT_MIRIAD_READ_HEADER>] for a
    list of defined attributes.
    [RETURNS] A pointer to the multi_array data structure on success, else
    NULL.
*/
{
    va_list argp;
    struct miriad_header header;
    multi_array *multi_desc;
    static char function_name[] = "foreign_miriad_read_header";

    va_start (argp, sanitise);
    if (channel == NULL)
    {
	fprintf (stderr, "NULL pointer passed\n");
	a_prog_bug (function_name);
    }
    FLAG_VERIFY (data_alloc);
    FLAG_VERIFY (sanitise);
    m_clear (&header, sizeof header);
    if ( !_foreign_miriad_read_header_items (channel, &header, argp) )
	return (NULL);
    multi_desc = _foreign_miriad_process_header (&header, data_alloc, sanitise,
						 argp);
    _foreign_miriad_read_free_header (header.first);
    va_end (argp);
    return (multi_desc);
}   /*  End Function foreign_miriad_read_header  */

/*EXPERIMENTAL_FUNCTION*/
multi_array *foreign_miriad_read_header2 (CONST char *dirname, flag data_alloc,
					  flag sanitise, ...)
/*  [SUMMARY] Read a Miriad Image or Visibility file header.
    [PURPOSE] This routine will read the header of a Miriad Image file from a
    channel. The data section is NOT read.
    <dirname> The directory name of the Miriad dataset name.
    <data_alloc> If TRUE, the data space is allocated.
    <sanitise> If TRUE, Miriad axes with length 1 are ignored. This is highly
    recommended.
    [VARARGS] The optional attributes are given as pairs of attribute-key
    attribute-value pairs. This list must be terminated with
    FA_MIRIAD_READ_HEADER_END. See [<foreign_ATT_MIRIAD_READ_HEADER>] for a
    list of defined attributes.
    [RETURNS] A pointer to the multi_array data structure on success, else
    NULL.
*/
{
    va_list argp;
    struct miriad_header header;
    multi_array *multi_desc;
    static char function_name[] = "foreign_miriad_read_header2";

    va_start (argp, sanitise);
    FLAG_VERIFY (data_alloc);
    FLAG_VERIFY (sanitise);
    m_clear (&header, sizeof header);
    header.vartable = _foreign_miriad_read_vartable (dirname);
    multi_desc = _foreign_miriad_read_header_argp (dirname, data_alloc,
						   sanitise, &header, argp);
    va_end (argp);
    _foreign_miriad_read_free_vartable (header.vartable);
    _foreign_miriad_read_free_header (header.first);
    if (header.polarisations.data) m_free (header.polarisations.data);
    if (header.times.data) m_free (header.times.data);
    if (header.baselines.data) m_free (header.baselines.data);
    return (multi_desc);
}   /*  End Function foreign_miriad_read_header2  */

/*PUBLIC_FUNCTION*/
multi_array *foreign_miriad_read (CONST char *dirname, flag sanitise, ...)
/*  [SUMMARY] Read a Miriad image file.
    <dirname> The directory name of the Miriad dataset name.
    <sanitise> If TRUE, Miriad axes with length 1 are ignored. This is highly
    recommended.
    [VARARGS] The optional attributes are given as pairs of attribute-key
    attribute-value pairs. This list must be terminated with
    FA_MIRIAD_READ_END. See [<foreign_ATT_MIRIAD_READ>] for a list of defined
    attributes.
    [RETURNS] A multi_array descriptor on success, else NULL.
*/
{
    va_list argp;
    KMiriadDataContext context;
    flag writable = FALSE;
    flag plain_read;
    unsigned int att_key;
    unsigned int mmap_option = 0;
    struct miriad_header header;
    unsigned long *blank_count = NULL;
    multi_array *multi_desc;
    static char function_name[] = "foreign_miriad_read";

    va_start (argp, sanitise);
    /*  Process attributes  */
    while ( ( att_key = va_arg (argp, unsigned int) ) != FA_MIRIAD_READ_END )
    {
	switch (att_key)
	{
	  case FA_MIRIAD_READ_NUM_BLANKS:
	    blank_count = va_arg (argp, unsigned long *);
	    break;
	  case FA_MIRIAD_READ_MMAP_OPTION:
	    mmap_option = va_arg (argp, unsigned int);
	    break;
	  case FA_MIRIAD_READ_WRITABLE:
	    writable = va_arg (argp, flag);
	    FLAG_VERIFY (writable);
	    break;
	  default:
	    fprintf (stderr, "Unknown attribute key: %u\n", att_key);
	    a_prog_bug (function_name);
	    break;
	}
    }
    va_end (argp);
    m_clear (&header, sizeof header);
    header.vartable = _foreign_miriad_read_vartable (dirname);
    if (header.vartable)
    {   /*  It's UV data  */
	if ( ( multi_desc =
	       _foreign_miriad_read_header (dirname, TRUE, sanitise, &header,
					    FA_MIRIAD_READ_HEADER_END) )
	     == NULL )
	{
	    _foreign_miriad_read_free_vartable (header.vartable);
	    return (NULL);
	}
	if ( !_foreign_miriad_read_visdata_data (dirname, &header,
						 multi_desc) )
	{
	    ds_dealloc_multi (multi_desc);
	    multi_desc = NULL;
	}
	if ( sanitise && (header.polarisations.num_values == 1) &&
	     (VISDATA_DIM_INDEX_POLARISATION == 0) )
	{
	    int count;
	    array_desc *arr_desc = header.vartable->vis.desc;
	    dim_desc *dim = arr_desc->dimensions[0];

	    m_free (dim->name);
	    m_free (dim->coordinates);
	    m_free (dim);
	    m_free (arr_desc->tile_lengths[0]);
	    m_free (arr_desc->offsets[0]);
	    for (count = 1; count < arr_desc->num_dimensions; ++count)
	    {
		arr_desc->dimensions[count - 1] = arr_desc->dimensions[count];
		arr_desc->tile_lengths[count-1] =arr_desc->tile_lengths[count];
		arr_desc->offsets[count - 1] = arr_desc->offsets[count];
	    }
	    --arr_desc->num_dimensions;
	}
	_foreign_miriad_read_free_vartable (header.vartable);
	_foreign_miriad_read_free_header (header.first);
	if (header.polarisations.data) m_free (header.polarisations.data);
	if (header.times.data) m_free (header.times.data);
	if (header.baselines.data) m_free (header.baselines.data);
	return (multi_desc);
    }
    /*  It's image data.
	Create data context first so we can determine if memory mapping is
	possible, and hence determine if array data should be allocated  */
    if ( ( context = foreign_miriad_create_data_context (dirname, mmap_option,
							 writable) ) == NULL )
    {
	fprintf (stderr, "Error creating KMiriadDataContext object\n");
	return (NULL);
    }
    if (ds_can_transfer_element_as_block (K_FLOAT) &&
	(context->mask_channel == NULL) && context->mmapped_ptr && !writable)
	plain_read = FALSE;
    else
    {
	plain_read = TRUE;
	if (context->mmapped_ptr)
	{   /* Unmap since we need to malloc(3), and may run out of VM space */
	    foreign_miriad_close_data_context (context);
	    context = NULL;
	}
    }
    multi_desc = _foreign_miriad_read_header (dirname, plain_read, sanitise,
					      &header,
					      FA_MIRIAD_READ_HEADER_END);
    _foreign_miriad_read_free_header (header.first);
    if (multi_desc == NULL)
    {
	if (context) foreign_miriad_close_data_context (context);
	return (NULL);
    }
    if (context == NULL)
    {   /*  Need to load context again  */
	if ( ( context = foreign_miriad_create_data_context (dirname,
							     mmap_option,
							     writable) )
	     == NULL )
	{
	    fprintf (stderr, "Error creating KMiriadDataContext object\n");
	    ds_dealloc_multi (multi_desc);
	    return (NULL);
	}
    }
    if (plain_read)
    {
	if ( !foreign_miriad_read_data (context, multi_desc, NULL, 0,
					FA_MIRIAD_READ_DATA_NUM_BLANKS,
					blank_count,
					FA_MIRIAD_READ_DATA_END) )
	{
	    fprintf (stderr, "Error reading KMiriadDataContext object\n");
	    ds_dealloc_multi (multi_desc);
	    foreign_miriad_close_data_context (context);
	    return (NULL);
	}
    }
    else
    {
	array_pointer arrayp;

	m_copy (&arrayp, multi_desc->data[0], sizeof arrayp);
	arrayp.array = (char *) context->mmapped_ptr;
	arrayp.free = ( void (*) () ) ch_close;
	arrayp.info = context->image_channel;
	arrayp.mem_type = K_ARRAY_MEM_TYPE_MMAP;
	arrayp.u.mmap.fd = ch_get_descriptor (context->image_channel);
	arrayp.u.mmap.size = ch_get_mem_size (context->image_channel);
	arrayp.u.mmap.base = ch_get_mmap_addr (context->image_channel);
	m_copy (multi_desc->data[0], &arrayp, sizeof arrayp);
	context->image_channel = NULL;
    }
    foreign_miriad_close_data_context (context);
    foreign_miriad_read_history (dirname, multi_desc);
    return (multi_desc);
}   /*  End Function foreign_miriad_read  */

/*EXPERIMENTAL_FUNCTION*/
KMiriadDataContext foreign_miriad_create_data_context
    (CONST char *dirname, unsigned int mmap_option, flag writable)
/*  [SUMMARY] Create a context suitable for reading Miriad Image data.
    [PURPOSE] This routine will create a context suitable for reading Miriad
    Image data. The [<foreign_miriad_read_data>] routine may be used to read
    data sequentially from the context.
    <dirname> The directory name of the Miriad dataset name.
    <mmap_option> Option to control memory mapping when reading from disc. See
    [<ch_MAP_CONTROLS>] for a list of legal values.
    <writable> If TRUE, the mapped structure will be writable. When the data
    structure data is modified these changes will be reflected in the disc
    file. The shape of the data structure cannot be changed though mapping.
    If FALSE and the structure is written to, a segmentation fault occurs.
    [RETURNS] A KMiriadDataContext object on success, else NULL (indicating the
    image file could not be read).
*/
{
    KMiriadDataContext context;
    unsigned long type;
    struct stat statbuf;
    char image_name[STRING_LENGTH];
    char mask_name[STRING_LENGTH];
    static char function_name[] = "foreign_miriad_create_data_context";

    if ( ( context = m_alloc (sizeof *context) ) == NULL )
	m_abort (function_name, "data context");
    sprintf (image_name, "%s/image", dirname);
    /*  Open the image file  */
    context->image_channel = ch_map_disc (image_name, mmap_option, writable,
					  TRUE);
    if (context->image_channel == NULL)
	context->image_channel = ch_open_compressed_file (image_name);
    if (context->image_channel == NULL)
    {
	fprintf (stderr, "%s: error opening: \"%s\"\t%s\n",
		 function_name, image_name, ERRSTRING);
	m_free (context);
	return (NULL);
    }
    if ( ch_test_for_mmap (context->image_channel) )
	context->mmapped_ptr = (float *)
	ch_get_mmap_addr (context->image_channel) + 1;
    else context->mmapped_ptr = NULL;
    context->first_time = TRUE;
    /*  The first 4 bytes of the file must contain an integer which defines the
	type of data stored in the image. Miriad only supports one type: IEEE
	single precision floating point. The value 4 is the magic value for
	float data  */
    if ( !pio_read32 (context->image_channel, &type) )
    {
	fprintf (stderr, "Error reading image data type\t%s\n",
		 ERRSTRING);
	ch_close (context->image_channel);
	return (NULL);
    }
    if (type != 4)
    {
	fprintf (stderr, "Image data type: %lu is not 4!\n", type);
	ch_close (context->image_channel);
	return (NULL);
    }
    /*  After the first 4 bytes, the rest of the image file is the same as FITS
	format. This saves me a bit of coding, but there is still the image
	mask to check for  */
    /*  Test if image mask present  */
    sprintf (mask_name, "%s/mask", dirname);
    if (stat (mask_name, &statbuf) != 0)
    {
	if (errno != ENOENT)
	{
	    /*  Strange error, but return data anyway  */
	    fprintf (stderr, "Error statting file: \"%s\"\t%s\n",
		     mask_name, ERRSTRING);
	}
	/*  Not present: good  */
	context->mask_channel = NULL;
	context->magic_number = MAGIC_NUMBER;
	return (context);
    }
    if ( ( context->mask_channel = ch_open_compressed_file (mask_name) )
	 == NULL )
    {
	/*  Strange error, but return data anyway  */
	context->mask_channel = NULL;
	context->magic_number = MAGIC_NUMBER;
	return (context);
    }
    /*  Image mask is present  */
    /*  The first 4 bytes of the file must contain an integer which defines the
	type of data stored in the mask. Miriad only supports one type:
	integer. The value 2 is the magic value for integer data  */
    type = 2;
    if ( !pio_read32 (context->mask_channel, &type) )
    {
	fprintf (stderr, "Error reading mask data type\t%s\n",
		 ERRSTRING);
	ch_close (context->image_channel);
	ch_close (context->mask_channel);
	m_free (context);
	return (NULL);
    }
    if (type != 2)
    {
	fprintf (stderr, "Mask data type: %lu is not 2!\n", type);
	ch_close (context->image_channel);
	ch_close (context->mask_channel);
	m_free (context);
	return (NULL);
    }
    /*  After the first 4 bytes, the rest of the mask file contains one bit per
	data value. For some obscure reason, every 32nd bit (starting from bit
	number 0) in the mask file is unused. Horrendous.
	Bit 0 (i.e. bit 0 of byte 3) corresponds to the first data value. A bit
	'1' indicates the corresponding data value is good, else it should be
	blanked.
	Hence, every 31 masks are stored in a 32 bit word, where the bits are
	stored on disc in MSB order. So for the first 32 bit word in the file,
	the bit 0x00000001 (MSB) corresponds to the first data value  */
    context->last_mask_bit_in_word = 31;
    context->magic_number = MAGIC_NUMBER;
    return (context);
}   /*  End Function foreign_miriad_create_data_context  */

/*EXPERIMENTAL_FUNCTION*/
flag foreign_miriad_read_data (KMiriadDataContext context,
			       multi_array *multi_desc,
			       char *data, uaddr num_values, ...)
/*  [SUMMARY] Read data in a Miriad Image file.
    [PURPOSE] This routine will read the data of a Miriad Image file from a
    KMiradDataContext object. The header section is NOT read.
    <context> The context to read from.
    <multi_desc> The Karma data structure to write the data into.
    <data> An alternate data array to write the data into. If this is NULL,
    the routine will write the data into the Karma data structure.
    <num_values> The number of values to write into the data array. This is
    used when data is not NULL or skipping is requested.
    [VARARGS] The optional attributes are given as pairs of attribute-key
    attribute-value pairs. This list must be terminated with
    FA_MIRIAD_READ_DATA_END. See [<foreign_ATT_MIRIAD_READ_DATA>] for a list
    of defined attributes.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    va_list argp;
    flag skip = FALSE;
    uaddr block_length;
    unsigned int att_key;
    unsigned long blank_count_local;
    unsigned long *blank_count = NULL;
    packet_desc *pack_desc;
    array_desc *arr_desc;
    static char function_name[] = "foreign_miriad_read_data";

    VERIFY_CONTEXT (context);
    va_start (argp, num_values);
    /*  Process attributes  */
    while ( ( att_key = va_arg (argp,unsigned int) )
	    != FA_MIRIAD_READ_DATA_END )
    {
	switch (att_key)
	{
	  case FA_MIRIAD_READ_DATA_NUM_BLANKS:
	    blank_count = va_arg (argp, unsigned long *);
	    break;
	  case FA_MIRIAD_READ_DATA_SKIP_BLOCKS:
	    skip = va_arg (argp, flag);
	    FLAG_VERIFY (skip);
	    break;
	  default:
	    fprintf (stderr, "Unknown attribute key: %u\n", att_key);
	    a_prog_bug (function_name);
	    break;
	}
    }
    va_end (argp);
    if (blank_count == NULL) blank_count = &blank_count_local;
    if (skip)
    {
	flag mask;
	unsigned long block, count;

	if ( !ch_skip (context->image_channel, 4 * num_values) )
	{
	    fprintf (stderr, "%s: error skipping\n", function_name);
	    return (FALSE);
	}
	if (context->mmapped_ptr) context->mmapped_ptr += num_values;
	if (context->mask_channel == NULL) return (TRUE);
	if (num_values > 62)
	{
	    block = num_values / 31;
	    block -= 2;
	    if ( !ch_skip (context->mask_channel, block * 4) )
	    {
		fprintf (stderr, "Error skipping in Miriad mask file\n");
		return (FALSE);
	    }
	    num_values -= block * 31;
	}
	for (count = 0; count < num_values; ++count)
	{
	    if ( !_foreign_miriad_read_get_next_mask_value (context, &mask) )
	    {
		fprintf (stderr, "Error reading Miriad mask file\n");
		return (FALSE);
	    }
	}
	return (TRUE);
    }
    if (data == NULL)
    {
	array_pointer arrayp;

	/*  Data is in multi_desc  */
	pack_desc = multi_desc->headers[0];
	arr_desc = (array_desc *) pack_desc->element_desc[0];
	m_copy (&arrayp, multi_desc->data[0], sizeof arrayp);
	data = arrayp.array;
	if (data == NULL)
	{
	    fprintf (stderr, "No array to write data into!\n");
	    a_prog_bug (function_name);
	}
	num_values = ds_get_array_size (arr_desc);
    }
    /*  Divide the transfer into multiple sections, which reduces the amount of
	disc activity due to swapping. If the sections are too large, the net
	effect is a massive disc-to-disc copy, since a write to virtual memory
	can result in a write to disc if there is insufficient RAM  */
    for (; num_values > 0;
	 num_values -= block_length, data += sizeof (float) * block_length)
    {
	block_length = (num_values > BLOCK_LENGTH) ? BLOCK_LENGTH : num_values;
	if ( !_foreign_miriad_read_read_data (context, data, block_length,
					      blank_count) )
	{
	    return (FALSE);
	}
	if ( s_check_for_int () )
	{
	    fprintf (stderr, "control_c abort: reading terminated\n");
	    return (FALSE);
	}
    }
    return (TRUE);
}   /*  End Function foreign_miriad_read_data  */

/*EXPERIMENTAL_FUNCTION*/
void foreign_miriad_close_data_context (KMiriadDataContext context)
/*  [SUMMARY] Close a KMiriadDataContext object.
    <context> The context.
    [RETURNS] Nothing.
*/
{
    static char function_name[] = "foreign_miriad_close_data_context";

    VERIFY_CONTEXT (context);
    if (context->image_channel) ch_close (context->image_channel);
    if (context->mask_channel) ch_close (context->mask_channel);
    context->magic_number = 0;
    m_free (context);
}   /*  End Function foreign_miriad_close_data_context  */

/*EXPERIMENTAL_FUNCTION*/
flag foreign_miriad_read_history (CONST char *dirname, multi_array *multi_desc)
/*  [SUMMARY] Read the history component of a Miriad Image file.
    <dirname> The directory name of the Miriad dataset name.
    <multi_desc> The multi_array header the history will be written to.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    Channel channel;
    struct stat statbuf;
    char history_name[STRING_LENGTH];
    char txt[STRING_LENGTH];
    static char function_name[] = "foreign_miriad_read_history";

    if ( (dirname == NULL) || (multi_desc == NULL) )
    {
	fprintf (stderr, "NULL pointer(s) passed\n");
	a_prog_bug (function_name);
    }
    sprintf (history_name, "%s/history", dirname);
    /*  Test if history file present  */
    if (stat (history_name, &statbuf) != 0)
    {
	if (errno != ENOENT)
	{
	    /*  Strange error  */
	    fprintf (stderr, "Error statting file: \"%s\"\t%s\n",
		     history_name, ERRSTRING);
	    return (FALSE);
	}
	/*  Not present: I can deal with that  */
	return (TRUE);
    }
    if ( ( channel = ch_open_compressed_file (history_name) ) == NULL )
	return (FALSE);
    while ( ch_getl (channel, txt, STRING_LENGTH) )
    {
	if ( !ds_history_append_string (multi_desc, txt, TRUE) )
	{
	    m_error_notify (function_name, "history string");
	    ch_close (channel);
	    return (FALSE);
	}
    }
    return ch_close (channel);
}   /*  End Function foreign_miriad_read_history  */

/*UNPUBLISHED_FUNCTION*/
double foreign_miriad_get_units_scale (CONST char *name)
/*  [SUMMARY] Determine the scale to convert Miriad to FITS value.
    <name> The name of the unit.
    [RETURNS] The scale value.
*/
{
    double scale;

    /*  NOTE: have to update _foreign_miriad_read_convert_units if entries are
	made to this table  */
    if (st_nicmp (name, "RA---", 5) == 0) scale = 180.0 / PI;
    else if (st_icmp (name, "RA") == 0) scale = 180.0 / PI;
    else if (st_nicmp (name, "DEC--", 5) == 0) scale = 180.0 / PI;
    else if (st_icmp (name, "DEC") == 0) scale = 180.0 / PI;
    else if (st_nicmp (name, "GLON-", 5) == 0) scale = 180.0 / PI;
    else if (st_icmp (name, "GLON") == 0) scale = 180.0 / PI;
    else if (st_nicmp (name, "GLAT-", 5) == 0) scale = 180.0 / PI;
    else if (st_icmp (name, "GLAT") == 0) scale = 180.0 / PI;
    else if (st_nicmp (name, "ELON-", 5) == 0) scale = 180.0 / PI;
    else if (st_icmp (name, "ELON") == 0) scale = 180.0 / PI;
    else if (st_nicmp (name, "ELAT-", 5) == 0) scale = 180.0 / PI;
    else if (st_icmp (name, "ELAT") == 0) scale = 180.0 / PI;
    else if (st_nicmp (name, "FREQ", 4) == 0) scale = 1e9;
    else if (st_nicmp (name, "VELO", 4) == 0) scale = 1e3;
    else if (st_nicmp (name, "FELO", 4) == 0) scale = 1e3;
    else if (st_nicmp (name, "VOBS", 4) == 0) scale = 1e3;
    else if (st_nicmp (name, "RESTFREQ", 8) == 0) scale = 1e9;
    else if (st_icmp (name, "ANGLE") == 0) scale = 180.0 / PI;
    else if (st_icmp (name, "BMIN") == 0) scale = 180.0 / PI;
    else if (st_icmp (name, "BMAJ") == 0) scale = 180.0 / PI;
    else scale = 1.0;
    return (scale);
}   /*  End Function foreign_miriad_get_units_scale  */

/*EXPERIMENTAL_FUNCTION*/
unsigned int foreign_miriad_read_get_alignment_padding (uaddr position,
							uaddr size)
/*  [SUMMARY] Compute padding to align to a word boundary.
    <position> The current position.
    <size> The size of the word in bytes.
    [RETURNS] The number of padding bytes required.
*/
{
    uaddr pad;
    static char function_name[] = "foreign_miriad_read_get_alignment_padding";

    if (size == 0)
    {
	fprintf (stderr, "zero size!\n");
	a_prog_bug (function_name);
    }
    pad = position % size;
    if (pad == 0) return (0);
    return (size - pad);
}   /*  End Function foreign_miriad_read_get_alignment_padding  */


/*  Private functions follow  */

multi_array *_foreign_miriad_read_header (CONST char *dirname, flag data_alloc,
					  flag sanitise, Header header, ...)
/*  [SUMMARY] Read a Miriad Image or Visibility file header.
    [PURPOSE] This routine will read the header of a Miriad Image file from a
    channel. The data section is NOT read.
    <dirname> The directory name of the Miriad dataset name.
    <data_alloc> If TRUE, the data space is allocated.
    <sanitise> If TRUE, Miriad axes with length 1 are ignored. This is highly
    recommended.
    <header> The header. This is modified.
    [VARARGS] The optional attributes are given as pairs of attribute-key
    attribute-value pairs. This list must be terminated with
    FA_MIRIAD_READ_HEADER_END. See [<foreign_ATT_MIRIAD_READ_HEADER>] for a
    list of defined attributes.
    [RETURNS] A pointer to the multi_array data structure on success, else
    NULL.
*/
{
    va_list argp;
    multi_array *multi_desc;
    /*static char function_name[] = "_foreign_miriad_read_header";*/

    va_start (argp, header);
    multi_desc = _foreign_miriad_read_header_argp (dirname, data_alloc,
						   sanitise, header, argp);
    va_end (argp);
    return (multi_desc);
}   /*  End Function _foreign_miriad_read_header  */

multi_array *_foreign_miriad_read_header_argp (CONST char *dirname,
					       flag data_alloc, flag sanitise,
					       Header header, va_list argp)
/*  [SUMMARY] Read a Miriad Image or Visibility file header.
    [PURPOSE] This routine will read the header of a Miriad Image file from a
    channel. The data section is NOT read.
    <dirname> The directory name of the Miriad dataset name.
    <data_alloc> If TRUE, the data space is allocated.
    <sanitise> If TRUE, Miriad axes with length 1 are ignored. This is highly
    recommended.
    <header> The header. This is modified.
    <argp> The optional attributes are given as pairs of attribute-key
    attribute-value pairs. This list must be terminated with
    FA_MIRIAD_READ_HEADER_END. See [<foreign_ATT_MIRIAD_READ_HEADER>] for a
    list of defined attributes.
    [RETURNS] A pointer to the multi_array data structure on success, else
    NULL.
*/
{
    Channel channel;
    multi_array *multi_desc;
    char filename[STRING_LENGTH];
    static char function_name[] = "_foreign_miriad_read_header_argp";

    sprintf (filename, "%s/header", dirname);
    if ( ( channel = ch_open_compressed_file (filename) ) == NULL )
    {
	fprintf (stderr, "%s: error opening: \"%s\"\t%s\n",
		 function_name, filename, ERRSTRING);
	return (NULL);
    }
    if ( !_foreign_miriad_read_header_items (channel, header, argp) )
    {
	ch_close (channel);
	return (NULL);
    }
    ch_close (channel);
    if ( !_foreign_miriad_read_visdata_header (dirname,header) ) return (NULL);
    multi_desc = _foreign_miriad_process_header (header, data_alloc, sanitise,
						 argp);
    return (multi_desc);
}   /*  End Function _foreign_miriad_read_header_argp  */

static flag _foreign_miriad_read_header_items (Channel channel, Header header,
					       va_list argp)
/*  [SUMMARY] Read a Miriad Image file header.
    [PURPOSE] This routine will read the header of a Miriad Image file from a
    channel. The data section is NOT read.
    <channel> The channel to read from.
    <header> The list of header items. This is modified.
    <argp> The optional attributes are given as pairs of attribute-key
    attribute-value pairs. This list must be terminated with
    FA_MIRIAD_READ_HEADER_END. See [<foreign_ATT_MIRIAD_READ_HEADER>] for a
    list of defined attributes.
    [RETURNS] A pointer to the multi_array data structure on success, else
    NULL.
*/
{
    flag keep_going;
    int item_size;
    unsigned int att_key;
    char *ch_ptr;
    char item_buf[MIN_ITEM_SIZE];
    extern char *data_type_names[NUMTYPES];
    static char function_name[] = "_foreign_miriad_read_header_items";

    if (first_time)
    {
	first_time = FALSE;
	if (r_getenv ("FOREIGN_MIRIAD_READ_DEBUG") != NULL) debug = TRUE;
    }
    if (channel == NULL)
    {
	fprintf (stderr, "NULL pointer passed\n");
	a_prog_bug (function_name);
    }
    /*  Process attributes  */
    while ( ( att_key = va_arg (argp, unsigned int) )
	    != FA_MIRIAD_READ_HEADER_END )
    {
	switch (att_key)
	{
	  default:
	    fprintf (stderr, "Unknown attribute key: %u\n", att_key);
	    a_prog_bug (function_name);
	    break;
	}
    }
    for ( keep_going = TRUE;
	  keep_going &&
	      (ch_read (channel, item_buf, MIN_ITEM_SIZE) >= MIN_ITEM_SIZE);
	  keep_going = ch_skip_to_boundary (channel, MIN_ITEM_SIZE) )
    {
	HeaderItem item, next;

	/*  Have a new item  */
	if ( ( item = m_calloc (sizeof *item) ) == NULL )
	    m_abort (function_name, "header item");
	item_size = item_buf[MIN_ITEM_SIZE - 1];
	item_buf[MIN_ITEM_SIZE - 1] = '\0';  /*  Replace with terminator  */
	if ( ( item->name = st_dup (item_buf) ) == NULL )
	    m_abort (function_name, "item name");
	/*  Make keywords FITS like (yuk)  */
	st_upr (item->name);
	/*  Ignore the "MOSTABLE" keyword which contains undescribed data (for
	    the mosaicing beams)
	    */
	if (strcmp (item->name, "MOSTABLE") == 0)
	{
	    if ( !ch_skip (channel, item_size) )
	    {
		fprintf (stderr, "%s: error skipping MOSTABLE item\n",
			 function_name);
		_foreign_miriad_read_free_header (item);
		return (FALSE);
	    }
	    continue;
	}
	if (item_size < 4)
	{
	    fprintf (stderr, "Item: \"%s\" size: %d too small\n",
		     item->name, item_size);
	    _foreign_miriad_read_free_header (item);
	    return (FALSE);
	}
	item_size -= 4;
	/*  Determine the data type  */
	if ( !_foreign_miriad_read_miriad_type (channel, &item->type) )
	{
	    fprintf (stderr, "%s: error reading type for item: \"%s\"\n",
		     function_name, item->name);
	    if ( !ch_skip (channel, item_size) )
	    {
		fprintf (stderr, "%s: error skipping item: \"%s\"\n",
			 function_name, item->name);
		_foreign_miriad_read_free_header (item);
		return (FALSE);
	    }
	    continue;
	}
	/*  Miriad uses it's byte type for strings. Weird  */
	if (item->type == K_BYTE) item->type = K_VSTRING;
	if ( !_foreign_miriad_read_header_value (channel, item, item_size) )
	{
	    _foreign_miriad_read_free_header (item);
	    return (FALSE);
	}
	/*  Process the item  */
	if (strcmp (item->name, "NAXIS") == 0)
	{
	    if (item->type != K_INT)
	    {
		fprintf (stderr, "NAXIS keyword not of type integer\n");
		_foreign_miriad_read_free_header (item);
		return (FALSE);
	    }
	    if ( (header->naxis = *(int *) item->data) < 1 )
	    {
		fprintf (stderr, "NAXIS value: %d negative!\n", header->naxis);
		_foreign_miriad_read_free_header (item);
		return (FALSE);
	    }
	}
	else if (strncmp (item->name, "NAXIS", 5) == 0)
	{
	    if (item->type == K_INT)
	    {
		if (*(int *) item->data > 1) ++header->num_dim;
	    }
	}
	else if (strncmp (item->name, "CTYPE", 5) == 0)
	{
	    if (item->type == K_VSTRING)
	    {
		ch_ptr = *(char **) item->data;
		if ( (strncmp (ch_ptr, "FREQ", 4) == 0) ||
		     (strncmp (ch_ptr, "FELO", 4) == 0) ||
		     (strncmp (ch_ptr, "VELO", 4) == 0) )
		{
		    header->have_frequency_axis = TRUE;
		}
	    }
	}
	else if (strcmp (item->name, "CELLSCAL") == 0)
	    header->have_cellscal = TRUE;
	if (debug)
	{
	    fprintf (stderr, "Hitem name: \"%s\"\ttype: %s\tvalue:\t",
		     item->name, data_type_names[item->type]);
	    if (item->type == NONE) fputs("unstructured binary data\n",stderr);
	    else dmp_element (stderr, item->type, NULL, item->data, FALSE);
	}
	if (item->type == NONE)
	    item = _foreign_miriad_read_split_binary_item (item);
	for (; item; item = next)
	{
	    next = item->next;
	    KLIST_PREPEND (header, item);
	}
	/*  Any padding is automatically read in the loop construct  */
    }
    return (TRUE);
}   /*  End Function _foreign_miriad_read_header_items  */

static multi_array *_foreign_miriad_process_header (Header header,
						    flag data_alloc,
						    flag sanitise,
						    va_list argp)
/*  [SUMMARY] Process a Miriad Image file header.
    [PURPOSE] This routine will read the header of a Miriad Image file from a
    channel. The data section is NOT read.
    <header> The header.
    <data_alloc> If TRUE, the data space is allocated.
    <sanitise> If TRUE, Miriad axes with length 1 are ignored. This is highly
    recommended.
    <append_items> A list of header items to be appended. These will not be
    freed. This may be NULL.
    <argp> The optional attributes are given as pairs of attribute-key
    attribute-value pairs. This list must be terminated with
    FA_MIRIAD_READ_HEADER_END. See [<foreign_ATT_MIRIAD_READ_HEADER>] for a
    list of defined attributes.
    [RETURNS] A pointer to the multi_array data structure on success, else
    NULL.
*/
{
    multi_array *multi_desc;
    /*static char function_name[] = "_foreign_miriad_process_header";*/

    if ( (header->naxis < 1) && (header->vartable == NULL) )
    {
	fprintf (stderr, "No NAXIS keyword found and no visibility data!\n");
	return (NULL);
    }
    /*  Convert units from Miriad to FITS  */
    if ( !_foreign_miriad_read_convert_units (header->first) )
    {
	return (NULL);
    }
    /*  Construct Karma data structure  */
    multi_desc = _foreign_miriad_read_make_desc (header, data_alloc, sanitise);
    if (multi_desc == NULL) return (NULL);
    if (header->have_frequency_axis && (header->num_dim == 3) &&
	!header->have_cellscal)
    {
	fprintf(stderr,
		"WARNING: Miriad spectral cube has no \"CELLSCAL\" keyword\n");
	fprintf (stderr,
		 "Assuming 1/F scaling for pixel cell size: adding keyword\n");
	if ( !ds_put_unique_named_string (multi_desc->headers[0],
					  multi_desc->data, "CELLSCAL", "1/F",
					  FALSE) )
	{
	    fprintf (stderr, "Error adding \"CELLSCAL\" keyword\n");
	    ds_dealloc_multi (multi_desc);
	    return (NULL);
	}
    }
    return (multi_desc);
}   /*  End Function _foreign_miriad_process_header  */

static flag _foreign_miriad_read_miriad_type (Channel channel,
					      unsigned int *type)
/*  [PURPOSE] This routine will read the 4 bytes in a Miriad file which contain
    a type value and will convert to Karma type values.
    <channel> The channel to read from.
    <type> The Karma type is written here.
    [RETURNS] TRUE on success, else FALSE on error.
*/
{
    unsigned long in_type;
    static char function_name[] = "_foreign_miriad_read_miriad_type";

    if ( !pio_read32 (channel, &in_type) ) return (FALSE);
    switch (in_type)
    {
      case 0:
	*type = NONE;
	break;
      case 1:
	*type = K_BYTE;
	break;
      case 2:
	*type = K_INT;
	break;
      case 3:
	*type = K_SHORT;
	break;
      case 4:
	*type = K_FLOAT;
	break;
      case 5:
	*type = K_DOUBLE;
	break;
      case 6:
	*type = K_VSTRING;
	break;
      case 7:
	*type = K_COMPLEX;
	break;
      default:
	fprintf (stderr, "%s: bad Miriad in_type: %lu\n",
		 function_name, in_type);
	return (FALSE);
	/*break;*/
    }
    return (TRUE);
}   /*  End Function _foreign_miriad_read_miriad_type  */

static flag _foreign_miriad_read_header_value (Channel channel,
					       HeaderItem item,
					       unsigned int length)
/*  [SUMMARY] Read a header value.
    <channel> The channel to read from.
    <item> The HeaderItem to fill.
    <length> The number of bytes to read, including any leading padding
    required for alignment purposes.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    static char function_name[] = "_foreign_miriad_read_header_value";

    if (item->type == NONE)
    {   /*  Unstructured binary data: just read it and deal with it later  */
	if (item->buf_len < length)
	{
	    if (item->data) m_free (item->data);
	    if ( ( item->data = m_alloc (length) ) == NULL )
		m_abort (function_name, "binary data");
	    item->buf_len = length;
	}
	if (ch_read (channel, item->data, length) < length)
	{
	    fprintf (stderr, "%s: error reading binary data\t%s\n",
		     function_name, ERRSTRING);
	    return (FALSE);
	}
	item->num_values = 1;
    }
    else if (item->type == K_VSTRING)
    {   /*  String data  */
	char *ch_ptr;

	if (item->data == NULL)
	{
	    if ( ( item->data = m_calloc (sizeof ch_ptr) ) == NULL )
		m_abort (function_name, "string pointer");
	}
	if (item->buf_len < length + 1)
	{
	    ch_ptr = *(char **) item->data;
	    if (ch_ptr) m_free (ch_ptr);
	    if ( ( ch_ptr = m_alloc (length + 1) ) == NULL )
		m_abort (function_name, "string");
	    item->buf_len = length + 1;
	    *(char **) item->data = ch_ptr;
	}
	else ch_ptr = *(char **) item->data;
	if (ch_read (channel, ch_ptr, length) < length)
	{
	    fprintf (stderr, "%s: error reading string\t%s\n",
		     function_name, ERRSTRING);
	    return (FALSE);
	}
	ch_ptr[length] = '\0';
	item->num_values = 1;
    }
    else
    {   /*  Normal data type  */
	unsigned int drain_size, net_bytes;
	packet_desc pack_desc;
	extern char host_type_sizes[NUMTYPES];
	extern char network_type_bytes[NUMTYPES];

	net_bytes = network_type_bytes[item->type];
	if (net_bytes == 0)
	{
	    fprintf (stderr, "Item type: %u has zero size\n", item->type);
	    return (FALSE);
	}
	drain_size = foreign_miriad_read_get_alignment_padding (4, net_bytes);
	if ( !ch_skip (channel, drain_size) )
	{
	    fprintf (stderr, "%s: error skipping to item data for: \"%s\"\n",
		     function_name, item->name);
	    return (FALSE);
	}
	length -= drain_size;
	if (net_bytes > length)
	{
	    fprintf (stderr, "Item: \"%s\" size: %d too small for data\n",
		     item->name, length);
	    return (FALSE);
	}
	item->num_values = length / net_bytes;
	if (item->buf_len < host_type_sizes[item->type] * item->num_values)
	{
	    if (item->data) m_free (item->data);
	    if ( ( item->data =
		   m_alloc (host_type_sizes[item->type] * item->num_values) )
		 == NULL )
		m_abort (function_name, "item data");
	    item->buf_len = host_type_sizes[item->type] * item->num_values;
	}
	/*  Read data values. Construct fake packet to optimise reading  */
	m_clear (&pack_desc, sizeof pack_desc);
	pack_desc.num_elements = 1;
	pack_desc.element_types = &item->type;
	if ( !dsrw_read_packets (channel, &pack_desc, item->data,
				 item->num_values) ) return (FALSE);
	length -= net_bytes * item->num_values;
	if (length > 0)
	{
	    fprintf (stderr, "Item: \"%s\", type: %u has: %u trailing bytes\n",
		     item->name, item->type, length);
	    if ( !ch_skip (channel, length) )
	    {
		fprintf (stderr, "%s: error skipping: %u bytes\n",
			 function_name, length);
		return (FALSE);
	    }
	}
    }
    return (TRUE);
}   /*  End Function _foreign_miriad_read_header_value  */

static multi_array *_foreign_miriad_read_make_desc (Header header,
						    flag data_alloc,
						    flag sanitise)
/*  [PURPOSE] This routine will make a Karma descriptor from a list of Miriad
    header items.
    <header> The header.
    <data_alloc> If TRUE, the data space is allocated.
    <sanitise> If TRUE, Miriad axes with length 1 are ignored. This is highly
    recommended.
    [RETURN] A multi array descriptor pointer on success, else NULL.
*/
{
    HeaderItem item;
    unsigned int count, num_items, num_arrays;
    packet_desc *top_pack_desc;
    array_desc *arr_desc[NUM_VISDATA_ARRAYS];
    char *top_packet, *element;
    multi_array *multi_desc;
    extern char host_type_sizes[NUMTYPES];
    static char function_name[] = "_foreign_miriad_read_make_desc";

    if (header->vartable == NULL)
    {
	arr_desc[0] =
	    _foreign_miriad_read_make_image_array_desc (header->first,
							header->naxis,
							sanitise);
	if (arr_desc[0] == NULL) return (NULL);
	num_arrays = 1;
    }
    else if ( !_foreign_miriad_read_make_visdata_array_descs (header, arr_desc,
							      &num_arrays) )
	return (NULL);
    if ( ( multi_desc = ds_alloc_multi (1) ) == NULL )
    {
	m_abort (function_name, "multi array header");
    }
    for (item = header->first, num_items = 0; item; item = item->next)
    {
	if (item->num_values < 1) continue;
	if (item->num_values != 1)
	{
	    if (sanitise) continue;
	    fprintf (stderr, "Item: \"%s\" has %u values, ignoring extras\n",
		     item->name, item->num_values);
	}
	++num_items;
    }
    if ( ( top_pack_desc = ds_alloc_packet_desc (num_items + num_arrays) )
	 == NULL )
	m_abort (function_name, "top level packet descriptor");
    multi_desc->headers[0] = top_pack_desc;
    /*  Fill in the packet descriptor  */
    for (count = 0; count < num_arrays; ++count)
    {
	top_pack_desc->element_desc[count] = (char *) arr_desc[count];
	top_pack_desc->element_types[count] = K_ARRAY;
    }
    for (item = header->first, num_items = 0; item; item = item->next)
    {
	if (item->num_values < 1) continue;
	if (sanitise && (item->num_values != 1) ) continue;
	top_pack_desc->element_types[num_items + num_arrays] = item->type;
	if ( ( top_pack_desc->element_desc[num_items + num_arrays] =
	      st_dup (item->name) ) == NULL )
	{
	    m_abort (function_name, "element name");
	}
	++num_items;
    }
    /*  Create and fill the packet  */
    if ( ( top_packet = ds_alloc_packet (top_pack_desc) ) == NULL )
    {
	m_abort (function_name, "top level packet");
    }
    multi_desc->data[0] = top_packet;
    if (data_alloc)
    {
	array_pointer arrayp;

	for (count = 0; count < num_arrays; ++count)
	{
	    if ( !ds_alloc_vm_array (&arrayp, arr_desc[count], FALSE, FALSE) &&
		 !ds_alloc_mmap_array (-1, 0, 0, TRUE, &arrayp,
				       arr_desc[count], FALSE, FALSE) )
	    {
		m_error_notify (function_name, "array");
		ds_dealloc_multi (multi_desc);
		return (NULL);
	    }
	    m_copy (multi_desc->data[0] +
		    ds_get_element_offset (top_pack_desc, count),
		    &arrayp, sizeof arrayp);
	}
    }
    for (item = header->first, num_items = 0; item; item = item->next)
    {
	if (item->num_values < 1) continue;
	if (sanitise && (item->num_values != 1) ) continue;
	element = top_packet + ds_get_element_offset (top_pack_desc,
						      num_items + num_arrays);
	m_copy (element, item->data, host_type_sizes[item->type]);
	++num_items;
    }
    return (multi_desc);
}   /*  End Function _foreign_miriad_read_make_desc  */

static array_desc *_foreign_miriad_read_make_image_array_desc
    (HeaderItem first_item, signed int naxis, flag sanitise)
/*  [SUMMARY] Make image array descriptor from a list of Miriad header items.
    <first_item> The first header item.
    <naxis> The value of the naxis keyword.
    <sanitise> If TRUE, Miriad axes with length 1 are ignored. This is highly
    recommended.
    [RETURN] An array descriptor pointer on success, else NULL.
*/
{
    HeaderItem item, crpix_item, crval_item, cdelt_item;
    signed int axis_length, axis_count;
    unsigned int num_dim;
    unsigned int elem_type = K_FLOAT;
    array_desc *arr_desc;
    char *elem_name, *name;
    uaddr *dim_lengths;
    char **dim_names;
    char txt[STRING_LENGTH];
    static char def_elem_name[] = "Data Value";
    static char function_name[] = "_foreign_miriad_read_make_image_array_desc";

    /*  Construct the lengths, first co-ord, last co-ord and names arrays  */
    if ( ( dim_lengths = m_alloc (sizeof *dim_lengths * naxis) ) == NULL )
    {
	m_error_notify (function_name, "dimension lengths");
	return (NULL);
    }
    /*  Create array of dimension name pointers  */
    if ( ( dim_names = m_alloc (sizeof *dim_names * naxis) ) == NULL )
    {
	m_error_notify (function_name, "dimension name pointers");
	m_free (dim_lengths);
	return (NULL);
    }
    /*  Grab dimension information from header  */
    for (axis_count = naxis, num_dim = 0; axis_count > 0; --axis_count)
    {
	/*  Search for this axis name  */
	sprintf (txt, "NAXIS%d", axis_count);
	if ( ( item = _foreign_miriad_read_find_item (first_item, txt) )
	     == NULL )
	{
	    fprintf (stderr, "%s: naxis: %u but no \"%s\" item found\n",
		     function_name, axis_count, txt);
	    m_free (dim_lengths);
	    m_free (dim_names);
	    return (NULL);
	}
	if (item->type != K_INT)
	{
	    fprintf (stderr, "%s: item: \"%s\" not integer type\n",
		     function_name, txt);
	    m_free (dim_lengths);
	    m_free (dim_names);
	    return (NULL);
	}
	axis_length = *(int *) item->data;
	if ( sanitise && (axis_length < 2) ) continue;
	/*  Put entry in arrays  */
	/*  First the name  */
	sprintf (txt, "CTYPE%d", axis_count);
	if ( ( item = _foreign_miriad_read_find_item (first_item, txt) )
	     == NULL )
	{
	    dim_names[num_dim] = NULL;
	}
	else if (item->type != K_VSTRING)
	{
	    fprintf (stderr, "%s: item: \"%s\" not vstring type\n",
		     function_name, txt);
	    m_free (dim_lengths);
	    m_free (dim_names);
	    return (NULL);
	}
	else
	{
	    /*  Axis name was found  */
	    dim_names[num_dim] = *(char **) item->data;
	    name = dim_names[num_dim];
	    if (st_icmp (name, "ANGLE") == 0)
	    {
		sprintf (txt, "Axis %u (deg)", num_dim);
		m_free (name);
		if ( ( name = st_dup (txt) ) == NULL )
		{
		    m_abort (function_name, "axis name");
		}
		dim_names[num_dim] = name;
		*(char **) item->data = name;
	    }
	    else if (st_icmp (name, "Radius (R\\dJ\\u)") == 0)
	    {
		sprintf (txt, "Radius %u (R\\dJ\\u)", num_dim);
		m_free (name);
		if ( ( name = st_dup (txt) ) == NULL )
		{
		    m_abort (function_name, "axis name");
		}
		dim_names[num_dim] = name;
		*(char **) item->data = name;
	    }
	}
	dim_lengths[num_dim] = axis_length;
	sprintf (txt, "CRPIX%d", axis_count);
	if ( ( crpix_item = _foreign_miriad_read_find_item (first_item, txt) )
	     != NULL )
	{
	    if (crpix_item->type == K_DOUBLE)
		crpix_item->d_data = *(double *) crpix_item->data;
	    else if (crpix_item->type == K_FLOAT)
		crpix_item->d_data = *(float *) crpix_item->data;
	    else
	    {
		fprintf (stderr, "%s: warning: \"%s\" not float or double\n",
			 function_name, txt);
		crpix_item = NULL;
	    }
	}
	sprintf (txt, "CRVAL%d", axis_count);
	if ( ( crval_item = _foreign_miriad_read_find_item (first_item, txt) )
	     != NULL )
	{
	    if (crval_item->type == K_DOUBLE)
		crval_item->d_data = *(double *) crval_item->data;
	    else if (crval_item->type == K_FLOAT)
		crval_item->d_data = *(float *) crval_item->data;
	    else
	    {
		fprintf (stderr, "%s: warning: \"%s\" not float or double\n",
			 function_name, txt);
		crval_item = NULL;
	    }
	}
	sprintf (txt, "CDELT%d", axis_count);
	if ( ( cdelt_item = _foreign_miriad_read_find_item (first_item, txt) )
	     != NULL )
	{
	    if (cdelt_item->type == K_DOUBLE)
		cdelt_item->d_data = *(double *) cdelt_item->data;
	    else if (cdelt_item->type == K_FLOAT)
		cdelt_item->d_data = *(float *) cdelt_item->data;
	    else
	    {
		fprintf (stderr, "%s: warning: \"%s\" not float or double\n",
			 function_name, txt);
		cdelt_item = NULL;
	    }
	}
	++num_dim;
    }
    sprintf (txt, "BUNIT");
    if ( ( item = _foreign_miriad_read_find_item (first_item, txt) ) == NULL )
    {
	elem_name = def_elem_name;
    }
    else elem_name = *(char **) item->data;
    if ( elem_name && (strlen (elem_name) < 1) )
    {
	fprintf (stderr, "WARNING: BUNIT is an empty string: defaulting to ");
	fprintf (stderr, "\"%s\"\n", def_elem_name);
	elem_name = def_elem_name;
    }
    /*  Create the array descriptor  */
    arr_desc = ds_easy_alloc_array_desc
	(num_dim, dim_lengths, NULL, NULL, (CONST double **) NULL,
	 (CONST char **) dim_names, 1, &elem_type, (CONST char **) &elem_name);
    m_free (dim_lengths);
    m_free (dim_names);
    return (arr_desc);
}   /*  End Function _foreign_miriad_read_make_image_array_desc  */

static flag _foreign_miriad_read_make_visdata_array_descs
    (Header header, array_desc **arr_descs, unsigned int *num_arrays)
/*  [SUMMARY] Make visdata array descriptors from vartable.
    <header> The header.
    <arr_descs> The array descriptors are written here.
    <num_arrays> The number of array descriptors is written here.
    [RETURN] TRUE on success, else FALSE.
*/
{
    HeaderItem item;
    unsigned int npol, nchan;
    dim_desc *dim;
    unsigned int elem_types[3];
    double value[2];
    uaddr dim_lengths[4];
    char *dim_names[4], *elem_names[3];
    static char function_name[] =
	"_foreign_miriad_read_make_visdata_array_descs";

    if ( ( item = _foreign_miriad_read_find_item (header->first,
						  "VISDATA:NPOL") )
	 == NULL )
    {
	fprintf (stderr, "%s: no npol item found\n", function_name);
	return (FALSE);
    }
    ds_get_element (item->data, item->type, value, NULL);
    npol = value[0];
    if (npol != header->polarisations.num_values)
    {
	fprintf (stderr, "%s: NPOL=%u but found %lu unique values\n",
		 function_name, npol, header->polarisations.num_values);
	return (FALSE);
    }
    if ( ( item = _foreign_miriad_read_find_item (header->first,
						  "VISDATA:NCHAN") )
	 == NULL )
    {
	fprintf (stderr, "%s: no nchan item found\n", function_name);
	return (FALSE);
    }
    ds_get_element (item->data, item->type, value, NULL);
    nchan = value[0];
    /*  Create the array descriptor for the visibility hypercube  */
    elem_types[0] = K_COMPLEX;
    elem_names[0] = "Visibility Value";
    dim_names[VISDATA_DIM_INDEX_POLARISATION] = "Polarisation";
    dim_names[VISDATA_DIM_INDEX_TIME] = "Time";
    dim_names[VISDATA_DIM_INDEX_BASELINE] = "Baseline";
    dim_names[VISDATA_DIM_INDEX_CHANNEL] = "Channel";
    dim_lengths[VISDATA_DIM_INDEX_POLARISATION] = npol;
    dim_lengths[VISDATA_DIM_INDEX_TIME] = header->times.num_values;
    dim_lengths[VISDATA_DIM_INDEX_BASELINE] = header->baselines.num_values;
    dim_lengths[VISDATA_DIM_INDEX_CHANNEL] = nchan;
    arr_descs[0] = ds_easy_alloc_array_desc
	(4, dim_lengths, NULL, NULL, (CONST double **) NULL,
	 (CONST char **) dim_names, 1, elem_types, (CONST char **) elem_names);
    dim = arr_descs[0]->dimensions[VISDATA_DIM_INDEX_POLARISATION];
    dim->coordinates = header->polarisations.data;
    header->polarisations.data = NULL;
    dim->first_coord = dim->coordinates[0];
    dim->last_coord = dim->coordinates[dim->length - 1];
    dim = arr_descs[0]->dimensions[VISDATA_DIM_INDEX_TIME];
    dim->coordinates = header->times.data;
    header->times.data = NULL;
    dim->first_coord = dim->coordinates[0];
    dim->last_coord = dim->coordinates[dim->length - 1];
    dim = arr_descs[0]->dimensions[VISDATA_DIM_INDEX_BASELINE];
    dim->coordinates = header->baselines.data;
    header->baselines.data = NULL;
    dim->first_coord = dim->coordinates[0];
    dim->last_coord = dim->coordinates[dim->length - 1];
    ds_autotile_array (arr_descs[0], FALSE);
    /*  Create the array descriptor for baseline index->antenna pair table  */
    elem_types[0] = K_UINT;
    elem_types[1] = K_UINT;
    elem_names[0] = "Antenna A";
    elem_names[1] = "Antenna B";
    dim_names[0] = "Baseline Index_Antenna Pair";
    dim_lengths[0] = header->baselines.num_values;
    arr_descs[1] = ds_easy_alloc_array_desc
	(1, dim_lengths, NULL, NULL, (CONST double **) NULL,
	 (CONST char **) dim_names, 2, elem_types, (CONST char **) elem_names);
    /*  Create the array descriptor for baseline/time index->coord table  */
    elem_types[0] = elem_types[1] = elem_types[2] = K_DOUBLE;
    elem_names[0] = "U Coord";
    elem_names[1] = "V Coord";
    elem_names[2] = "W Coord";
    dim_names[VISUVW_DIM_INDEX_TIME] = "Time Index_Coords";
    dim_names[VISUVW_DIM_INDEX_BASELINE] = "Baseline Index_Coords";
    dim_lengths[VISUVW_DIM_INDEX_TIME] = header->times.num_values;
    dim_lengths[VISUVW_DIM_INDEX_BASELINE] = header->baselines.num_values;
    arr_descs[2] = ds_easy_alloc_array_desc
	(2, dim_lengths, NULL, NULL, (CONST double **) NULL,
	 (CONST char **) dim_names, 3, elem_types, (CONST char **) elem_names);
    *num_arrays = NUM_VISDATA_ARRAYS;
    return (TRUE);
}   /*  End Function _foreign_miriad_read_make_visdata_array_descs  */

static HeaderItem _foreign_miriad_read_find_item (HeaderItem first_item,
						  CONST char *name)
/*  [PURPOSE] This routine will find an item in the header with given name.
    <first_item> The first item in the header.
    <name> The name of the item.
    [RETURNS] The item on sucess, else NULL.
*/
{
    HeaderItem item;

    for (item = first_item; item; item = item->next)
    {
	if (st_icmp (item->name, name) == 0) return (item);
    }
    return (NULL);
}   /*  End Function _foreign_miriad_read_find_item  */

static void _foreign_miriad_read_free_header (HeaderItem first_item)
/*  [PURPOSE] This routine will free a header.
    <first_item> The first item in the header.
    [RETURNS] Nothing.
*/
{
    HeaderItem next;

    while (first_item)
    {
	m_free (first_item->name);
	if (first_item->data) m_free (first_item->data);
	next = first_item->next;
	m_free (first_item);
	first_item = next;
    }
}   /*  End Function _foreign_miriad_read_free_header  */

static flag _foreign_miriad_read_get_next_mask_value
    (KMiriadDataContext context, flag *mask)
/*  [SUMMARY] Get next mask value.
    <context> The context to get the mask from.
    <mask> The mask value is written here. This will be TRUE if mask is set
    (data is valid), else FALSE indicating the data is invalid.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    if (++context->last_mask_bit_in_word < 31)
    {
	/*  Have at least one more mask bit available  */
	*mask = (context->mask_word & 0x01) ? TRUE : FALSE;
	context->mask_word = context->mask_word >> 1;
	return (TRUE);
    }
    if ( !pio_read32 (context->mask_channel, &context->mask_word) )
	return (FALSE);
    *mask = (context->mask_word & 0x01) ? TRUE : FALSE;
    context->mask_word = context->mask_word >> 1;
    context->last_mask_bit_in_word = 0;
    return (TRUE);
}   /*  End Function _foreign_miriad_read_get_next_mask_value  */

static flag _foreign_miriad_read_convert_units (HeaderItem first_item)
/*  [SUMMARY] Convert header units from Miriad to FITS.
    <first_item> A pointer to the first header item.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    HeaderItem item;
    HeaderItem lon_name_item = NULL;
    HeaderItem lon_cdelt_item = NULL;
    HeaderItem lat_name_item = NULL;
    unsigned int axis_count;
    double scale;
    double orig_lat_crval = TOOBIG;
    char txt[STRING_LENGTH];
    static char function_name[] = "_foreign_miriad_read_convert_units";

    /*  First process "CRVALn" and "CDELTn" keywords associated with each
	"CTYPEn" keyword found  */
    for (axis_count = 1; TRUE; ++axis_count)
    {
	CONST char *name;

	sprintf (txt, "CTYPE%u", axis_count);
	if ( ( item = _foreign_miriad_read_find_item (first_item, txt) )
	     == NULL )
	    break;
	else if (item->type != K_VSTRING)
	{
	    fprintf (stderr, "%s: item: \"%s\" not vstring type\n",
		     function_name, txt);
	    return (FALSE);
	}
	name = *(CONST char **) item->data;
	/*  Get information on bogus Miriad "CAR" projection for later fix  */
	if ( (st_icmp (name, "RA---CAR") == 0) ||
	     (st_icmp (name, "GLON-CAR") == 0) )
	{
	    lon_name_item = item;
	    sprintf (txt, "CDELT%u", axis_count);
	    lon_cdelt_item = _foreign_miriad_read_find_item (first_item, txt);
	}
	else if ( (st_icmp (name, "DEC--CAR") == 0) ||
		  (st_icmp (name, "GLAT-CAR") == 0) )
	{
	    HeaderItem tmp;

	    lat_name_item = item;
	    sprintf (txt, "CRVAL%u", axis_count);
	    tmp = _foreign_miriad_read_find_item (first_item, txt);
	    if (tmp->type == K_DOUBLE) orig_lat_crval = *(double *) tmp->data;
	}
	scale = foreign_miriad_get_units_scale (name);
	if (scale == 1.0)
	{
	    /*  No scaling: ignore  */
	    continue;
	}
	/*  Update "CRVALn" if present  */
	sprintf (txt, "CRVAL%u", axis_count);
	if ( !_foreign_miriad_read_scale_item (first_item, txt, scale) )
	    return (FALSE);
	/*  Update "CDELTn" if present  */
	sprintf (txt, "CDELT%u", axis_count);
	if ( !_foreign_miriad_read_scale_item (first_item, txt, scale) )
	    return (FALSE);
    }
    if ( lon_name_item && lon_cdelt_item && lat_name_item &&
	 DOUBLE_IS_VALID (orig_lat_crval) &&
	 (lon_cdelt_item->type == K_DOUBLE) )
    {
	/*  Correct the bogus Miriad "CAR" projection  */
	*(double *) lon_cdelt_item->data /= cos (orig_lat_crval);
	*( strchr (*(CONST char **) lon_name_item->data, '-') ) = '\0';
	*( strchr (*(CONST char **) lat_name_item->data, '-') ) = '\0';
    }
    /*  Update "RESTFREQ" if present  */
    scale = foreign_miriad_get_units_scale ("RESTFREQ");
    if ( !_foreign_miriad_read_scale_item (first_item, "RESTFREQ", scale) )
	return (FALSE);
    /*  Update "BMIN" if present  */
    scale = foreign_miriad_get_units_scale ("BMIN");
    if ( !_foreign_miriad_read_scale_item (first_item, "BMIN", scale) )
	return (FALSE);
    /*  Update "BMAJ" if present  */
    scale = foreign_miriad_get_units_scale ("BMAJ");
    if ( !_foreign_miriad_read_scale_item (first_item, "BMAJ", scale) )
	return (FALSE);
    /*  Update "VOBS" if present  */
    scale = foreign_miriad_get_units_scale ("VOBS");
    if ( !_foreign_miriad_read_scale_item (first_item, "VOBS", scale) )
	return (FALSE);
    return (TRUE);
}   /*  End Function _foreign_miriad_read_convert_units  */

static flag _foreign_miriad_read_scale_item (HeaderItem first_item,
					     CONST char *name, double scale)
/*  [SUMMARY] Scale an atomic element.
    <first_item> A pointer to the first header item.
    <name> The name of the element to scale.
    <scale> The scale value to apply.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    HeaderItem item;
    static char function_name[] = "_foreign_miriad_read_scale_item";

    if ( ( item = _foreign_miriad_read_find_item (first_item, name) )
	 == NULL ) return (TRUE);
    if (item->type == K_DOUBLE) *(double *) item->data *= scale;
    else if (item->type == K_FLOAT) *(float *) item->data *= scale;
    else if (item->type == K_INT)
    {
	*(int *) item->data *= scale;
	fprintf (stderr,
		 "%s: warning: \"%s\" is type integer: Miriad should not do this\n",
		 function_name, name);
    }
    else
    {
	fprintf (stderr, "%s: warning: \"%s\" not float or double\n",
		 function_name, name);
	return (FALSE);
    }
    return (TRUE);
}   /*  End Function _foreign_miriad_read_scale_item  */

static flag _foreign_miriad_read_read_data (KMiriadDataContext context,
					    char *data, uaddr num_values,
					    unsigned long *blank_count)
/*  [SUMMARY] Read data in a Miriad Image file.
    [PURPOSE] This routine will read the data of a Miriad Image file from a
    KMiradDataContext object. The header section is NOT read.
    <context> The context to read from.
    <data> The data array to write the data into.
    <num_values> The number of values to write into the data array.
    <blank_count> A pointer to the number of blanks. This is updated.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    flag mask;
    uaddr value_count;
    unsigned long blank_count_local = 0;
    float *f_ptr;
#if defined(HAS_MADVISE) && defined(MADV_SEQUENTIAL)
    static char function_name[] = "_foreign_miriad_read_read_data";
#endif

    if (context->first_time)
    {
	context->first_time = FALSE;
#if defined(HAS_MADVISE) && defined(MADV_SEQUENTIAL)
	if (context->mmapped_ptr)
	{
	    if (madvise (ch_get_mmap_addr (context->image_channel),
			 ch_get_mem_size (context->image_channel),
			 MADV_SEQUENTIAL) != 0)
	    {
		fprintf (stderr, "%s: error calling madvise(2)\t%s\n",
			 function_name, ERRSTRING);
	    }
	}
#endif
    }
    /*  Read section of data  */
    if ( context->mmapped_ptr && ds_can_transfer_element_as_block (K_FLOAT) )
    {
	/*  Is memory mapped and can do block transfer  */
	m_copy (data, context->mmapped_ptr, sizeof (float) * num_values);
	context->mmapped_ptr += num_values;
    }
    else
    {
	if (pio_read_floats (context->image_channel, num_values,
			     (float *) data, NULL) < num_values)
	{
	    fprintf (stderr, "Error reading image data\t%s\n", ERRSTRING);
	    return (FALSE);
	}
    }
    if (context->mask_channel == NULL) return (TRUE);
    /*  Read section of mask and apply to data  */
    /*  Read mask one bit at a time  */
    f_ptr = (float *) data;
    for (value_count = 0; value_count < num_values; value_count++, ++f_ptr)
    {
	if ( !_foreign_miriad_read_get_next_mask_value (context, &mask) )
	{
	    fprintf (stderr, "Error reading mask data\t%s\n",
		     ERRSTRING);
	    return (FALSE);
	}
	if (!mask)
	{
	    /*  Mask is clear: bad data, so blank the value  */
	    *f_ptr = TOOBIG;
	    ++blank_count_local;
	}
    }
    *blank_count += blank_count_local;
    return (TRUE);
}   /*  End Function _foreign_miriad_read_read_data  */

static HeaderItem _foreign_miriad_read_split_binary_item (HeaderItem item)
/*  [SUMMARY] Split unstructured binary items.
    [PURPOSE] Split Miriad's stupid unstructured (application-specific) binary
    data into multiple typed items.
    <item> The item to split.
    [RETURNS] The list of split items. This is in reverse order.
*/
{
    HeaderItem first;
    /*static char function_name[] = "_foreign_miriad_read_split_binary_item";*/

    if (strcmp (item->name, "FREQS") == 0)
    {
	signed long lv;
	unsigned int types[3] = {K_DOUBLE, K_DOUBLE, K_INT};
	CONST char *names[3] = {"FREQS/INTERVAL", "FREQS/BASE","FREQS/NSCHAN"};

	first = _foreign_miriad_read_alloc_items (3, names, types);
	p_read_buf32s (item->data + 4, &lv);
	*(int *) first->next->next->data = lv;
	p_read_buf_double (item->data + 12, (double *) first->next->data);
	p_read_buf_double (item->data + 20, (double *) first->data);
	_foreign_miriad_read_free_header (item);
	return (first);
    }
    else return (item);
}   /*  End Function _foreign_miriad_read_split_binary_item  */

static HeaderItem _foreign_miriad_read_alloc_items (unsigned int num,
						    CONST char **names,
						    CONST unsigned int *types)
/*  [SUMMARY] Create a list of items.
    <num> The number of items to allocate.
    <names> The names of the items.
    <types> The types of the items.
    [RETURNS] The the first item.
*/
{
    unsigned int count;
    HeaderItem first = NULL;
    HeaderItem last = NULL;
    HeaderItem new;
    extern char host_type_sizes[NUMTYPES];
    static char function_name[] = "_foreign_miriad_read_alloc_items";

    for (count = 0; count < num; ++count)
    {
	if ( ( new = m_calloc (sizeof *new) ) == NULL )
	    m_abort (function_name, "item");
	new->type = types[count];
	if ( ( new->name = st_dup (names[count]) ) == NULL )
	    m_abort (function_name, "item name");
	if ( ( new->data = m_alloc (host_type_sizes[new->type]) ) == NULL )
	    m_abort (function_name, "item data");
	new->num_values = 1;
	new->buf_len = 1;
	KLIST_APPEND2 (first, last, new);
    }
    return (first);
}   /*  End Function _foreign_miriad_read_alloc_items  */

static VarTable _foreign_miriad_read_vartable (CONST char *dir)
/*  [SUMMARY] Read a vartable.
    <dir> The directory.
    [RETURNS] The VarTable on success, else NULL.
*/
{
    Channel channel;
    VarTable table;
    VartableItem vt_item;
    int len;
    unsigned int count;
    uaddr nread;
    char buffer[BUF_SIZE];
    char filename[STRING_LENGTH];
    static char function_name[] = "_foreign_miriad_read_vartable";

    sprintf (filename, "%s/vartable", dir);
    if ( ( channel = ch_open_compressed_file (filename) ) == NULL )
	return (NULL);
    if ( ( table = m_alloc (sizeof *table) ) == NULL )
	m_abort (function_name, "vartable header");
    table->num_items = 0;
    table->indices.polarisation = -1;
    table->indices.corr = -1;
    table->indices.time = -1;
    table->indices.baseline = -1;
    table->indices.coord = -1;
    /*  First count number of newlines (number of entries)  */
    while ( ( nread = ch_read (channel, buffer, BUF_SIZE) ) > 0 )
    {
	for (count = 0; count < nread; ++count)
	    if (buffer[count] == '\n') ++table->num_items;
    }
    ch_seek (channel, 0);
    if ( ( table->items =
	   m_alloc (sizeof vt_item * table->num_items) ) == NULL )
	m_abort (function_name, "vartable item pointer array");
    for (count = 0; count < table->num_items; ++count)
    {
	if ( !ch_getl (channel, buffer, BUF_SIZE) )
	{
	    fprintf (stderr, "%s: error reading line\t%s\n",
		     function_name, ERRSTRING);
	    _foreign_miriad_read_free_vartable (table);
	    ch_close (channel);
	    return (NULL);
	}
	len = strlen (buffer + 2);
	if ( ( vt_item = m_calloc (sizeof *vt_item + len + 8) ) == NULL )
	{
	    m_error_notify (function_name, "vt_item");
	    _foreign_miriad_read_free_vartable (table);
	    ch_close (channel);
	    return (NULL);
	}
	table->items[count] = vt_item;
	st_upr (buffer + 2);
	strcpy (vt_item->name, "VISDATA:");
	strcat (vt_item->name, buffer + 2);
	switch (buffer[0])
	{
	  case 'r':
	    vt_item->type = K_FLOAT;
	    break;
	  case 'i':
	    vt_item->type = K_INT;
	    break;
	  case 'a':
	    vt_item->type = K_VSTRING;
	    break;
	  case 'd':
	    vt_item->type = K_DOUBLE;
	    break;
	  default:
	    fprintf (stderr, "%s: illegal type: %c\n",
		     function_name, buffer[0]);
	    _foreign_miriad_read_free_vartable (table);
	    ch_close (channel);
	    return (NULL);
	    /*break;*/
	}
	if (strcmp (buffer + 2, "CORR") == 0) table->indices.corr = count;
	else if (strcmp (buffer + 2, "TIME") == 0) table->indices.time = count;
	else if (strcmp (buffer + 2, "BASELINE") == 0)
	    table->indices.baseline = count;
	else if (strcmp (buffer + 2, "POL") == 0)
	    table->indices.polarisation = count;
	else if (strcmp (buffer + 2, "COORD") == 0)
	    table->indices.coord = count;
    }
    ch_close (channel);
    if (table->indices.corr < 0)
	fprintf (stderr, "%s: no corr item\n", function_name);
    else if (table->indices.time < 0)
	fprintf (stderr, "%s: no time item\n", function_name);
    else if (table->indices.baseline < 0)
	fprintf (stderr, "%s: no baseline item\n", function_name);
    else if (table->indices.polarisation < 0)
	fprintf (stderr, "%s: no polarisation item\n", function_name);
    else if (table->indices.coord < 0)
	fprintf (stderr, "%s: no coord item\n", function_name);
    else if (table->items[table->indices.corr]->type != K_FLOAT)
	fprintf (stderr, "%s: bad type: %u for corr item\n",
		 function_name, table->items[table->indices.corr]->type);
    else if (table->items[table->indices.time]->type != K_DOUBLE)
	fprintf (stderr, "%s: bad type: %u for time item\n",
		 function_name, table->items[table->indices.time]->type);
    else if (table->items[table->indices.baseline]->type != K_FLOAT)
	fprintf (stderr, "%s: bad type: %u for baseline item\n",
		 function_name, table->items[table->indices.baseline]->type);
    else if (table->items[table->indices.polarisation]->type != K_INT)
	fprintf (stderr, "%s: bad type: %u for polarisation item\n",
		 function_name,
		 table->items[table->indices.polarisation]->type);
    else if (table->items[table->indices.coord]->type != K_DOUBLE)
	fprintf (stderr, "%s: bad type: %u for coord item\n",
		 function_name, table->items[table->indices.coord]->type);
    else
    {   /*  Success case  */
	return (table);
    }
    /*  Common failure path  */
    _foreign_miriad_read_free_vartable (table);
    return (NULL);
}   /*  End Function _foreign_miriad_read_vartable  */

static void _foreign_miriad_read_free_vartable (VarTable vt)
/*  [SUMMARY] Deallocate a vartable.
    <vt> The VarTable.
    [RETURNS] Nothing.
*/
{
    unsigned int count;

    if (vt == NULL) return;
    for (count = 0; count < vt->num_items; ++count)
	if (vt->items[count]) m_free (vt->items[count]);
    m_free (vt);
}   /*  End Function _foreign_miriad_read_free_vartable  */

static flag _foreign_miriad_read_visdata_header (CONST char *dirname,
						 Header header)
/*  [SUMMARY] Read Visibility header data.
    <dirname> The Miriad dataset name.
    <header> The header. This is modified.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    Channel channel;
    VarTable vartable = header->vartable;
    flag scan_all = FALSE;
    unsigned int count;
    struct stat statbuf;
    time_t visdata_mtime, auxilary_mtime;
    char visdata_filename[STRING_LENGTH], auxilary_filename[STRING_LENGTH];
    static char function_name[] = "_foreign_miriad_read_visdata_header";

    if (vartable == NULL) return (TRUE);
    sprintf (visdata_filename, "%s/visdata", dirname);
    sprintf (auxilary_filename, "%s/pol+time+bl", dirname);
    if (stat (visdata_filename, &statbuf) == 0)
	visdata_mtime = statbuf.st_mtime;
    else visdata_mtime = 0;
    if (stat (auxilary_filename, &statbuf) == 0)
	auxilary_mtime = statbuf.st_mtime;
    else auxilary_mtime = 0;
    if (auxilary_mtime > visdata_mtime)
    {
	if ( ( channel = ch_open_file (auxilary_filename, "r") ) == NULL )
	    return (FALSE);
	if ( !_foreign_miriad_read_sized_array (channel,
						&header->polarisations) ||
	     !_foreign_miriad_read_sized_array (channel,
						&header->times) ||
	     !_foreign_miriad_read_sized_array (channel,
						&header->baselines) )
	{
	    ch_close (channel);
	    return (FALSE);
	}
	ch_close (channel);
    }
    if (header->times.data == NULL)
    {
	fprintf (stderr,
		 "Time and baseline data missing, scanning entire file...\n");
	if ( ( header->polarisations.data =
	       m_alloc (sizeof *header->polarisations.data * 4) ) == NULL )
	    m_abort (function_name, "polarisation list");
	header->polarisations.buflen = 4;
	if ( ( header->times.data =
	       m_alloc (sizeof *header->times.data * 256) ) == NULL )
	    m_abort (function_name, "time list");
	header->times.buflen = 256;
	if ( ( header->baselines.data =
	       m_alloc (sizeof *header->baselines.data * 512) ) == NULL )
	    m_abort (function_name, "baseline list");
	header->baselines.buflen = 512;
	scan_all = TRUE;
    }
    channel = ch_map_disc (visdata_filename, K_CH_MAP_IF_AVAILABLE, FALSE,
			   FALSE);
    if (channel == NULL) channel = ch_open_file (visdata_filename, "r");
    if (channel == NULL) return (FALSE);
    /*  Read visdata file  */
    while (TRUE)
    {
	flag eof;
	int index;
	unsigned int entry_type;
	double value;
	double *ptr;
	SizedArray sarray;

	if ( !_foreign_miriad_read_visdata_chunk (channel, vartable, &eof,
						  &entry_type, &index,
						  vartable->indices.corr) )
	{
	    ch_close (channel);
	    return (FALSE);
	}
	if (eof) break;
	if ( (entry_type == VISDATA_ENTRY_TYPE_END_OF_RECORD) && !scan_all )
	    break;
	if (entry_type != VISDATA_ENTRY_TYPE_DATA) continue;
	/*  If pol, time or baseline entry, append value if not yet listed  */
	if (index == vartable->indices.polarisation)
	{
	    value = *(int *) vartable->items[index]->header_item->data;
	    sarray = &header->polarisations;
	}
	else if (index == vartable->indices.time)
	{
	    value = *(double *) vartable->items[index]->header_item->data;
	    sarray = &header->times;
	}
	else if (index == vartable->indices.baseline)
	{
	    value = *(float *) vartable->items[index]->header_item->data;
	    sarray = &header->baselines;
	}
	else continue;
	for (count = 0; count < sarray->num_values; ++count)
	    if (value == sarray->data[count]) break;
	if (count >= sarray->num_values)
	{
	    if (sarray->num_values >= sarray->buflen)
	    {
		if ( ( ptr = m_alloc (sizeof *ptr * sarray->buflen * 2) )
		     == NULL )
		    m_abort (function_name, "expanded list");
		m_copy (ptr, sarray->data, sizeof *ptr * sarray->num_values);
		m_free (sarray->data);
		sarray->data = ptr;
		sarray->buflen *= 2;
	    }
	    sarray->data[count] = value;
	    ++sarray->num_values;
	}
    }
    ch_close (channel);
    if (scan_all)
    {
	/*  Construct and append timelist header entry  */
	if ( ( channel = ch_open_file (auxilary_filename, "w") ) == NULL )
	{
	    fprintf (stderr, "Error opening: \"%s\" for write\t%s\n",
		     auxilary_filename, ERRSTRING);
	    fprintf (stderr, "Header will not be updated\n");
	}
	else
	{
	    _foreign_miriad_read_write_sized_array (channel,
						    &header->polarisations);
	    _foreign_miriad_read_write_sized_array (channel, &header->times);
	    _foreign_miriad_read_write_sized_array (channel,
						    &header->baselines);
	    ch_close (channel);
	}
    }
    for (count = 0; count < vartable->num_items; ++count)
    {
	VartableItem vt_item = vartable->items[count];
	HeaderItem item = vt_item->header_item;

	if (!item) continue;
	/*  Ignore multiple items, but still accept items duplicated once  */
	if (vt_item->count > 2) continue;
	KLIST_APPEND (header, item);
    }
    return (TRUE);
}   /*  End Function _foreign_miriad_read_visdata_header  */

static flag _foreign_miriad_read_sized_array (Channel channel,
					      SizedArray array)
/*  [SUMMARY] Read a sized array.
    <channel> The channel to write to.
    <array> The array. This is modified.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    static char function_name[] = "_foreign_miriad_read_sized_array";

    if ( !pio_read32 (channel, &array->num_values) ) return (FALSE);
    if (array->num_values > array->buflen)
    {
	if (array->data) m_free (array->data);
	array->data = NULL;
	if ( ( array->data =
	       m_alloc (sizeof *array->data * array->num_values) ) == NULL )
	{
	    m_error_notify (function_name, "array data");
	    m_clear (array, sizeof *array);
	    return (FALSE);
	}
    }
    if (pio_read_doubles (channel, array->num_values, array->data, NULL)
	< array->num_values)
    {
	array->num_values = 0;
	return (FALSE);
    }
    return (TRUE);
}   /*  End Function _foreign_miriad_read_sized_array  */

static flag _foreign_miriad_read_write_sized_array (Channel channel,
						    SizedArray array)
/*  [SUMMARY] Write a sized array.
    <channel> The channel to write to.
    <array> The array to write.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    unsigned int count;

    if ( !pio_write32 (channel, array->num_values) ) return (FALSE);
    for (count = 0; count < array->num_values; ++count)
	if ( !pio_write_double (channel, array->data[count]) ) return (FALSE);
    return (TRUE);
}   /*  End Function _foreign_miriad_read_write_sized_array  */

static flag _foreign_miriad_read_visdata_chunk (Channel channel,
						VarTable vartable,
						flag *end_of_file,
						unsigned int *entry_type,
						int *index, int skip_index)
/*  [SUMMARY] Read a chunk (entry) in the Visibility data.
    <channel> The channel to read from.
    <vartable> The vartable.
    <end_of_file> If the end of the file is reached, TRUE is written here, else
    FALSE is written here. This may be NULL.
    <entry_type> The type of the entry will be written here. This may be NULL.
    <index> The vartable index of the chuck read will be written here. This may
    be NULL.
    <skip_index> The index of the variable to skip (not read). If this is less
    than 0, no variable is skipped.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    HeaderItem item;
    VartableItem vt_item;
    unsigned int etype = 10000;
    unsigned char var_index;
    char header[4];
    static char function_name[] = "_foreign_miriad_read_visdata_chunk";

    if (ch_read (channel, header, 4) < 4)
    {
	if (end_of_file) *end_of_file = TRUE;
	return (TRUE);
    }
    var_index = header[0];
    if (var_index >= vartable->num_items)
    {
	fprintf (stderr, "%s: illegal variable index: %u (max: %u)\n",
		 function_name, var_index, vartable->num_items - 1);
	return (FALSE);
    }
    vt_item = vartable->items[var_index];
    if (header[2] == 0)
    {   /*  Set the length for a value  */
	unsigned long length;

	etype = VISDATA_ENTRY_TYPE_DESCRIPTOR;
	if ( !pio_read32 (channel, &length) )
	{
	    fprintf (stderr, "%s: error reading length for: \"%s\"\t%s\n",
		     function_name, vt_item->name, ERRSTRING);
	    return (FALSE);
	}
	vt_item->length = length;
	if (vt_item->header_item == NULL)
	{
	    if ( ( item = m_calloc (sizeof *item) ) == NULL )
		m_abort (function_name, "item");
	    if ( ( item->name = st_dup (vt_item->name) ) == NULL )
		m_abort (function_name, "item name");
	    vt_item->header_item = item;
	    item->type = vt_item->type;
	}
	else
	{
	    item = vt_item->header_item;
	    if (item->type != vt_item->type)
	    {
		fprintf (stderr,
			 "%s: item: \"%s\" type changed from: %u to %u\n",
			 function_name, vt_item->name, vt_item->type,
			 item->type);
		return (FALSE);
	    }
	}
    }
    else if (header[2] == 1)
    {   /*  Data for a value starts  */
	unsigned char pad;
	extern char network_type_bytes[NUMTYPES];
	extern char *data_type_names[NUMTYPES];

	etype = VISDATA_ENTRY_TYPE_DATA;
	if ( (item = vt_item->header_item) == NULL )
	{
	    fprintf (stderr, "%s: header item for: \"%s\" missing\n",
		     function_name, vt_item->name);
	    return (FALSE);
	}
	if (item->type == K_VSTRING) pad = 0;
	else pad = foreign_miriad_read_get_alignment_padding
		 (4, network_type_bytes[item->type]);
	if (skip_index == (int) var_index)
	{
	    if ( !ch_skip (channel, vt_item->length + pad) ) return (FALSE);
	}
	else
	{
	    if ( !_foreign_miriad_read_header_value (channel, item,
						     vt_item->length + pad) )
		return (FALSE);
	}
	++vt_item->count;
	if (debug)
	{
	    fprintf (stderr, "Vitem[%d] \"%s\"\ttype: %s\tvalue:\t",
		     var_index, item->name, data_type_names[item->type]);
	    if (item->type == NONE) fputs("unstructured binary data\n",stderr);
	    else if (!item->data)
		fprintf (stderr, "array of %u bytes\n", vt_item->length);
	    else dmp_element (stderr, item->type, NULL, item->data, FALSE);
	}
    }
    else if (header[2] == 2)
    {
	etype = VISDATA_ENTRY_TYPE_END_OF_RECORD;
	if (debug) fprintf (stderr, "EndOfRecord for: %d ?\n", header[0]);
    }
    else
    {
	fprintf (stderr, "Bad flag: %d\n", header[2]);
	return (FALSE);
    }
    ch_skip_to_boundary (channel, 8);
    if (end_of_file) *end_of_file = FALSE;
    if (entry_type) *entry_type = etype;
    if (index) *index = var_index;
    return (TRUE);
}   /*  End Function _foreign_miriad_read_visdata_chunk  */

struct searchable_dim_struct
{
    dim_desc *dim;
    uaddr offset;
    unsigned int pos;
    unsigned int dim_index;
};

#define INIT_SDIM(sdim, index) \
{ \
    (sdim).dim = vtable->vis.desc->dimensions[(index)]; \
    (sdim).offset = 0; \
    (sdim).pos = 0; \
    (sdim).dim_index = (index); \
}

static flag _foreign_miriad_read_visdata_data (CONST char *dirname,
					       Header header,
					       multi_array *multi_desc)
/*  [SUMMARY] Read Visibility data.
    <dirname> The Miriad dataset name.
    <header> The header.
    <multi_desc> The multi_array descriptor containing the header. The data
    space will be allocated.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    Channel channel;
    Channel flags_channel;
    uaddr uvw_time_off = 0;
    uaddr uvw_baseline_off = 0;
    VarTable vtable = header->vartable;
    signed int mask_index = 3;
    unsigned char mask = 0x01;
    unsigned int count, time_pos;
    double toobig = TOOBIG;
    struct searchable_dim_struct sdim_pol, sdim_time, sdim_bl;
    packet_desc *top_pack_desc = multi_desc->headers[0];
    dim_desc *dim, *dim_time, *dim_bl;
    char *top_packet = multi_desc->data[0];
    unsigned char *flags_buffer;
    unsigned int *u_ptr;
    struct array_struct *arrays[NUM_VISDATA_ARRAYS];
    char filename[STRING_LENGTH];
    static char function_name[] = "_foreign_miriad_read_visdata_data";

    /*  Allocate and initialise data for arrays  */
    arrays[0] = &vtable->vis;
    arrays[1] = &vtable->ant;
    arrays[2] = &vtable->uvw;
    for (count = 0; count < top_pack_desc->num_elements; ++count)
    {
	array_pointer arrayp;
	char *element;
	array_desc *arr_desc;

	if (top_pack_desc->element_types[count] != K_ARRAY) continue;
	arr_desc = (array_desc *) top_pack_desc->element_desc[count];
	if ( !ds_compute_array_offsets (arr_desc) )
	    m_abort (function_name, "offset arrays");
	element = top_packet + ds_get_element_offset (top_pack_desc, count);
	if ( !ds_alloc_array (arr_desc, element, FALSE, TRUE) )
	{
	    m_error_notify (function_name, "data");
	    return (FALSE);
	}
	m_copy (&arrayp, element, sizeof arrayp);
	arrays[count]->data = arrayp.array;
	arrays[count]->desc = arr_desc;
    }
    /*  Fill in BL -> antenna pair array  */
    dim = vtable->vis.desc->dimensions[VISDATA_DIM_INDEX_BASELINE];
    u_ptr = (unsigned int *) vtable->ant.data;
    for (count = 0; count < dim->length; ++count, u_ptr += 2)
    {   /*  Code derived from busant.for in Miriad sources  */
	long ant1, ant2, mant;

	ant2 = dim->coordinates[count];
	if (ant2 > 65536)
	{
	    ant2 -= 65536;
	    mant = 2048;
	}
	else mant = 256;
	ant1 = ant2 / mant;
	ant2 = ant2 - ant1 * mant;
	u_ptr[0] = ant1;
	u_ptr[1] = ant2;
    }
    /*  Initialise UVW table with blanks, so that missing data can easily be
	checked. This saves having to initialise the main array  */
    m_fill (vtable->uvw.data, sizeof (double), &toobig, sizeof (double),
	    ds_get_array_size (vtable->uvw.desc) * 3);
    sprintf (filename, "%s/visdata", dirname);
    channel = ch_map_disc (filename, K_CH_MAP_IF_AVAILABLE, FALSE, FALSE);
    if (channel == NULL) channel = ch_open_file (filename, "r");
    if (channel == NULL) return (FALSE);
    /*  Set up flags array  */
    sprintf (filename, "%s/flags", dirname);
    flags_channel = ch_map_disc2 (filename, K_CH_MAP_IF_AVAILABLE);
    if (flags_channel)
    {
	flags_buffer = (unsigned char *) ch_get_memory_addr (flags_channel) +4;
	ch_register_close_func (channel, ( void (*) () ) ch_close,
				flags_channel);
    }
    else flags_buffer = NULL;
    INIT_SDIM (sdim_pol, VISDATA_DIM_INDEX_POLARISATION);
    INIT_SDIM (sdim_time, VISDATA_DIM_INDEX_TIME);
    INIT_SDIM (sdim_bl, VISDATA_DIM_INDEX_BASELINE);
    /*  Read the data  */
    while (TRUE)
    {
	flag eof;
	flag found = FALSE;
	HeaderItem item;
	uaddr *offset;
	int var_index;
	unsigned int entry_type, dim_index, ncomplex;
	struct searchable_dim_struct *sdim;
	char *data;
	double value[2];

	if ( !_foreign_miriad_read_visdata_chunk (channel, vtable, &eof,
						  &entry_type,&var_index, -1) )
	{
	    ch_close (channel);
	    return (FALSE);
	}
	if (eof) break;
	item = vtable->items[var_index]->header_item;
	if (entry_type == VISDATA_ENTRY_TYPE_END_OF_RECORD)
	{   /*  Write correlation data to main array  */
	    array_desc *arr_desc = vtable->vis.desc;
	    uaddr *channel_offsets =
		arr_desc->offsets[VISDATA_DIM_INDEX_CHANNEL];
	    CONST float *source;

	    if (var_index != vtable->indices.corr) continue;
	    /*  Copy correlation data to output array  */
	    data = vtable->vis.data +
		sdim_pol.offset + sdim_time.offset + sdim_bl.offset;
	    source = (CONST float *) item->data;
	    ncomplex = item->num_values >> 1;
	    for (count = 0; count < ncomplex; ++count, source += 2)
	    {
		float *dest = (float *) (data + channel_offsets[count]);

		if ( flags_buffer && !(mask & flags_buffer[mask_index]) )
		{
		    dest[0] = TOOBIG;
		    dest[1] = TOOBIG;
		}
		else
		{
		    dest[0] = source[0];
		    dest[1] = source[1];
		}
		if (flags_buffer)
		{
		    mask = mask << 1;
		    if (!mask)
		    {
			mask = 0x01;
			--mask_index;
		    }
		    else if ( (mask == 0x80) && (mask_index == 0) )
		    {
			flags_buffer += 4;
			mask = 0x01;
			mask_index = 3;
		    }
		}
	    }
	    /*  Copy U,V,W co-ordinate data to output array  */
	    item = vtable->items[vtable->indices.coord]->header_item;
	    m_copy (vtable->uvw.data + uvw_time_off + uvw_baseline_off,
		    item->data, sizeof (double) * item->num_values);
	    continue;
	}
	if (entry_type != VISDATA_ENTRY_TYPE_DATA) continue;
	/*  Set up visibility array offsets  */
	if (var_index == vtable->indices.polarisation) sdim = &sdim_pol;
	else if (var_index == vtable->indices.time) sdim = &sdim_time;
	else if (var_index == vtable->indices.baseline) sdim = &sdim_bl;
	else continue;
	dim = sdim->dim;
	ds_get_element (item->data, item->type, value, NULL);
	count = sdim->pos;
	do
	{
	    if (dim->coordinates[count] == value[0])
	    {
		found = TRUE;
		break;
	    }
	    if (++count >= dim->length) count = 0;
	}
	while (count != sdim->pos);
	if (!found)
	{
	    fprintf (stderr,
		     "%s: co-ordinate: %.14g not found in dimension: \"%s\"\n",
		     function_name, value[0], dim->name);
	    ch_close (channel);
	    return (FALSE);
	}
	sdim->pos = count;
	sdim->offset = vtable->vis.desc->offsets[sdim->dim_index][count];
	/*  Set up Time,BL -> U,V,W array offsets. Note how <<count>> is still
	    valid  */
	if (var_index == vtable->indices.time)
	{
	    dim_index = VISUVW_DIM_INDEX_TIME;
	    offset = &uvw_time_off;
	}
	else if (var_index == vtable->indices.baseline)
	{
	    dim_index = VISUVW_DIM_INDEX_BASELINE;
	    offset = &uvw_baseline_off;
	}
	else continue;
	*offset = vtable->uvw.desc->offsets[dim_index][count];
    }
    ch_close (channel);
    /*  Now scan the Time,BL -> U,V,W array, applying blanks to corresponding
	pol,channel planes in the main array. This allows applications to
	process the main array without having to check the UVW table for blanks
	*/
    dim_time = vtable->uvw.desc->dimensions[VISUVW_DIM_INDEX_TIME];
    dim_bl = vtable->uvw.desc->dimensions[VISUVW_DIM_INDEX_BASELINE];
    for (time_pos = 0; time_pos < dim_time->length; ++time_pos)
    {
	unsigned int bl_pos;
	char *u_coord = vtable->uvw.data +
	    vtable->uvw.desc->offsets[VISUVW_DIM_INDEX_TIME][time_pos];

	for (bl_pos = 0; bl_pos < dim_bl->length; ++bl_pos)
	{
	    uaddr offset;
	    double u = *(double *)
		(u_coord +
		 vtable->uvw.desc->offsets[VISUVW_DIM_INDEX_BASELINE][bl_pos]);

	    if ( DOUBLE_IS_VALID (u) ) continue;
	    offset =
		vtable->vis.desc->offsets[VISDATA_DIM_INDEX_TIME][time_pos] +
		vtable->vis.desc->offsets[VISDATA_DIM_INDEX_BASELINE][bl_pos];
	    blank_plane (vtable->vis.desc, vtable->vis.data + offset,
			 VISDATA_DIM_INDEX_POLARISATION,
			 VISDATA_DIM_INDEX_CHANNEL);
	}
    }
    return (TRUE);
}   /*  End Function _foreign_miriad_read_visdata_data  */

static void blank_plane (CONST array_desc *arr_desc, char *data,
			 unsigned int y_dim_index, unsigned int x_dim_index)
/*  [SUMMARY] Fill a plane with blank values.
    <arr_desc> The array descriptor.
    <data> The start of the array.
    <y_dim_index> The Y dimension index.
    <x_dim_index> The X dimension index.
    [RETURNS] Nothing.
*/
{
    unsigned int y, x;
    uaddr *yoffsets = arr_desc->offsets[y_dim_index];
    uaddr *xoffsets = arr_desc->offsets[x_dim_index];
    dim_desc *ydim = arr_desc->dimensions[y_dim_index];
    dim_desc *xdim = arr_desc->dimensions[x_dim_index];

    for (y = 0; y < ydim->length; ++y)
    {
	uaddr yoffset = yoffsets[y];

	for (x = 0; x < xdim->length; ++x)
	{
	    float *f_ptr = (float *) (data + yoffset + xoffsets[x]);

	    f_ptr[0] = TOOBIG;
	    f_ptr[1] = TOOBIG;
	}
    }
}   /*  End Function blank_plane  */
