/*LINTLIBRARY*/
/*  matrix.c

    This code provides matrix operations for Intelligent Arrays.

    Copyright (C) 2003-2004  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This file contains matrix operations code for Intelligent Arrays.


    Written by      Richard Gooch   20-SEP-2003

    Last updated by Richard Gooch   28-JUN-2004: Use uaddr where needed.


*/
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <karma.h>
#include <karma_iarray.h>
#include <karma_mt.h>
#include <karma_a.h>
#include <karma_m.h>


#define MAGIC_NUMBER 939032982  /*  The other files depend on this too!  */

#define VERIFY_IARRAY(array) if (array == NULL) \
{fprintf (stderr, "NULL iarray passed\n"); a_prog_bug (function_name); } \
if (array->magic_number != MAGIC_NUMBER) \
{fprintf (stderr, "Invalid iarray\n"); a_prog_bug (function_name); }


/*  Private structures  */
struct struct_info
{
    iarray in1;
    iarray in2;
    iarray out;
};


/*  Private functions  */
STATIC_FUNCTION (void _iarray_matrix_multiply_job,
		 (void *pool_info, void *call_info1,
		  void *call_info2, void *call_info3,
		  void *call_info4, void *thread_info) );


/*  Public functions follow  */

/*EXPERIMENTAL_FUNCTION*/
flag iarray_matrix_multiply (iarray out, iarray in1, iarray in2, flag thread)
/*  [SUMMARY] Perform a matrix multiply.
    <out> The output array (matrix).
    <in1> The first input array.
    <in2> The second input array.
    <thread> If TRUE, the operation may be threaded, else it will not be.
    [NOTE] Only double precision arrays are supported.
    [MT-LEVEL] Unsafe.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    KThreadPool pool;
    unsigned int num_threads, clen, yblock, count;
    uaddr ylen;
    struct struct_info info;
    double *vec2;
    static char function_name[] = "iarray_matrix_multiply";

    VERIFY_IARRAY (out);
    VERIFY_IARRAY (in1);
    VERIFY_IARRAY (in2);
    if ( (iarray_type (in1) != K_DOUBLE) || (iarray_type (in2) != K_DOUBLE) ||
	 (iarray_type (out) != K_DOUBLE) )
    {
	fprintf (stderr,
		 "Only double precision floating-point arrays supported\n");
	return (FALSE);
    }
    if ( (iarray_num_dim (in1) != 2) || (iarray_num_dim (in2) != 2) ||
	 (iarray_num_dim (out) != 2) )
    {
	fprintf (stderr, "Only 2-dimensional arrays supported\n");
	return (FALSE);
    }
    if ( ( iarray_dim_length (in1, 1) != iarray_dim_length (in2, 0) )
	 || ( iarray_dim_length (out, 1) != iarray_dim_length (in2, 1) )
	 || ( iarray_dim_length (out,0) != iarray_dim_length (in1, 0) ) )
    {
	fprintf (stderr, "Size mismatch\n");
	return (FALSE);
    }
    clen = iarray_dim_length (in1, 1);
    ylen = iarray_dim_length (in1, 0);
    if (clen + iarray_dim_length (in2, 1) + ylen < 100)
    {
	thread = FALSE;
	pool = NULL;
	num_threads = 1;
    }
    else
    {
	pool = mt_get_shared_pool ();
	num_threads = mt_num_threads (pool);
    }
    info.in1 = in1;
    info.in2 = in2;
    info.out = out;
    if (ylen <= num_threads) num_threads = 1, thread = FALSE;
    vec2 = m_alloc_scratch (clen * sizeof *vec2 * num_threads, function_name);
    if (thread)
    {
	yblock = ylen / num_threads;
	for (count = 0; count < num_threads; ++count, vec2 += clen)
	{
	    uaddr ystart = count * yblock;
	    uaddr ystop = ystart + yblock;

	    if (count == num_threads - 1) ystop = ylen;
	    mt_launch_job (pool, _iarray_matrix_multiply_job, &info,
			   (void *) ystart, (void *) ystop, vec2);
	}
	mt_wait_for_all_jobs (pool);
    }
    else _iarray_matrix_multiply_job (NULL, &info,
				      (void *) 0, (void *) ylen, vec2,
				      NULL);
    m_free_scratch ();
    return (TRUE);
}   /*  End Function iarray_matrix_multiply  */


/*  Private functions follow  */

static void _iarray_matrix_multiply_job (void *pool_info, void *call_info1,
					 void *call_info2, void *call_info3,
					 void *call_info4, void *thread_info)
/*  [SUMMARY] Perform a job.
    <pool_info> The arbitrary pool information pointer.
    <call_info1> The first arbitrary call information pointer.
    <call_info2> The second arbitrary call information pointer.
    <call_info3> The third arbitrary call information pointer.
    <call_info4> The fourth arbitrary call information pointer.
    <thread_info> A pointer to arbitrary, per thread, information. This
    information is private to the thread.
    [RETURNS] Nothing.
*/
{
    iarray in1, in2, out;
    unsigned int clen, xlen, ccount, xcount, ycount;
    unsigned int ystart = (uaddr) call_info2;
    unsigned int ystop = (uaddr) call_info3;
    struct struct_info *info = call_info1;
    double *vec2 = call_info4;

    in1 = info->in1;
    in2 = info->in2;
    out = info->out;
    clen = iarray_dim_length (in1, 1);
    xlen = iarray_dim_length (in2, 1);
    for (xcount = 0; xcount < xlen; ++xcount)
    {
	for (ccount = 0; ccount < clen; ++ccount)
	    vec2[ccount] = D2 (in2, ccount, xcount);
	for (ycount = ystart; ycount < ystop; ++ycount)
	{
	    double sum = 0.0;

	    if (in1->contiguous[1])
	    {
		CONST double *vec1 = &D2 (in1, ycount, 0);

		for (ccount = 0; ccount < clen; ++ccount)
		    sum += *vec1++ * vec2[ccount];
		D2 (out, ycount, xcount) = sum;
		continue;
	    }
	    for (ccount = 0; ccount < clen; ++ccount)
		sum += D2 (in1, ycount, ccount) * vec2[ccount];
	    D2 (out, ycount, xcount) = sum;
	}
    }
}   /*  End Function _iarray_matrix_multiply_job  */
