/*LINTLIBRARY*/
/*  moment.c

    This code provides miscellaneous Intelligent Array moment routines.

    Copyright (C) 1996-2004  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This file contains all routines needed to compute various moments of
    n-dimensional data structures.


    Written by      Richard Gooch   20-JUN-2004: Moved code from
  source/karmawidgets/MomentGenerator.c

    Last updated by Richard Gooch   22-JUN-2004: Moved more code.


*/
#include <unistd.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <karma.h>
#include <karma_iarray.h>
#include <karma_wcs.h>
#include <karma_ds.h>
#include <karma_a.h>
#include <karma_m.h>


/*  Public data  */
char *iarray_moment_algorithm_alternatives[IARRAY_NUM_MOMENT_CHOICES] =
{
    "weighted mean",
    "median",
};


/*  Private data  */
static char *keywords[] =
{
    "OBJECT",
    "DATE-OBS",
    "OBSRA",
    "OBSDEC",
    "EPOCH",
    "TELESCOP",
    "INSTRUME",
    "EQUINOX",
    "BTYPE",
    "BMAJ",
    "BMIN",
    "BPA",
    "RESTFREQ",
    "VELREF",
    "PBFWHM",
    NULL
};


/*  Private functions  */
STATIC_FUNCTION (double _iarray_moment_get_aberration, (iarray cube) );
STATIC_FUNCTION (flag _iarray_moment_copy_header_info,
		 (iarray out, iarray in) );


/*  Public functions follow  */

/*EXPERIMENTAL_FUNCTION*/
flag iarray_moment_compute (iarray mom0, iarray mom1, iarray cube,
			    float lower_clip, float sum_clip,
			    unsigned int moment1_algorithm,
			    unsigned int start_z, unsigned int stop_z,
			    float *mom0_min, float *mom0_max)
/*  [SUMMARY] Compute the 0th and 1st moments along the Z axis of a cube.
    <mom0> The 0th moment array.
    <mom1> The 1st moment array.
    <cube> The cube.
    <lower_clip> Values in the cube lower than this value are not used in the
    computation of the moments.
    <sum_clip> Values in the 0th moment map lower than this value are not used
    in the computation of the 1st moment.
    <moment1_algorithm> The 1st moment algorithm.
    <start_z> The start channel.
    <stop_z> The stop channel.
    <mom0_min> The minimum value in the 0th moment image is written here.
    <mom0_max> The maximum value in the 0th moment image is written here.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    KwcsAstro cube_ap;
    int x, y, z, xlen, ylen, zlen;
    float val, weighted_sum, sum, half_mom0_val, index;
    float mom0_multiplier = 1.0;
    float mom0_val = 0;   /*  Initialised to keep compiler happy  */
    double mom1_val = 0;  /*  Initialised to keep compiler happy  */
    float mom1_min = TOOBIG;
    float mom1_max = -TOOBIG;
    CONST char *elem_name, *z_name;
    CONST char *names[3];
    char txt[STRING_LENGTH];
    extern char module_name[STRING_LENGTH + 1];
    extern char module_version_date[STRING_LENGTH + 1];
    extern char module_lib_version[STRING_LENGTH + 1];
    extern char karma_library_version[STRING_LENGTH + 1];
    /*static char function_name[] = "iarray_moment_compute";*/

    cube_ap = iarray_get_associated_wcs_astro (cube);
    xlen = iarray_dim_length (cube, 2);
    ylen = iarray_dim_length (cube, 1);
    zlen = iarray_dim_length (cube, 0);
    z_name = iarray_dim_name (cube, 0);
    elem_name = iarray_value_name (cube);
    if ( (strcmp (z_name, "FELO") == 0) || (strcmp (z_name, "VELO") == 0) )
    {
	double vel[2];

	sprintf (txt, "%s.M/S", elem_name);
	vel[0] = 0.0;
	vel[1] = 1.0;
	wcs_astro_transform (cube_ap, FALSE, 2,
			     NULL, FALSE, NULL, FALSE, vel, FALSE,
			     0, NULL, NULL);
	if (strcmp (iarray_value_name (mom0), txt) == 0)
	    mom0_multiplier = fabs (vel[1] - vel[0]);
    }
    *mom0_min = TOOBIG;
    *mom0_max = -TOOBIG;
    /*  Append a bit of history  */
    sprintf (txt, "%s: Module version %s  Karma v%s  compiled with v%s",
	     module_name, module_version_date,
	     karma_library_version, module_lib_version);
    iarray_append_history_string (mom0, txt, TRUE);
    iarray_append_history_string (mom1, txt, TRUE);
    sprintf (txt, "%s: 0th moment map", module_name);
    iarray_append_history_string (mom0, txt, TRUE);
    sprintf (txt, "%s: 1st moment map  low clip %e", module_name, lower_clip);
    iarray_append_history_string (mom1, txt, TRUE);
    sprintf (txt, "%s: sum_clip: %e  algorithm: %s",
	     module_name, sum_clip,
	     iarray_moment_algorithm_alternatives[moment1_algorithm]);
    iarray_append_history_string (mom1, txt, TRUE);
    names[0] = iarray_dim_name (cube, 0);
    names[1] = iarray_dim_name (cube, 1);
    names[2] = iarray_dim_name (cube, 2);
    /*  Loop through image pixels  */
    for (y = 0; y < ylen; ++y) for (x = 0; x < xlen; ++x)
    {
	switch (moment1_algorithm)
	{
	  case IARRAY_MOMENT1_ALGORITHM_WEIGHTED_MEAN:
	    mom0_val = 0.0;
	    weighted_sum = 0.0;
	    for (z = start_z; z < stop_z; ++z)
	    {
		if ( ( val = F3 (cube, z, y, x) ) < TOOBIG )
		{
		    if (val < lower_clip) continue;
		    mom0_val += val;
		    weighted_sum += val * (float) z;
		}
		else continue;
	    }
	    if (mom0_val < lower_clip)
	    {
		mom0_val = 0.0;
		mom1_val = TOOBIG;
		break;
	    }
	    if (mom0_val < sum_clip)
	    {
		mom1_val = TOOBIG;
		break;
	    }
	    /*  Now compute 1st moment  */
	    z = (weighted_sum / mom0_val);
	    if (z < 0) z = 0;
	    else if (z > zlen - 1) z = zlen - 1;
	    mom1_val = z;  /*  Co-ordinate index  */
	    break;
	  case IARRAY_MOMENT1_ALGORITHM_MEDIAN:
	    /*  First compute 0th moment  */
	    mom0_val = 0.0;
	    for (z = start_z; z < stop_z; ++z)
	    {
		if ( ( val = F3 (cube, z, y, x) ) < TOOBIG )
		{
		    if (val < lower_clip) continue;
		    mom0_val += val;
		}
		else continue;
	    }
	    if (mom0_val < lower_clip)
	    {
		mom0_val = 0.0;
		mom1_val = TOOBIG;
		break;
	    }
	    if (mom0_val < sum_clip)
	    {
		mom1_val = TOOBIG;
		break;
	    }
	    /*  Now compute 1st moment  */
	    half_mom0_val = mom0_val * 0.5;
	    sum = 0.0;
	    index = -1.0;
	    for (z = start_z; z < stop_z; ++z)
	    {
		if ( ( val = F3 (cube, z, y, x) ) < TOOBIG )
		{
		    if (val < lower_clip) continue;
		    sum += val;
		}
		else continue;
		if (sum >= half_mom0_val)
		{
		    index = (float) z + (half_mom0_val - sum + val) /val - 0.5;
		    break;
		}
	    }
	    if (index < 0.0) mom1_val = TOOBIG;
	    else mom1_val = index;
	    break;
	  default:
	    break;
	}
	mom0_val *= mom0_multiplier;
	F2 (mom0, y, x) = mom0_val;
	if (mom0_val < *mom0_min) *mom0_min = mom0_val;
	if (mom0_val > *mom0_max) *mom0_max = mom0_val;
	if ( DOUBLE_IS_VALID (mom1_val) )
	{
	    if (cube_ap == NULL)
		mom1_val = iarray_get_coordinate (cube, 0, mom1_val);
	    else
	    {
		double x_d = x;
		double y_d = y;

		wcs_astro_transform3 (cube_ap, FALSE, 1,
				      names[0], &mom1_val, FALSE,
				      names[1], &y_d, FALSE,
				      names[2], &x_d, FALSE,
				      0, NULL, NULL);
	    }
	    F2 (mom1, y, x) = mom1_val;
	    if (mom1_val < mom1_min) mom1_min = mom1_val;
	    if (mom1_val > mom1_max) mom1_max = mom1_val;
	}
	else F2 (mom1, y, x) = TOOBIG;
    }
    iarray_set_data_range (mom0, *mom0_min, *mom0_max);
    iarray_set_data_range (mom1, mom1_min, mom1_max);
    return (TRUE);
}   /*  End Function iarray_moment_compute  */

/*EXPERIMENTAL_FUNCTION*/
flag iarray_moment_create (iarray *mom0, iarray *mom1, iarray cube)
/*  [SUMMARY] Create 2-D arrays to contain computed moment maps.
    <mom0> The Intelligent Array for the 0th moment is written here.
    <mom1> The Intelligent Array for the 1st moment is written here.
    <cube> The cube from which the moment maps are created.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    unsigned int count, zlen;
    double aberration;
    CONST char *elem_name, *z_name;
    unsigned long dim_lengths[2];
    CONST char *dim_names[2];
    char txt[STRING_LENGTH];
    static char function_name[] = "iarray_moment_create";

    aberration = _iarray_moment_get_aberration (cube);
    if (aberration > 0.5)
    {
	fprintf (stderr,
		 "WARNING: %s: cube has significant chromatic aberration\n",
		 function_name);
	fprintf (stderr, "There is more than %e pixel offset due to this\n",
		 aberration);
    }
    z_name = iarray_dim_name (cube, 0);
    zlen = iarray_dim_length (cube, 0);
    dim_lengths[0] = iarray_dim_length (cube, 1);
    dim_lengths[1] = iarray_dim_length (cube, 2);
    dim_names[0] = iarray_dim_name (cube, 1);
    dim_names[1] = iarray_dim_name (cube, 2);
    elem_name = iarray_value_name (cube);
    /*  Create the 0th moment map array  */
    if ( (strcmp (z_name, "FELO") == 0) || (strcmp (z_name, "VELO") == 0) )
	sprintf (txt, "%s.M/S", elem_name);
    else strcpy (txt, elem_name);
    if ( ( *mom0 = iarray_create (K_FLOAT, 2, dim_names, dim_lengths, txt,
				  NULL) )
	 == NULL )
    {
	m_error_notify (function_name, "0th moment array");
	return (FALSE);
    }
    iarray_put_named_string (*mom0, "BUNIT", txt);
    /*  Copy over FITS co-ordinate information for the RA and DEC axes  */
    for (count = 0; count < 2; ++count)
    {
	unsigned int axis_num;
	double value[2];

	if ( ( axis_num = iarray_get_fits_axis (cube, count + 1) ) == 0 )
	    continue;
	/*  Set the "CTYPEn" keyword with the dimension name so that the <wcs>
	    package can deal with it  */
	sprintf (txt, "CTYPE%u", 2 - count);
	iarray_put_named_string ( *mom0, txt,
				  iarray_dim_name (cube, count + 1) );
	sprintf (txt, "CRVAL%u", axis_num);
	if ( iarray_get_named_value (cube, txt, NULL, value) )
	{
	    sprintf (txt, "CRVAL%u", 2 - count);
	    iarray_put_named_value (*mom0, txt, K_DOUBLE, value);
	}
	sprintf (txt, "CRPIX%u", axis_num);
	if ( iarray_get_named_value (cube, txt, NULL, value) )
	{
	    sprintf (txt, "CRPIX%u", 2 - count);
	    iarray_put_named_value (*mom0, txt, K_DOUBLE, value);
	}
	sprintf (txt, "CDELT%u", axis_num);
	if ( iarray_get_named_value (cube, txt, NULL, value) )
	{
	    sprintf (txt, "CDELT%u", 2 - count);
	    iarray_put_named_value (*mom0, txt, K_DOUBLE, value);
	}
	sprintf (txt, "CROTA%u", axis_num);
	if ( iarray_get_named_value (cube, txt, NULL, value)
	     && (value[0] != 0.0) )
	{
	    sprintf (txt, "CROTA%u", 2 - count);
	    iarray_put_named_value (*mom0, txt, K_DOUBLE, value);
	}
    }
    /*  Copy over some other header information  */
    _iarray_moment_copy_header_info (*mom0, cube);
    /*  Create the 1st moment array  */
    /*  The value name should reflect the z-axis name  */
    if (strncmp (z_name, "FREQ", 4) == 0) elem_name = "HZ";
    else if ( (strncmp (z_name, "VELO", 4) == 0)
	      || (strncmp (z_name, "FELO", 4) == 0) )
	elem_name = "M/S";
    else elem_name = z_name;
    if ( ( *mom1 = iarray_create (K_FLOAT, 2, dim_names, dim_lengths,
				  elem_name, NULL) )
	 == NULL )
    {
	m_error_notify (function_name, "1st moment array");
	return (FALSE);
    }
    /*  Copy over FITS co-ordinate information for the RA and DEC axes  */
    for (count = 0; count < 2; ++count)
    {
	unsigned int axis_num;
	double value[2];

	if ( ( axis_num = iarray_get_fits_axis (cube, count + 1) ) == 0 )
	    continue;
	/*  Set the "CTYPEn" keyword with the dimension name so that the <wcs>
	    package can deal with it  */
	sprintf (txt, "CTYPE%u", 2 - count);
	iarray_put_named_string ( *mom1, txt,
				  iarray_dim_name (cube, count + 1) );
	sprintf (txt, "CRVAL%u", axis_num);
	if ( iarray_get_named_value (cube, txt, NULL, value) )
	{
	    sprintf (txt, "CRVAL%u", 2 - count);
	    iarray_put_named_value (*mom1, txt, K_DOUBLE, value);
	}
	sprintf (txt, "CRPIX%u", axis_num);
	if ( iarray_get_named_value (cube, txt, NULL, value) )
	{
	    sprintf (txt, "CRPIX%u", 2 - count);
	    iarray_put_named_value (*mom1, txt, K_DOUBLE, value);
	}
	sprintf (txt, "CDELT%u", axis_num);
	if ( iarray_get_named_value (cube, txt, NULL, value) )
	{
	    sprintf (txt, "CDELT%u", 2 - count);
	    iarray_put_named_value (*mom1, txt, K_DOUBLE, value);
	}
	sprintf (txt, "CROTA%u", axis_num);
	if ( iarray_get_named_value (cube, txt, NULL, value)
	     && (value[0] != 0.0) )
	{
	    sprintf (txt, "CROTA%u", 2 - count);
	    iarray_put_named_value (*mom1, txt, K_DOUBLE, value);
	}
    }
    iarray_put_named_string (*mom1, "BUNIT", elem_name);
    /*  Copy over some other header information  */
    _iarray_moment_copy_header_info (*mom1, cube);
    return (TRUE);
}   /*  End Function iarray_moment_create  */


/*  Private functions follow  */

static double _iarray_moment_get_aberration (iarray cube)
/*  [SUMMARY] Check for chromatic aberration.
    [PURPOSE] This routine will check for chromatic aberration in a cube. This
    aberretion is due to the changing wavelength in different channels in a
    cube. The RA/DEC cell size changes with wavelength. If the wavelength
    difference between different channels is sufficient, then a pixel in one
    channel is is displaced by more the 0.5 pixels in another channel. The
    further the pixel from the phase centre of the synthesised map the greater
    the pixel offset between channels.
    <cube> The datacube.
    [RETURNS] The pixel offset due to aberration.
*/
{
    KwcsAstro cube_ap;
    unsigned int count;
    dim_desc *xdim, *ydim, *zdim;
    double z;
    double aberration = 0.0;
    double px[2], py[2], wx[2], wy[2];

    cube_ap = iarray_get_associated_wcs_astro (cube);
    if (cube_ap == NULL) return (0.0);
    if ( !wcs_astro_has_chromatic_aberration (cube_ap) ) return (0.0);
    xdim = iarray_get_dim_desc (cube, 2);
    ydim = iarray_get_dim_desc (cube, 1);
    zdim = iarray_get_dim_desc (cube, 0);
    px[0] = 0.0;
    py[0] = 0.0;
    px[1] = xdim->length - 1;
    py[1] = ydim->length - 1;
    for (count = 0; count < 2; ++count)
    {
	/*  Convert to linear co-ordinates  */
	wx[count] = ds_get_coordinate (xdim, px[count]);
	wy[count] = ds_get_coordinate (ydim, py[count]);
	/*  Convert to world co-ordinates for the first channel  */
	z = zdim->first_coord;
	wcs_astro_transform3 (cube_ap, FALSE, 1,
			      xdim->name, wx + count, FALSE,
			      ydim->name, wy + count, FALSE,
			      zdim->name, &z, FALSE,
			      0, NULL, NULL);
	/*  Convert back to linear co-ordinates for the last channel  */
	z = zdim->last_coord;
	wcs_astro_transform3 (cube_ap, FALSE, 1,
			      xdim->name, wx + count, TRUE,
			      ydim->name, wy + count, TRUE,
			      zdim->name, &z, FALSE,
			      0, NULL, NULL);
	/*  Convert back to pixel co-ordinates  */
	ds_convert_coordinates (xdim, 1, wx + count, TRUE, FALSE, FALSE);
	ds_convert_coordinates (ydim, 1, wy + count, TRUE, FALSE, FALSE);
	/*  Now compare  */
	if (fabs (wx[count] - px[count]) > aberration)
	    aberration = fabs (wx[count] - px[count]);
	if (fabs (wy[count] - py[count]) > aberration)
	    aberration = fabs (wy[count] - py[count]);
    }
    return (aberration);
}   /*  End Function get_aberration  */

static flag _iarray_moment_copy_header_info (iarray out, iarray in)
/*  [SUMMARY] Copy some header information from one array to another.
    <out> The output array.
    <in> The input array.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    unsigned int count;
    static char function_name[] = "_iarray_moment_copy_header_info";

    for (count = 0; keywords[count]; ++count)
    {
	if ( !iarray_copy_named_element (out, in, keywords[count],
					 FALSE, FALSE, TRUE) )
	{
	    fprintf (stderr, "%s: Failed to copy header keyword %s\n",
		     function_name, keywords[count]);
	    return (FALSE);
	}
    }
    return (TRUE);
}   /*  End Function copy_header_info  */
