/*LINTLIBRARY*/
/*  scroll.c

    This code provides miscellaneous Intelligent Array scrolling routines.

    Copyright (C) 2004  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This file contains all routines needed for various scrolling of
    n-dimensional data structures.


    Written by      Richard Gooch   14-APR-2004

    Last updated by Richard Gooch   25-APR-2004: Worked around gcc 2.95.4
  optimiser bug.


*/
#include <stdio.h>
#include <math.h>
#include <karma.h>
#include <karma_iarray.h>
#include <karma_ds.h>
#include <karma_a.h>
#include <karma_m.h>


/*  Public functions follow  */

/*EXPERIMENTAL_FUNCTION*/
void iarray_scroll (iarray array, unsigned long num_to_scroll,
		    unsigned int scroll_mode)
/*  [SUMMARY] Scroll an Intelligent Array.
    [PURPOSE] This routine will scroll "down" an array by a specified number of
    co-ordinates along the major dimension (dimension 0).
    <array> The Intelligent Array to scroll.
    <num_to_scroll> The number of co-ordinates to scroll by.
    <scroll_mode> The scroll mode.
    [RETURNS] Nothing.
*/
{
    flag do_in_bulk;
    unsigned int count, elem_size;
    unsigned long scroll_elements, scroll_bytes, copy_bytes;
    double blank = TOOBIG;
    array_desc *arr_desc;
    dim_desc *major_dim;
    unsigned char *buffer = NULL;
    void *blank_ptr = &blank;  /*  Work around gcc 2.95.4 optimiser bug  */
    extern char host_type_sizes[NUMTYPES];
    static char function_name[] = "iarray_scroll";

    if ( !iarray_is_full_array (array, &do_in_bulk) ) do_in_bulk = FALSE;
    if (!do_in_bulk)
    {
	fprintf (stderr, "%s: need full, contiguous array\n", function_name);
	return;
    }
    arr_desc = iarray_array_desc (array);
    major_dim = arr_desc->dimensions[0];
    if (num_to_scroll >= major_dim->length) return;
    elem_size = host_type_sizes[iarray_type (array)];
    scroll_elements = 1;
    for (count = 1; count < arr_desc->num_dimensions; ++count)
	scroll_elements *= arr_desc->dimensions[count]->length;
    scroll_bytes = scroll_elements * elem_size;
    copy_bytes = (major_dim->length - num_to_scroll) * scroll_bytes;
    scroll_elements *= num_to_scroll;
    scroll_bytes *= num_to_scroll;
    if (scroll_mode == SCROLL_CIRCULAR)
    {   /*  Save for later  */
	buffer = m_alloc_scratch (scroll_bytes, function_name);
	m_copy (buffer, array->data, scroll_bytes);
    }
    m_copy (array->data, array->data + scroll_bytes, copy_bytes);
    switch (scroll_mode)
    {
      case SCROLL_FILL_WITH_BLANKS:
	switch ( iarray_type (array) )
	{
	  case K_FLOAT:
	    *(float *) blank_ptr = TOOBIG;
	    break;
	  case K_SHORT:
	    *(signed short *) blank_ptr = -32768;
	    break;
	  case K_INT:
	    *(signed int *) blank_ptr = -2147483647 - 1;
	    break;
	}
	m_fill (array->data + copy_bytes, elem_size, blank_ptr, elem_size,
		scroll_elements);
	break;
      case SCROLL_CIRCULAR:  /*  Copy saved data to the top  */
	m_copy (array->data + copy_bytes, buffer, scroll_bytes);
	break;
    }
    if (buffer) m_free_scratch (buffer);
}   /*  End Function iarray_scroll  */
