/*LINTLIBRARY*/
/*  zoom.c

    This code provides miscellaneous zooming routines for world canvases.

    Copyright (C) 2001-2002  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This file contains miscellaneous routines for world canvases.


    Written by      Richard Gooch   3-DEC-2001

    Last updated by Richard Gooch   3-JAN-2002: Bound mouse buttons 4 and 5 to
  zoom out/zoom in.


*/
#include <stdio.h>
#include <math.h>
#include <ctype.h>
#define NEW_WIN_SCALE
#include <karma.h>
#include <k_event_codes.h>
#include <karma_canvas.h>
#include <karma_kwin.h>
#include <karma_m.h>


/*  Private structures  */
struct canvas_and_limits
{
    KWorldCanvas canvas;
    CONST KWorldCanvasLimits *limits;
};


/*  Private functions  */
STATIC_FUNCTION (flag _canvas_zoom_position_func,
		 (KWorldCanvas canvas, double x, double y,
		  unsigned int event_code, void *e_info,
		  void **f_info, double x_lin, double y_lin,
		  int x_pix, int y_pix) );
STATIC_FUNCTION (flag _canvas_zoom_geom_func,
		 (KPixCanvasEditorContext context, KPixCanvas canvas,
		  unsigned int object_type,
		  unsigned int num_coords, CONST int *x, CONST int *y,
		  unsigned int raw_event_code,
		  void *raw_e_info, void **f_info) );


/*  Public functions follow  */

/*EXPERIMENTAL_FUNCTION*/
KCallbackFunc canvas_zoom_handle_keys (KWorldCanvas canvas,
				       CONST KWorldCanvasLimits *limits)
/*  [SUMMARY] Install callbacks to handle zoom key bindings for a world canvas.
    <canvas> The world canvas object.
    <limits> The canvas limits. This is referenced when the appropriate keys
    are pressed.
    [RETURNS] A KCallbackFunc object.
*/
{
    return canvas_register_position_event_func (canvas,
						_canvas_zoom_position_func,
						(void *) limits);
}   /*  End Function canvas_zoom_handle_keys  */

/*EXPERIMENTAL_FUNCTION*/
KPixCanvasEditorContext canvas_zoom_handle_mouse (KWorldCanvas canvas,
						  CONST KWorldCanvasLimits *limits)
/*  [SUMMARY] Install callbacks to handle left mouse zooming for world canvas.
    <canvas> The world canvas object.
    <limits> The canvas limits. This is referenced when the appropriate keys
    are pressed.
    [RETURNS] The KPixCanvasEditorContext created for rubber-banding. Extra key
    bindings may be added to this context.
*/
{
    KPixCanvas pixcanvas;
    KPixCanvasEditorContext ctx;
    unsigned int visual;
    unsigned long pixel_value;
    struct canvas_and_limits *info;
    static char function_name[] = "canvas_zoom_handle_mouse";

    pixcanvas = canvas_get_pixcanvas (canvas);
    kwin_get_attributes (pixcanvas,
			 KWIN_ATT_VISUAL, &visual,
			 KWIN_ATT_END);
    if (visual == KWIN_VISUAL_PSEUDOCOLOUR)
    {
	if ( !kwin_get_colour (pixcanvas, "green", &pixel_value,
			       NULL, NULL, NULL) )
	{
	    /*  Can't get green: try white instead  */
	    if ( !kwin_get_colour (pixcanvas, "white", &pixel_value,
				   NULL, NULL, NULL) )
	    {
		/*  Can't even get white, so just pick 0  */
		pixel_value = 0;
	    }
	}
    }
    else
    {
	kwin_get_attributes (pixcanvas,
			     KWIN_ATT_PIX_GREEN_MASK, &pixel_value,
			     KWIN_ATT_END);
    }
    if ( ( info = m_alloc (sizeof *info) ) == NULL )
	m_abort (function_name, "canvas and limits structure");
    info->canvas = canvas;
    info->limits = limits;
    /*  Create the dragzoom editor context  */
    ctx = kwin_editor_create_context (pixcanvas,
				      KWIN_EDITOR_CONTEXT_ATT_MOUSE,
				      K_CANVAS_EVENT_LEFT_MOUSE_MASK,
				      KWIN_EDITOR_CONTEXT_ATT_EDIT_OBJECT,
				      KWIN_EDIT_OBJECT_RECTANGLE,
				      KWIN_EDITOR_CONTEXT_ATT_PIXEL_VALUE,
				      pixel_value,
				      KWIN_EDITOR_CONTEXT_ATT_CLEAR_OBJECT,
				      FALSE,
				      KWIN_EDITOR_CONTEXT_ATT_END);
    if (ctx == NULL) m_abort (function_name, "editor context");
    kwin_editor_register_event_func (ctx, _canvas_zoom_geom_func, info);
    canvas_register_destroy_func (canvas, TRUE,
				  kwin_editor_destroy_context, ctx);
    canvas_register_destroy_func (canvas, FALSE, m_free, info);
    return (ctx);
}   /*  End Function canvas_zoom_handle_mouse  */

flag canvas_zoom (KWorldCanvas canvas, CONST KWorldCanvasLimits *limits,
		  char action, double x, double y)
/*  [SUMMARY] Zoom a canvas.
    <canvas> The canvas on which the event occurred.
    <limits> The canvas limits.
    <action> The action to take, such as: 'u', 'i', 'o' or 'p'.
    <x> The horizontal linear world co-ordinate.
    <y> The vertical linear world co-ordinate.
    [RETURNS] TRUE if the action is supported, else FALSE.
*/
{
    double tmp, xl, xr, yb, yt;
    CONST struct _worldcanvas_limit_struct *limits_to_apply = &limits->soft;

    if (action == 'u')
    {
	xl = limits->soft.left_x;
	xr = limits->soft.right_x;
	yb = limits->soft.bottom_y;
	yt = limits->soft.top_y;
    }
    else if (action == 'U')
    {
	limits_to_apply = &limits->hard;
	tmp = (limits->soft.right_x - limits->soft.left_x) * 0.025;
	xl = limits->soft.left_x - tmp;
	xr = limits->soft.right_x + tmp;
	tmp = (limits->soft.top_y - limits->soft.bottom_y) * 0.025;
	yb = limits->soft.bottom_y - tmp;
	yt = limits->soft.top_y + tmp;
    }
    else if ( (tolower (action) == 'i') || (tolower (action) == 'o') ||
	      (action == 'p') )
    {
	double scale = 0.0;  /*  Initialised to keep compiler happy  */

	if ( isupper (action) ) limits_to_apply = &limits->hard;
	canvas_get_attributes (canvas,
			       CANVAS_ATT_LEFT_X, &xl,
			       CANVAS_ATT_RIGHT_X, &xr,
			       CANVAS_ATT_BOTTOM_Y, &yb,
			       CANVAS_ATT_TOP_Y, &yt,
			       CANVAS_ATT_END);
	switch ( tolower (action) )
	{
	  case 'i':
	    scale = 0.5;
	    limits_to_apply = &limits->hard;  /*  For 'O' followed by 'i'  */
	    break;
	  case 'o':
	    scale = 2.0;
	    x = (xl + xr) * 0.5;
	    y = (yb + yt) * 0.5;
	    break;
	  case 'p':
	    scale = 1.0;
	    break;
	}
	tmp = (xr - xl) * scale;
	xl = x - tmp * 0.5;
	xr = xl + tmp;
	tmp = (yt - yb) * scale;
	yb = y - tmp * 0.5;
	yt = yb + tmp;
    }
    else return (FALSE);
    if ( DOUBLE_IS_VALID (limits_to_apply->left_x) &&
	 DOUBLE_IS_VALID (limits_to_apply->right_x) )
    {
	tmp = xr - xl;
	if (limits_to_apply->left_x < limits_to_apply->right_x)
	{
	    if (xl < limits_to_apply->left_x)
	    {
		xl = limits_to_apply->left_x;
		xr = xl + tmp;
	    }
	    if (xr > limits_to_apply->right_x)
	    {
		xr = limits_to_apply->right_x;
		xl = xr - tmp;
	    }
	    if (xl < limits_to_apply->left_x) xl = limits_to_apply->left_x;
	}
	else
	{
	    if (xl > limits_to_apply->left_x)
	    {
		xl = limits_to_apply->left_x;
		xr = xl + tmp;
	    }
	    if (xr < limits_to_apply->right_x)
	    {
		xr = limits_to_apply->right_x;
		xl = xr - tmp;
	    }
	    if (xl > limits_to_apply->left_x) xl = limits_to_apply->left_x;
	}
    }
    if ( DOUBLE_IS_VALID (limits_to_apply->bottom_y) &&
	 DOUBLE_IS_VALID (limits_to_apply->top_y) )
    {
	tmp = yt - yb;
	if (limits_to_apply->bottom_y < limits_to_apply->top_y)
	{
	    if (yb < limits_to_apply->bottom_y)
	    {
		yb = limits_to_apply->bottom_y;
		yt = yb + tmp;
	    }
	    if (yt > limits_to_apply->top_y)
	    {
		yt = limits_to_apply->top_y;
		yb = yt - tmp;
	    }
	    if (yb < limits_to_apply->bottom_y) yb = limits_to_apply->bottom_y;
	}
	else
	{
	    if (yb > limits_to_apply->bottom_y)
	    {
		yb = limits_to_apply->bottom_y;
		yt = yb + tmp;
	    }
	    if (yt < limits_to_apply->top_y)
	    {
		yt = limits_to_apply->top_y;
		yb = yt - tmp;
	    }
	    if (yb > limits_to_apply->bottom_y) yb = limits_to_apply->bottom_y;
	}
    }
    canvas_set_attributes (canvas,
			   CANVAS_ATT_LEFT_X, xl,
			   CANVAS_ATT_RIGHT_X, xr,
			   CANVAS_ATT_BOTTOM_Y, yb,
			   CANVAS_ATT_TOP_Y, yt,
			   CANVAS_ATT_END);
    kwin_refresh_if_visible (canvas_get_pixcanvas (canvas), TRUE);
    if ( (action == 'i') || (action == 'o') )
	canvas_warp_pointer (canvas,
			     (xl + xr) * 0.5, (yb + yt) * 0.5, TRUE, TRUE);
    return (TRUE);
}   /*  End Function canvas_zoom  */


/*  Private functions follow  */

static flag _canvas_zoom_position_func (KWorldCanvas canvas,
					double x, double y,
					unsigned int event_code, void *e_info,
					void **f_info,
					double x_lin, double y_lin,
					int x_pix, int y_pix)
/*  [SUMMARY] Position event callback.
    [PURPOSE] This routine is a position event consumer for a world canvas.
    <canvas> The canvas on which the event occurred.
    <x> The horizontal world co-ordinate of the event.
    <y> The vertical world co-ordinate of the event.
    <event_code> The arbitrary event code.
    <e_info> A pointer to arbitrary event information.
    <f_info> A pointer to an arbitrary function information pointer.
    <x_lin> The horizontal linear world co-ordinate prior to the transform
    function being called.
    <y_lin> The vertical linear world co-ordinate prior to the transform
    function being called.
    <x_pix> The horizontal pixel co-ordinate.
    <y_pix> The vertical pixel co-ordinate.
    [NOTE] The world co-ordinate values will have been transformed by the
    registered transform function (see [<canvas_register_transforms_func>]).
    [RETURNS] TRUE if the event was consumed, else FALSE indicating that
    the event is still to be processed.
*/
{
    CONST char *string = e_info;

    switch (event_code)
    {
      case K_CANVAS_EVENT_MIDDLE_MOUSE_RELEASE:
      case K_CANVAS_EVENT_MOUSE_BUTTON5_RELEASE:
	return canvas_zoom (canvas, *f_info, 'i', x_lin, y_lin);
	/*break;*/
      case K_CANVAS_EVENT_RIGHT_MOUSE_RELEASE:
      case K_CANVAS_EVENT_MOUSE_BUTTON4_RELEASE:
	return canvas_zoom (canvas, *f_info, 'o', x_lin, y_lin);
	/*break;*/
      case K_CANVAS_EVENT_PLAIN_KEY_PRESS:
	if (string[1] != '\0') return (FALSE);
	switch ( tolower (string[0]) )
	{
	  case 'u':
	  case 'i':
	  case 'o':
	    return canvas_zoom (canvas, *f_info, string[0], x_lin, y_lin);
	    /*break;*/
	}
	break;
    }
    return (FALSE);
}   /*  End Function _canvas_zoom_position_func  */

static flag _canvas_zoom_geom_func (KPixCanvasEditorContext context,
				    KPixCanvas canvas,
				    unsigned int object_type,
				    unsigned int num_coords,
				    CONST int *x, CONST int *y,
				    unsigned int raw_event_code,
				    void *raw_e_info, void **f_info)
/*  [SUMMARY] Process a geometry editor event on a pixel canvas.
    <context> The KPixCanvasEditorContext.
    <canvas> The pixel canvas on which the event occurred.
    <object_type> The type of the geometric object.
    <num_coords> The number of co-ordinates in the geometric object.
    <x> The array of horizontal co-ordinates for the geometric object.
    <y> The array of vertical co-ordinates for the geometric object.
    <raw_event_code> The raw arbitrary event code.
    <raw_e_info> A pointer to the raw arbitrary event information.
    <f_info> A pointer to an arbitrary function information pointer.
    [RETURNS] TRUE if the event was consumed, else FALSE indicating that
    the event is still to be processed.
*/
{
    struct canvas_and_limits *info = *f_info;
    double lx[2], ly[2];
    static char function_name[] = "_canvas_zoom_geom_func";

    if (raw_event_code != K_CANVAS_EVENT_LEFT_MOUSE_RELEASE) return (FALSE);
    lx[0] = x[0];
    ly[0] = y[0];
    lx[1] = x[1];
    ly[1] = y[1];
    canvas_convert_to_canvas_coords (info->canvas, TRUE, 2,
				     lx, ly, lx, ly, NULL, NULL);
    if ( (x[0] == x[1]) && (y[0] == y[1]) )
    {
	/*  Pan  */
	canvas_zoom (info->canvas, info->limits, 'p', lx[0], ly[0]);
	return (TRUE);
    }
    if ( (x[0] == x[1]) || (y[0] == y[1]) ) return (TRUE);
    if ( (lx[0] == lx[1]) || (ly[0] == ly[1]) ) return (TRUE);
    canvas_set_attributes (info->canvas,
			   CANVAS_ATT_LEFT_X, lx[0],
			   CANVAS_ATT_RIGHT_X, lx[1],
			   CANVAS_ATT_BOTTOM_Y, ly[1],
			   CANVAS_ATT_TOP_Y, ly[0],
			   CANVAS_ATT_END);
    if ( !kwin_resize (canvas, TRUE, 0, 0, 0, 0) )
	fprintf (stderr, "%s: Error refreshing window\n", function_name);
    return (TRUE);
}   /*  End Function _canvas_zoom_geom_func  */
