/*LINTLIBRARY*/
/*  array_main.c

    This code provides generic KConstructArrayManager objects.

    Copyright (C) 2004  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This file contains all routines needed for managing the construction of
    array data.


    Written by      Richard Gooch   14-MAR-2004

    Last updated by Richard Gooch   31-MAR-2004: Added add&subtract algorithm.


*/
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <karma.h>
#include <karma_construct.h>
#include <karma_dispdata.h>
#include <karma_listw.h>
#include <karma_m.h>
#include <karma_c.h>
#include <karma_a.h>


#define MAGIC_NUMBER         2679017514U

#define VERIFY_MANAGER(manager) {if (manager == NULL) \
{fprintf (stderr, "NULL manager passed\n"); \
 a_prog_bug (function_name); } \
if (manager->magic_number != MAGIC_NUMBER) \
{fprintf (stderr, "Invalid manager object\n"); \
 a_prog_bug (function_name); } }


struct construct_array_manager_type
{
    unsigned int magic_number;
    KListWidget listw;
    KDisplayDataClass dataclass;
    KCallbackFunc new_data_func;
    KCallbackFunc del_data_func;
    void (*change_name) (void *info, CONST char *name);
    void *change_name_info;
    /*  Algorithm hooks  */
    void *algorithm_info;
    void (*cleanup_func) (void *algorithm_info);
    flag (*position_func) (void *algorithm_info, unsigned int index,
			   unsigned int event_code, void *e_info);
    flag (*data_create_func) (void *algorithm_info,
			      KDisplayDataObject dataobject);
    flag (*data_destroy_func) (void *algorithm_info);
    void (*create) (void *algorithm_info, CONST char *name);
};


/*  Public data  */
char *construct_array_algorithm_choices[KCONSTRUCTARRAY_NUM_ALGORITHMS] =
{
    "RGB", "Arithmetic",
};


/*  External functions  */

/*  File: array_rgb.c  */
EXTERN_FUNCTION (void *construct_array_rgb_init,
		 (KConstructArrayManager manager, KListWidget list) );
EXTERN_FUNCTION (void construct_array_rgb_cleanup, (void *algorithm_info) );
EXTERN_FUNCTION (flag construct_array_rgb_position,
		 (void *algorithm_info, unsigned int index,
		  unsigned int event_code, void *e_info) );
EXTERN_FUNCTION (flag construct_array_rgb_data_create,
		 (void *algorithm_info, KDisplayDataObject dataobject) );
EXTERN_FUNCTION (flag construct_array_rgb_data_destroy,
		 (void *algorithm_info) );
EXTERN_FUNCTION (void construct_array_rgb_create,
		 (void *algorithm_info, CONST char *name) );

/*  File: array_addsub.c  */
EXTERN_FUNCTION (void *construct_array_addsub_init,
		 (KConstructArrayManager manager, KListWidget list) );
EXTERN_FUNCTION (void construct_array_addsub_cleanup, (void *algorithm_info) );
EXTERN_FUNCTION (flag construct_array_addsub_position,
		 (void *algorithm_info, unsigned int index,
		  unsigned int event_code, void *e_info) );
EXTERN_FUNCTION (flag construct_array_addsub_data_create,
		 (void *algorithm_info, KDisplayDataObject dataobject) );
EXTERN_FUNCTION (flag construct_array_addsub_data_destroy,
		 (void *algorithm_info) );
EXTERN_FUNCTION (void construct_array_addsub_create,
		 (void *algorithm_info, CONST char *name) );


/*  Private functions  */
STATIC_FUNCTION (void _construct_array_data_create_func,
		 (void *info, KDisplayDataObject dataobject) );
STATIC_FUNCTION (void _construct_array_data_destroy_func, (void *info) );
STATIC_FUNCTION (flag _construct_array_list_position_func,
		 (KListWidget list, unsigned int index,
		  unsigned int event_code, void *e_info, void **f_info) );


/*  Private data  */


/* Public functions follow */

/*EXPERIMENTAL_FUNCTION*/
KConstructArrayManager construct_array_create_manager (KListWidget listw,
						       void (*change_name) (),
						       void *info)
/*  [SUMMARY] Create a management object for creating array data.
    <listw> The list widget.
    <change_name> The function to call when the default name changes.
    <info> An arbitrary information pointer to pass to <<change_name>>.
    [RETURNS] A KConstructArrayManager object on success, else NULL.
*/
{
    KConstructArrayManager manager;
    static char function_name[] = "construct_array_create_manager";

    if ( ( manager = m_calloc (sizeof *manager) ) == NULL )
    {
	m_error_notify (function_name, "manager");
	return (NULL);
    }
    manager->magic_number = MAGIC_NUMBER;
    manager->listw = listw;
    manager->dataclass = dispdata_array_create_dataclass ();
    manager->new_data_func =
	dispdata_register_data_create_func (manager->dataclass,
					    _construct_array_data_create_func,
					    manager);
    manager->del_data_func =
	dispdata_register_data_destroy_func(manager->dataclass,
					    _construct_array_data_destroy_func,
					    manager);
    manager->change_name = change_name;
    manager->change_name_info = info;
    listw_register_position_event_func
	(listw,
	 ( flag (*) () ) _construct_array_list_position_func,
	 manager);
    listw_set_attributes (listw, FALSE,
			  LISTW_ATT_MOUSE_MODE, LISTW_MOUSE_MODE_HIGHLIGHT,
			  LISTW_ATT_END);
    return (manager);
}   /*  End Function construct_array_create_manager  */

/*EXPERIMENTAL_FUNCTION*/
void construct_array_destroy_manager (KConstructArrayManager manager)
/*  [SUMMARY] Destroy a management object for creating array data.
    <manager> The management object.
    [RETURNS] Nothing.
*/
{
    static char function_name[] = "construct_array_destroy_manager";

    VERIFY_MANAGER (manager);
    if (manager->cleanup_func)
	(*manager->cleanup_func) (manager->algorithm_info);
    c_unregister_callback2 (manager->new_data_func);
    c_unregister_callback2 (manager->del_data_func);
    manager->magic_number = 0;
    m_free (manager);
}   /*  End Function construct_array_destroy_manager  */

/*EXPERIMENTAL_FUNCTION*/
void construct_array_select_algorithm (KConstructArrayManager manager,
				       unsigned int index)
/*  [SUMMARY] Select an algorithm for constructing data.
    <manager> The management object.
    <index> The index into <<construct_array_algorithm_choices>>.
    [RETURNS] Nothing.
*/
{
    static char function_name[] = "construct_array_select_algorithm";

    VERIFY_MANAGER (manager);
    if (index >= KCONSTRUCTARRAY_NUM_ALGORITHMS)
    {
	fprintf (stderr, "Invalid index: %u\n", index);
	a_prog_bug (function_name);
    }
    if (manager->cleanup_func)
	(*manager->cleanup_func) (manager->algorithm_info);
    switch (index)
    {
      case KCONSTRUCTARRAY_ALGORITHM_RGB:
	manager->algorithm_info =
	    construct_array_rgb_init (manager, manager->listw);
	manager->cleanup_func = construct_array_rgb_cleanup;
	manager->position_func = construct_array_rgb_position;
	manager->data_create_func = construct_array_rgb_data_create;
	manager->data_destroy_func = construct_array_rgb_data_destroy;
	manager->create = construct_array_rgb_create;
	break;
      case KCONSTRUCTARRAY_ALGORITHM_ADDSUB:
	manager->algorithm_info =
	    construct_array_addsub_init (manager, manager->listw);
	manager->cleanup_func = construct_array_addsub_cleanup;
	manager->position_func = construct_array_addsub_position;
	manager->data_create_func = construct_array_addsub_data_create;
	manager->data_destroy_func = construct_array_addsub_data_destroy;
	manager->create = construct_array_addsub_create;
	break;
    }
}   /*  End Function construct_array_select_algorithm  */

/*EXPERIMENTAL_FUNCTION*/
void construct_array_change_name (KConstructArrayManager manager,
				  CONST char *name)
/*  [SUMMARY] Change the default output name for constructed data.
    <manager> The management object.
    <name> The new default name.
    [RETURNS] Nothing.
*/
{
    static char function_name[] = "construct_array_change_name";

    VERIFY_MANAGER (manager);
    (*manager->change_name) (manager->change_name_info, name);
}   /*  End Function construct_array_change_name  */

/*EXPERIMENTAL_FUNCTION*/
void construct_array_create_data (KConstructArrayManager manager,
				  CONST char *name)
/*  [SUMMARY] Construct an array.
    <manager> The management object.
    <name> The name to give the new array.
    [RETURNS] Nothing.
*/
{
    static char function_name[] = "construct_array_create_data";

    VERIFY_MANAGER (manager);
    (*manager->create) (manager->algorithm_info, name);
}   /*  End Function construct_array_create_data  */


/*  Private functions follow  */

static void _construct_array_data_create_func (void *info,
					       KDisplayDataObject dataobject)
{
    KConstructArrayManager manager = info;

    (*manager->data_create_func) (manager->algorithm_info, dataobject);
}   /*  End Function _construct_array_data_create_func  */

static void _construct_array_data_destroy_func (void *info)
{
    KConstructArrayManager manager = info;

    (*manager->data_destroy_func) (manager->algorithm_info);
}   /*  End Function _construct_array_data_create_func  */

static flag _construct_array_list_position_func (KListWidget list,
						 unsigned int index,
						 unsigned int event_code,
						 void *e_info, void **f_info)
{
    KConstructArrayManager manager = *f_info;

    return (*manager->position_func) (manager->algorithm_info, index,
				      event_code, e_info);
}   /*  End Function _construct_array_list_position_func  */
