/*LINTLIBRARY*/
/*  orthogonal.c

    This code supports orthogonal 1-D plotting facilities

    Copyright (C) 2002-2004  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This file contains all routines needed for plotting orthogonal profiles
    from n-D arrays.


    Written by      Richard Gooch   26-DEC-2002: Copied some code from
  TracePopup widget.

    Updated by      Richard Gooch   27-DEC-2002: Basic features completed.

    Updated by      Richard Gooch   28-DEC-2002: Removed <<min>> and <<max>>
  parameters from <graph1d_orthogonal_create> as they are not currently needed.

    Updated by      Richard Gooch   29-DEC-2002: Create empty KViewableGraph
  in <graph1d_orthogonal_create> so that colour and scale changes are recorded.

    Updated by      Richard Gooch   20-JAN-2003: Removed code which set canvas
  horizontal range in <graph1d_orthogonal_configure_canvas>.

    Updated by      Richard Gooch   1-JAN-2004: Minor code migration from
  <graph1d_orthogonal_save> to <graph1d_save_ascii>.

    Updated by      Richard Gooch   2-JAN-2004: Added fallback to last vgraph
  in <graph1d_orthogonal_save>.

    Updated by      Richard Gooch   6-MAY-2004: No longer enable attribute
  GRAPH1D_ VGRAPH_ATT_USE_COLOURMAP: upper levels should do it.

    Updated by      Richard Gooch   15-MAY-2004: Use world canvas horizonal
  range to control format unit name.

    Updated by      Richard Gooch   17-MAY-2004: Added support for data with
  chromatic aberration.

    Updated by      Richard Gooch   21-MAY-2004: Allow calling function to
  disable correction for chromatic aberration.

    Last updated by Richard Gooch   3-AUG-2004: Fixed data-dependent alignment
  problems in <graph1d_orthogonal_show_traces>.


*/
#include <unistd.h>
#include <stdio.h>
#include <math.h>
#include <stdarg.h>
#include <string.h>
#include <errno.h>
#include <karma.h>
#include <k_list.h>
#define NEW_WIN_SCALE
#include <karma_graph1d.h>
#include <karma_canvas.h>
#include <karma_iarray.h>
#include <karma_kwin.h>
#include <karma_wcs.h>
#include <karma_ds.h>
#include <karma_ch.h>
#include <karma_a.h>
#include <karma_m.h>
#include <karma_c.h>


#define MAGIC_NUMBER  (unsigned int)  2101234939
#define CANVAS_KEY    &_graph1d_orthogonal_canvas_profile_key

#define VERIFY_PROFILE(profile) if (profile == NULL) \
{fprintf (stderr, "NULL orthogonal profile passed\n"); \
 a_prog_bug (function_name); } \
if (profile->magic_number != MAGIC_NUMBER) \
{fprintf (stderr, "Invalid orthogonal profile object\n"); \
 a_prog_bug (function_name); }


struct orthogonal_profile_type
{
    unsigned int       magic_number;
    KWorldCanvas       canvas;
    KPixCanvas         pixcanvas;
    /*  Attributes  */
    /*  Input data  */
    iarray             array;
    /*  Other information  */
    KCallbackFunc      iarr_destroy_callback;
    KCallbackFunc      canvas_destroy_callback;
    KwcsAstro          ap;
    double             scale;
    double             offset;
    int                trace_dim;
    KViewableGraph     vgraph;           /*  Automatically NULLed on destroy */
    flag               valid;
    CONST char         *value_name;
    double             trace_minimum;          /*  Scaled                    */
    double             trace_maximum;          /*  Scaled                    */
    unsigned int       num_restr_allocated;
    CONST char         **restr_names;
    double             *restr_values;
    unsigned int       coord_indices_buf_len;
    uaddr              *coord_indices;
    unsigned int       coord_counters_buf_len;
    uaddr              *coord_counters;
    unsigned int       num_traces;             /*  For the <<save_cbk>>   */
    /*  Internal management  */
    KCallbackList      destroy_list;
};


/*  Private data  */
static char _graph1d_orthogonal_canvas_profile_key;          /*  Unique key  */


/*  Private functions  */
STATIC_FUNCTION (void _graph1d_orthogonal_iarr_destroy_callback,
		 (iarray arr, KOrthogonalProfile profile) );


/*  Public functions follow  */

/*EXPERIMENTAL_FUNCTION*/
KOrthogonalProfile graph1d_orthogonal_create (KWorldCanvas canvas,iarray array,
					      KViewableGraph *vgraph)
/*  [SUMMARY] Create an object for later viewing of orthogonal profiles.
    <canvas> The world canvas object where profiles will be drawn.
    <array> The Intelligent Array which contains the data.
    <vgraph> The KViewableGraph that is created is written here. This may be
    NULL.
    [NOTE] The KOrthogonalProfile will be automatically destroyed if the canvas
    or Intelligent Array are destroyed.
    [RETURNS] A KOrthogonalProfile object on success, else NULL.
*/
{
    KOrthogonalProfile profile;
    static char function_name[] = "graph1d_orthogonal_create";

    if (canvas == NULL)
    {
	fprintf (stderr, "NULL world canvas passed\n");
	a_prog_bug (function_name);
    }
    if ( ( profile = m_calloc (sizeof *profile) ) == NULL )
    {
	m_error_notify (function_name, "orthogonal profile");
	return (NULL);
    }
    profile->canvas = canvas;
    profile->pixcanvas = canvas_get_pixcanvas (canvas);
    profile->ap = NULL;
    profile->trace_dim = -1;
    profile->array = array;
    profile->iarr_destroy_callback = 
	iarray_register_destroy_func
	(profile->array,
	 ( flag (*) () ) _graph1d_orthogonal_iarr_destroy_callback,
	 profile);
    profile->ap = iarray_get_associated_wcs_astro (array);
    iarray_get_data_scaling (profile->array, &profile->scale,&profile->offset);
    profile->canvas_destroy_callback = 
	canvas_register_destroy_func (canvas, FALSE,
				      graph1d_orthogonal_destroy, profile);
    profile->vgraph = graph1d_create (profile->canvas, 0, NULL, NULL, 0,
				      GRAPH1D_VGRAPH_ATT_END);
    profile->magic_number = MAGIC_NUMBER;
    if (!profile->iarr_destroy_callback || !profile->canvas_destroy_callback ||
	!profile->vgraph ||
	!graph1d_register_destroy_func (profile->vgraph, c_write_null_func,
					&profile->vgraph) )
    {
	graph1d_orthogonal_destroy (profile);
	return (NULL);
    }
    if (vgraph) *vgraph = profile->vgraph;
    return (profile);
}   /*  End Function graph1d_orthogonal_create  */

/*EXPERIMENTAL_FUNCTION*/
void graph1d_orthogonal_destroy (KOrthogonalProfile profile)
/*  [SUMMARY] Destroy an orthogonal profile.
    <profile> The orthogonal profile.
    [RETURNS] Nothing.
*/
{
    KOrthogonalProfile master;
    static char function_name[] = "graph1d_orthogonal_destroy";

    VERIFY_PROFILE (profile);
    master = canvas_get_associated_object (profile->canvas, CANVAS_KEY);
    if (master == profile)
	canvas_associate_object (profile->canvas, CANVAS_KEY, NULL, TRUE);
    c_call_callbacks (profile->destroy_list, NULL);
    c_destroy_list (profile->destroy_list);
    c_unregister_callback (profile->iarr_destroy_callback);
    c_unregister_callback (profile->canvas_destroy_callback);
    if (profile->vgraph) graph1d_destroy (profile->vgraph);
    if (profile->restr_names) m_free (profile->restr_names);
    if (profile->restr_values) m_free(profile->restr_values);
    if (profile->coord_indices) m_free (profile->coord_indices);
    if (profile->coord_counters) m_free (profile->coord_counters);
    profile->magic_number = 0;
    m_free (profile);
}   /*  End Function graph1d_orthogonal_destroy  */

/*EXPERIMENTAL_FUNCTION*/
KCallbackFunc graph1d_orthogonal_register_destroy_func
    (KOrthogonalProfile profile, void (*func) (), void *info)
/*  [SUMMARY] Register destroy callback.
    [PURPOSE] This routine will register a destroy function for an orthogonal
    profile. Many destroy functions may be registered per profile. The first
    function registered is the first function called upon destruction.
    <profile> The orthogonal profile.
    <func> The function that is called when the profile is destroyed. The
    prototype function is [<graph1d_orthogonal_PROTO_destroy_func>].
    <info> The arbitrary function information pointer.
    [RETURNS] A KCallbackFunc object. On failure, the process aborts.
*/
{
    static char function_name[] = "graph1d_orthogonal_register_destroy_func";

    VERIFY_PROFILE (profile);
    return c_register_callback (&profile->destroy_list, ( flag (*) () ) func,
				info, NULL, FALSE, NULL, FALSE, FALSE);
}   /*  End Function graph1d_orthogonal_register_destroy_func  */

/*EXPERIMENTAL_FUNCTION*/
void graph1d_orthogonal_invalidate (KOrthogonalProfile profile)
/*  [SUMMARY] Invalidate an orthogonal profile.
    <profile> The orthogonal profile.
    [RETURNS] Nothing.
*/
{
    static char function_name[] = "graph1d_orthogonal_invalidate";

    VERIFY_PROFILE (profile);
    graph1d_set_active (profile->vgraph, FALSE, FALSE, FALSE, FALSE, 0);
    graph1d_set_size (profile->vgraph, 0);
    profile->valid = FALSE;
}   /*  End Function graph1d_orthogonal_invalidate  */

/*EXPERIMENTAL_FUNCTION*/
void graph1d_orthogonal_show_trace (KOrthogonalProfile profile,
				    CONST unsigned int *dim_indices,
				    flag correct_aberration,
				    CONST uaddr *coord_indices, uaddr refpos)
/*  [SUMMARY] Show a trace.
    <profile> The orthogonal profile.
    <dim_indices> The dimension indicies. The length of this array must be one
    less than the number of dimensions in the array.
    <correct_aberration> If TRUE, attempt to correct for chromatic aberration.
    <coord_indices> The co-ordinate indices of the trace. The length of this
    array must be one less than the number of dimensions in the array.
    <refpos> The reference position along the profile dimension.
    [RETURNS] Nothing.
*/
{
    static char function_name[] = "graph1d_orthogonal_show_trace";

    VERIFY_PROFILE (profile);
    graph1d_orthogonal_show_traces (profile, dim_indices, correct_aberration,
				    TRUE, 1.0, 1, coord_indices, refpos, NULL);
}   /*  End Function graph1d_orthogonal_show_trace  */

/*EXPERIMENTAL_FUNCTION*/
void graph1d_orthogonal_show_traces (KOrthogonalProfile profile,
				     CONST unsigned int *dim_indices,
				     flag correct_aberration,
				     flag sum, float global_weight,
				     unsigned int num_traces,
				     CONST uaddr *coord_indices, uaddr refpos,
				     CONST float *weights)
/*  [SUMMARY] Show a weighted, summed trace.
    <profile> The orthogonal profile.
    <dim_indices> The dimension indicies. The length of this array must be one
    less than the number of dimensions in the array.
    <correct_aberration> If TRUE, attempt to correct for chromatic aberration.
    <sum> If TRUE, the data are summed rather than averaged.
    <global_weight> This weight value is applied to each trace as it is
    combined.
    <num_traces> The number of traces to combine.
    <coord_indices> The co-ordinate indices for the traces. This is grouped
    into <<num_traces>> contiguous blocks of co-ordinate traces. Each block
    must contain <<num_dim>> - 1 co-ordinate indices.
    <refpos> The reference position along the profile dimension.
    <weights> The weights to apply to each trace as it is combined. If this is
    NULL no weights are applied.
    [RETURNS] Nothing.
*/
{
    int trace_dim = -1;
    unsigned int trace_count, count, dim_count, num_dim, type;
    float weight, val;
    double scale, offset, min, max;
    double value[2];
    dim_desc *dim;
    iaddr *offsets;
    float *myweights;
    double *x, *y, *trace_pos;
    static double **coords;
    static unsigned int num_coord_pointers_allocated = 0;
    static char function_name[] = "graph1d_orthogonal_show_traces";

    VERIFY_PROFILE (profile);
    profile->valid = FALSE;
    graph1d_set_active (profile->vgraph, FALSE, FALSE, FALSE, FALSE, 0);
    profile->num_traces = num_traces;  /*  For <<save_cbk>>  */
    num_dim = iarray_num_dim (profile->array);
    if (num_dim > num_coord_pointers_allocated)
    {
	m_free2 (coords);
	if ( ( coords = m_alloc (sizeof *coords * num_dim) ) == NULL )
	{
	    m_error_notify (function_name, "co-ordinate pointer array");
	    return;
	}
	num_coord_pointers_allocated = num_dim;
    }
    /*  Find trace dimension  */
    for (dim_count = 0; (dim_count < num_dim) && (trace_dim < 0); ++dim_count)
    {
	trace_dim = dim_count;
	for (count = 0; count < num_dim - 1; ++count)
	    if (dim_indices[count] == dim_count) trace_dim = -1;
    }
    if (trace_dim < 0)
    {
	fprintf (stderr, "No trace dimension!\n");
	a_prog_bug (function_name);
    }
    dim = iarray_get_dim_desc (profile->array, trace_dim);
    if (refpos >= dim->length)
    {
	fprintf (stderr, "Reference index: %lu is greater than length: %lu\n",
		 refpos, dim->length);
	a_prog_bug (function_name);
    }
    type = iarray_type (profile->array);
    /*  Ensure restriction arrays are long enough  */
    if (num_dim > profile->num_restr_allocated)
    {
	if (profile->restr_names) m_free (profile->restr_names);
	if ( ( profile->restr_names =
	       m_alloc (num_dim *sizeof *profile->restr_names) ) == NULL )
	    m_abort (function_name, "array of name pointers");
	if (profile->restr_values) m_free (profile->restr_values);
	if ( ( profile->restr_values =
	       m_alloc (num_dim * sizeof *profile->restr_values) )
	     == NULL )
	    m_abort (function_name, "array of values");
	profile->num_restr_allocated = num_dim;
    }
    /*  Compute restrictions  */
    for (dim_count = 0; dim_count < num_dim - 1; ++dim_count)
    {
	unsigned int min, max;

	profile->restr_names[dim_count] =
	    iarray_dim_name (profile->array, dim_indices[dim_count]);
	/*  Find minimum and maximum index for this dimension across traces  */
	min = max = 0;
	--min;
	for (trace_count = 0; trace_count < num_traces; ++trace_count)
	{
	    unsigned int index;

	    index = coord_indices[trace_count * (num_dim - 1) + dim_count];
	    if (index < min) min = index;
	    if (index > max) max = index;
	}
	profile->restr_values[dim_count] =
	    iarray_get_coordinate (profile->array, dim_indices[dim_count],
				   (max + min) >> 1);
    }
    profile->value_name = iarray_value_name (profile->array);
    if ( sum && (num_traces > 1) &&
	 (strcmp (profile->value_name, "JY/PIX") == 0) )
	profile->value_name = "JY";
    else if ( sum && (num_traces > 1) &&
	      (strcmp (profile->value_name, "JY/BEAM") == 0) &&
	      (num_dim == 3) )
    {
	weight = iarray_get_beam_correction (profile->array,
					     dim_indices[0], dim_indices[1]);
	if ( DOUBLE_IS_VALID (weight) )
	{
	    global_weight *= weight;
	    profile->value_name = "JY";
	}
    }
    /*  The following allocation is very fragile, and depends on certain type
	sizes and alignment requirements. Sparc and MIPS are susceptible  */
    trace_pos = m_alloc_scratch (sizeof *trace_pos * num_dim +
				 dim->length * (sizeof **coords * num_dim +
						sizeof *offsets +
						sizeof *myweights),
				 function_name);
    coords[0] = (double *) (trace_pos + num_dim);
    offsets = (iaddr *) (coords[0] + dim->length * num_dim);
    myweights = (float *) (offsets + dim->length);
    for (count = 1; count < num_dim; ++count)
	coords[count] = coords[count - 1] + dim->length;
    /*  Ensure co-ordinate buffer is big enough  */
    if ( !graph1d_allocate_coords (profile->vgraph, dim->length, &x, &y) )
	m_abort (function_name, "coord array");
    /*  Initialise the trace X and Y co-ordinates and weights buffers  */
    for (count = 0; count < dim->length; ++count)
    {
	x[count] = iarray_get_coordinate (profile->array, trace_dim, count);
	y[count] = 0.0;
	myweights[count] = 0.0;
    }
    if ( !profile->ap || !wcs_astro_test_if_velocity (profile->ap, dim->name)
	 || !wcs_astro_has_chromatic_aberration (profile->ap) )
	correct_aberration = FALSE;
    if (profile->ap)
	wcs_astro_transform3 (profile->ap, FALSE, dim->length,
			      dim->name, x, FALSE,
			      NULL, NULL, FALSE,
			      NULL, NULL, FALSE,
			      num_dim - 1,
			      profile->restr_names, profile->restr_values);
    scale = profile->scale;
    offset = profile->offset;
    /*  Loop over the traces, computing trace Y co-ordinates  */
    for (trace_count = 0; trace_count < num_traces;
	 ++trace_count, coord_indices += num_dim - 1)
    {
	weight = global_weight;
	if (weights) weight *= weights[trace_count];
	/*  Compute offsets for each position along trace dimension  */
	if (correct_aberration)
	{
	    for (count = 0; count < num_dim - 1; ++count)
		trace_pos[count] = iarray_get_coordinate(profile->array,
							 dim_indices[count],
							 coord_indices[count]);
	    trace_pos[count] = iarray_get_coordinate (profile->array,
						      trace_dim, refpos);
	    wcs_astro_transform3 (profile->ap, FALSE, 1,
				  iarray_dim_name (profile->array,
						   dim_indices[0]),
				  trace_pos, FALSE,
				  iarray_dim_name (profile->array,
						   dim_indices[1]),
				  trace_pos + 1, FALSE,
				  iarray_dim_name (profile->array, trace_dim),
				  trace_pos + 2, FALSE,
				  0, NULL, NULL);
	    for (dim_count = 0; dim_count < num_dim - 1; ++dim_count)
		for (count = 0; count < dim->length; ++count)
		    coords[dim_count][count] = trace_pos[dim_count];
	    for (count = 0; count < dim->length; ++count)
		coords[num_dim - 1][count] = count;
	    wcs_astro_transform3 (profile->ap, FALSE, dim->length,
				  iarray_dim_name (profile->array,
						   dim_indices[0]),
				  coords[0], TRUE,
				  iarray_dim_name (profile->array,
						   dim_indices[1]),
				  coords[1], TRUE,
				  iarray_dim_name (profile->array, trace_dim),
				  coords[2], FALSE,
				  0, NULL, NULL);
	    for (dim_count = 0; dim_count < num_dim - 1; ++dim_count)
		ds_convert_coordinates
		    (iarray_get_dim_desc (profile->array,
					  dim_indices[dim_count]),
		     dim->length, coords[dim_count], TRUE, TRUE, TRUE);
	    for (count = 0; count < dim->length; ++count)
	    {
		uaddr offset = 0;
		flag blank_it = FALSE;

		for (dim_count = 0; dim_count < num_dim - 1; ++dim_count)
		{
		    unsigned int dim_index = dim_indices[dim_count];
		    double coord = coords[dim_count][count];

		    if ( DOUBLE_IS_BLANK (coord) )
		    {
			blank_it = TRUE;
			break;
		    }
		    offset += profile->array->offsets[dim_index][(long) coord];
		}
		if (blank_it) offsets[count] = -1;
		else offsets[count] = offset +
			 profile->array->offsets[trace_dim][count];
	    }
	}
	else
	{   /*  No chromatic aberration: much easier  */
	    uaddr offset = 0;

	    for (count = 0; count < num_dim - 1; ++count)
	    {
		unsigned int dim_index = dim_indices[count];

		if (coord_indices[count] >=
		    profile->array->lengths[dim_index])
		{
		    fprintf(stderr,
			    "Axis: %u co-ordinate: %lu not less than length: %lu\n",
			    dim_count, coord_indices[count],
			    profile->array->lengths[dim_index]);
		    m_free_scratch ();
		    return;
		}
		offset += profile->array->offsets[dim_index][coord_indices[count]];
	    }
	    for (count = 0; count < dim->length; ++count)
		offsets[count] = (offset +
				  profile->array->offsets[trace_dim][count]);
	}
	/*  Step down a trace and add to the trace array  */
	if (type == K_FLOAT)
	{
	    for (count = 0; count < dim->length; ++count)
	    {
		if (offsets[count] < 0) continue;
		val = *(float *) (profile->array->data +
				  (long) offsets[count]);
		if ( DOUBLE_IS_BLANK (val) ) continue;
		val = val * scale + offset;
		y[count] += weight * val;
		myweights[count] += weight;
	    }
	}
	else
	{
	    /*  Do it the hard way  */
	    for (count = 0; count < dim->length; ++count)
	    {
		if (offsets[count] < 0) continue;
		if ( !ds_get_element (profile->array->data +
				      (long) offsets[count],
				      type, value, NULL) )
		{
		    fprintf (stderr, "Error converting value\n");
		    a_prog_bug (function_name);
		}
		if ( DOUBLE_IS_BLANK (value[0]) ) continue;
		val = value[0] * scale + offset;
		y[count] += weight * val;
		myweights[count] += weight;
	    }
	}
    }
    min = TOOBIG;
    max = -TOOBIG;
    for (count = 0; count < dim->length; ++count)
    {
	if (myweights[count] == 0.0) y[count] = TOOBIG;
	else
	{
	    if (!sum) y[count] /= myweights[count];
	    if (y[count] < min) min = y[count];
	    if (y[count] > max) max = y[count];
	}
    }
    m_free_scratch ();
    profile->trace_minimum = min;
    profile->trace_maximum = max;
    profile->trace_dim = trace_dim;
    graph1d_set_size (profile->vgraph, dim->length);
    graph1d_set_active (profile->vgraph, TRUE, FALSE, FALSE, FALSE, 0);
    profile->valid = TRUE;
}   /*  End Function graph1d_orthogonal_show_traces  */

/*EXPERIMENTAL_FUNCTION*/
void graph1d_orthogonal_show_boxed_trace (KOrthogonalProfile profile,
					  CONST unsigned int *dim_indices,
					  flag sum,
					  CONST uaddr *start_indices,
					  CONST uaddr *stop_indices,
					  uaddr refpos)
/*  [SUMMARY] Show a boxed trace.
    <profile> The orthogonal profile.
    <dim_indices> The dimension indicies. The length of this array must be one
    less than the number of dimensions in the array.
    <sum> If TRUE the traces are summed, else they are averaged.
    <start_indices> The starting co-ordinate indices of the box. The length of
    this array must be one less than the number of dimensions in the array.
    <stop_indices> The stopping co-ordinate indices of the box. The length of
    this array must be one less than the number of dimensions in the array.
    <refpos> The reference position along the profile dimension.
    [NOTE] This function will attempt to correct for chromatic aberration.
    [RETURNS] Nothing.
*/
{
    flag keep_going;
    int count;
    unsigned int ndim, num_traces, trace_count;
    uaddr *indices, *counters;
    static char function_name[] = "graph1d_orthogonal_show_boxed_trace";

    VERIFY_PROFILE (profile);
    ndim = iarray_num_dim (profile->array) - 1;
    /*  Ensure co-ordinate indices array is large enough  */
    for (count = 0, num_traces = 1; count < ndim; ++count)
	num_traces *= stop_indices[count] - start_indices[count];
    if (num_traces * ndim > profile->coord_indices_buf_len)
    {
	m_free2 (profile->coord_indices);
	if ( ( profile->coord_indices =
	       m_alloc (sizeof *profile->coord_indices * ndim *
			num_traces) ) == NULL )
	{
	    m_error_notify (function_name, "co-ordinate indices buffer");
	    return;
	}
	profile->coord_indices_buf_len = num_traces * ndim;
    }
    /*  Ensure co-ordinate counter array is large enough  */
    if (profile->coord_counters_buf_len < ndim)
    {
	m_free2 (profile->coord_counters);
	if ( ( profile->coord_counters =
	       m_alloc (sizeof *profile->coord_counters *ndim) ) == NULL )
	{
	    m_error_notify (function_name, "co-ordinate counters buffer");
	    return;
	}
	profile->coord_counters_buf_len = ndim;
    }
    counters = profile->coord_counters;
    /*  Initialise co-ordinate counters  */
    for (count = 0; count < ndim; ++count)
	counters[count] = start_indices[count];
    /*  Loop over co-ordinates  */
    keep_going = TRUE;
    trace_count = 0;
    for (indices = profile->coord_indices; keep_going;
	 indices += ndim, ++trace_count)
    {
	/*  Copy co-ordinate counter contents to indices array  */
	for (count = 0; count <ndim; ++count) indices[count] = counters[count];
	/*  Now increment counter to next co-ordinate  */
	keep_going = TRUE;
	for (count = ndim - 1; keep_going && (count >= 0); )
	{
	    if (++counters[count] < stop_indices[count]) keep_going = FALSE;
	    else
	    {
		counters[count] = start_indices[count];
		--count;
	    }
	}
	if (count < 0) keep_going = FALSE;
	else keep_going = TRUE;
    }
    if (trace_count != num_traces)
    {
	fprintf (stderr, "trace_count: %u is not num_traces: %u\n",
		 trace_count, num_traces);
	a_prog_bug (function_name);
    }
    graph1d_orthogonal_show_traces (profile, dim_indices, TRUE, sum, 1.0,
				    num_traces, profile->coord_indices,
				    refpos, NULL);
}   /*  End Function graph1d_orthogonal_show_boxed_trace  */

/*EXPERIMENTAL_FUNCTION*/
void graph1d_orthogonal_configure_canvas (KOrthogonalProfile profile)
/*  [SUMMARY] Configure a canvas using an orthogonal profile object.
    <profile> The orthogonal profile.
    [RETURNS] Nothing.
*/
{
    dim_desc *dim;
    static char function_name[] = "graph1d_orthogonal_configure_canvas";

    VERIFY_PROFILE (profile);
    if (!profile->valid) return;
    dim = iarray_get_dim_desc (profile->array, profile->trace_dim);
    canvas_specify (profile->canvas, dim->name, profile->value_name,
		    iarray_num_dim (profile->array) - 1,
		    (CONST char **) profile->restr_names,
		    profile->restr_values);
    graph1d_set_canvas_attributes
	(profile->canvas,
	 GRAPH1D_CANVAS_ATT_LEFT_X, dim->first_coord,
	 GRAPH1D_CANVAS_ATT_RIGHT_X, dim->last_coord,
	 GRAPH1D_CANVAS_ATT_MINIMUM_Y, profile->trace_minimum,
	 GRAPH1D_CANVAS_ATT_MAXIMUM_Y, profile->trace_maximum,
	 GRAPH1D_CANVAS_ATT_END);
    canvas_associate_object (profile->canvas, CANVAS_KEY, profile, TRUE);
}   /*  End Function graph1d_orthogonal_configure_canvas  */

/*EXPERIMENTAL_FUNCTION*/
void graph1d_orthogonal_save (KWorldCanvas canvas, CONST char *filename)
/*  [SUMMARY] Save orthogonal profile data to a file.
    <canvas> The world canvas.
    <filename> The filename to save to.
    [NOTE] If there is no orthogonal profile data, the KViewableGraph that is
    drawn last (i.e. on top) is used instead.
    [RETURNS] Nothing.
*/
{
    KOrthogonalProfile profile;
    KViewableGraph vgraph;
    Channel ch;
    unsigned int count, num_restr;
    multi_array *multi_desc;
    char *tmp = NULL;
    char *xlabel, *ylabel;
    char **restr_names;
    double *restr_values;
    static char function_name[] = "graph1d_orthogonal_save";

    profile = canvas_get_associated_object (canvas, CANVAS_KEY);
    if (profile)
    {
	VERIFY_PROFILE (profile);
	if (!profile->valid) profile = NULL;
    }
    if (profile)
    {
	vgraph = profile->vgraph;
	multi_desc = iarray_multi_array (profile->array);
	ds_get_named_object (multi_desc, multi_desc->num_arrays,
			     "filename", (void **) &tmp);
    }
    else
    {   /*  Fallback to last vgraph  */
	vgraph = graph1d_get_last (canvas);
	if (!vgraph) return;
	if (graph1d_get_size (vgraph, NULL, NULL) < 1) return;
	multi_desc = NULL;
	tmp = NULL;
    }
    if ( ( ch = ch_open_file (filename, "w") ) == NULL )
    {
	fprintf (stderr, "Error opening file: \"%s\"\t%s\n",
		 filename, ERRSTRING);
	return;
    }
    if (tmp == NULL) ch_printf (ch, "# ASCII dump of profile\n");
    else ch_printf (ch, "# ASCII dump of profile. Original file: %s\n", tmp);
    if (!multi_desc)
    {
	graph1d_save_ascii (vgraph, ch);
	ch_close (ch);
	return;
    }
    canvas_get_specification (canvas, &xlabel, &ylabel,
			      &num_restr, &restr_names, &restr_values);
    if (num_restr > 1)
    {
	ch_printf (ch, "# Position of profile (%u dimensions):\n", num_restr);
	for (count = 0; count < num_restr; ++count)
	{
	    double value = restr_values[count];
	    double dummy = 0.0;
	    char txt[STRING_LENGTH];

	    if (profile->ap == NULL)
	    {
		ch_printf (ch, "# Axis: %s  co-ordinate: %g\n",
			   restr_names[count], value);
		continue;
	    }
	    if ( !wcs_astro_transform3 (profile->ap, FALSE, 1,
					restr_names[count], &value, FALSE,
					xlabel, &dummy, FALSE,
					NULL, NULL, FALSE,
					num_restr, (CONST char **) restr_names,
					restr_values) )
	    {
		ch_close (ch);
		unlink (filename);
		fprintf
		  (stderr,
		   "Error saving profile header, axis: \"%s\" value: \"%g\"\n",
		   restr_names[count], restr_values[count]);
		return;
	    }
	    wcs_astro_format (profile->ap, restr_names[count], txt, value);
	    ch_printf (ch, "# %-40s  linear co-oridinate: %g\n",
		       txt, restr_values[count]);
	}
    }
    ch_printf (ch,
	       "# %u %s to produce this profile\n",
	       profile->num_traces,
	       (profile->num_traces < 2) ?
	       "individual profile was used" : "individual profiles were combined");
    graph1d_save_ascii (vgraph, ch);
    ch_close (ch);
}   /*  End Function graph1d_orthogonal_save  */

/*EXPERIMENTAL_FUNCTION*/
flag graph1d_orthogonal_track_compute (KWorldCanvas canvas, double x, double y,
				       double x_lin,
				       char x_string[STRING_LENGTH],
				       char y_string[STRING_LENGTH])
/*  [SUMMARY] Compute track strings.
    <canvas> The world canvas.
    <x> The horizontal world co-ordinate of the event.
    <y> The vertical world co-ordinate of the event.
    <x_lin> The horizontal linear co-ordinate of the event.
    <x_string> The X co-ordinate information will be written here.
    <y_string> The Y co-ordinate information will be written here.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    KOrthogonalProfile profile;
    unsigned long px;
    dim_desc *dim;
    char *ptr, *xlabel, *ylabel;
    char unit_string[STRING_LENGTH];
    static char function_name[] = "graph1d_orthogonal_track_compute";

    profile = canvas_get_associated_object (canvas, CANVAS_KEY);
    if (!profile) return (FALSE);
    VERIFY_PROFILE (profile);
    if (!profile->valid) return (FALSE);
    /*  First compute the x co-ordinate information  */
    canvas_get_specification (canvas, &xlabel, &ylabel, NULL, NULL, NULL);
    dim = iarray_get_dim_desc (profile->array, profile->trace_dim);
    /*  Convert linear world co-ordinates to array indices and display  */
    px = ds_get_coord_num (dim, x_lin, SEARCH_BIAS_CLOSEST);
    if (profile->ap == NULL)
    {
	double wx[2], wy[2];

	sprintf (x_string, "x: %lu pixel  ", px);
	ptr = strstr (x_string, "pixel");
	canvas_get_attributes (canvas,
			       CANVAS_ATT_LEFT_X, wx,
			       CANVAS_ATT_RIGHT_X, wx + 1,
			       CANVAS_ATT_BOTTOM_Y, wy,
			       CANVAS_ATT_TOP_Y, wy + 1,
			       CANVAS_ATT_END);
	canvas_coords_transform (canvas, 2, wx, FALSE, wy, FALSE);
	ds_format_value (ptr + 7, x, xlabel, 1.0, 0.0,
			 fabs (wx[0]) + fabs (wx[1]), NULL, NULL);
    }
    else
    {
	wcs_astro_format (profile->ap, xlabel, unit_string, x);
	sprintf (x_string, "x: %lu pixel  %s", px, unit_string);
    }
    /*  Now compute the y co-ordinate information. Unscale and pass scale so
	that raw value can be shown  */
    y = (y - profile->offset) / profile->scale;
    ds_format_value (y_string, y, ylabel, profile->scale, profile->offset,
		     y, profile->array->top_pack_desc,
		     *profile->array->top_packet);
    return (TRUE);
}   /*  End Function graph1d_orthogonal_track_compute  */


/*  Private functions follow  */

static void _graph1d_orthogonal_iarr_destroy_callback
    (iarray arr, KOrthogonalProfile profile)
/*  [SUMMARY] Register the destruction of the Intelligent Array.
    <arr> The Intelligent Array.
    <profile> The KOrthogonalProfile object.
    [RETURNS] Nothing.
*/
{
    graph1d_orthogonal_destroy (profile);
}   /*  End Function iarr_destroy_callback  */

/*
    outy = Lm - Ls * (Hm - Lm) / (Hs - Ls) + (Hm - Lm) / (Hs - Ls) * y
*/
