/*LINTLIBRARY*/
/*  main.c

    This code provides generic KListWidget objects.

    Copyright (C) 2004  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This file contains all routines needed for displaying lists of strings.


    Written by      Richard Gooch   7-MAR-2004

    Updated by      Richard Gooch   10-MAR-2004: Initial tweaking complete.

    Updated by      Richard Gooch   11-MAR-2004: Added Home/End/PgUp/PgDn
  controls.

    Last updated by Richard Gooch   30-OCT-2004: Bug fix in function
  <_listw_canvas_position_event>: up arrow on short list resulted in
  wrap-around.


*/
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <errno.h>
#include <karma.h>
#include <k_event_codes.h>
#include <karma_listw.h>
#include <karma_kwin.h>
#include <karma_a.h>
#include <karma_m.h>
#include <karma_c.h>


#define MAGIC_NUMBER         (unsigned int) 3872327

#define VERIFY_LIST(list) {if (list == NULL) \
{fprintf (stderr, "NULL list passed\n"); \
 a_prog_bug (function_name); } \
if (list->magic_number != MAGIC_NUMBER) \
{fprintf (stderr, "Invalid list object\n"); \
 a_prog_bug (function_name); } }


/*  Private structures follow  */

struct listwidget_type
{
    unsigned int magic_number;
    KPixCanvas canvas;
    /*  Callback lists where others register an interest in this list   */
    KCallbackList position_list;
    KCallbackList att_notify_list;
    /*  Callback functions this widget has registered with other objects  */
    KCallbackFunc refresh_func;
    KCallbackFunc position_func;
    KCallbackFunc destroy_func;
    int height;
    int ascent;
    int descent;
    unsigned int event_index;
    /*  Attributes  */
    unsigned int num_strings;
    CONST char **strings;
    unsigned int top_position;
    unsigned int reverse_video_index;
    unsigned int highlight_index;
    unsigned long foreground_pixel;
    unsigned long highlight_pixel;
    unsigned int mouse_mode;
};

struct position_struct
{
    unsigned int index;
    unsigned int event_code;
    void *e_info;
};

typedef flag (*PositionFunc) (KListWidget list, unsigned int index,
			      unsigned int event_code, void *e_info,
			      void **f_info);


/*  Private data  */


/*  Private functions  */
STATIC_FUNCTION (void _listw_refresh_line,
		 (KListWidget list, unsigned int index, int width, int height,
		  flag always_clear) );
STATIC_FUNCTION (void _listw_canvas_refresh_func,
		 (KPixCanvas canvas, int width, int height, void **info,
		  PostScriptPage pspage,
		  unsigned int num_areas, KPixCanvasRefreshArea *areas,
		  flag *honoured_areas) );
STATIC_FUNCTION (flag _listw_canvas_position_event,
		 (KPixCanvas pixcanvas, int x, int y, unsigned int event_code,
		  void *event_info, void **f_info) );
STATIC_FUNCTION (flag _listw_position_event_func,
		 (void *object, void *client1_data,
		  void *call_data, void *client2_data) );


/* Public functions follow */

/*EXPERIMENTAL_FUNCTION*/
KListWidget listw_create ()
/*  [SUMMARY] Create a list widget.
    [RETURNS] A list widget object on success, else NULL.
*/
{
    KListWidget list;
    static char function_name[] = "listw_create";

    if ( ( list = m_calloc (sizeof *list) ) == NULL )
    {
	m_error_notify (function_name, "list widget");
	return (NULL);
    }
    list->magic_number = MAGIC_NUMBER;
    return (list);
}   /*  End Function listw_create  */

/*EXPERIMENTAL_FUNCTION*/
void listw_bind (KListWidget list, KPixCanvas canvas)
/*  [SUMMARY] Bind a list widget to a pixel canvas.
    <canvas> The pixel canvas object.
    [NOTE] Only one list widget may be bound per pixel canvas.
    [RETURNS] Nothing. On failure, the process aborts.
*/
{
    KPixCanvasFont font;
    static char function_name[] = "listw_bind";

    VERIFY_LIST (list);
    if (canvas == NULL)
    {
	fprintf (stderr, "NULL KPixCanvas passed\n");
	a_prog_bug (function_name);
    }
    list->canvas = canvas;
    /*  Process refreshes from lower down  */
    list->refresh_func =
	kwin_register_refresh_func (canvas, _listw_canvas_refresh_func, list);
    /*  Process position events from lower down  */
    list->position_func =
	kwin_register_position_event_func (canvas,
					   _listw_canvas_position_event,
					   list);
    /*  Process destroy events from lower down  */
    list->destroy_func =
	kwin_register_destroy_func (canvas, listw_destroy, list);
    kwin_get_attributes (canvas,
			 KWIN_ATT_FONT, &font,
			 KWIN_ATT_END);
    if ( !kwin_get_string_size (font, "aAg|",
				KWIN_STRING_HEIGHT, &list->height,
				KWIN_STRING_ASCENT, &list->ascent,
				KWIN_STRING_DESCENT, &list->descent,
				KWIN_STRING_END) )
	m_abort (function_name, "font info");
    list->height += 1;
}   /*  End Function listw_bind  */

/*EXPERIMENTAL_FUNCTION*/
void listw_destroy (KListWidget list)
/*  [SUMMARY] Destroy a list widget
    <list> The list widget.
    [RETURNS] Nothing.
*/
{
    static char function_name[] = "listw_destroy";

    VERIFY_LIST (list);
    c_unregister_callback (list->refresh_func);
    c_unregister_callback (list->position_func);
    c_unregister_callback (list->destroy_func);
    c_destroy_list (list->position_list);
    c_destroy_list (list->att_notify_list);
    list->magic_number = 0;
    m_free (list);
}   /*  End Function listw_destroy  */

/*EXPERIMENTAL_FUNCTION*/
KCallbackFunc listw_register_position_event_func (KListWidget list,
						  flag (*func) (),
						  void *f_info)
/*  [SUMMARY] Register position event callback.
    [PURPOSE] This routine will register a position event function for a list
    widget. The position event function will be called whenever a position
    event on the list widget has not been consumed. Many position event
    functions may be registered per list widget. The first function registered
    is the first function called upon a position event.
    <list> The list widget.
    <func> The function that is called when a position event occurs. The
    prototype function is [<listw_PROTO_position_func>].
    <f_info> The initial arbitrary function information pointer.
    [RETURNS] A KCallbackFunc object.
*/
{
    static char function_name[] = "listw_register_position_event_func";

    VERIFY_LIST (list);
    return c_register_callback (&list->position_list,
				_listw_position_event_func,
				list, f_info, TRUE, (void *) func, FALSE,
				TRUE);
}   /*  End Function listw_register_position_event_func  */

/*EXPERIMENTAL_FUNCTION*/
void listw_get_attributes (KListWidget list, ...)
/*  [SUMMARY] Get list widget attributes.
    <list> The list widget.
    [VARARGS] The optional attributes are given as pairs of attribute-key
    attribute-value pointer pairs. The list must be terminated with
    LISTW_ATT_END. See [<listw_ATTRIBUTES>] for the list of attributes.
    [RETURNS] Nothing.
*/
{
    va_list argp;
    int height;
    unsigned int att_key;
    double shown_fraction;
    static char function_name[] = "listw_get_attributes";

    va_start (argp, list);
    VERIFY_LIST (list);
    while ( ( att_key = va_arg (argp, unsigned int) ) != LISTW_ATT_END )
    {
	switch (att_key)
	{
	  case LISTW_ATT_NUM_STRINGS:
	    *( va_arg (argp, unsigned int *) ) = list->num_strings;
	    break;
	  case LISTW_ATT_STRINGS:
	    *( va_arg (argp, CONST char ***) ) = list->strings;
	    break;
	  case LISTW_ATT_TOP_POSITION:
	    *( va_arg (argp, unsigned int *) ) = list->top_position;
	    break;
	  case LISTW_ATT_REVERSE_VIDEO_INDEX:
	    *( va_arg (argp, unsigned int *) ) = list->reverse_video_index;
	    break;
	  case LISTW_ATT_HIGHLIGHT_INDEX:
	    *( va_arg (argp, unsigned int *) ) = list->highlight_index;
	    break;
	  case LISTW_ATT_FOREGROUND_PIXEL:
	    *( va_arg (argp, unsigned long *) ) = list->foreground_pixel;
	    break;
	  case LISTW_ATT_HIGHLIGHT_PIXEL:
	    *( va_arg (argp, unsigned long *) ) = list->foreground_pixel;
	    break;
	  case LISTW_ATT_MOUSE_MODE:
	    *( va_arg (argp, unsigned int *) ) = list->mouse_mode;
	    break;
	  case LISTW_ATT_SHOWN_FRACTION:
	    if (list->canvas && list->num_strings)
	    {
		kwin_get_size (list->canvas, NULL, &height);
		height = (height - 2) / list->height;  /*  Integral strings  */
		if (height > list->num_strings) height = list->num_strings;
		shown_fraction = (double) height / (double) list->num_strings;
	    }
	    else shown_fraction = 1.0;
	    *( va_arg (argp, double *) ) = shown_fraction;
	    break;
	  default:
	    fprintf (stderr, "Unknown attribute key: %u\n", att_key);
	    a_prog_bug (function_name);
	    break;
	}
    }
    va_end (argp);
}   /*  End Function listw_get_attributes  */

/*EXPERIMENTAL_FUNCTION*/
flag listw_set_attributes (KListWidget list, flag refresh_if_needed, ...)
/*  [SUMMARY] Set list widget attributes.
    <list> The list widget.
    <refresh_if_needed> If TRUE, the list display is refreshed if something has
    changed.
    [VARARGS] The optional attributes are given as pairs of attribute-key
    attribute-value pairs. The list must be terminated with
    LISTW_ATT_END. See [<listw_ATTRIBUTES>] for the list of attributes.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    va_list argp;
    flag everything_changed = FALSE;
    flag retval = TRUE;
    unsigned int att_key, uival, old_reverse_video_index, old_highlight_index;
    unsigned long ulval;
    static char function_name[] = "listw_set_attributes";

    va_start (argp, refresh_if_needed);
    VERIFY_LIST (list);
    old_reverse_video_index = list->reverse_video_index;
    old_highlight_index = list->highlight_index;
    while ( ( att_key = va_arg (argp, unsigned int) ) != LISTW_ATT_END )
    {
	switch (att_key)
	{
	  case LISTW_ATT_NUM_STRINGS:
	    list->num_strings = va_arg (argp, unsigned int);
	    everything_changed = TRUE;
	    break;
	  case LISTW_ATT_STRINGS:
	    list->strings = va_arg (argp, CONST char **);
	    everything_changed = TRUE;
	    break;
	  case LISTW_ATT_TOP_POSITION:
	    uival = va_arg (argp, unsigned int);
	    if (list->top_position == uival) break;
	    list->top_position = uival;
	    everything_changed = TRUE;
	    break;
	  case LISTW_ATT_REVERSE_VIDEO_INDEX:
	    uival = va_arg (argp, unsigned int);
	    if (list->reverse_video_index == uival) break;
	    list->reverse_video_index = uival;
	    break;
	  case LISTW_ATT_HIGHLIGHT_INDEX:
	    uival = va_arg (argp, unsigned int);
	    if (list->highlight_index == uival) break;
	    list->highlight_index = uival;
	    break;
	  case LISTW_ATT_FOREGROUND_PIXEL:
	    ulval = va_arg (argp, unsigned long);
	    if (list->foreground_pixel == ulval) break;
	    list->foreground_pixel = ulval;
	    everything_changed = TRUE;
	    break;
	  case LISTW_ATT_HIGHLIGHT_PIXEL:
	    ulval = va_arg (argp, unsigned long);
	    if (list->highlight_pixel == ulval) break;
	    list->highlight_pixel = ulval;
	    everything_changed = TRUE;
	    break;
	  case LISTW_ATT_MOUSE_MODE:
	    uival = va_arg (argp, unsigned int);
	    if (list->mouse_mode == uival) break;
	    list->mouse_mode = uival;
	    everything_changed = TRUE;
	    break;
	  default:
	    fprintf (stderr, "Unknown attribute key: %u\n", att_key);
	    a_prog_bug (function_name);
	    break;
	}
    }
    va_end (argp);
    if (refresh_if_needed && list->canvas)
    {
	int width, height, top_position, lines_showable;

	kwin_get_size (list->canvas, &width, &height);
	lines_showable = (height - 2) / list->height;  /*  Integral strings  */
	top_position = list->top_position;
	if (top_position + lines_showable > list->num_strings)
	    top_position = (int) list->num_strings - lines_showable;
	if (top_position < 0) top_position = 0;
	list->top_position = top_position;
	if (everything_changed)
	    retval = kwin_refresh_if_visible (list->canvas, TRUE);
	else
	{
	    if (old_reverse_video_index != list->reverse_video_index)
	    {
		_listw_refresh_line (list, old_reverse_video_index,
				     width, height, TRUE);
		_listw_refresh_line (list, list->reverse_video_index,
				     width, height, TRUE);
	    }
	    if (old_highlight_index != list->highlight_index)
	    {
		_listw_refresh_line (list, old_highlight_index,
				     width, height, TRUE);
		_listw_refresh_line (list, list->highlight_index,
				     width, height, TRUE);
	    }
	}
    }
    c_call_callbacks (list->att_notify_list, NULL);
    return (retval);
}   /*  End Function listw_set_attributes  */

/*EXPERIMENTAL_FUNCTION*/
KCallbackFunc listw_register_att_notify_func (KListWidget list,
					      void (*func) (), void *info)
/*  [SUMMARY] Register an attribute change function.
    [PURPOSE] This routine will register an attribute notify function for a
    list widget. The attribute notify function will be called whenever
    the list widget attributes are changed (see [<listw_set_attributes>]).
    Many attribute notify functions may be registered per list widget. The
    first function registered is the first function called upon attribute
    changes.
    <list> The list widget.
    <func> The function which is called when the attributes are changed.
    The prototype function is [<listw_PROTO_att_notify_func>].
    <info> The initial arbitrary information pointer.
    [RETURNS] A KCallbackFunc object.
*/
{
    static char function_name[] = "listw_register_att_notify_func";

    VERIFY_LIST (list);
    return c_register_callback (&list->att_notify_list,
				( flag (*) () ) func,
				list, info, TRUE, NULL, FALSE,
				FALSE);
}   /*  End Function listw_register_att_notify_func  */


/*  Private functions follow  */

static void _listw_refresh_line (KListWidget list, unsigned int index,
				 int width, int height, flag always_clear)
/*  [SUMMARY] Refresh a single line.
    <list> The list widget.
    <index> The index of the line to refresh.
    <width> The width of the canvas.
    <height> The height of the canvas.
    [RETURNS] Nothing.
*/
{
    unsigned int y;
    unsigned long fg_pixel = list->foreground_pixel;
    unsigned long bg_pixel, box_pixel;
    CONST char *string = list->strings[index];
    /*static char function_name[] = "_listw_refresh_line";*/

    if (index >= list->num_strings) return;
    if (index < list->top_position) return;
    y = (index - list->top_position) * list->height + list->ascent + 2;
    if (y > height + list->height) return;
    kwin_get_attributes (list->canvas,
			 KWIN_ATT_BACKGROUND_PIXEL, &bg_pixel,
			 KWIN_ATT_END);
    box_pixel = bg_pixel;
    if (index == list->highlight_index) box_pixel = list->highlight_pixel;
    if (index == list->reverse_video_index)
    {
	fg_pixel = box_pixel;
	box_pixel = list->foreground_pixel;
    }
    if ( always_clear || (box_pixel != bg_pixel) )
	kwin_fill_rectangle (list->canvas, 2, y - list->ascent, width - 4,
			     list->height, box_pixel);
    kwin_draw_string (list->canvas, 3, y, string, fg_pixel, FALSE);
}   /*  End Function _listw_refresh_line  */

static void _listw_canvas_refresh_func (KPixCanvas canvas,
					int width, int height,
					void **info, PostScriptPage pspage,
					unsigned int num_areas,
					KPixCanvasRefreshArea *areas,
					flag *honoured_areas)
/*  [PURPOSE] This routine is a refresh event consumer for a pixel canvas.
    <canvas> The pixel canvas.
    <width> The width of the canvas in pixels.
    <height> The height of the canvas in pixels.
    <info> A pointer to the arbitrary information pointer.
    <pspage> The PostScriptPage object. If this is NULL, the refresh is *not*
    destined for a PostScript page.
    <num_areas> The number of areas that need to be refreshed. If this is
    0 then the entire pixel canvas needs to be refreshed.
    <areas> The list of areas that need to be refreshed.
    <honoured_areas> If the value TRUE is written here it is assumed the
    routine honoured the list of refresh areas and did not write outside
    these areas and hence the list of areas will be passed to subsequent
    registered refresh routines. If FALSE is written here (or nothing is
    written here), implying the routine refreshed the entire pixel canvas,
    subsequent refresh routines will be told to refresh the entire canvas.
    [RETURNS] Nothing.
*/
{
    KListWidget list = *info;
    unsigned int count, lines_showable, stop;
    static char function_name[] = "_listw_canvas_refresh_func";

    VERIFY_LIST (list);
    if (canvas != list->canvas)
    {
	fprintf (stderr, "Pixel canvases do not match\n");
	a_prog_bug (function_name);
    }
    lines_showable = (height - 2) / list->height + 1;
    stop = list->top_position + lines_showable;
    if (stop > list->num_strings) stop = list->num_strings;
    for (count = list->top_position; count < stop; ++count)
	_listw_refresh_line (list, count, width, height, FALSE);
}   /*  End Function _listw_canvas_refresh_func  */

static flag _listw_canvas_position_event (KPixCanvas canvas,
					  int x, int y,
					  unsigned int event_code,
					  void *event_info, void **f_info)
/*  [SUMMARY] Process a position event on a pixel canvas.
    <canvas> The pixel canvas on which the event occurred.
    <x> The horizontal position of the event, relative to the canvas origin
    <y> The vertical position of the event, relative to the canvas origin.
    <event_code> The arbitrary event code.
    <event_info> A pointer to arbitrary event information.
    <f_info> A pointer to an arbitrary function information pointer.
    [RETURNS] TRUE if the event was consumed, else FALSE indicating that
    the event is still to be processed.
*/
{
    KListWidget list = *f_info;
    int width, height, index, lines_showable, top_position, factor;
    unsigned long mask = 0;
    struct position_struct data;
    char *key;
    static char function_name[] = "_listw_canvas_position_event";

    VERIFY_LIST (list);
    if (list->num_strings < 1) return (FALSE);
    if ( (x < 0) || (y < 0) ) return (FALSE);
    kwin_get_size (list->canvas, &width, &height);
    if ( (x >= width) || (y >= height) ) return (FALSE);
    index = y / list->height + list->top_position;
    lines_showable = (height - 2) / list->height;      /*  Integral strings  */
    top_position = list->top_position;
    switch (event_code)
    {
      case K_CANVAS_EVENT_KEYSYM_PRESS:
	key = event_info;
	m_copy (&mask, key + strlen (key) + 1, sizeof mask);
	if (strncmp (key, "KP_", 3) == 0) key += 3;
	index = list->event_index;
	factor = (mask & K_CANVAS_EVENT_CONTROL_KEY_MASK) ? 10 : 1;
	if (strcmp (key, "Down") == 0)
	{
	    ++index;
	    if (index >= top_position + lines_showable) ++top_position;
	}
	else if (strcmp (key, "Up") == 0)
	{
	    --index;
	    if (index < top_position) --top_position;
	}
	else if (strcmp (key, "Home") == 0)
	{
	    if (mask & K_CANVAS_EVENT_CONTROL_KEY_MASK) top_position = 0;
	    index = top_position;
	}
	else if (strcmp (key, "End") == 0)
	{
	    if (mask & K_CANVAS_EVENT_CONTROL_KEY_MASK)
		top_position = (int) list->num_strings - lines_showable;
	    index = top_position + lines_showable - 1;
	}
	else if (strcmp (key, "Prior") == 0)
	{
	    top_position -= lines_showable * factor;
	    index -= lines_showable * factor;
	}
	else if (strcmp (key, "Next") == 0)
	{
	    top_position += lines_showable * factor;
	    index += lines_showable * factor;
	}
	else break;
	event_code = K_CANVAS_EVENT_POINTER_MOVE;
	/*  Fall through  */
      case K_CANVAS_EVENT_POINTER_MOVE:
	if (top_position > (int) list->num_strings - lines_showable)
	    top_position = (int) list->num_strings - lines_showable;
	if (top_position < 0) top_position = 0;
	if (index < top_position) index = top_position;
	if (index >= top_position + lines_showable)
	    index = top_position + lines_showable - 1;
	if (index >= list->num_strings) index = list->num_strings - 1;
	if (list->event_index == index) return (TRUE);
	list->event_index = index;
	break;
      case K_CANVAS_EVENT_POINTER_ENTER:
	if (index >= list->num_strings) index = list->num_strings - 1;
	list->event_index = index;
	break;
      case K_CANVAS_EVENT_POINTER_LEAVE:
	index = list->num_strings;
	list->event_index = index;
	break;
    }
    switch (list->mouse_mode)
    {
      case LISTW_MOUSE_MODE_REVERSE_VIDEO:
	listw_set_attributes (list, TRUE,
			      LISTW_ATT_TOP_POSITION, top_position,
			      LISTW_ATT_REVERSE_VIDEO_INDEX, list->event_index,
			      LISTW_ATT_END);
	break;
      case LISTW_MOUSE_MODE_HIGHLIGHT:
	listw_set_attributes (list, TRUE,
			      LISTW_ATT_TOP_POSITION, top_position,
			      LISTW_ATT_HIGHLIGHT_INDEX, list->event_index,
			      LISTW_ATT_END);
	break;
    }
    if (list->event_index >= list->num_strings) return (FALSE);
    /*  Call event consumer functions  */
    data.event_code = event_code;
    data.index = list->event_index;
    data.e_info = event_info;
    return c_call_callbacks (list->position_list, &data);
}   /*  End Function _listw_canvas_position_event  */

static flag _listw_position_event_func (void *object, void *client1_data,
					void *call_data, void *client2_data)
/*  [PURPOSE] This routine is called when object callbacks are called.
    <object> The object information pointer.
    <client1_data> The first client information pointer.
    <call_data> The call information pointer.
    <client2_data> The second client information pointer.
    [RETURNS] TRUE if further callbacks should not be called, else FALSE.
*/
{
    KListWidget list = object;
    PositionFunc func = (PositionFunc) client2_data;
    struct position_struct *data = call_data;
    static char function_name[] = "_listw_position_event_func";

    VERIFY_LIST (list);
    return (*func) (list, data->index, data->event_code,
		    data->e_info, client1_data);
}   /*  End Function _listw_position_event_func  */
