/*LINTLIBRARY*/
/*  main.c

    This code provides KScatterPlot objects.

    Copyright (C) 2001-2003  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This file contains all routines needed for manipulating scatter plots.


    Written by      Richard Gooch   22-NOV-2001: First crude implementation.

    Updated by      Richard Gooch   5-DEC-2001: Added experimental function
  <scatplt_register_destroy_func>.

    Updated by      Richard Gooch   10-DEC-2001: Made crude optimsation when
  drawing areas.

    Updated by      Richard Gooch   11-DEC-2001: Implemented bitmask for faster
  drawing.

    Updated by      Richard Gooch   12-DEC-2001: Removed unused position list.
  Created <scatplt_find_closest>.

    Updated by      Richard Gooch   13-DEC-2001: Created function
  <scatplt_find_closest_all>.

    Updated by      Richard Gooch   14-DEC-2001: Fixed bug in function
  <scatplt_find_closest>.

    Updated by      Richard Gooch   18-DEC-2001: Force colour to be black for
  paper colours.

    Updated by      Richard Gooch   4-FEB-2002: Added CONST to some casts.

    Updated by      Richard Gooch   27-MAY-2002: Made use of new KLIST_ macros.

    Updated by      Richard Gooch   14-AUG-2002: Added #include <stdlib.h>.

    Updated by      Richard Gooch   17-AUG-2002: Created experimental function
  <scatplt_get_worldcanvas>.

    Updated by      Richard Gooch   23-MAR-2003: Fixed bug in function
  <scatplt_find_closest_all>: initial <<best_distance2>> needs TOOBIG * TOOBIG.

    Last updated by Richard Gooch   7-JUN-2003: Added #include <karma_c.h>


*/
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <stdarg.h>
#include <karma.h>
#include <k_list.h>
#define NEW_WIN_SCALE
#include <karma_scatplt.h>
#include <karma_canvas.h>
#include <karma_iarray.h>
#include <karma_kwin.h>
#include <karma_wcs.h>
#include <karma_ds.h>
#include <karma_st.h>
#include <karma_a.h>
#include <karma_m.h>
#include <karma_c.h>


#define SCATPLT_USED_MAGIC_NUMBER (unsigned int) 1781293278
#define SCATPLT_FREE_MAGIC_NUMBER (unsigned int) 1943802292
#define HOLDER_MAGIC_NUMBER       (unsigned int) 1932832932
#define DEFAULT_COLOUR "green"
#define BUF_SIZE 4096
#define MASK_OF_SHIFT(shift) ((1 << (shift)) - 1)

#define VERIFY_HOLDER(holder) {if (holder == NULL) \
{fprintf (stderr, "NULL canvas holder passed\n"); \
 a_prog_bug (function_name); } \
if (holder->magic_number != HOLDER_MAGIC_NUMBER) \
{fprintf (stderr, "Invalid canvas holder object\n"); \
 a_prog_bug (function_name); } }

#define VERIFY_USED_SCATPLT(scatplt) if (scatplt == NULL) \
{fprintf (stderr, "NULL scatter plot passed\n"); \
 a_prog_bug (function_name); } \
if (scatplt->magic_number != SCATPLT_USED_MAGIC_NUMBER) \
{fprintf (stderr, "Invalid scatter plot object\n"); \
 a_prog_bug (function_name); }

#define VERIFY_FREE_SCATPLT(scatplt) if (scatplt == NULL) \
{fprintf (stderr, "NULL scatter plot passed\n"); \
 a_prog_bug (function_name); } \
if (scatplt->magic_number != SCATPLT_FREE_MAGIC_NUMBER) \
{fprintf (stderr, "Invalid freed scatter plot object\n"); \
 a_prog_bug (function_name); }


typedef struct canvas_holder_type * CanvasHolder;

struct canvas_holder_type
{
    unsigned int magic_number;
    KWorldCanvas canvas;
    KScatterPlot first;
    KScatterPlot last;
    CanvasHolder prev;
    CanvasHolder next;
    char *colourname;
    unsigned long previous_background;
    unsigned long pixel_value;
    float previous_intensity;
};

struct raster_struct     /*  Array entry is one screen pixel. Always         */
{
    int shift;           /*  Shift (right) to convert position to word index */
    unsigned int buflen; /*  Number of words in bit buffer                   */
    double left_wx;
    double right_wx;
    double bottom_wy;
    double top_wy;
    unsigned int *bits;      /*  One bit per screen pixel                    */
    unsigned int *indices;   /*  Index into original array of points         */
};

struct scatter_plot_type
{
    unsigned int magic_number;
    CanvasHolder canvas_holder;
    flag active;
    /*  Input data  */
    unsigned int num_points;
    CONST double *world_x;
    CONST double *world_y;
    /*  Other information  */
    uaddr key;                      /*  Key for grouping  */
    /*  Cached data  */
    double test_x[5];
    double test_y[5];
    struct raster_struct raster;
    /*  Internal management  */
    KCallbackList destroy_list;
    KScatterPlot prev;
    KScatterPlot next;
};


/*  Private data  */
static CanvasHolder first_canvas_holder = NULL;
static CanvasHolder last_canvas_holder = NULL;
static KScatterPlot first_free = NULL;


/*  Private functions  */
STATIC_FUNCTION (CanvasHolder _scatplt_get_canvas_holder,
		 (KWorldCanvas canvas, char *func_name) );
STATIC_FUNCTION (CanvasHolder _scatplt_alloc_canvas_holder,
		 (KWorldCanvas canvas) );
STATIC_FUNCTION (void _scatplt_free_canvas_holder, (CanvasHolder holder) );
STATIC_FUNCTION (void process_canvas_attributes,
		 (CanvasHolder holder, va_list argp) );
STATIC_FUNCTION (void _scatplt_worldcanvas_precompute_func,
		 (KWorldCanvas canvas, int width, int height,
		  CONST struct win_scale_type *win_scale,
		  Kcolourmap cmap, flag cmap_resize,
		  void **info, PostScriptPage pspage,
		  unsigned int num_areas, KPixCanvasRefreshArea *areas) );
STATIC_FUNCTION (void _scatplt_precompute_plot,
		 (KScatterPlot plot, CONST struct win_scale_type *win_scale,
		  unsigned int num_areas, KPixCanvasRefreshArea *areas) );
STATIC_FUNCTION (void _scatplt_worldcanvas_refresh_func,
		 (KWorldCanvas canvas, int width, int height,
		  struct win_scale_type *win_scale,
		  Kcolourmap cmap, flag cmap_resize,
		  void **info, PostScriptPage pspage,
		  unsigned int num_areas, KPixCanvasRefreshArea *areas,
		  flag *honoured_areas) );
STATIC_FUNCTION (void _scatplt_refresh_plot,
		 (KScatterPlot plot, struct win_scale_type *win_scale,
		  unsigned int num_areas, KPixCanvasRefreshArea *areas) );
STATIC_FUNCTION (void _scatplt_refresh_plot_fast,
		 (KScatterPlot plot, struct win_scale_type *win_scale,
		  unsigned int num_areas, KPixCanvasRefreshArea *areas) );
STATIC_FUNCTION (flag _scatplt_find_closest_slow,
		 (KScatterPlot plot,
		  double xin, double yin, double *xout, double *yout,
		  uaddr *index, flag subpixel, double search_radius) );


/* Public functions follow */

/*EXPERIMENTAL_FUNCTION*/
void scatplt_init (KWorldCanvas canvas, ...)
/*  [SUMMARY] Initialise the package for a particular canvas.
    [PURPOSE] This routine will initialise the [<scatplt>] package for a
    particular world canvas. Calling this routine causes a number of callback
    routines internal to the package to be registered with the canvas (such
    as refresh and position event callbacks). This routine must be called
    before creating scatter plots.
    <canvas> The world canvas object.
    [VARARGS] The list of parameter attribute-key attribute-value-ptr pairs
    must follow. This list must be terminated with the SCATPLT_CANVAS_ATT_END.
    See [<scatplt_CANVAS_ATTRIBUTES>] for the list of attributes.
    [RETURNS] Nothing.
*/
{
    va_list argp;
    CanvasHolder holder;
    static char function_name[] = "scatplt_init";

    va_start (argp, canvas);
    if (canvas == NULL)
    {
	fprintf (stderr, "NULL world canvas passed\n");
	a_prog_bug (function_name);
    }
    if ( _scatplt_get_canvas_holder (canvas, NULL) )
    {
	fprintf (stderr, "Canvas already initialised\n");
	a_prog_bug (function_name);
    }
    if ( ( holder = _scatplt_alloc_canvas_holder (canvas) ) == NULL )
	m_abort (function_name, "canvas holder");
    process_canvas_attributes (holder, argp);
    va_end (argp);
}   /*  End Function scatplt_init  */

/*EXPERIMENTAL_FUNCTION*/
KScatterPlot scatplt_create (KWorldCanvas canvas, unsigned int num_points,
			     CONST double *x, CONST double *y, uaddr key)
/*  [SUMMARY] Create scatter plot.
    [PURPOSE] This routine will create a scatter plot object from two arrays of
    values. At a later time, this scatter plot may be made visible. This
    routine will not cause the canvas to be refreshed.
    <canvas> The world canvas onto which the scatter plot may be drawn. Many
    scatter plots may be associated with a single canvas.
    <num_points> The number of points in this scatter plot.
    <x> The array of horizontal values.
    <y> The array of vertical values.
    <key> The key to associate with this scatter plot.
    [NOTE] The KScatterPlot will be automatically destroyed if the canvas is
    destroyed.
    [RETURNS] A KScatterPlot object on success, NULL.
*/
{
    CanvasHolder holder;
    KScatterPlot plot;
    static char function_name[] = "scatplt_create";

    if (canvas == NULL)
    {
	fprintf (stderr, "NULL world canvas passed\n");
	a_prog_bug (function_name);
    }
    if ( (x == NULL) || (y == NULL) )
    {
	fprintf (stderr, "NULL array pointer(s) passed\n");
	a_prog_bug (function_name);
    }
    holder = _scatplt_get_canvas_holder (canvas, function_name);
    if (first_free == NULL)
    {
	if ( ( plot = m_alloc (sizeof *plot) ) == NULL ) return (NULL);
    }
    else
    {
	plot = first_free;
	first_free = plot->next;
	VERIFY_FREE_SCATPLT (plot);
    }
    m_clear (plot, sizeof *plot);
    plot->magic_number = SCATPLT_USED_MAGIC_NUMBER;
    plot->active = FALSE;
    plot->canvas_holder = holder;
    plot->num_points = num_points;
    plot->world_x = x;
    plot->world_y = y;
    plot->key = key;
    switch (sizeof *plot->raster.bits)
    {
      case 1:
	plot->raster.shift = 3;
	break;
      case 2:
	plot->raster.shift = 4;
	break;
      case 4:
	plot->raster.shift = 5;
	break;
      case 8:
	plot->raster.shift = 6;
	break;
      default:
	fprintf (stderr, "Illegal word size: %d\n",
		 (int) sizeof *plot->raster.bits);
	a_prog_bug (function_name);
	break;
    }
    plot->raster.left_wx = TOOBIG;
    plot->raster.right_wx = TOOBIG;
    plot->raster.bottom_wy = TOOBIG;
    plot->raster.top_wy = TOOBIG;
    plot->test_x[0] = TOOBIG;
    KLIST_APPEND (holder, plot);
    return (plot);
}   /*  End Function scatplt_create  */

/*EXPERIMENTAL_FUNCTION*/
flag scatplt_set_active (KScatterPlot plot, flag active, flag force_refresh,
			 flag refresh_if_changed, flag exclusive, uaddr key)
/*  [SUMMARY] Set active state for a scatter plot controlled refresh.
    [PURPOSE] This routine will make a scatter plot active or inactive.
    <plot> The scatter plot.
    <active> If TRUE, the scatter plot is made active, else it is made
    inactive.
    <force_refresh> If TRUE, the canvas is always refreshed.
    <refresh_if_changed> If TRUE, the canvas is refreshed if the scatter plot
    active state changed.
    <exclusive> If TRUE and <<active>> is TRUE, make this scatter plot the only
    active one for this canvas. Other scatter plots for this canvas will be
    made inactive.
    <key> The key to use for inactivation of other scatter plots. If this is 0
    then all other scatter plots will be made inactive.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    CanvasHolder holder;
    KScatterPlot curr;
    static char function_name[] = "scatplt_set_active";

    VERIFY_USED_SCATPLT (plot);
    FLAG_VERIFY (active);
    FLAG_VERIFY (force_refresh);
    FLAG_VERIFY (refresh_if_changed);
    holder = plot->canvas_holder;
    if (active != plot->active)
    {
	plot->active = active;
	if (refresh_if_changed) force_refresh = TRUE;
    }
    if (exclusive && active)
    {
	/*  Make others inactive  */
	for (curr = holder->first; curr; curr = curr->next)
	{
	    if (curr == plot) continue;
	    if ( (key == 0) || (key == curr->key) )
	    {
		if (curr->active && refresh_if_changed) force_refresh = TRUE;
		curr->active = FALSE;
	    }
	}
    }
    if (force_refresh) return canvas_resize (holder->canvas, NULL, FALSE);
    return (TRUE);
}   /*  End Function scatplt_set_active  */

/*EXPERIMENTAL_FUNCTION*/
flag scatplt_set_all_inactive (KWorldCanvas canvas, flag force_refresh,
			       flag refresh_if_changed, uaddr key)
/*  [SUMMARY] Clear active state all scatter plots controlled refresh.
    [PURPOSE] This routine will make all scatter plots associated with a
    world canvas inactive.
    <canvas> The world canvas.
    <force_refresh> If TRUE, the canvas is always refreshed.
    <refresh_if_changed> If TRUE, the canvas is refreshed if the active state
    of any scatter plot changed.
    <key> The key to use for inactivation of scatter plots. If this is 0 then
    all scatter plots will be made inactive.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    CanvasHolder holder;
    KScatterPlot curr;
    static char function_name[] = "scatplt_set_all_inactive";

    holder = _scatplt_get_canvas_holder (canvas, function_name);
    FLAG_VERIFY (force_refresh);
    FLAG_VERIFY (refresh_if_changed);
    /*  Make all inactive  */
    for (curr = holder->first; curr; curr = curr->next)
    {
	if ( (key != 0) && (key != curr->key) ) continue;
	if (curr->active && refresh_if_changed) force_refresh = TRUE;
	curr->active = FALSE;
    }
    if (force_refresh) return canvas_resize (holder->canvas, NULL, FALSE);
    return (TRUE);
}   /*  End Function scatplt_set_all_inactive  */

/*EXPERIMENTAL_FUNCTION*/
void scatplt_destroy (KScatterPlot plot)
/*  [SUMMARY] Destroy scatter plot.
    [PURPOSE] This routine will destroy a scatter plot.
    <plot> The scatter plot.
    [RETURNS] Nothing.
*/
{
    CanvasHolder holder;
    static char function_name[] = "scatplt_destroy";

    VERIFY_USED_SCATPLT (plot);
    holder = plot->canvas_holder;
    c_call_callbacks (plot->destroy_list, NULL);
    c_destroy_list (plot->destroy_list);
    KLIST_REMOVE (holder, plot);
    plot->magic_number = SCATPLT_FREE_MAGIC_NUMBER;
    plot->next = first_free;
    first_free = plot;
}   /*  End Function scatplt_destroy  */

/*EXPERIMENTAL_FUNCTION*/
KCallbackFunc scatplt_register_destroy_func (KScatterPlot plot,
					     void (*func) (), void *info)
/*  [SUMMARY] Register destroy callback.
    [PURPOSE] This routine will register a destroy function for a scatter plot.
    The function will be called when the canvas is destroyed. Many destroy
    functions may be registered per plot. The first function registered is
    the first function called upon destruction.
    <plot> The scatter plot.
    <func> The function that is called when the plot is destroyed. The
    prototype function is [<scatplt_PROTO_destroy_func>].
    <info> The arbitrary function information pointer.
    [RETURNS] A KCallbackFunc object. On failure, the process aborts.
*/
{
    static char function_name[] = "scatplt_register_destroy_func";

    VERIFY_USED_SCATPLT (plot);
    return c_register_callback (&plot->destroy_list, ( flag (*) () ) func,
				info, NULL, FALSE, NULL, FALSE, FALSE);
}   /*  End Function scatplt_register_destroy_func  */

/*EXPERIMENTAL_FUNCTION*/
KWorldCanvas scatplt_get_worldcanvas (KScatterPlot plot)
/*  [SUMMARY] Get the world canvas for a scatter plot.
    <plot> The scatter plot.
    [RETURNS] The KWorldCanvas object.
*/
{
    static char function_name[] = "scatplt_get_worldcanvas";

    VERIFY_USED_SCATPLT (plot);
    return (plot->canvas_holder->canvas);
}   /*  End Function scatplt_get_worldcanvas  */

/*EXPERIMENTAL_FUNCTION*/
flag scatplt_find_closest (KScatterPlot plot, double xin, double yin,
			   double *xout, double *yout, uaddr *index,
			   flag subpixel, double search_radius)
/*  [SUMMARY] Find the closest point to a specified position.
    <plot> The scatter plot to search.
    <xin> The X world co-ordinate to search from.
    <yin> The Y world co-ordinate to search from.
    <xout> The exact X co-ordinate of the point found is written here.
    <yout> The exact Y co-ordinate of the point found is written here.
    <index> The index of the point is written here.
    <subpixel> If TRUE, search with subpixel resolution so that multiple points
    lying on the same screen pixel will be searched to find the closest point.
    If FALSE, the search will terminate early on the first point which lies on
    the same screen pixel. It is highly recommended that subpixel resolution is
    not used, as it may be orders of magnitude slower.
    <search_radius> The maximum search radius. Points further from the search
    position will be ignored.
    [RETURNS] TRUE if a close point was found, else FALSE.
*/
{
    int startx, stopx, starty, stopy, xin_p, yin_p;
    int x, y, i_xo, i_xp, i_yo, i_yp, shift;
    int closest_distance2 = 2147483647;
    unsigned int closest_index, mask;
    static char function_name[] = "scatplt_find_closest";

    VERIFY_USED_SCATPLT (plot);
    if ( subpixel || DOUBLE_IS_BLANK (plot->test_x[0]) )
	return _scatplt_find_closest_slow (plot, xin, yin, xout, yout, index,
					   subpixel, search_radius);
    /*  Use raster to search  */
    shift = plot->raster.shift;
    mask = MASK_OF_SHIFT (plot->raster.shift);
    closest_index = plot->num_points;
    canvas_get_attributes (plot->canvas_holder->canvas,
			   CANVAS_ATT_X_OFFSET, &i_xo,
			   CANVAS_ATT_Y_OFFSET, &i_yo,
			   CANVAS_ATT_X_PIXELS, &i_xp,
			   CANVAS_ATT_Y_PIXELS, &i_yp,
			   CANVAS_ATT_END);
    canvas_convert_from_canvas_coords (plot->canvas_holder->canvas,
				       FALSE, FALSE,
				       1, &xin, &yin, xout, yout);
    xin_p = *xout + 1e-3;
    yin_p = *yout + 1e-3;
    if ( (xin_p < i_xo) || (xin_p >= i_xo + i_xp) ||
	 (yin_p < i_yo) || (yin_p >= i_yo + i_yp) ) return (FALSE);
    if ( DOUBLE_IS_VALID (search_radius) )
    {
	startx = xin_p - (int) search_radius;
	stopx = xin_p + (int) search_radius;
	starty = yin_p - (int) search_radius;
	stopy = yin_p + (int) search_radius;
	if (startx < i_xo) startx = i_xo;
	if (stopx > i_xo + i_xp) stopx = i_xo + i_xp;
	if (starty < i_yo) starty = i_yo;
	if (stopy > i_yo + i_yp) stopy = i_yo + i_yp;
    }
    else
    {
	startx = i_xo;
	stopx = i_xo + i_xp;
	starty = i_yo;
	stopy = i_yo + i_yp;
    }
    for (y = starty; y < stopy; ++y) for (x = startx; x < stopx; ++x)
    {
	int distance2;
	int pos = (y - i_yo) * i_xp + (x - i_xo);

	if ( !( plot->raster.bits[pos >> shift] & ( 1 << (pos & mask) ) ) )
	    continue;
	distance2 = (x - xin_p) * (x - xin_p) + (y - yin_p) * (y - yin_p);
	if (distance2 >= closest_distance2) continue;
	closest_distance2 = distance2;
	closest_index = plot->raster.indices[pos];
    }
    if (closest_index >= plot->num_points) return (FALSE);
    if ( DOUBLE_IS_VALID (search_radius) &&
	 (closest_distance2 > search_radius * search_radius) ) return (FALSE);
    *index = closest_index;
    *xout = plot->world_x[closest_index];
    *yout = plot->world_y[closest_index];
    return (TRUE);
}   /*  End Function scatplt_find_closest  */

/*EXPERIMENTAL_FUNCTION*/
flag scatplt_find_closest_all (KWorldCanvas canvas, double xin, double yin,
			       double *xout, double *yout, uaddr *index,
			       KScatterPlot *plot,
			       flag subpixel, double search_radius)
/*  [SUMMARY] Find the closest point to a specified position.
    <canvas> The canvas to search. All active scatter plots are searched.
    <xin> The X world co-ordinate to search from.
    <yin> The Y world co-ordinate to search from.
    <xout> The exact X co-ordinate of the point found is written here.
    <yout> The exact Y co-ordinate of the point found is written here.
    <index> The index of the point is written here.
    <plot> The KScatterPlot the point was found in is written here.
    <subpixel> If TRUE, search with subpixel resolution so that multiple points
    lying on the same screen pixel will be searched to find the closest point.
    If FALSE, the search will terminate early on the first point which lies on
    the same screen pixel. It is highly recommended that subpixel resolution is
    not used, as it may be orders of magnitude slower.
    <search_radius> The maximum search radius. Points further from the search
    position will be ignored.
    [RETURNS] TRUE if a close point was found, else FALSE.
*/
{
    CanvasHolder holder;
    KScatterPlot curr;
    flag found = FALSE;
    double best_distance2 = TOOBIG * TOOBIG;
    static char function_name[] = "scatplt_find_closest_all";

    holder = _scatplt_get_canvas_holder (canvas, function_name);
    for (curr = holder->first; curr; curr = curr->next)
    {
	uaddr l_index;
	double l_xout, l_yout, tmp;

	if (!curr->active) continue;
	if ( !scatplt_find_closest (curr, xin, yin, &l_xout, &l_yout, &l_index,
				    subpixel, search_radius) ) continue;
	tmp = (l_xout - xin) * (l_xout - xin) + (l_yout - yin) * (l_yout -yin);
	if (tmp >= best_distance2) continue;
	best_distance2 = tmp;
	*xout = l_xout;
	*yout = l_yout;
	*index = l_index;
	*plot = curr;
	found = TRUE;
    }
    return (found);
}   /*  End Function scatplt_find_closest_all  */


/*  Private functions follow  */

static CanvasHolder _scatplt_get_canvas_holder (KWorldCanvas canvas,
						char *func_name)
/*  [SUMMARY] Get the canvas holder for a world canvas.
    <canvas> The canvas.
    <func_name> The name of the calling function. If this is not NULL, the
    process will abort if a canvas holder is not found.
    [RETURNS] The canvas holder on success, else NULL.
*/
{
    CanvasHolder holder;

    /*  First search for existing canvas holder  */
    for (holder = first_canvas_holder; holder; holder = holder->next)
	if (canvas == holder->canvas) return (holder);
    if (func_name)
    {
	fprintf (stderr, "CanvasHolder not found!\n");
	a_prog_bug (func_name);
    }
    return (NULL);
}   /*  End Function get_canvas_holder  */

static CanvasHolder _scatplt_alloc_canvas_holder (KWorldCanvas canvas)
/*  This routine will allocate a canvas holder for a world canvas.
    The canvas must be given by  canvas  .
    The routine returns the canvas holder on success, else it returns NULL.
*/
{
    CanvasHolder holder;
    KPixCanvas pixcanvas;
    unsigned short red, green, blue;
    double r, g, b;
    static char function_name[] = "_scatplt_alloc_canvas_holder";

    /*  Must allocate holder  */
    if ( ( holder = m_calloc (sizeof *holder) ) == NULL )
    {
	m_error_notify (function_name, "canvas holder");
	return (NULL);
    }
    holder->magic_number = HOLDER_MAGIC_NUMBER;
    holder->canvas = canvas;
    KLIST_APPEND2 (first_canvas_holder, last_canvas_holder, holder);
    canvas_register_precompute_func (canvas,
				     _scatplt_worldcanvas_precompute_func,
				     holder);
    canvas_register_refresh_func (canvas, _scatplt_worldcanvas_refresh_func,
				  holder);
    canvas_register_destroy_func (canvas, FALSE, _scatplt_free_canvas_holder,
				  holder);
    pixcanvas = canvas_get_pixcanvas (canvas);
    kwin_get_attributes (pixcanvas,
			 KWIN_ATT_BACKGROUND_PIXEL,
			 &holder->previous_background,
			 KWIN_ATT_END);
    kwin_get_pixel_RGB_values (pixcanvas,
			       &holder->previous_background,
			       &red, &green, &blue, 1);
    r = red;
    g = green;
    b = blue;
    holder->previous_intensity = sqrt (r * r + g * g + b * b) / 65535.0;
    return (holder);
}   /*  End Function alloc_canvas_holder  */

static void _scatplt_free_canvas_holder (CanvasHolder holder)
/*  [SUMMARY] Free a canvas holder (and any associated KScatterPlot objects).
    <holder> The canvas holder.
    [RETURNS] Nothing.
*/
{
    static char function_name[] = "_scatplt_free_canvas_holder";

    VERIFY_HOLDER (holder);
    while (holder->first) scatplt_destroy (holder->first);
    KLIST_REMOVE2 (first_canvas_holder, last_canvas_holder, holder);
    if (holder->colourname) m_free (holder->colourname);
    holder->magic_number = 0;
    m_free (holder);
}   /*  End Function _scatplt_free_canvas_holder  */

static void process_canvas_attributes (CanvasHolder holder, va_list argp)
/*  [PURPOSE] This routine will set the attributes for a canvas.
    <holder> The canvas holder.
    <argp> The optional list of parameter attribute-key attribute-value
    pairs.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    unsigned int att_key;
    char *colour;
    static char function_name[] = "_scatplt_process_canvas_attributes";

    while ( ( att_key = va_arg (argp, unsigned int) ) !=
	    SCATPLT_CANVAS_ATT_END )
    {
	switch (att_key)
	{
	  case SCATPLT_CANVAS_ATT_COLOURNAME:
	    colour = va_arg (argp, char *);
	    if (colour == NULL) colour = st_dup (DEFAULT_COLOUR);
	    else colour = st_dup (colour);
	    if (colour == NULL) m_abort (function_name, "colour name");
	    if (holder->colourname) m_free (holder->colourname);
	    holder->colourname = colour;
	    break;
	  default:
	    fprintf (stderr, "Illegal attribute key: %u\n", att_key);
	    a_prog_bug (function_name);
	}
    }
}   /*  End Function process_canvas_attributes  */

static void _scatplt_worldcanvas_precompute_func
    (KWorldCanvas canvas, int width, int height,
     CONST struct win_scale_type *win_scale, Kcolourmap cmap, flag cmap_resize,
     void **info, PostScriptPage pspage,
     unsigned int num_areas, KPixCanvasRefreshArea *areas)
/*  [PURPOSE] This routine processes a precompute event for a world canvas.
    <canvas> The world canvas.
    <width> The width of the canvas in pixels.
    <height> The height of the canvas in pixels.
    <win_scale> A pointer to the window scaling information.
    <cmap> The canvas colourmap.
    <cmap_resize> If TRUE, the refresh is due to a colourmap resize.
    <info> A pointer to the arbitrary information pointer.
    <pspage> The PostScriptPage object. If this is NULL, the refresh is *not*
    destined for a PostScript page.
    <num_areas> The number of areas that need to be refreshed. If this is
    0 then the entire pixel canvas needs to be refreshed.
    <areas> The list of areas that need to be refreshed. Note that these
    areas are given in pixel co-ordinates.
    [RETURNS] Nothing.
*/
{
    CanvasHolder holder = *info;
    KScatterPlot plot;
    KPixCanvas pixcanvas;
    unsigned long background_pixel;
    double resolution;
    CONST char *colourname;
    static char function_name[] = "_scatplt_worldcanvas_precompute_func";

    VERIFY_HOLDER (holder);
    if (canvas != holder->canvas)
    {
	fprintf (stderr, "Different canvas in canvas holder object\n");
	a_prog_bug (function_name);
    }
    pixcanvas = canvas_get_pixcanvas (canvas);
    kwin_get_attributes (pixcanvas,
			 KWIN_ATT_RESOLUTION, &resolution,
			 KWIN_ATT_BACKGROUND_PIXEL, &background_pixel,
			 KWIN_ATT_END);
    if (holder->previous_background != background_pixel)
    {
	unsigned short red, green, blue;
	double r, g, b;

	holder->previous_background = background_pixel;
	kwin_get_pixel_RGB_values (pixcanvas, &background_pixel,
				   &red, &green, &blue, 1);
	r = red;
	g = green;
	b = blue;
	holder->previous_intensity = sqrt (r * r + g * g + b * b) / 65535.0;
    }
    if ( (holder->previous_intensity >= 0.5) || pspage )
    {
	/*  Light background  */
	colourname = "black";
    }
    else
    {
	/*  Dark background  */
	colourname = holder->colourname ? holder->colourname : "white";
    }
    kwin_get_colour (pixcanvas, colourname, &holder->pixel_value,
		     NULL, NULL, NULL);
    if (resolution < 1.0) return;
    /*  Show each active plot  */
    for (plot = holder->first; plot; plot = plot->next)
    {
	if (plot->active) _scatplt_precompute_plot (plot, win_scale,
						    num_areas, areas);
    }
}   /*  End Function _scatplt_worldcanvas_precompute_func  */

static void _scatplt_precompute_plot (KScatterPlot plot,
				      CONST struct win_scale_type *win_scale,
				      unsigned int num_areas,
				      KPixCanvasRefreshArea *areas)
/*  [PURPOSE] This routine processes a precompute event for a world canvas.
    <plot> The KScatterPlot object.
    <win_scale> A pointer to the window scaling information.
    <num_areas> The number of areas that need to be refreshed. If this is
    0 then the entire pixel canvas needs to be refreshed.
    <areas> The list of areas that need to be refreshed. Note that these
    areas are given in pixel co-ordinates.
    [RETURNS] Nothing.
*/
{
    KWorldCanvas canvas = plot->canvas_holder->canvas;
    int i_xo, i_xp, i_yo, i_yp;
    unsigned int count, npixels, nwords;
    unsigned int num_points = plot->num_points;
    double p_xleft, p_xright, p_ytop, p_ybottom;
    CONST double *wx = plot->world_x;
    CONST double *wy = plot->world_y;
    double px[BUF_SIZE], py[BUF_SIZE], test_x[5], test_y[5];
    static char function_name[] = "_scatplt_precompute_plot";

    if (plot->num_points < 1000) return;
    /*  Check if the non-linear to pixel co-ordinate transformation has changed
     */
    for (count = 0, npixels = 0; (npixels < 5) && count < plot->num_points;
	 ++count)
    {
	if ( DOUBLE_IS_BLANK (plot->world_x[count]) ||
	     DOUBLE_IS_BLANK (plot->world_y[count]) ) continue;
	test_x[npixels] = plot->world_x[count];
	test_y[npixels] = plot->world_y[count];
	++npixels;
    }
    canvas_convert_from_canvas_coords (canvas, FALSE, FALSE, 5,
				       test_x, test_y, test_x, test_y);
    if ( m_cmp (plot->test_x, test_x, sizeof *test_x * 5) &&
	 m_cmp (plot->test_y, test_y, sizeof *test_y * 5) ) return;
    /*  Compute raster  */
    plot->test_x[0] = TOOBIG;
    canvas_get_attributes (plot->canvas_holder->canvas,
			   CANVAS_ATT_X_OFFSET, &i_xo,
			   CANVAS_ATT_Y_OFFSET, &i_yo,
			   CANVAS_ATT_X_PIXELS, &i_xp,
			   CANVAS_ATT_Y_PIXELS, &i_yp,
			   CANVAS_ATT_END);
    npixels = i_xp * i_yp;
    nwords = (npixels >> plot->raster.shift) + 1;
    if (nwords >= plot->raster.buflen)
    {
	if (plot->raster.bits) m_free (plot->raster.bits);
	if (plot->raster.indices) m_free (plot->raster.indices);
	plot->raster.indices = NULL;
	if ( ( plot->raster.bits =
	       m_alloc (nwords * sizeof *plot->raster.bits) ) == NULL )
	{
	    m_error_notify (function_name, "raster");
	    return;
	}
	if ( ( plot->raster.indices =
	       m_alloc (npixels * sizeof *plot->raster.indices) ) == NULL )
	{
	    m_error_notify (function_name, "array of indices");
	    return;
	}
	plot->raster.buflen = nwords;
    }
    m_clear (plot->raster.bits, nwords * sizeof *plot->raster.bits);
    p_xleft = i_xo;
    p_xright = i_xo + i_xp - 1;
    p_ytop = i_yo;
    p_ybottom = i_yo + i_yp - 1;
    while (num_points > 0)
    {
	unsigned int block_size;

	block_size = (num_points > BUF_SIZE) ? BUF_SIZE : num_points;
	canvas_convert_from_canvas_coords (plot->canvas_holder->canvas,
					   FALSE, FALSE, block_size,
					   wx, wy, px, py);
	for (count = 0; count < block_size; ++count)
	{
	    int shift = plot->raster.shift;
	    unsigned int mask = MASK_OF_SHIFT (plot->raster.shift);
	    unsigned int pos;
	    double x, y;

	    x = px[count];
	    y = py[count];
	    if ( (x < p_xleft) || (x > p_xright) ||
		 (y < p_ytop) || (y > p_ybottom) ) continue;
	    pos = ( (int) y - p_ytop ) * i_xp + (int) x - p_xleft;
	    plot->raster.bits[pos >> shift] |= 1 << (pos & mask);
	    plot->raster.indices[pos] = plot->num_points - num_points + count;
	}
	wx += block_size;
	wy += block_size;
	num_points -= block_size;
    }
    /*  Tag bitmap as valid  */
    m_copy (plot->test_x, test_x, sizeof *test_x * 5);
    m_copy (plot->test_y, test_y, sizeof *test_y * 5);
}   /*  End Function _scatplt_precompute_plot  */

static void _scatplt_worldcanvas_refresh_func(KWorldCanvas canvas,
					      int width, int height,
					      struct win_scale_type *win_scale,
					      Kcolourmap cmap,
					      flag cmap_resize,
					      void **info,
					      PostScriptPage pspage,
					      unsigned int num_areas,
					      KPixCanvasRefreshArea *areas,
					      flag *honoured_areas)
/*  [PURPOSE] This routine processes a refresh event for a world canvas.
    <canvas> The world canvas.
    <width> The width of the canvas in pixels.
    <height> The height of the canvas in pixels.
    <win_scale> A pointer to the window scaling information.
    <cmap> The canvas colourmap.
    <cmap_resize> If TRUE, the refresh is due to a colourmap resize.
    <info> A pointer to the arbitrary information pointer.
    <pspage> The PostScriptPage object. If this is NULL, the refresh is *not*
    destined for a PostScript page.
    <num_areas> The number of areas that need to be refreshed. If this is
    0 then the entire pixel canvas needs to be refreshed.
    <areas> The list of areas that need to be refreshed. Note that these
    areas are given in pixel co-ordinates.
    <honoured_areas> If the value TRUE is written here it is assumed the
    routine honoured the list of refresh areas and did not write outside
    these areas and hence the list of areas will be passed to subsequent
    registered refresh routines. If FALSE is written here (or nothing is
    written here), implying the routine refreshed the entire canvas,
    subsequent refresh routines will be told to refresh the entire canvas.
    [RETURNS] Nothing.
*/
{
    CanvasHolder holder = *info;
    KScatterPlot plot;
    static char function_name[] = "_scatplt_worldcanvas_refresh_func";

    VERIFY_HOLDER (holder);
    if (canvas != holder->canvas)
    {
	fprintf (stderr, "Different canvas in canvas holder object\n");
	a_prog_bug (function_name);
    }
    /*  Show each active plot  */
    for (plot = holder->first; plot; plot = plot->next)
    {
	if (plot->active)
	    _scatplt_refresh_plot (plot, win_scale, num_areas, areas);
    }
    *honoured_areas = TRUE;
}   /*  End Function _scatplt_worldcanvas_refresh_func  */

static void _scatplt_refresh_plot (KScatterPlot plot,
				   struct win_scale_type *win_scale,
				   unsigned int num_areas,
				   KPixCanvasRefreshArea *areas)
/*  [PURPOSE] This routine processes a refresh event for a world canvas.
    <plot> The KScatterPlot object.
    <win_scale> A pointer to the window scaling information.
    <num_areas> The number of areas that need to be refreshed. If this is
    0 then the entire pixel canvas needs to be refreshed.
    <areas> The list of areas that need to be refreshed. Note that these
    areas are given in pixel co-ordinates.
    [RETURNS] Nothing.
*/
{
    KPixCanvas pixcanvas;
    int i_xo, i_xp, i_yo, i_yp;
    unsigned int num_points = plot->num_points;
    double p_xleft, p_xright, p_ytop, p_ybottom, resolution;
    CONST double *wx = plot->world_x;
    CONST double *wy = plot->world_y;
    double tx[BUF_SIZE], ty[BUF_SIZE], px[BUF_SIZE], py[BUF_SIZE];
    /*static char function_name[] = "_scatplt_refresh_plot";*/

    pixcanvas = canvas_get_pixcanvas (plot->canvas_holder->canvas);
    kwin_get_attributes (pixcanvas,
			 KWIN_ATT_RESOLUTION, &resolution,
			 KWIN_ATT_END);
    if ( (resolution >= 1.0) && DOUBLE_IS_VALID (plot->test_x[0]) )
    {
	_scatplt_refresh_plot_fast (plot, win_scale, num_areas, areas);
	return;
    }
    canvas_get_attributes (plot->canvas_holder->canvas,
			   CANVAS_ATT_X_OFFSET, &i_xo,
			   CANVAS_ATT_Y_OFFSET, &i_yo,
			   CANVAS_ATT_X_PIXELS, &i_xp,
			   CANVAS_ATT_Y_PIXELS, &i_yp,
			   CANVAS_ATT_END);
    p_xleft = i_xo;
    p_xright = i_xo + i_xp - 1;
    p_ytop = i_yo;
    p_ybottom = i_yo + i_yp - 1;
    while (num_points > 0)
    {
	unsigned int count, num_to_draw;
	unsigned int block_size;

	block_size = (num_points > BUF_SIZE) ? BUF_SIZE : num_points;
	canvas_convert_from_canvas_coords (plot->canvas_holder->canvas,
					   FALSE, FALSE, block_size,
					   wx, wy, tx, ty);
	for (count = 0, num_to_draw = 0; count < block_size; ++count)
	{
	    double x, y;

	    x = tx[count];
	    y = ty[count];
	    if ( (x < p_xleft) || (x > p_xright) ||
		 (y < p_ytop) || (y > p_ybottom) ) continue;
	    if (num_areas > 0)
	    {
		int xi = x;
		int yi = y;
		unsigned int area_count;

		for (area_count = 0; area_count < num_areas; ++area_count)
		{
		    if ( (areas[area_count].startx <= xi) &&
			 (xi <= areas[area_count].endx) &&
			 (areas[area_count].starty <= yi) &&
			 (yi <= areas[area_count].endy) ) break;
		}
		if (area_count >= num_areas) continue;
	    }
	    px[num_to_draw] = x;
	    py[num_to_draw] = y;
	    ++num_to_draw;
	}
	kwin_draw_points (pixcanvas, px, py, num_to_draw,
			  plot->canvas_holder->pixel_value);
	wx += block_size;
	wy += block_size;
	num_points -= block_size;
    }
}   /*  End Function refresh_plot  */

static void _scatplt_refresh_plot_fast (KScatterPlot plot,
					struct win_scale_type *win_scale,
					unsigned int num_areas,
					KPixCanvasRefreshArea *areas)
/*  [PURPOSE] Fast refresh using raster.
    <plot> The KScatterPlot object.
    <win_scale> A pointer to the window scaling information.
    <num_areas> The number of areas that need to be refreshed. If this is
    0 then the entire pixel canvas needs to be refreshed.
    <areas> The list of areas that need to be refreshed. Note that these
    areas are given in pixel co-ordinates.
    [RETURNS] Nothing.
*/
{
    KPixCanvas pixcanvas;
    int i_xo, i_xp, i_yo, i_yp, x, y, xi, yi;
    int shift = plot->raster.shift;
    unsigned int mask = MASK_OF_SHIFT (plot->raster.shift);
    unsigned int num_points = 0;
    double px[BUF_SIZE], py[BUF_SIZE];
    /*static char function_name[] = "_scatplt_refresh_plot_fast";*/

    pixcanvas = canvas_get_pixcanvas (plot->canvas_holder->canvas);
    canvas_get_attributes (plot->canvas_holder->canvas,
			   CANVAS_ATT_X_OFFSET, &i_xo,
			   CANVAS_ATT_Y_OFFSET, &i_yo,
			   CANVAS_ATT_X_PIXELS, &i_xp,
			   CANVAS_ATT_Y_PIXELS, &i_yp,
			   CANVAS_ATT_END);
    if (num_areas > 0)
    {
	unsigned int area_count;
	KPixCanvasRefreshArea *area = areas;

	for (area_count = 0; area_count < num_areas; ++area_count, ++area)
	{
	    for (y = area->starty; y <= area->endy; ++y)
		for (x = area->startx; x <= area->endx; ++x)
		{
		    int pos;

		    if ( (x < i_xo) || (y < i_yo) ) continue;
		    xi = x - i_xo;
		    yi = y - i_yo;
		    if ( (xi >= i_xp) || (yi >= i_yp) ) continue;
		    pos = yi * i_xp + xi;
		    if ( !( plot->raster.bits[pos >> shift] &
			    ( 1 << (pos & mask) ) ) )
			continue;
		    if (num_points >= BUF_SIZE)
		    {
			kwin_draw_points (pixcanvas, px, py, num_points,
					  plot->canvas_holder->pixel_value);
			num_points = 0;
		    }
		    px[num_points] = x;
		    py[num_points] = y;
		    ++num_points;
		}
	}
    }
    else
    {
	for (yi = 0; yi < i_yp; ++yi) for (xi = 0; xi < i_xp; ++xi)
	{
	    int pos;

	    x = xi + i_xo;
	    y = yi + i_yo;
	    pos = yi * i_xp + xi;
	    if ( !( plot->raster.bits[pos >> shift] &
		    ( 1 << (pos & mask) ) ) ) continue;
	    if (num_points >= BUF_SIZE)
	    {
		kwin_draw_points (pixcanvas, px, py, num_points,
				  plot->canvas_holder->pixel_value);
		num_points = 0;
	    }
	    px[num_points] = x;
	    py[num_points] = y;
	    ++num_points;
	}
    }
    if (num_points > 0)
	kwin_draw_points (pixcanvas, px, py, num_points,
			  plot->canvas_holder->pixel_value);
    }   /*  End Function refresh_plot_fast  */

static flag _scatplt_find_closest_slow (KScatterPlot plot,
					double xin, double yin,
					double *xout, double *yout,
					uaddr *index,
					flag subpixel, double search_radius)
/*  [SUMMARY] Find the closest point to a specified position.
    <plot> The scatter plot.
    <xin> The X world co-ordinate to search from.
    <yin> The Y world co-ordinate to search from.
    <xout> The exact X co-ordinate of the point found.
    <yout> The exact Y co-ordinate of the point found.
    <subpixel> If TRUE, search with subpixel resolution so that multiple points
    lying on the same screen pixel will be searched to find the closest point.
    If FALSE, the search will terminate early on the first point which lies on
    the same screen pixel.
    <search_radius> The maximum search radius. Points further from the search
    position will be ignored.
    [RETURNS] TRUE if a close point was found, else FALSE.
*/
{
    unsigned int closest_index = plot->num_points;
    unsigned int num_points = plot->num_points;
    double closest_distance2 = TOOBIG;
    double xin_p, yin_p;
    CONST double *wx = plot->world_x;
    CONST double *wy = plot->world_y;
    double px[BUF_SIZE], py[BUF_SIZE];
    /*static char function_name[] = "_scatplt_find_closest_slow";*/

    if ( DOUBLE_IS_VALID (search_radius) ) search_radius *= search_radius;
    canvas_convert_from_canvas_coords (plot->canvas_holder->canvas,
				       FALSE, FALSE,
				       1, &xin, &yin, &xin_p, &yin_p);
    while (num_points > 0)
    {
	unsigned int block_size, count;

	block_size = (num_points > BUF_SIZE) ? BUF_SIZE : num_points;
	canvas_convert_from_canvas_coords (plot->canvas_holder->canvas,
					   FALSE, FALSE, block_size,
					   wx, wy, px, py);
	for (count = 0; count < block_size; ++count)
	{
	    double x, y, distance2;

	    x = px[count];
	    y = py[count];
	    if ( DOUBLE_IS_BLANK (x) || DOUBLE_IS_BLANK (y) ) continue;
	    distance2 = (x - xin_p) * (x - xin_p) + (y - yin_p) * (y - yin_p);
	    if (distance2 > search_radius) continue;
	    if (distance2 >= closest_distance2) continue;
	    closest_distance2 = distance2;
	    closest_index = plot->num_points - num_points + count;
	    if ( !subpixel &&
		 (fabs (x - xin_p) < 1.0) && (fabs (y - yin_p) < 1.0) )
	    {
		*xout = wx[count];
		*yout = wy[count];
		*index = closest_index;
		return (TRUE);
	    }
	}
	wx += block_size;
	wy += block_size;
	num_points -= block_size;
    }
    if (closest_index >= plot->num_points) return (FALSE);
    *index = closest_index;
    *xout = plot->world_x[closest_index];
    *yout = plot->world_y[closest_index];
    return (TRUE);
}   /*  End Function _scatplt_find_closest_slow  */
