/*LINTLIBRARY*/
/*  misc.c

    This code provides miscellaneous routines for scatter plots.

    Copyright (C) 2001-2002  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This file contains miscellaneous routines for scatter plots.


    Written by      Richard Gooch   5-DEC-2001

    Updated by      Richard Gooch   6-DEC-2001: Took account of interface
  change to <wcs_astro_transform*>. Made use of new function
  <wcs_compare_grids_2D>.

    Updated by      Richard Gooch   10-DEC-2001: Documented interface.

    Updated by      Richard Gooch   14-DEC-2001: Created function
  <scatplt_create_smart_cursor>.

    Updated by      Richard Gooch   14-DEC-2001: Deal with event
  K_CANVAS_EVENT_POINTER_LEAVE in function <_scatplt_cursor_position_func>.

    Updated by      Richard Gooch   19-DEC-2001: Discard duplicate events in
  function <_scatplt_cursor_position_func>. Added scaling parameters to
  function <scatplt_create_vv>.

    Updated by      Richard Gooch   23-JUL-2002: Trap for no overlap or all
  blanks in <scatplt_create_vv>.

    Last updated by Richard Gooch   13-DEC-2002: Check for failed transforms.
  Changed interface to function <scatplt_create_vv> so that restriction
  information is passed in.


*/
#include <stdio.h>
#include <math.h>
#include <unistd.h>
#include <karma.h>
#include <k_event_codes.h>
#define NEW_WIN_SCALE
#include <karma_scatplt.h>
#include <karma_canvas.h>
#include <karma_iarray.h>
#include <karma_kwin.h>
#include <karma_wcs.h>
#include <karma_ds.h>
#include <karma_st.h>
#include <karma_a.h>
#include <karma_m.h>


#define CURSOR_RADIUS 4


/*  Private structures  */

struct smart_cursor_struct
{
    flag (*func) (KScatterPlot plot, double x, double y, uaddr index,
		  void **info);
    void *info;
    int old_x, old_y;
    flag old_found;
    flag old_index;
    double search_radius;
};


/*  Private functions  */
STATIC_FUNCTION (flag _scatplt_cursor_position_func,
		 (KWorldCanvas canvas, double x, double y,
		  unsigned int event_code, void *e_info, void **f_info,
		  double x_lin, double y_lin, int x_pix, int y_pix) );


/* Public functions follow */

/*EXPERIMENTAL_FUNCTION*/
KScatterPlot scatplt_create_vv (KWorldCanvas canvas,
				CONST char *array1,CONST array_desc *arr_desc1,
				unsigned int xdim1, unsigned int ydim1,
				unsigned int type1,
				double scale1, double offset1, KwcsAstro ap1,
				unsigned int num_restr1,
				CONST char **restr_names1,
				CONST double *restr_values1,
				double *min1, double *max1,
				uaddr xstart, uaddr xlen,
				uaddr ystart, uaddr ylen,
				CONST char *array2,CONST array_desc *arr_desc2,
				unsigned int xdim2, unsigned int ydim2,
				unsigned int type2,
				double scale2, double offset2, KwcsAstro ap2,
				unsigned int num_restr2,
				CONST char **restr_names2,
				CONST double *restr_values2,
				double *min2, double *max2)
/*  [SUMMARY] Create a value-value scatter plot from two 2D sub-arrays.
    <canvas> The world canvas onto which the scatter plot may be drawn.
    <array1> The data for the first array.
    <arr_desc1> The array descriptor for the first array.
    <xdim1> The X dimension index for the first array.
    <ydim1> The Y dimension index for the first array.
    <type1> The data type for the first array.
    <scale1> The data scaling for the first array.
    <offset1> The data offset for the first array.
    <ap1> The KwcsAstro object for the first array.
    <num_restr1> The number of restrictions for the first array.
    <restr_names1> The array of restriction names for the first array.
    <restr_values1> The restriction values for the first array.
    <min1> The minimum value from the first array. This is updated.
    <max1> The maximum value from the first array. This is updated.
    <xstart> The X dimension start of the sub-array in the first array.
    <xlen> The X dimension length of the sub-array in the first array.
    <ystart> The Y dimension start of the sub-array in the first array.
    <ylen> The Y dimension length of the sub-array in the first array.
    <array2> The data for the second array.
    <arr_desc2> The array descriptor for the second array.
    <xdim2> The X dimension index for the second array.
    <ydim2> The Y dimension index for the second array.
    <type2> The data type for the second array.
    <scale2> The data scaling for the second array.
    <offset2> The data offset for the second array.
    <ap2> The KwcsAstro object for the second array.
    <num_restr> The number of restrictions for the second array.
    <restr_names> The array of restriction names for the second array.
    <restr_values> The restriction values for the second array.
    <min2> The minimum value from the second array. This is updated.
    <max2> The maximum value from the second array. This is updated.
    [NOTE] The KScatterPlot will be automatically destroyed if the canvas is
    destroyed.
    [RETURNS] A KScatterPlot object on success, NULL.
*/
{
    KScatterPlot plot;
    flag have_overlap = FALSE;
    flag same_grid;
    unsigned int xi, yi;
    double xp, yp;
    dim_desc *xd1, *yd1, *xd2, *yd2;
    double *dbuffer, *xcoords, *ycoords, *values;
    double *values_buf, *values1, *values2;
    static char function_name[] = "scatplt_create_vv";

    if ( (xlen < 2) || (ylen < 2) )
    {
	fprintf (stderr, "%s: box1 (%lu*%lu) too small\n",
		 function_name, xlen, ylen);
	return (NULL);
    }
    xd1 = arr_desc1->dimensions[xdim1];
    yd1 = arr_desc1->dimensions[ydim1];
    xd2 = arr_desc2->dimensions[xdim2];
    yd2 = arr_desc2->dimensions[ydim2];
    if ( !wcs_compare_grids_2D (ap1, xd1, yd1, num_restr1, restr_names1,
				restr_values1,
				ap2, xd2, yd2, num_restr2, restr_names2,
				restr_values2,
				&same_grid) )
	return (NULL);
    if ( ( values_buf = m_alloc (sizeof *values1 * xlen * ylen * 2) ) == NULL )
    {
	m_error_notify (function_name, "scatter plot values");
	return (NULL);
    }
    values1 = values_buf;
    values2 = values1 + xlen * ylen;
    dbuffer = m_alloc_scratch (sizeof *dbuffer * xlen * 4, function_name);
    xcoords = dbuffer;
    ycoords = xcoords + xlen;
    values = ycoords + xlen;
    /*  Construct list of value pairs from 2D sub-arrays, by iterating over
	array1 and finding matching point in array2, then adding to list  */
    if (same_grid)
    {   /*  Same grids: optimise  */
	for (yi = 0; yi < ylen; ++yi, values1 += xlen, values2 += xlen)
	{
	    ds_get_scattered_elements
		(array1 + arr_desc1->offsets[ydim1][ystart + yi],
		 type1,
		 arr_desc1->offsets[xdim1] + xstart, dbuffer, NULL, xlen);
	    ds_get_scattered_elements
		(array2 + arr_desc2->offsets[ydim2][ystart + yi],
		 type2,
		 arr_desc2->offsets[xdim2] + xstart, values, NULL, xlen);
	    for (xi = 0; xi < xlen; ++xi)
	    {
		xp = dbuffer[xi * 2];
		yp = values[xi * 2];
		if ( DOUBLE_IS_BLANK (xp) || DOUBLE_IS_BLANK (yp) )
		{
		    values1[xi] = TOOBIG;
		    values2[xi] = TOOBIG;
		    continue;
		}
		xp = scale1 * xp + offset1;
		yp = scale2 * yp + offset2;
		values1[xi] = xp;
		values2[xi] = yp;
		if (xp < *min1) *min1 = xp;
		if (xp > *max1) *max1 = xp;
		if (yp < *min2) *min2 = yp;
		if (yp > *max2) *max2 = yp;
	    }
	}
	have_overlap = TRUE;
    }
    else
    {
	for (yi = 0; yi < ylen; ++yi, values1 += xlen, values2 += xlen)
	{
	    ds_get_scattered_elements
		(array1 + arr_desc1->offsets[ydim1][ystart + yi],
		 type1,
		 arr_desc1->offsets[xdim1] + xstart, values, NULL, xlen);
	    for (xi = 0; xi < xlen; ++xi)
	    {
		double xp = values[xi * 2];

		xcoords[xi] = xi + xstart;
		ycoords[xi] = yi + ystart;
		if ( DOUBLE_IS_BLANK (xp) ) values1[xi] = TOOBIG;
		else values1[xi] = scale1 * xp + offset1;
	    }
	    ds_convert_coordinates (xd1, xlen, xcoords, FALSE, FALSE, FALSE);
	    ds_convert_coordinates (yd1, xlen, ycoords, FALSE, FALSE, FALSE);
	    if (ap1)  /*  Convert pixel coords in array1 to world coords  */
		if ( !wcs_astro_transform3 (ap1, FALSE, xlen,
					    xd1->name, xcoords, FALSE,
					    yd1->name, ycoords, FALSE,
					    NULL, NULL, FALSE,
					    num_restr1, restr_names1,
					    restr_values1) )
		{
		    m_free_scratch (dbuffer);
		    m_free (values_buf);
		    fprintf (stderr,
			     "%s: failure transforming to world coords\n",
			     function_name);
		    return (NULL);
		}
	    if (ap2)  /*  Convert world coords to pixel coords in array2  */
		if ( !wcs_astro_transform3 (ap2, FALSE, xlen,
					    xd2->name, xcoords, TRUE,
					    yd2->name, ycoords, TRUE,
					    NULL, NULL, FALSE,
					    num_restr2, restr_names2,
					    restr_values2) )
		{
		    m_free_scratch (dbuffer);
		    m_free (values_buf);
		    fprintf (stderr,
			     "%s: failure transforming from world coords\n",
			     function_name);
		    return (NULL);
		}
	    ds_convert_coordinates (xd2, xlen, xcoords, TRUE, TRUE, TRUE);
	    ds_convert_coordinates (yd2, xlen, ycoords, TRUE, TRUE, TRUE);
	    for (xi = 0; xi < xlen; ++xi)
	    {
		xp = xcoords[xi];
		yp = ycoords[xi];
		if ( DOUBLE_IS_BLANK (xp) || DOUBLE_IS_BLANK (yp) )
		{
		    values1[xi] = TOOBIG;
		    values2[xi] = TOOBIG;
		    continue;
		}
		have_overlap = TRUE;
		if ( DOUBLE_IS_BLANK (values1[xi]) )
		{
		    values1[xi] = TOOBIG;
		    values2[xi] = TOOBIG;
		    continue;
		}
		ds_get_element (array2 +
				arr_desc2->offsets[ydim2][(long) yp] +
				arr_desc2->offsets[xdim2][(long) xp],
				type2,
				values, NULL);
		if ( DOUBLE_IS_BLANK (values[0]) )
		{   /*  Convert possible NaN to TOOBIG  */
		    values1[xi] = TOOBIG;
		    values2[xi] = TOOBIG;
		    continue;
		}
		values2[xi] = scale2 * values[0] + offset2;
		if (values1[xi] < *min1) *min1 = values1[xi];
		if (values1[xi] > *max1) *max1 = values1[xi];
		if (values2[xi] < *min2) *min2 = values2[xi];
		if (values2[xi] > *max2) *max2 = values2[xi];
	    }
	}
    }
    m_free_scratch (dbuffer);
    if (!have_overlap)
    {
	fprintf (stderr, "%s: no overlap\n", function_name);
	m_free (values_buf);
	return (NULL);
    }
    if (*max1 <= *min1) 
    {
	fprintf (stderr, "%s: no values\n", function_name);
	m_free (values_buf);
	return (NULL);
    }
    values1 = values_buf;
    values2 = values1 + xlen * ylen;
    plot = scatplt_create (canvas, xlen * ylen, values1, values2, 0);
    scatplt_register_destroy_func (plot, m_free, values_buf);
    return (plot);
}   /*  End Function scatplt_create_vv  */

/*EXPERIMENTAL_FUNCTION*/
void scatplt_create_smart_cursor (KWorldCanvas canvas, double search_radius,
				  flag (*func) (), void *info)
/*  [SUMMARY] Create a smart cursor.
    <canvas> The world canvas.
    <search_radius> The maximum search radius. Points further from the search
    position will be ignored.
    <func> The function that is called when the plot is destroyed. The
    prototype function is [<scatplt_PROTO_cursor_func>].
    <info> The arbitrary function information pointer.
    [RETURNS] Nothing.
*/
{
    struct smart_cursor_struct *cursor;
    static char function_name[] = "scatplt_create_smart_cursor";

    if ( ( cursor = m_calloc (sizeof *cursor) ) == NULL )
	m_abort (function_name, "smart cursor data");
    cursor->func = func;
    cursor->info = info;
    cursor->old_x = -CURSOR_RADIUS;
    cursor->old_y = -CURSOR_RADIUS;
    cursor->search_radius = search_radius;
    canvas_register_destroy_func (canvas, FALSE, m_free, cursor);
    canvas_register_position_event_func (canvas, _scatplt_cursor_position_func,
					 cursor);
}   /*  End Function scatplt_create_smart_cursor  */


/*  Private functions follow  */

static flag _scatplt_cursor_position_func (KWorldCanvas canvas,
					   double x, double y,
					   unsigned int event_code,
					   void *e_info, void **f_info,
					   double x_lin, double y_lin,
					   int x_pix, int y_pix)
/*  [SUMMARY] Position event callback.
    [PURPOSE] This routine is a position event consumer for a world canvas.
    <canvas> The canvas on which the event occurred.
    <x> The horizontal world co-ordinate of the event.
    <y> The vertical world co-ordinate of the event.
    <event_code> The arbitrary event code.
    <e_info> A pointer to arbitrary event information.
    <f_info> A pointer to an arbitrary function information pointer.
    <x_lin> The horizontal linear world co-ordinate prior to the transform
    function being called.
    <y_lin> The vertical linear world co-ordinate prior to the transform
    function being called.
    <x_pix> The horizontal pixel co-ordinate.
    <y_pix> The vertical pixel co-ordinate.
    [NOTE] The world co-ordinate values will have been transformed by the
    registered transform function (see [<canvas_register_transforms_func>]).
    [RETURNS] TRUE if the event was consumed, else FALSE indicating that
    the event is still to be processed.
*/
{
    KScatterPlot plot;
    KPixCanvasRefreshArea area;
    KPixCanvas pixcanvas;
    flag found = FALSE;
    uaddr index;
    unsigned long pixel_value;
    double xout, yout;
    struct smart_cursor_struct *cursor = *f_info;

    if ( (event_code != K_CANVAS_EVENT_POINTER_MOVE) &&
	 (event_code != K_CANVAS_EVENT_POINTER_WARP) &&
	 (event_code != K_CANVAS_EVENT_POINTER_LEAVE) ) return (FALSE);
    pixcanvas = canvas_get_pixcanvas (canvas);
    /*  Precompute as much as possible before clearing  */
    if (event_code != K_CANVAS_EVENT_POINTER_LEAVE)
    {
	found = scatplt_find_closest_all (canvas, x, y, &xout, &yout, &index,
					  &plot, FALSE, cursor->search_radius);
    }
    if (found)
    {   /*  Discard event on same point  */
	if ( cursor->old_found && (cursor->old_index == index) ) return FALSE;
	canvas_convert_from_canvas_coords (canvas, FALSE, FALSE, 1,
					   &xout, &yout, &x, &y);
	kwin_get_colour (pixcanvas, "red", &pixel_value, NULL, NULL, NULL);
    }
    else if (!cursor->old_found) return (FALSE);  /* Discard redundant event */
    cursor->old_found = found;
    cursor->old_index = index;
    if (cursor->old_x > -CURSOR_RADIUS)
    {
	m_clear (&area, sizeof area);
	area.startx = cursor->old_x - CURSOR_RADIUS;
	area.endx = cursor->old_x + CURSOR_RADIUS;
	area.starty = cursor->old_y - CURSOR_RADIUS;
	area.endy = cursor->old_y + CURSOR_RADIUS;
	area.geom_clear = TRUE;
	kwin_partial_refresh (pixcanvas, 1, &area, FALSE);
	cursor->old_x = -CURSOR_RADIUS;
    }
    if (!found) return (*cursor->func) (NULL, TOOBIG, TOOBIG, 0,&cursor->info);
    cursor->old_x = x + 1e-3;
    cursor->old_y = y + 1e-3;
    kwin_draw_ellipse (pixcanvas, cursor->old_x, cursor->old_y,
		       CURSOR_RADIUS, CURSOR_RADIUS, pixel_value);
    return (*cursor->func) (plot, xout, yout, index, &cursor->info);
}   /*  End Function _scatplt_cursor_position_func  */
