/*LINTLIBRARY*/
/*  colourbar.c

    This code provides colour support for ViewableImage objects.

    Copyright (C) 2002  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This file contains convenience routines which implement a colourbar.


    Written by      Richard Gooch   31-JAN-2002

    Updated by      Richard Gooch   5-FEB-2002: First working version.

    Updated by      Richard Gooch   8-FEB-2002: Applied data scaling. Disable
  intensity scale for non-linear intensity scale functions.

    Last updated by Richard Gooch   3-SEP-2002: Added call do <kwin_draw_point>
  to flush any pending geom_clear.


*/
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <karma.h>
#define NEW_WIN_SCALE
#include <karma_viewimg.h>
#include <karma_canvas.h>
#include <karma_kcmap.h>
#include <karma_kwin.h>
#include <karma_kwin_hooks.h>
#include <karma_psw_def.h>
#include <karma_c.h>
#include <karma_m.h>


#define COLOURBAR_OFFSET    10
#define COLOURBAR_WIDTH     17
#define BUFLEN            4096

/*  Private structures follow  */

struct colourbar_struct
{
    int cache_width;
    int cache_height;
    KPixCanvasImageCache cache;
    flag cache_valid;
    KWorldCanvas main_wc;
};


/*  Private functions  */
STATIC_FUNCTION (void _viewimg_colourbar_reservation_func,
		 (KWorldCanvas canvas, void **info,
		  int *left_margin, int *right_margin,
		  int *bottom_margin, int *top_margin) );
STATIC_FUNCTION (void _viewimg_colourbar_update_attributes,
		 (KWorldCanvas colourbar_wc, KWorldCanvas main_wc) );
STATIC_FUNCTION (void _viewimg_colourbar_size_control_func,
		 (KWorldCanvas canvas, int width, int height,
		  struct win_scale_type *win_scale, void **info,
		  flag *boundary_clear, flag *force) );
STATIC_FUNCTION (void _viewimg_colourbar_precompute_func,
		 (KWorldCanvas canvas, int width, int height,
		  CONST struct win_scale_type *win_scale,
		  Kcolourmap cmap, flag cmap_resize,
		  void **info, PostScriptPage pspage,
		  unsigned int num_areas, KPixCanvasRefreshArea *areas) );
STATIC_FUNCTION (void _viewimg_colourbar_refresh_func,
		 (KWorldCanvas canvas, int width, int height,
		  CONST struct win_scale_type *win_scale,
		  Kcolourmap cmap, flag cmap_resize, void **info,
		  PostScriptPage pspage,
		  unsigned int num_areas,
		  KPixCanvasRefreshArea *areas,
		  flag *honoured_areas) );
STATIC_FUNCTION (void _viewimg_colourbar_destroy_func, (void *info) );


/*  Public functions follow  */

/*EXPERIMENTAL_FUNCTION*/
KWorldCanvas viewimg_colourbar_create (KWorldCanvas main_wc)
/*  [SUMMARY] Create a colourbar world canvas for another world canvas.
    <main_wc> The main world canvas to create the colourbar world canvas for.
    [RETURNS] The colourbar world canvas on success, else NULL. This will be
    automatically destroyed when the main world canvas is destroyed.
*/
{
    KWorldCanvas colourbar_wc;
    KPixCanvas main_pc;
    KCallbackFunc cb;
    struct colourbar_struct *info;
    static char function_name[] = "viewimg_colourbar_create";

    main_pc = canvas_get_pixcanvas (main_wc);
    colourbar_wc = canvas_create (main_pc, canvas_get_cmap (main_wc), NULL);
    if (colourbar_wc == NULL) return (NULL);
    canvas_register_destroy_func (main_wc, TRUE, canvas_destroy, colourbar_wc);
    cb = canvas_register_legend_reservation_func
	(main_wc, _viewimg_colourbar_reservation_func, colourbar_wc);
    canvas_register_destroy_func (colourbar_wc, TRUE, c_unregister_callback,
				  cb);
    canvas_set_dressing (colourbar_wc,
			 KCD_ATT_DISPLAY, TRUE,
			 KCD_ATT_LEFT_TICKS, TRUE,
			 KCD_ATT_RIGHT_TICKS, TRUE,
			 KCD_ATT_SHOW_LABELS, TRUE,
			 KCD_ATT_RIGHT_SCALE, TRUE,
			 KCD_ATT_MAJOR_PIXELS, 4,
			 KCD_ATT_MINOR_PIXELS, 2,
			 KCD_ATT_END);
    canvas_register_size_control_func (colourbar_wc,
				       _viewimg_colourbar_size_control_func,
				       main_wc);
    
    if ( ( info = m_calloc (sizeof *info) ) == NULL )
	m_abort (function_name, "struct_colourbar");
    info->main_wc = main_wc;
    canvas_register_precompute_func (colourbar_wc,
				     _viewimg_colourbar_precompute_func,
				     info);
    canvas_register_refresh_func (colourbar_wc,
				  _viewimg_colourbar_refresh_func, info);
    canvas_register_destroy_func (colourbar_wc, FALSE,
				  _viewimg_colourbar_destroy_func, info);
    return (colourbar_wc);
}   /*  End Function viewimg_colourbar_create  */


/*  Private functions follow  */


static void _viewimg_colourbar_reservation_func (KWorldCanvas canvas,
						 void **info,
						 int *left_margin,
						 int *right_margin,
						 int *bottom_margin,
						 int *top_margin)
/*  [SUMMARY] Legend space reservation callback.
    [PURPOSE] This routine reserves space on the main canvas for the colourbar.
    <canvas> The canvas on which the event occurred.
    <info> A pointer to the arbitrary transform information pointer.
    <left_margin> The number of pixels required for the left margin must be
    written here. If nothing is written, 0 is assumed.
    <right_margin> The number of pixels required for the right margin must be
    written here. If nothing is written, 0 is assumed.
    <bottom_margin> The number of pixels required for the bottom margin must be
    written here. If nothing is written, 0 is assumed.
    <top_margin> The number of pixels required for the top margin must be
    written here. If nothing is written, 0 is assumed.
    [RETURNS] Nothing.
*/
{
    KWorldCanvas colourbar_wc = *info;
    int left, right, bottom, top;

    /*  Must ensure that colourbar canvas attributes are up to date before
	computing its size. Thus it must be done here  */
    _viewimg_colourbar_update_attributes (colourbar_wc, canvas);
    canvas_get_margins (colourbar_wc, &left, &right, &bottom, &top);
    *right_margin = left + right + COLOURBAR_OFFSET + COLOURBAR_WIDTH;
}   /*  End Function _viewimg_colourbar_reservation_func  */

static void _viewimg_colourbar_update_attributes (KWorldCanvas colourbar_wc,
						  KWorldCanvas main_wc)
/*  [SUMMARY] Update colourbar canvas attributes.
    <colourbar_wc> The colourbar world canvas.
    <main_wc> The main world canvas.
    [RETURNS] Nothing.
*/
{
    ViewableImage vimage;
    flag truecolour, linear;
    unsigned int index;
    double bottom_y, top_y, scale, offset;
    array_desc *arr_desc;
    char *xlabel;
    void *iscale_func;
    /*static char function_name[] = "_viewimg_colourbar_update_attributes";*/

    vimage = viewimg_get_active (main_wc);
    if (vimage == NULL) return;
    viewimg_get_attributes (vimage,
			    VIEWIMG_VATT_ARRAY_DESC, &arr_desc,
			    VIEWIMG_VATT_TRUECOLOUR, &truecolour,
			    VIEWIMG_VATT_END);
    if (truecolour) return;
    canvas_get_attributes (main_wc,
			   CANVAS_ATT_VALUE_MIN, &bottom_y,
			   CANVAS_ATT_VALUE_MAX, &top_y,
			   CANVAS_ATT_ISCALE_FUNC, &iscale_func,
			   CANVAS_ATT_END);
    viewimg_get_attributes (vimage,
			    VIEWIMG_VATT_PSEUDO_INDEX, &index,
			    VIEWIMG_VATT_DATA_SCALE, &scale,
			    VIEWIMG_VATT_DATA_OFFSET, &offset,
			    VIEWIMG_VATT_END);
    bottom_y = bottom_y * scale + offset;
    top_y = top_y * scale + offset;
    canvas_set_attributes (colourbar_wc,
			   CANVAS_ATT_BOTTOM_Y, bottom_y,
			   CANVAS_ATT_TOP_Y, top_y,
			   CANVAS_ATT_END);
    canvas_get_specification (colourbar_wc, &xlabel, NULL, NULL, NULL, NULL);
    if ( (xlabel == NULL) ||
	 (strcmp (xlabel, arr_desc->packet->element_desc[index]) != 0) )
	canvas_specify (colourbar_wc, NULL,
			arr_desc->packet->element_desc[index], 0, NULL, NULL);
    linear = iscale_func ? FALSE : TRUE;
    canvas_set_dressing (colourbar_wc,
			 KCD_ATT_DISPLAY, TRUE,
			 KCD_ATT_LEFT_TICKS, linear,
			 KCD_ATT_RIGHT_TICKS, linear,
			 KCD_ATT_SHOW_LABELS, linear,
			 KCD_ATT_END);
}   /*  End Function _viewimg_colourbar_update_attributes  */

static void _viewimg_colourbar_size_control_func
    (KWorldCanvas canvas, int width, int height,
     struct win_scale_type *win_scale, void **info, flag *boundary_clear,
     flag *force)
/*  [SUMMARY] Modify the window scaling information for a world canvas.
    <canvas> The world canvas.
    <width> The width of the canvas in pixels.
    <height> The height of the canvas in pixels.
    <win_scale> The window scaling information. This may be modified.
    <info> A pointer to arbitrary canvas information.
    <boundary_clear> If the value TRUE is written here then the canvas_resize
    routine will attempt to clear only the boundary between the pixel canvas
    and the world canvas. If the value FALSE is written here or nothing is
    written here, the canvas_resize routine will clear the entire pixel canvas
    as appropriate.
    <force> If the value TRUE is written here, the location is forced. If the
    value FALSE is written here, or nothing is written here, the canvas
    boundary may be centred.
    [RETURNS] Nothing.
*/
{
    KWorldCanvas main_wc = *info;
    int xoffset, yoffset, xpixels, ypixels;
    /*static char function_name[] = "_viewimg_colourbar_size_control_func";*/

    canvas_get_attributes (main_wc,
			   CANVAS_ATT_X_OFFSET, &xoffset,
			   CANVAS_ATT_Y_OFFSET, &yoffset,
			   CANVAS_ATT_X_PIXELS, &xpixels,
			   CANVAS_ATT_Y_PIXELS, &ypixels,
			   CANVAS_ATT_END);
    win_scale->x_offset = xoffset + xpixels + COLOURBAR_OFFSET;
    win_scale->x_pixels = COLOURBAR_WIDTH;
    win_scale->y_offset = yoffset;
    win_scale->y_pixels = ypixels;
    *force = TRUE;
}   /*  End Function _viewimg_colourbar_size_control_func  */

static void _viewimg_colourbar_precompute_func
    (KWorldCanvas canvas, int width, int height,
     CONST struct win_scale_type *win_scale, Kcolourmap cmap, flag cmap_resize,
     void **info, PostScriptPage pspage,
     unsigned int num_areas, KPixCanvasRefreshArea *areas)
/*  [PURPOSE] This routine processes a precompute event for a world canvas.
    <canvas> The world canvas.
    <width> The width of the canvas in pixels.
    <height> The height of the canvas in pixels.
    <win_scale> A pointer to the window scaling information.
    <cmap> The canvas colourmap.
    <cmap_resize> If TRUE, the refresh is due to a colourmap resize.
    <info> A pointer to the arbitrary information pointer.
    <pspage> The PostScriptPage object. If this is NULL, the refresh is *not*
    destined for a PostScript page.
    <num_areas> The number of areas that need to be refreshed. If this is
    0 then the entire pixel canvas needs to be refreshed.
    <areas> The list of areas that need to be refreshed. Note that these
    areas are given in pixel co-ordinates.
    [RETURNS] Nothing.
*/
{
    KPixCanvas pixcanvas;
    unsigned int count, num_pixels;
    uaddr hoffset = 0;
    double font_scale;
    char *grid_colour, *scale_colour, *label_colour;
    struct colourbar_struct *colourbar = *info;
    void *draw_cached_image_func;
    unsigned long *pixel_values;
    unsigned short array[BUFLEN];
    uaddr voffsets[BUFLEN];

    canvas_get_dressing (colourbar->main_wc,
			 KCD_ATT_GRID_COLOUR, &grid_colour,
			 KCD_ATT_SCALE_COLOUR, &scale_colour,
			 KCD_ATT_LABEL_COLOUR, &label_colour,
			 KCD_ATT_FONT_SCALE, &font_scale,
			 KCD_ATT_END);
    canvas_set_dressing (canvas,
			 KCD_ATT_GRID_COLOUR, grid_colour,
			 KCD_ATT_SCALE_COLOUR, scale_colour,
			 KCD_ATT_LABEL_COLOUR, label_colour,
			 KCD_ATT_FONT_SCALE, font_scale,
			 KCD_ATT_END);
    if ( (win_scale->x_pixels == colourbar->cache_width) &&
	 (win_scale->y_pixels == colourbar->cache_height) && !cmap_resize )
	return;
    /*  Need to compute cache (colourbar image)  */
    colourbar->cache_valid = FALSE;
    if (win_scale->y_pixels > BUFLEN) return;
    pixcanvas = canvas_get_pixcanvas (canvas);
    kwin_get_attributes (pixcanvas,
			 KWIN_FUNC_DRAW_CACHED_IMAGE, &draw_cached_image_func,
			 KWIN_ATT_END);
    if (draw_cached_image_func == NULL) return;
    num_pixels = kcmap_get_pixels (cmap, &pixel_values);
    for (count = 0; count < win_scale->y_pixels; ++count)
    {
	array[count] = count * (num_pixels - 1) / (win_scale->y_pixels - 1);
	voffsets[count] = count * sizeof (short);
    }
    kwin_compute_pc_subimage (pixcanvas, 0, 0,
			      win_scale->x_pixels, win_scale->y_pixels,
			      win_scale->x_pixels, win_scale->y_pixels,
			      (char *) array, &hoffset, voffsets,
			      1, win_scale->y_pixels,
			      K_USHORT, 0,
			      num_pixels, pixel_values, 0, 0, 0,
			      0.0, num_pixels - 1,
			      NULL, NULL,
			      &colourbar->cache);
    colourbar->cache_width = win_scale->x_pixels;
    colourbar->cache_height = win_scale->y_pixels;
    colourbar->cache_valid = TRUE;
}   /*  End Function _viewimg_colourbar_precompute_func  */

static void _viewimg_colourbar_refresh_func
   (KWorldCanvas canvas, int width, int height,
    CONST struct win_scale_type *win_scale,
    Kcolourmap cmap, flag cmap_resize, void **info,
    PostScriptPage pspage,
    unsigned int num_areas,
    KPixCanvasRefreshArea *areas,
    flag *honoured_areas)
/*  [SUMMARY] Refresh event callback.
    [PURPOSE] This routine is a refresh event consumer for a world canvas.
    <canvas> The world canvas being refreshed.
    <width> The width of the canvas in pixels.
    <height> The height of the canvas in pixels.
    <win_scale> A pointer to the window scaling information.
    <cmap> The colourmap associated with the canvas.
    <cmap_resize> TRUE if the refresh function was called as a result of a
    colourmap resize, else FALSE.
    <info> A pointer to the arbitrary canvas information pointer.
    <pspage> If not NULL, the PostScriptPage object the refresh is
    redirected to.
    <num_areas> The number of areas that need to be refreshed. If this is
    0 then the entire pixel canvas needs to be refreshed.
    <areas> The list of areas that need to be refreshed. Note that these
    areas are given in pixel co-ordinates.
    <honoured_areas> If the value TRUE is written here it is assumed the
    routine honoured the list of refresh areas and did not write outside
    these areas and hence the list of areas will be passed to subsequent
    registered refresh routines. If FALSE is written here (or nothing is
    written here), implying the routine refreshed the entire canvas,
    subsequent refresh routines will be told to refresh the entire canvas.
    [RETURNS] Nothing.
*/
{
    KPixCanvas pixcanvas;
    unsigned int count, num_pixels;
    uaddr hoffset = 0;
    struct colourbar_struct *colourbar = *info;
    unsigned long *pixel_values;
    void *draw_cached_image_func;
    unsigned short array[BUFLEN];
    uaddr voffsets[BUFLEN];

    pixcanvas = canvas_get_pixcanvas (canvas);
    /*  Draw a point so than any pending geom_clear is executed now, before the
	image is drawn  */
    kwin_draw_point (pixcanvas, win_scale->x_offset, win_scale->y_offset, 0);
    kwin_get_attributes (pixcanvas,
			 KWIN_FUNC_DRAW_CACHED_IMAGE, &draw_cached_image_func,
			 KWIN_ATT_END);
    if ( colourbar->cache_valid && (draw_cached_image_func != NULL) &&
	 kwin_draw_cached_image (colourbar->cache,
				 win_scale->x_offset, win_scale->y_offset) )
	return;
    /*  Need to draw it the hard way  */
    num_pixels = kcmap_get_pixels (cmap, &pixel_values);
    for (count = 0; count < win_scale->y_pixels; ++count)
    {
	array[count] = count * (num_pixels - 1) / (win_scale->y_pixels - 1);
	voffsets[count] = count * sizeof (short);
    }
    kwin_draw_pc_image (pixcanvas, win_scale->x_offset, win_scale->y_offset,
			win_scale->x_pixels, win_scale->y_pixels,
			(char *) array, &hoffset, voffsets,
			1, win_scale->y_pixels,
			K_USHORT, 0,
			num_pixels, pixel_values, 0, 0, 0,
			0.0, num_pixels - 1, NULL, NULL,
			&colourbar->cache);
}   /*  End Function _viewimg_colourbar_refresh_func  */

static void _viewimg_colourbar_destroy_func (void *info)
/*  [SUMMARY] Process a destroy event on a world canvas.
    <info> An arbitrary function information pointer.
    [RETURNS] Nothing.
*/
{
    struct colourbar_struct *colourbar = info;

    kwin_free_cache_data (colourbar->cache);
    m_free (info);
}   /*  End Function _viewimg_colourbar_destroy_func  */
