/*LINTLIBRARY*/
/*  DataBrowser.c

    This code provides a generic data browser for Xt.

    Copyright (C) 2002-2004  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This file contains all routines needed for manipulating a popup data
    browser widget for Xt.


    Written by      Richard Gooch   23-NOV-2002

    Updated by      Richard Gooch   11-DEC-2002: First cut of code. Array data
  manipulation and new <kvis> application ready for beta testing. 

    Updated by      Richard Gooch   12-DEC-2002: Basic overlay data
  manipulation implemented.

    Updated by      Richard Gooch   14-DEC-2002: Added movie and profile menus.

    Updated by      Richard Gooch   16-DEC-2002: Implemented "some" mode.

    Updated by      Richard Gooch   20-DEC-2002: Automatically select image to
  be master movie and profile if user has not made a selection.

    Updated by      Richard Gooch   28-DEC-2002: Added profile colour controls.

    Updated by      Richard Gooch   16-JAN-2003: Stretch only list when user
  stretches the window.

    Updated by      Richard Gooch   25-JAN-2003: Added data replacement control

    Updated by      Richard Gooch   3-FEB-2003: Changed  <XkwDataBrowserPopup>
  to popup the current browser when in "one" mode. This allows the user to
  pop up the browser irrespective of which window is used to summon it.

    Updated by      Richard Gooch   8-FEB-2003: Activate contours when selected
  via menu. Previous behaviour (requiring separate activation) was less
  convenient and more confusing.

    Updated by      Richard Gooch   10-FEB-2003: Popup first browser when first
  array dataset is loaded.

    Updated by      Richard Gooch   15-FEB-2003: Bugfix in <list_event_handler>
  where button events were also being handled by key event handling code.

    Updated by      Richard Gooch   17-FEB-2003: Removed calls to ViewDatasets
  widget to invalidate title, since <dispdata> package now handles auto titles.

    Updated by      Richard Gooch   20-FEB-2003: Added Renzogram support.

    Updated by      Richard Gooch   18-MAR-2003: Popup first browser in
  <Initialise> method, so that it doesn't pop up after command-line files have
  been loaded (which is annoying for virtual desktops).

    Updated by      Richard Gooch   19-MAY-2003: Support control+shift+i to
  cycle through image in reverse order.

    Updated by      Richard Gooch   30-MAY-2003: Fixed bug in <copy_cbk>: was
  creating new browser for parent window rather than current window.

    Updated by      Richard Gooch   7-JUN-2003: Fixed bug in function
  <array_process_event>: was not updating movie controls when image change also
  changed movie master.

    Updated by      Richard Gooch   11-JUN-2003: Added window identifier.

    Updated by      Richard Gooch   13-JUN-2003: Fixed bug in <Destroy> method
  which could lead to an Xt error or dangling pointers when using "some" or
  "one" mode. Very subtle and requires some special conditions. Moved identity
  display to title bar.

    Updated by      Richard Gooch   21-JUL-2003: Added advanced control panel.

    Updated by      Richard Gooch   3-AUG-2003: Added clip mode menu. Added
  "Range changes affect all arrays" toggle.

    Updated by      Richard Gooch   25-AUG-2003: Changed <show_source> to
  handle composite datasets.

    Updated by      Richard Gooch   3-OCT-2003: Took account of interface
  change for Dataclip widget.

    Updated by      Richard Gooch   9-MAR-2004: Switched to new ScrollableList
  widget.

    Updated by      Richard Gooch   14-MAR-2004: No longer indiscriminantly
  reset selected index when list grows/shrinks.

    Updated by      Richard Gooch   16-MAR-2004: Added "Make Data" button.

    Updated by      Richard Gooch   7-APR-2004: Added movie slave and activate
  annotation toggles in advanced control panel.

    Updated by      Richard Gooch   3-MAY-2004: Fixed bug in function
  <canvas_position_func>: do not forward events if browser not in array mode.

    Updated by      Richard Gooch   18-MAY-2004: Added toggle to slave H-range
  for profiles.

    Updated by      Richard Gooch   9-JUN-2004: Fixed size calculation for RGB
  arrays.

    Updated by      Richard Gooch   27-JUN-2004: Use defaults if empty contour
  or Renzogram levels string is entered.

    Updated by      Richard Gooch   7-JUL-2004: Added XkwNpopupFirstDataBrowser

    Last updated by Richard Gooch   14-JUL-2004: Added "Show new data as Image"
  toggle to advanced control panel. Fixed bugs in <array_process_event> and
  <overlay_process_event>: was not skipping if list is empty, which caused a
  bad pointer dereference.


*/
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <ctype.h>
#include <sys/param.h>
#include <errno.h>
#include <Xkw/DataBrowserP.h>
#include <X11/Xos.h>
#include <X11/StringDefs.h>
#include <X11/Xaw/Command.h>    
#include <X11/Xaw/Form.h>
#include <karma.h>
#include <k_event_codes.h>
#include <karma_dispdata.h>
#include <karma_foreign.h>
#include <karma_iarray.h>
#include <karma_xtmisc.h>
#include <karma_listw.h>
#include <karma_ext.h>
#include <karma_chs.h>
#include <karma_ch.h>
#include <karma_ds.h>
#include <karma_st.h>
#include <karma_m.h>
#include <karma_c.h>
#include <karma_r.h>
#include <karma_a.h>
#include <Xkw/Ktoggle.h>
#include <Xkw/ImageDisplay.h>
#include <Xkw/ExclusiveMenu.h>
#include <Xkw/Dataclip.h>
#include <Xkw/SimpleTextEntry.h>
#include <Xkw/SimpleSlider.h>
#include <Xkw/ViewDatasets.h>
#include <Xkw/ScrollableList.h>
#include <Xkw/MakeDataBrowser.h>


#define INIT_CONTOUR_STRING   "%10:90+10"
#define INIT_RENZOGRAM_STRING       "%80"
#define WIDTH                         430

#define CHAR_INDEX_REPLACE              0

#define CHAR_INDEX_ARRAY_IMAGE          1
#define CHAR_INDEX_ARRAY_CONTOUR        2
#define CHAR_INDEX_ARRAY_MOVIE          3
#define CHAR_INDEX_ARRAY_PROFILE        4
#define NUM_RESERVED_CHAR_ARRAY         7

#define CHAR_INDEX_OVERLAY_ACTIVE       1
#define NUM_RESERVED_CHAR_OVERLAY       4

#define EVENT_CODE_LEFT_ARROW        1000
#define EVENT_CODE_RIGHT_ARROW       1001
#define EVENT_CODE_MOUSE_MOVE        1002
#define EVENT_CODE_LEFT_MOUSE        1003
#define EVENT_CODE_MIDDLE_MOUSE      1004
#define EVENT_CODE_RIGHT_MOUSE       1005
#define EVENT_CODE_CONTOUR_OFF       1006
#define EVENT_CODE_CONTOUR_NORMAL    1007
#define EVENT_CODE_CONTOUR_RENZOGRAM 1008
#define EVENT_CODE_MOVIE_OFF         1009
#define EVENT_CODE_MOVIE_MASTER      1010
#define EVENT_CODE_MOVIE_SLAVE       1011
#define EVENT_CODE_PROFILE_OFF       1012
#define EVENT_CODE_PROFILE_MASTER    1013
#define EVENT_CODE_PROFILE_SLAVE     1014

#define REFRESH_ARRAY_RANGE          (1<<0)
#define REFRESH_ARRAY_CLEVELS        (1<<1)
#define REFRESH_ARRAY_NEG_WIDTH      (1<<2)
#define REFRESH_ARRAY_POS_WIDTH      (1<<3)
#define REFRESH_ARRAY_NEG_COLOUR     (1<<4)
#define REFRESH_ARRAY_POS_COLOUR     (1<<5)
#define REFRESH_ARRAY_RENZO_LEVELS   (1<<6)
#define REFRESH_ARRAY_RENZO_CHANNELS (1<<7)
#define REFRESH_ARRAY_DARK_COLOUR    (1<<8)
#define REFRESH_ARRAY_LIGHT_COLOUR   (1<<9)

#define REFRESH_ARRAY_ALL        0xffffffff

#define IMAGE_MODE_OFF                  0
#define IMAGE_MODE_MAIN                 1
#define IMAGE_MODE_ALTERNATE            2
#define NUM_IMAGE_MODE_CHOICES          3

static char *image_mode_choices[NUM_IMAGE_MODE_CHOICES] =
{
    "off", "main", "alt",
};

#define CONTOUR_MODE_OFF                0
#define CONTOUR_MODE_NORMAL             1
#define CONTOUR_MODE_RENZOGRAM          2
#define NUM_CONTOUR_MODE_CHOICES        3

static char *contour_mode_choices[NUM_CONTOUR_MODE_CHOICES] =
{
    "off", "map", "renzo",
};

#define MOVIE_MODE_OFF                  0
#define MOVIE_MODE_MASTER               1
#define MOVIE_MODE_SLAVE                2
#define NUM_MOVIE_MODE_CHOICES          3

static char *movie_mode_choices[NUM_MOVIE_MODE_CHOICES] =
{
    "off", "master", "slave",
};

#define PROFILE_MODE_OFF                0
#define PROFILE_MODE_MASTER             1
#define PROFILE_MODE_SLAVE              2
#define NUM_PROFILE_MODE_CHOICES        3

static char *profile_mode_choices[NUM_PROFILE_MODE_CHOICES] =
{
    "off", "master", "slave",
};

#define BROWSER_MODE_ALL                0
#define BROWSER_MODE_SOME               1
#define BROWSER_MODE_ONE                2
#define NUM_BROWSER_MODE_CHOICES        3

static char *browser_mode_choices[NUM_BROWSER_MODE_CHOICES] =
{
    "all", "some", "one",
};


struct array_list
{
    struct blinkstate *parent;
    KDisplayDataClass dataclass;
    KDisplayDataClassBlinkState class_blinkstate;
    KDisplayDataPrivateNameList namelist;
    unsigned int num_entries;
    char **strings;
    CONST KDisplayDataBlinkEntry *blinkentries;
    struct array_blink_entry *main_image;
    struct array_blink_entry *alt_image;
    struct array_blink_entry *movie_control;
    struct array_blink_entry *main_profile;
    char user_selected_movie_control;
    char user_selected_main_profile;
};

struct overlay_list
{
    struct blinkstate *parent;
    KDisplayDataClass dataclass;
    KDisplayDataClassBlinkState class_blinkstate;
    KDisplayDataPrivateNameList namelist;
    unsigned int num_entries;
    char **strings;
    CONST KDisplayDataBlinkEntry *blinkentries;
};

struct blinkstate
{
    DataBrowserWidget owner;    /*  Parent widget  */
    DataBrowserWidget viewer;   /*  Widget used for display (may be NULL)  */
    KDisplayDataBlinkState generic_blinkstate;
    KDisplayDataWindow window;
    KCallbackFunc destroy_func;
    struct array_list array_list;
    struct overlay_list overlay_list;
    struct blinkstate *prev;
    struct blinkstate *next;
};

struct array_blink_entry
{
    struct array_list *parent;
    KDisplayDataBlinkEntry generic_blinkentry;
    unsigned int index;
};

struct overlay_blink_entry
{
    struct overlay_list *parent;
    KDisplayDataBlinkEntry generic_blinkentry;
    unsigned int index;
};


/*  Private functions  */
STATIC_FUNCTION (void _DataBrowser_ClassInitialise, () );
STATIC_FUNCTION (void _DataBrowser_Initialise, (Widget request, Widget new) );
STATIC_FUNCTION (void _DataBrowser_Destroy, (Widget W) );
STATIC_FUNCTION (void _DataBrowser_listcanvas_realise_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void _DataBrowser_close_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void _DataBrowser_destroy_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void _DataBrowser_copy_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void _DataBrowser_active_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void _DataBrowser_browser_mode_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void _DataBrowser_dataclass_mode_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void _DataBrowser_refresh, (DataBrowserWidget w) );
STATIC_FUNCTION (void _DataBrowser_set_image_display_bg,
		 (DataBrowserWidget w, flag active) );
STATIC_FUNCTION (void _DataBrowser_set_identity,
		 (struct blinkstate *blinkstate) );
STATIC_FUNCTION (void _DataBrowser_set_identities,
		 (KDisplayDataWindow window) );
STATIC_FUNCTION (flag _DataBrowser_listcanvas_position_func,
		 (KPixCanvas canvas, int x, int y,
		  unsigned int event_code, void *e_info, void **f_info) );
STATIC_FUNCTION (flag _DataBrowser_list_position_func,
		 (KListWidget list, unsigned int index,
		  unsigned int event_code, void *e_info, void **f_info) );
STATIC_FUNCTION (void _DataBrowser_select_entry,
		 (DataBrowserWidget w, unsigned int index) );
STATIC_FUNCTION (void _DataBrowser_show_source,
		 (KDisplayDataObject dataobject, Widget lbl) );
STATIC_FUNCTION (void _DataBrowser_array_refresh_func,
		 (void *info, KDisplayDataClass dataclass) );
STATIC_FUNCTION (void *_DataBrowser_array_create_blink_entry,
		 (void *info, KDisplayDataBlinkEntry be) );
STATIC_FUNCTION (void _DataBrowser_array_destroy_blink_entry, (void *info) );
STATIC_FUNCTION (void _DataBrowser_array_copy_blink_entry,
		 (void *dest_p, void *source_p) );
STATIC_FUNCTION (void _DataBrowser_array_process_event,
		 (DataBrowserWidget w, int index, int event_code,
		  unsigned long mask) );
STATIC_FUNCTION (void _DataBrowser_array_show_details, (DataBrowserWidget w) );
STATIC_FUNCTION (void _DataBrowser_array_show_settings,(DataBrowserWidget w) );
STATIC_FUNCTION (void _DataBrowser_array_settings_refresh_one,
		 (DataBrowserWidget w, unsigned int mask) );
STATIC_FUNCTION (void _DataBrowser_array_settings_refresh_all,
		 (KDisplayDataObject dataobject, unsigned int mask) );
STATIC_FUNCTION (void _DataBrowser_histogram_btn_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void _DataBrowser_array_new_data,
		 (void *info, KDisplayDataObject dataobject) );
STATIC_FUNCTION (void _DataBrowser_array_clone_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void _DataBrowser_array_unload_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void _DataBrowser_dataclip_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void _DataBrowser_image_mode_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void _DataBrowser_contour_mode_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void _DataBrowser_movie_mode_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void _DataBrowser_profile_mode_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void _DataBrowser_minmax_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void _DataBrowser_string_cbk,
		 (DataBrowserWidget top, String string,
		  unsigned int attribute, unsigned int mask) );
STATIC_FUNCTION (void _DataBrowser_clevels_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void _DataBrowser_neg_width_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void _DataBrowser_pos_width_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void _DataBrowser_neg_colour_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void _DataBrowser_pos_colour_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void _DataBrowser_renzo_levels_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void _DataBrowser_renzo_channels_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void _DataBrowser_dark_colour_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void _DataBrowser_light_colour_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void _DataBrowser_active_change,
		 (void *blinkstate_p, flag active) );
STATIC_FUNCTION (void _DataBrowser_set_movie_controls,
		 (struct blinkstate *blinkstate) );
STATIC_FUNCTION (flag _DataBrowser_canvas_position_func,
		 (KWorldCanvas canvas, double x, double y,
		  unsigned int event_code, void *e_info,
		  void **f_info, double x_lin, double y_lin,
		  int x_pix, int y_pix) );
STATIC_FUNCTION (void _DataBrowser_overlay_refresh_func,
		 (void *info, KDisplayDataClass dataclass) );
STATIC_FUNCTION (void *_DataBrowser_overlay_create_blink_entry,
		 (void *info, KDisplayDataBlinkEntry be) );
STATIC_FUNCTION (void _DataBrowser_overlay_destroy_blink_entry, (void *info) );
STATIC_FUNCTION (void _DataBrowser_overlay_copy_blink_entry,
		 (void *dest_p, void *source_p) );
STATIC_FUNCTION (void _DataBrowser_overlay_process_event,
		 (DataBrowserWidget w, int index, int event_code,
		  unsigned long mask) );
STATIC_FUNCTION (void _DataBrowser_overlay_show_details,
		 (DataBrowserWidget w) );
STATIC_FUNCTION (void _DataBrowser_overlay_show_settings,
		 (DataBrowserWidget w) );
STATIC_FUNCTION (void _DataBrowser_overlay_unload_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void _DataBrowser_overlay_active_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (Widget _DataBrowser_create_advanced_control, (Widget w) );
STATIC_FUNCTION (void _DataBrowser_advanced_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void _DataBrowser_clip_mode_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void _DataBrowser_construct_cbk,
		 (Widget W, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void _DataBrowser_slave_profile_h_range_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (KDisplayDataBlinkState _DataBrowser_get_next_blinkstate,
		 (KDisplayDataBlinkState blinkstate) );


#define offset(field) XtOffset(DataBrowserWidget, dataBrowser.field)
#define coreOffset(field) XtOffset(DataBrowserWidget, core.field)

static XtResource DataBrowserResources[] = 
{
    {XkwNpopupFirstDataBrowser, XkwCPopupFirstDataBrowser, XtRBool,
     sizeof (Bool),
     offset (popupFirstDataBrowser), XtRImmediate, (XtPointer) True},
    {XkwNcreateAtInit, XkwCCreateAtInit, XtRBool, sizeof (Bool),
     offset (createAtInit), XtRImmediate, (XtPointer) FALSE},
};

#undef offset


DataBrowserClassRec dataBrowserClassRec = 
{
  {        /* CoreClassPart */
    (WidgetClass)&topLevelShellClassRec,  /* superclass */
    "DataBrowser",                   /* class_name */
    sizeof(DataBrowserRec),          /* widget_size */
    _DataBrowser_ClassInitialise,    /* class_initialise */
    NULL,                          /* class_part_initialise */
    FALSE,                         /* class_init */
    (XtInitProc)_DataBrowser_Initialise,        /* initialise */
    NULL,                          /* initialise_hook */
    XtInheritRealize,              /* realize */
    NULL,                          /* actions */
    0,                             /* num_actions */
    DataBrowserResources,            /* resources */
    XtNumber(DataBrowserResources),  /* num_resources */
    NULLQUARK,                     /* xrm_class */
    TRUE,                          /* compress_motion */
    TRUE,                          /* compress_exposure */
    TRUE,                          /* compress_enterleave */
    TRUE,                          /* visible_interest */
    _DataBrowser_Destroy,            /* destroy */
    XtInheritResize,               /* resize */
    NULL,                          /* expose */
    NULL,                          /* set_values */
    NULL,                          /* set_values_hook */
    XtInheritSetValuesAlmost,      /* set_values_almost */
    NULL,                          /* get_values_hook */
    NULL,                          /* accept_focus */
    XtVersion,                     /* version */
    NULL,                          /* callback_private */
    NULL,                          /* tm_translations */
    NULL,                          /* query_geometry  */
    NULL,
    NULL,
  },
  {     /* CompositeClassPart */
    /* geometry_manager	  */	XtInheritGeometryManager,
    /* change_managed	  */	XtInheritChangeManaged,
    /* insert_child	  */	XtInheritInsertChild,
    /* delete_child	  */	XtInheritDeleteChild,
    /* extension	  */	NULL
  },{ /* Shell */
    /* extension	  */	NULL
  },{ /* WMShell */
    /* extension	  */	NULL
  },{ /* VendorShell */
    /* extension	  */	NULL
  },{ /* TopLevelShell */
    /* extension	  */	NULL
  },
  {  /* DataBrowserClassPart */
    0 /* empty */
  }
};

WidgetClass dataBrowserWidgetClass = (WidgetClass) &dataBrowserClassRec;


static unsigned int num_classes = 0;
static KDisplayDataClass *dataclasses = NULL;
static char **classnames = NULL;
static int browser_mode = BROWSER_MODE_ALL;
static char *histogram_key = "DataclipWidget";
static DataBrowserWidget current_one_browser = NULL;
static Widget change_all_ranges_tgl = NULL;
static Widget slave_movies_on_load_tgl = NULL;
static Widget activate_annotations_on_load_tgl = NULL;
static Widget show_new_data_as_image_tgl = NULL;
static KDisplayDataBlinkState round_robin_bs = NULL;

static KPixCanvasBinding main_bindings[] =
{
  {"i, control",    "make next image in list active"},
  {NULL,           NULL}
};


static void _DataBrowser_ClassInitialise ()
{
    KDisplayDataClass dataclass;

    dispdata_init ();
    num_classes = dispdata_get_classlist (&dataclasses, &classnames);
    dataclass = dispdata_array_create_dataclass ();
    dispdata_register_data_create_func (dataclass, _DataBrowser_array_new_data,
					NULL);
}   /*  End Function ClassInitialise  */

static void _DataBrowser_Initialise (Widget Request, Widget New)
{
    KDisplayDataBlinkState active;
    flag only_for_window;
    /*DataBrowserWidget request = (DataBrowserWidget) Request;*/
    DataBrowserWidget new = (DataBrowserWidget) New;
    Widget image_display, form, w, list, upper_widget;
    struct blinkstate *blinkstate;
    static flag first_time = TRUE;
    static char function_name[] = "DataBrowser::Initialise";

    w = image_display = XtParent (New);
    if ( !XtIsImageDisplay (image_display) )
    {
	fprintf (stderr, "Parent widget is not ImageDisplay class\n");
	a_prog_bug (function_name);
    }
    new->dataBrowser.image_display_widget = w;
    w = XtNameToWidget (image_display, "applicationBox");
    XtVaGetValues (w, XtNbackground, &new->dataBrowser.image_display_bg, NULL);
    w = XtNameToWidget (image_display, "viewcontrolpopup");
    new->dataBrowser.viewdatasets = w;
    w = XtNameToWidget (w, "animatepopup");
    if ( !XtIsAnimateControl (w) )
    {
	fprintf (stderr, "Movie widget is not AnimateControl class\n");
	a_prog_bug (function_name);
    }
    new->dataBrowser.animatecontrol = w;
    new->dataBrowser.user_popped_up = FALSE;
    new->dataBrowser.dataclass_mode = 0;
    new->dataBrowser.active_mode = TRUE;
    if ( ( blinkstate = m_calloc (sizeof *blinkstate) ) == NULL )
	m_abort (function_name, "blinkstate structure");
    blinkstate->window = XkwImageDisplayGetWindow (image_display);
    blinkstate->destroy_func =
	dispdata_register_window_destroy_func (blinkstate->window,
					       c_write_null_func,
					       &blinkstate->window);
    only_for_window = (dispdata_get_num_blinkstates (blinkstate->window) ==
		       0) ? TRUE : FALSE;
    new->dataBrowser.my_blinkstate = blinkstate;
    blinkstate->owner = new;
    /*  Make this browser viewable so that refresh event can set up canvas, if
	required. Later, this may be changed, depending on browser mode  */
    new->dataBrowser.current_blinkstate = blinkstate;
    blinkstate->viewer = new;
    /*  Create common widgets  */
    form = XtVaCreateManagedWidget ("form", formWidgetClass, New,
				    XtNborderWidth, 0,
				    NULL);
    new->dataBrowser.main_form = form;
    w = XtVaCreateManagedWidget ("closeButton", commandWidgetClass, form,
				 XtNlabel, "Close",
				 /*XtNtop, XtChainTop,*/
				 XtNbottom, XtChainTop,
				 NULL);
    upper_widget = w;
    XtAddCallback (w, XtNcallback, _DataBrowser_close_cbk, New);
    w = XtVaCreateManagedWidget ("destroyButton", commandWidgetClass, form,
				 XtNlabel, "Destroy",
				 XtNfromHoriz, w,
				 XtNtop, XtChainTop,
				 XtNbottom, XtChainTop,
				 NULL);
    new->dataBrowser.destroy_btn = w;
    XtAddCallback (w, XtNcallback, _DataBrowser_destroy_cbk, New);
    w = XtVaCreateManagedWidget ("copyButton", commandWidgetClass, form,
				 XtNlabel, "Copy",
				 XtNfromHoriz, w,
				 XtNtop, XtChainTop,
				 XtNbottom, XtChainTop,
				 NULL);
    XtAddCallback (w, XtNcallback, _DataBrowser_copy_cbk, New);
    w = XtVaCreateManagedWidget ("activeLabel", commandWidgetClass, form,
				 XtNfromHoriz, w,
				 XtNtop, XtChainTop,
				 XtNbottom, XtChainTop,
				 NULL);
    new->dataBrowser.active_btn = w;
    XtAddCallback (w, XtNcallback, _DataBrowser_active_cbk, New);
    w = XtVaCreateManagedWidget ("browserModeMenu", exclusiveMenuWidgetClass,
				 form,
				 XtNmenuName, "browserModeMenu",
				 XkwNchoiceName, "Show",
				 XtNfromHoriz, w,
				 XtNtop, XtChainTop,
				 XtNbottom, XtChainTop,
				 XkwNnumItems, NUM_BROWSER_MODE_CHOICES,
				 XkwNitemStrings, browser_mode_choices,
				 NULL);
    new->dataBrowser.browser_mode_menu = w;
    XtAddCallback (w, XkwNselectCallback, _DataBrowser_browser_mode_cbk, New);
    w = XtVaCreateManagedWidget ("displayModeMenu", exclusiveMenuWidgetClass,
				 form,
				 XtNmenuName, "displayModeMenu",
				 XkwNchoiceName, "Data",
				 XtNfromVert, upper_widget,
				 XtNtop, XtChainTop,
				 XtNbottom, XtChainTop,
				 XkwNnumItems, num_classes,
				 XkwNitemStrings, classnames,
				 NULL);
    XtAddCallback (w, XkwNselectCallback, _DataBrowser_dataclass_mode_cbk,New);
    w = XtVaCreateManagedWidget ("advancedButton", commandWidgetClass, form,
				 XtNlabel,
				 strcmp (r_getenv ("HOME"), "/u/gibson") ? "Advanced" : "Steve",
				 XtNfromVert, upper_widget,
				 XtNfromHoriz, w,
				 XtNtop, XtChainTop,
				 XtNbottom, XtChainTop,
				 NULL);
    XtAddCallback (w, XtNcallback, _DataBrowser_advanced_cbk, New);
    w = XtVaCreateManagedWidget ("constructButton", commandWidgetClass, form,
				 XtNlabel, "Make Data",
				 XtNfromVert, upper_widget,
				 XtNfromHoriz, w,
				 XtNtop, XtChainTop,
				 XtNbottom, XtChainTop,
				 NULL);
    XtAddCallback (w, XtNcallback, _DataBrowser_construct_cbk, New);
    upper_widget = w;
    list = XtVaCreateManagedWidget ("list", scrollableListWidgetClass, form,
				    XtNborderWidth, 3,
				    XtNhorizDistance, 0,
				    XtNwidth, WIDTH,
				    XtNheight, 150,
				    XtNfromVert, upper_widget,
				    XtNtop, XtChainTop,
				    XtNbottom, XtChainBottom,
				    NULL);
    new->dataBrowser.list_widget = list;
    XtVaGetValues (list, XkwNlist, &new->dataBrowser.listw, NULL);
    XtAddCallback (XtNameToWidget (list, "listCanvas"), XkwNrealiseCallback,
		   _DataBrowser_listcanvas_realise_cbk, New);
    listw_set_attributes (new->dataBrowser.listw, FALSE,
			  LISTW_ATT_MOUSE_MODE, LISTW_MOUSE_MODE_HIGHLIGHT,
			  LISTW_ATT_END);
    listw_register_position_event_func (new->dataBrowser.listw,
					( flag (*) () ) _DataBrowser_list_position_func,
					new);
    /*  Create array data-specific widgets  */
    w = XtVaCreateManagedWidget ("arrayForm", formWidgetClass, form,
				 XtNborderWidth, 0,
				 XtNhorizDistance, 0,
				 XtNvertDistance, 0,
				 XtNfromVert, list,
				 XtNtop, XtChainBottom,
				 XtNbottom, XtChainBottom,
				 NULL);
    new->dataBrowser.array.form = w;
    w = XtVaCreateManagedWidget ("sourceLabel", labelWidgetClass,
				 new->dataBrowser.array.form,
				 XtNborderWidth, 0,
				 XtNwidth, WIDTH,
				 XtNhorizDistance, 0,
				 XtNvertDistance, 0,
				 XtNleft, XawChainLeft,
				 XtNright, XawChainRight,
				 XtNjustify, XtJustifyLeft,
				 XtNresize, False,
				 NULL);
    new->dataBrowser.array.source_lbl = w;
    w = XtVaCreateManagedWidget ("objectNDateLabel", labelWidgetClass,
				 new->dataBrowser.array.form,
				 XtNborderWidth, 0,
				 XtNwidth, WIDTH,
				 XtNhorizDistance, 0,
				 XtNvertDistance, 0,
				 XtNfromVert, w,
				 XtNleft, XawChainLeft,
				 XtNright, XawChainRight,
				 XtNjustify, XtJustifyLeft,
				 XtNresize, False,
				 NULL);
    new->dataBrowser.array.objectNdate_lbl = w;
    w = XtVaCreateManagedWidget ("unitsLabel", labelWidgetClass,
				 new->dataBrowser.array.form,
				 XtNborderWidth, 0,
				 XtNwidth, WIDTH,
				 XtNhorizDistance, 0,
				 XtNvertDistance, 0,
				 XtNfromVert, w,
				 XtNleft, XawChainLeft,
				 XtNright, XawChainRight,
				 XtNjustify, XtJustifyLeft,
				 XtNresize, False,
				 NULL);
    new->dataBrowser.array.fu_lbl = w;
    w = XtVaCreateManagedWidget ("sizeLabel", labelWidgetClass,
				 new->dataBrowser.array.form,
				 XtNborderWidth, 0,
				 XtNwidth, WIDTH,
				 XtNhorizDistance, 0,
				 XtNvertDistance, 0,
				 XtNfromVert, w,
				 XtNleft, XawChainLeft,
				 XtNright, XawChainRight,
				 XtNjustify, XtJustifyLeft,
				 XtNresize, False,
				 NULL);
    new->dataBrowser.array.size_lbl = w;
    w = XtVaCreateManagedWidget ("rangeLabel", labelWidgetClass,
				 new->dataBrowser.array.form,
				 XtNborderWidth, 0,
				 XtNwidth, WIDTH,
				 XtNhorizDistance, 0,
				 XtNvertDistance, 0,
				 XtNfromVert, w,
				 XtNleft, XawChainLeft,
				 XtNright, XawChainRight,
				 XtNjustify, XtJustifyLeft,
				 XtNresize, False,
				 NULL);
    new->dataBrowser.array.minmax_lbl = w;
    upper_widget = w;
    w = XtVaCreateManagedWidget ("button", commandWidgetClass,
				 new->dataBrowser.array.form,
				 XtNlabel, "Histogram",
				 XtNfromVert, upper_widget,
				 NULL);
    new->dataBrowser.array.histogram_btn = w;
    XtAddCallback (w, XtNcallback, _DataBrowser_histogram_btn_cbk, new);
    w = XtVaCreateManagedWidget ("button", commandWidgetClass,
				 new->dataBrowser.array.form,
				 XtNlabel, "Clone",
				 XtNfromVert, upper_widget,
				 XtNfromHoriz, w,
				 NULL);
    XtAddCallback (w, XtNcallback, _DataBrowser_array_clone_cbk, new);
    w = XtVaCreateManagedWidget ("destroyButton", commandWidgetClass,
				 new->dataBrowser.array.form,
				 XtNlabel, "Unload",
				 XtNfromVert, upper_widget,
				 XtNfromHoriz, w,
				 NULL);
    XtAddCallback (w, XtNcallback, _DataBrowser_array_unload_cbk, new);
    w = XtVaCreateManagedWidget ("divider", simpleWidgetClass,
				 new->dataBrowser.array.form,
				 XtNborderWidth, 0,
				 XtNwidth, WIDTH,
				 XtNheight, 2,
				 XtNhorizDistance, 0,
				 XtNfromVert, w,
				 XtNleft, XawChainLeft,
				 XtNright, XawChainRight,
				 XtVaTypedArg,
				      XtNbackground,
				      XtRString,
				      "black",
				      6,
				 NULL);
    upper_widget = w;
    w = XtVaCreateManagedWidget ("imageModeMenu", exclusiveMenuWidgetClass,
				 new->dataBrowser.array.form,
				 XtNmenuName, "imageModeMenu",
				 XkwNchoiceName, "Image",
				 XtNfromVert, upper_widget,
				 XkwNnumItems, NUM_IMAGE_MODE_CHOICES,
				 XkwNitemStrings, image_mode_choices,
				 NULL);
    new->dataBrowser.array.image_mode_menu = w;
    XtAddCallback (w, XkwNselectCallback, _DataBrowser_image_mode_cbk, New);
    w = XtVaCreateManagedWidget ("contourModeMenu", exclusiveMenuWidgetClass,
				 new->dataBrowser.array.form,
				 XtNmenuName, "contourModeMenu",
				 XkwNchoiceName, "Contour",
				 XtNfromVert, upper_widget,
				 XtNfromHoriz, w,
				 XkwNnumItems, NUM_CONTOUR_MODE_CHOICES,
				 XkwNitemStrings, contour_mode_choices,
				 NULL);
    new->dataBrowser.array.contour_mode_menu = w;
    XtAddCallback (w, XkwNselectCallback, _DataBrowser_contour_mode_cbk, New);
    w = XtVaCreateManagedWidget ("movieModeMenu", exclusiveMenuWidgetClass,
				 new->dataBrowser.array.form,
				 XtNmenuName, "movieModeMenu",
				 XkwNchoiceName, "Movie",
				 XtNfromVert, upper_widget,
				 XtNfromHoriz, w,
				 XkwNnumItems, NUM_MOVIE_MODE_CHOICES,
				 XkwNitemStrings, movie_mode_choices,
				 NULL);
    upper_widget = w;
    new->dataBrowser.array.movie_mode_menu = w;
    XtAddCallback (w, XkwNselectCallback, _DataBrowser_movie_mode_cbk, New);
    w = XtVaCreateManagedWidget ("lowClipLabel", labelWidgetClass,
				 new->dataBrowser.array.form,
				 XtNborderWidth, 0,
				 XtNhorizDistance, 0,
				 XtNfromVert, upper_widget,
				 XtNlabel, "Low:",
				 NULL);
    w = XtVaCreateManagedWidget ("lowClipText", simpleTextEntryWidgetClass,
				 new->dataBrowser.array.form,
				 XtNhorizDistance, 0,
				 XtNfromVert, upper_widget,
				 XtNfromHoriz, w,
				 XtNwidth, (WIDTH / 2) - 50,
				 XtNeditType, XawtextEdit,
				 NULL);
    new->dataBrowser.array.low_clip_txt = w;
    XtAddCallback (w, XkwNnewLineCallback, _DataBrowser_minmax_cbk, New);
    w = XtVaCreateManagedWidget ("highClipLabel", labelWidgetClass,
				 new->dataBrowser.array.form,
				 XtNborderWidth, 0,
				 XtNhorizDistance, 0,
				 XtNfromVert, upper_widget,
				 XtNfromHoriz, w,
				 XtNlabel, "High:",
				 NULL);
    w = XtVaCreateManagedWidget ("highClipText", simpleTextEntryWidgetClass,
				 new->dataBrowser.array.form,
				 XtNhorizDistance, 0,
				 XtNfromVert, upper_widget,
				 XtNfromHoriz, w,
				 XtNwidth, (WIDTH / 2) - 50,
				 XtNeditType, XawtextEdit,
				 NULL);
    upper_widget = w;
    new->dataBrowser.array.high_clip_txt = w;
    XtAddCallback (w, XkwNnewLineCallback, _DataBrowser_minmax_cbk, New);
    w = XtVaCreateManagedWidget ("contourLevelsLabel", labelWidgetClass,
				 new->dataBrowser.array.form,
				 XtNborderWidth, 0,
				 XtNhorizDistance, 0,
				 XtNfromVert, upper_widget,
				 XtNlabel, "Map Contour Levels:",
				 NULL);
    w = XtVaCreateManagedWidget ("clevelsText", simpleTextEntryWidgetClass,
				 new->dataBrowser.array.form,
				 XtNhorizDistance, 0,
				 XtNfromVert, upper_widget,
				 XtNfromHoriz, w,
				 XtNwidth, WIDTH - 180,
				 XtNeditType, XawtextEdit,
				 NULL);
    upper_widget = w;
    new->dataBrowser.array.contour_levels_txt = w;
    XtAddCallback (w, XkwNnewLineCallback, _DataBrowser_clevels_cbk, New);
    w = XtVaCreateManagedWidget ("slider", simpleSliderWidgetClass,
				 new->dataBrowser.array.form,
				 XtNhorizDistance, 0,
				 XtNfromVert, upper_widget,
				 XtNlabel, "-ve Cont. Width",
				 XtNwidth, (WIDTH / 2) - 5,
				 XkwNminimum, 0,
				 XkwNmaximum, 100,
				 XkwNmodifier, 1,
				 XtNvalue, 0,
				 XkwNscaledFormat, "%.1f",
				 NULL);
    new->dataBrowser.array.contour_neg_width_sld = w;
    XkwSimpleSliderSetScale (w, 1e-1, 0.0, FALSE, TRUE);
    XtAddCallback (w, XkwNvalueChangeCallback, _DataBrowser_neg_width_cbk,New);
    w = XtVaCreateManagedWidget ("slider", simpleSliderWidgetClass,
				 new->dataBrowser.array.form,
				 XtNfromVert, upper_widget,
				 XtNfromHoriz, w,
				 XtNlabel, "+ve Cont. Width",
				 XtNwidth, (WIDTH / 2) - 5,
				 XkwNminimum, 0,
				 XkwNmaximum, 100,
				 XkwNmodifier, 1,
				 XtNvalue, 0,
				 XkwNscaledFormat, "%.1f",
				 NULL);
    upper_widget = w;
    new->dataBrowser.array.contour_pos_width_sld = w;
    XkwSimpleSliderSetScale (w, 1e-1, 0.0, FALSE, TRUE);
    XtAddCallback (w, XkwNvalueChangeCallback, _DataBrowser_pos_width_cbk,New);
    w = XtVaCreateManagedWidget ("contourColourLabel", labelWidgetClass,
				 new->dataBrowser.array.form,
				 XtNborderWidth, 0,
				 XtNhorizDistance, 0,
				 XtNfromVert, upper_widget,
				 XtNlabel, "-Colour:",
				 NULL);
    w = XtVaCreateManagedWidget ("colourText", simpleTextEntryWidgetClass,
				 new->dataBrowser.array.form,
				 XtNhorizDistance, 0,
				 XtNfromVert, upper_widget,
				 XtNfromHoriz, w,
				 XtNwidth, 115,
				 XtNeditType, XawtextEdit,
				 NULL);
    new->dataBrowser.array.contour_neg_colour_txt = w;
    XtAddCallback (w, XkwNnewLineCallback, _DataBrowser_neg_colour_cbk, New);
    w = XtVaCreateManagedWidget ("contourColourLabel", labelWidgetClass,
				 new->dataBrowser.array.form,
				 XtNborderWidth, 0,
				 XtNhorizDistance, 0,
				 XtNfromVert, upper_widget,
				 XtNfromHoriz, w,
				 XtNlabel, "+Colour:",
				 NULL);
    w = XtVaCreateManagedWidget ("colourText", simpleTextEntryWidgetClass,
				 new->dataBrowser.array.form,
				 XtNhorizDistance, 0,
				 XtNfromVert, upper_widget,
				 XtNfromHoriz, w,
				 XtNwidth, 115,
				 XtNeditType, XawtextEdit,
				 NULL);
    upper_widget = w;
    new->dataBrowser.array.contour_pos_colour_txt = w;
    XtAddCallback (w, XkwNnewLineCallback, _DataBrowser_pos_colour_cbk, New);
    w = XtVaCreateManagedWidget ("renzogramLevelsLabel", labelWidgetClass,
				 new->dataBrowser.array.form,
				 XtNborderWidth, 0,
				 XtNhorizDistance, 0,
				 XtNfromVert, upper_widget,
				 XtNlabel, "Renzo Contour Levels:",
				 NULL);
    w = XtVaCreateManagedWidget ("renzoLevelsText", simpleTextEntryWidgetClass,
				 new->dataBrowser.array.form,
				 XtNhorizDistance, 0,
				 XtNfromVert, upper_widget,
				 XtNfromHoriz, w,
				 XtNwidth, WIDTH - 200,
				 XtNeditType, XawtextEdit,
				 NULL);
    upper_widget = w;
    new->dataBrowser.array.renzogram_levels_txt = w;
    XtAddCallback (w, XkwNnewLineCallback, _DataBrowser_renzo_levels_cbk, New);
    w = XtVaCreateManagedWidget ("renzogramChannelsLabel", labelWidgetClass,
				 new->dataBrowser.array.form,
				 XtNborderWidth, 0,
				 XtNhorizDistance, 0,
				 XtNfromVert, upper_widget,
				 XtNlabel, "Renzogram Channels:",
				 NULL);
    w = XtVaCreateManagedWidget ("renzoChannelsText",
				 simpleTextEntryWidgetClass,
				 new->dataBrowser.array.form,
				 XtNhorizDistance, 0,
				 XtNfromVert, upper_widget,
				 XtNfromHoriz, w,
				 XtNwidth, WIDTH - 200,
				 XtNeditType, XawtextEdit,
				 NULL);
    upper_widget = w;
    new->dataBrowser.array.renzogram_channels_txt = w;
    XtAddCallback (w, XkwNnewLineCallback, _DataBrowser_renzo_channels_cbk,
		   New);
    w = XtVaCreateManagedWidget ("profileModeMenu", exclusiveMenuWidgetClass,
				 new->dataBrowser.array.form,
				 XtNmenuName, "profileModeMenu",
				 XkwNchoiceName, "Profile",
				 XtNfromVert, upper_widget,
				 XkwNnumItems, NUM_PROFILE_MODE_CHOICES,
				 XkwNitemStrings, profile_mode_choices,
				 NULL);
    new->dataBrowser.array.profile_mode_menu = w;
    XtAddCallback (w, XkwNselectCallback, _DataBrowser_profile_mode_cbk, New);
    w = XtVaCreateManagedWidget ("profileColoursLabel", labelWidgetClass,
				 new->dataBrowser.array.form,
				 XtNborderWidth, 0,
				 XtNfromVert, upper_widget,
				 XtNfromHoriz, w,
				 XtNlabel, " Profile Colours:",
				 NULL);
    upper_widget = w;
    w = XtVaCreateManagedWidget ("darkColourLabel", labelWidgetClass,
				 new->dataBrowser.array.form,
				 XtNborderWidth, 0,
				 XtNhorizDistance, 0,
				 XtNfromVert, upper_widget,
				 XtNlabel, "Dark:",
				 NULL);
    w = XtVaCreateManagedWidget ("colourText", simpleTextEntryWidgetClass,
				 new->dataBrowser.array.form,
				 XtNhorizDistance, 0,
				 XtNfromVert, upper_widget,
				 XtNfromHoriz, w,
				 XtNwidth, 115,
				 XtNeditType, XawtextEdit,
				 NULL);
    new->dataBrowser.array.profile_dark_colour_txt = w;
    XtAddCallback (w, XkwNnewLineCallback, _DataBrowser_dark_colour_cbk, New);
    w = XtVaCreateManagedWidget ("lightColourLabel", labelWidgetClass,
				 new->dataBrowser.array.form,
				 XtNborderWidth, 0,
				 XtNhorizDistance, 0,
				 XtNfromVert, upper_widget,
				 XtNfromHoriz, w,
				 XtNlabel, "Light:",
				 NULL);
    w = XtVaCreateManagedWidget ("colourText", simpleTextEntryWidgetClass,
				 new->dataBrowser.array.form,
				 XtNhorizDistance, 0,
				 XtNfromVert, upper_widget,
				 XtNfromHoriz, w,
				 XtNwidth, 115,
				 XtNeditType, XawtextEdit,
				 NULL);
    upper_widget = w;
    new->dataBrowser.array.profile_light_colour_txt = w;
    XtAddCallback (w, XkwNnewLineCallback, _DataBrowser_light_colour_cbk, New);
    /*  Create overlay data-specific widgets  */
    w = XtVaCreateManagedWidget ("overlayForm", formWidgetClass, form,
				 XtNborderWidth, 0,
				 XtNhorizDistance, 0,
				 XtNvertDistance, 0,
				 XtNfromVert, list,
				 XtNtop, XtChainBottom,
				 XtNbottom, XtChainBottom,
				 NULL);
    new->dataBrowser.overlay.form = w;
    w = XtVaCreateManagedWidget ("sourceLabel", labelWidgetClass,
				 new->dataBrowser.overlay.form,
				 XtNborderWidth, 0,
				 XtNwidth, WIDTH,
				 XtNhorizDistance, 0,
				 XtNvertDistance, 0,
				 XtNleft, XawChainLeft,
				 XtNright, XawChainRight,
				 XtNjustify, XtJustifyLeft,
				 XtNresize, False,
				 NULL);
    upper_widget = w;
    new->dataBrowser.overlay.source_lbl = w;
    w = XtVaCreateManagedWidget ("destroyButton", commandWidgetClass,
				 new->dataBrowser.overlay.form,
				 XtNlabel, "Unload",
				 XtNfromVert, upper_widget,
				 NULL);
    XtAddCallback (w, XtNcallback, _DataBrowser_overlay_unload_cbk, new);
    w = XtVaCreateManagedWidget ("divider", simpleWidgetClass,
				 new->dataBrowser.overlay.form,
				 XtNborderWidth, 0,
				 XtNwidth, WIDTH,
				 XtNheight, 2,
				 XtNhorizDistance, 0,
				 XtNfromVert, w,
				 XtNleft, XawChainLeft,
				 XtNright, XawChainRight,
				 XtVaTypedArg,
				      XtNbackground,
				      XtRString,
				      "black",
				      6,
				 NULL);
    upper_widget = w;
    w = XtVaCreateManagedWidget ("toggle", ktoggleWidgetClass,
				 new->dataBrowser.overlay.form,
				 XtNlabel, "Active",
				 XkwNcrosses, False,
				 XtNfromVert, upper_widget,
				 NULL);
    new->dataBrowser.overlay.active_tgl = w;
    XtAddCallback (w, XtNcallback, _DataBrowser_overlay_active_cbk, new);
    /*  Create generic blinkstate  */
    blinkstate->generic_blinkstate =
	dispdata_create_blinkstate (blinkstate->window,
				    _DataBrowser_active_change, blinkstate);
    /*  Create array data-specific blinkstate  */
    m_clear (&blinkstate->array_list, sizeof blinkstate->array_list);
    blinkstate->array_list.parent = blinkstate;
    blinkstate->array_list.dataclass = dispdata_array_create_dataclass ();
    blinkstate->array_list.class_blinkstate =
	dispdata_get_blinkstate_class (blinkstate->generic_blinkstate,
				       blinkstate->array_list.dataclass);
    dispdata_register_blinkstate_funcs
	(blinkstate->array_list.class_blinkstate,
	 _DataBrowser_array_refresh_func,
	 _DataBrowser_array_create_blink_entry,
	 _DataBrowser_array_destroy_blink_entry,
	 _DataBrowser_array_copy_blink_entry,
	 &blinkstate->array_list);
    /*  Create overlay data-specific blinkstate  */
    m_clear (&blinkstate->overlay_list, sizeof blinkstate->overlay_list);
    blinkstate->overlay_list.parent = blinkstate;
    blinkstate->overlay_list.dataclass = dispdata_overlay_create_dataclass ();
    blinkstate->overlay_list.class_blinkstate =
	dispdata_get_blinkstate_class (blinkstate->generic_blinkstate,
				       blinkstate->overlay_list.dataclass);
    dispdata_register_blinkstate_funcs
	(blinkstate->overlay_list.class_blinkstate,
	 _DataBrowser_overlay_refresh_func,
	 _DataBrowser_overlay_create_blink_entry,
	 _DataBrowser_overlay_destroy_blink_entry,
	 _DataBrowser_overlay_copy_blink_entry,
	 &blinkstate->overlay_list);
    /*  Generic code follows  */
    active = dispdata_get_active_blinkstate (blinkstate->window, NULL);
    if (active)
	XtVaSetValues (new->dataBrowser.active_btn, XtNlabel, "Inactive",NULL);
    else
    {
	KWorldCanvas pseudo, direct, true;

	dispdata_activate_blinkstate (blinkstate->generic_blinkstate);
	XtVaGetValues (image_display,
		       XkwNpseudoColourCanvas, &pseudo,
		       XkwNdirectColourCanvas, &direct,
		       XkwNtrueColourCanvas, &true,
		       NULL);
	if (pseudo)
	{
	    canvas_register_position_event_func
		(pseudo, _DataBrowser_canvas_position_func,blinkstate->window);
	    kwin_register_bindings (canvas_get_pixcanvas (pseudo),
				    main_bindings, FALSE);
	}
	if (direct)
	{
	    canvas_register_position_event_func
		(direct, _DataBrowser_canvas_position_func,blinkstate->window);
	    kwin_register_bindings (canvas_get_pixcanvas (direct),
				    main_bindings, FALSE);
	}
	if (true)
	{
	    canvas_register_position_event_func
		(true, _DataBrowser_canvas_position_func, blinkstate->window);
	    kwin_register_bindings (canvas_get_pixcanvas (true),
				    main_bindings, FALSE);
	}
    }
    new->dataBrowser.num_entries = &blinkstate->array_list.num_entries;
    new->dataBrowser.show_details = _DataBrowser_array_show_details;
    new->dataBrowser.process_event = _DataBrowser_array_process_event;
    _DataBrowser_refresh (new);
    switch (browser_mode)
    {
      case BROWSER_MODE_ALL:
	break;
      case BROWSER_MODE_SOME:
	if (only_for_window) break;
	/*  Fall through  */
      case BROWSER_MODE_ONE:
	new->dataBrowser.current_blinkstate = NULL;
	blinkstate->viewer = NULL;
	break;
    }
    if (only_for_window)
	dispdata_register_window_renumber_func (blinkstate->window,
						_DataBrowser_set_identities,
						blinkstate->window);
    _DataBrowser_set_identities (blinkstate->window);
    if (first_time)
    {
	first_time = FALSE;
	if (new->dataBrowser.popupFirstDataBrowser)
	    XkwDataBrowserPopup (blinkstate->generic_blinkstate);
	round_robin_bs = blinkstate->generic_blinkstate;
    }
    if (new->dataBrowser.createAtInit)
	_DataBrowser_create_advanced_control (New);
}   /*  End Function Initialise  */

static void _DataBrowser_Destroy (Widget W)
{
    DataBrowserWidget w = (DataBrowserWidget) W;
    DataBrowserWidget viewer = w->dataBrowser.my_blinkstate->viewer;
    struct blinkstate *first_blinkstate, *blinkstate;

    /*  The order of destroy callbacks (both from the <dispdata> package and
	Xt) is subtle. Do lots of thinking before changing this function  */
    dispdata_get_first_blinkstate (dispdata_get_first_window (NULL),
				   (void **) &first_blinkstate);
    if ( !w->dataBrowser.my_blinkstate->window &&
	 (browser_mode == BROWSER_MODE_ONE) )
	_DataBrowser_set_image_display_bg (first_blinkstate->owner, TRUE);
    if ( viewer && (viewer != w) )
    {   /*  Someone else is showing me: make them show themself  */
	blinkstate = viewer->dataBrowser.my_blinkstate;
	blinkstate->viewer = viewer;
	viewer->dataBrowser.current_blinkstate = blinkstate;
	_DataBrowser_refresh (viewer);
    }
    if ( w->dataBrowser.current_blinkstate &&
	 (browser_mode != BROWSER_MODE_ALL) )
    {   /*  I'm showing something: make someone else show a blinkstate  */
	blinkstate = w->dataBrowser.current_blinkstate;  /*  Prefer same one */
	if (blinkstate->window)
	{   /*  Show a blinkstate for their window  */
	    if (w->dataBrowser.my_blinkstate == blinkstate)
	    {   /*  I'm showing myself, so show first blinkstate for window  */
		dispdata_get_first_blinkstate (blinkstate->window,
					       (void **) &blinkstate);
	    }
	}
	else if (browser_mode == BROWSER_MODE_ONE)  /*  Their window died  */
	    blinkstate = first_blinkstate;
	else blinkstate = NULL;  /*  Nothing to show  */
	/*  Disconnect me so that I'm not referenced again. This has to be done
	    if the window has died, optional otherwise  */
	w->dataBrowser.current_blinkstate->viewer = NULL;
	if (blinkstate)
	{
	    blinkstate->viewer = blinkstate->owner;
	    blinkstate->owner->dataBrowser.current_blinkstate = blinkstate;
	    _DataBrowser_refresh (blinkstate->owner);
	    if (blinkstate->owner->dataBrowser.user_popped_up)
		XtPopup ( (Widget) blinkstate->owner, XtGrabNone );
	    if (browser_mode == BROWSER_MODE_ONE)
		current_one_browser = blinkstate->owner;
	}
    }
    blinkstate = w->dataBrowser.my_blinkstate;
    if (blinkstate->window)
    {
	if (round_robin_bs == blinkstate->generic_blinkstate)
	    round_robin_bs = _DataBrowser_get_next_blinkstate (round_robin_bs);
	dispdata_destroy_blinkstate (blinkstate->generic_blinkstate);
	c_unregister_callback (blinkstate->destroy_func);
	_DataBrowser_set_identities (blinkstate->window);
    }
    else if (round_robin_bs == blinkstate->generic_blinkstate)
	round_robin_bs = first_blinkstate->generic_blinkstate;
    dispdata_free_namelist (blinkstate->array_list.namelist);
    dispdata_free_namelist (blinkstate->overlay_list.namelist);
    m_free (blinkstate);
}   /*  End Function Destroy  */

static void _DataBrowser_listcanvas_realise_cbk (Widget w,
						 XtPointer client_data,
						 XtPointer call_data)
/*  [SUMMARY] Canvas realise callback.
    [PURPOSE] This routine is called when a Canvas widget is realised.
    <w> The Canvas widget that has just been realised.
    <client_data> The ScrollableList widget.
    <call_data> The pixel canvas(es).
    [RETURNS] Nothing.
*/
{
    KCallbackFunc canvas_callback;

    canvas_callback =
	kwin_register_position_event_func (call_data,
					   ( flag (*) () ) _DataBrowser_listcanvas_position_func,
					   client_data);
    c_move_callback (canvas_callback, FALSE);
}   /*  End Function listcanvas_realise_cbk   */

static void _DataBrowser_close_cbk (Widget w, XtPointer client_data,
				    XtPointer call_data)
{
    DataBrowserWidget top = client_data;

    XtPopdown ( (Widget) top );
    top->dataBrowser.user_popped_up = FALSE;
}   /*  End Function close_cbk  */

static void _DataBrowser_destroy_cbk (Widget w, XtPointer client_data,
				      XtPointer call_data)
{
    KDisplayDataWindow window;
    KDisplayDataBlinkState bs;
    DataBrowserWidget top = client_data;
    struct blinkstate *blinkstate = top->dataBrowser.current_blinkstate;

    bs = blinkstate->generic_blinkstate;
    if ( dispdata_test_blinkstate_active (bs) )
    {
	window = dispdata_get_window_from_blinkstate (bs);
	bs = dispdata_get_first_blinkstate (window, (void **) &blinkstate);
	dispdata_activate_blinkstate (bs);
    }
    XtDestroyWidget ( (Widget) top->dataBrowser.current_blinkstate->owner );
}   /*  End Function destroy_cbk  */

static void _DataBrowser_copy_cbk (Widget w, XtPointer client_data,
				   XtPointer call_data)
{
    DataBrowserWidget top = client_data;
    DataBrowserWidget new;
    Widget image_display, NEW;
    struct blinkstate *blinkstate = top->dataBrowser.current_blinkstate;

    image_display = blinkstate->owner->dataBrowser.image_display_widget;
    NEW = XtVaCreatePopupShell ("dataBrowserShell", dataBrowserWidgetClass,
				image_display, NULL);
    new = (DataBrowserWidget) NEW;
    blinkstate = new->dataBrowser.my_blinkstate;
    dispdata_copy_blinkstate
	(blinkstate->generic_blinkstate,
	 top->dataBrowser.current_blinkstate->generic_blinkstate);
    switch (browser_mode)
    {
      case BROWSER_MODE_ALL:
	dispdata_refresh_blinkstate (blinkstate->generic_blinkstate);
	XtPopup (NEW, XtGrabNone);
	new->dataBrowser.user_popped_up = TRUE;
	break;
      case BROWSER_MODE_SOME:
      case BROWSER_MODE_ONE:
	blinkstate->viewer = NULL;
	new->dataBrowser.current_blinkstate = NULL;
	break;
    }
}   /*  End Function copy_cbk  */

static void _DataBrowser_active_cbk (Widget w, XtPointer client_data,
				     XtPointer call_data)
{
    DataBrowserWidget top = client_data;
    struct blinkstate *blinkstate = top->dataBrowser.current_blinkstate;

    dispdata_activate_blinkstate (blinkstate->generic_blinkstate);
}   /*  End Function active_cbk  */

static void _DataBrowser_browser_mode_cbk (Widget w, XtPointer client_data,
					   XtPointer call_data)
/*  This is the browser mode callback.
*/
{
    KDisplayDataWindow window, win;
    KDisplayDataBlinkState bs;
    int new_mode = *(int *) call_data;
    DataBrowserWidget top = client_data;
    struct blinkstate *blinkstate = top->dataBrowser.current_blinkstate;

    if (browser_mode == new_mode) return;
    browser_mode = new_mode;
    window = top->dataBrowser.my_blinkstate->window;
    switch (new_mode)
    {
      case BROWSER_MODE_ALL:
	for ( win = dispdata_get_first_window (NULL);
	      win;
	      win = dispdata_get_next_window (win, FALSE, NULL) )
	{
	    bs = dispdata_get_first_blinkstate (win, (void **) &blinkstate);
	    _DataBrowser_set_image_display_bg (blinkstate->owner, FALSE);
	    for ( ; bs;
		  bs = dispdata_get_next_blinkstate (bs, FALSE,
						     (void **) &blinkstate) )
	    {
		blinkstate->viewer = blinkstate->owner;
		blinkstate->owner->dataBrowser.current_blinkstate = blinkstate;
		_DataBrowser_refresh (blinkstate->owner);
		if (blinkstate->owner->dataBrowser.user_popped_up)
		    XtPopup ( (Widget) blinkstate->owner, XtGrabNone );
	    }
	}
	current_one_browser = NULL;
	break;
      case BROWSER_MODE_SOME:
	for ( win = dispdata_get_first_window (NULL);
	      win;
	      win = dispdata_get_next_window (win, FALSE, NULL) )
	{
	    bs = dispdata_get_first_blinkstate (win, (void **) &blinkstate);
	    _DataBrowser_set_image_display_bg (blinkstate->owner, FALSE);
	    for ( ; bs;
		  bs = dispdata_get_next_blinkstate (bs, FALSE,
						     (void **) &blinkstate) )
	    {
		if (blinkstate->owner == top)
		{   /*  Show my own blinkstate  */
		    blinkstate->viewer = blinkstate->owner;
		    top->dataBrowser.current_blinkstate = blinkstate;
		    _DataBrowser_refresh (blinkstate->owner);
		}
		else if ( (win != window) &&
			  dispdata_test_blinkstate_active (bs) )
		{
		    blinkstate->viewer = blinkstate->owner;
		    blinkstate->owner->dataBrowser.current_blinkstate = blinkstate;
		    _DataBrowser_refresh (blinkstate->owner);
		    if (blinkstate->owner->dataBrowser.user_popped_up)
			XtPopup ( (Widget) blinkstate->owner, XtGrabNone );
		}
		else
		{
		    blinkstate->viewer = NULL;
		    blinkstate->owner->dataBrowser.current_blinkstate = NULL;
		    XtPopdown ( (Widget) blinkstate->owner );
		}
	    }
	}
	current_one_browser = NULL;
	break;
      case BROWSER_MODE_ONE:
	for ( win = dispdata_get_first_window (NULL);
	      win;
	      win = dispdata_get_next_window (win, FALSE, NULL) )
	{
	    bs = dispdata_get_first_blinkstate (win, (void **) &blinkstate);
	    if (blinkstate->owner != top)
		_DataBrowser_set_image_display_bg (blinkstate->owner, FALSE);
	    for ( ; bs;
		  bs = dispdata_get_next_blinkstate (bs, FALSE,
						     (void **) &blinkstate) )
	    {
		if (blinkstate->owner == top) continue;
		blinkstate->viewer = NULL;
		blinkstate->owner->dataBrowser.current_blinkstate = NULL;
		XtPopdown ( (Widget) blinkstate->owner );
	    }
	}
	_DataBrowser_set_image_display_bg (top, TRUE);
	current_one_browser = top;
	break;
    }
}   /*  End Function browser_mode_cbk  */

static void _DataBrowser_dataclass_mode_cbk (Widget w, XtPointer client_data,
					     XtPointer call_data)
/*  This is the display data mode callback.
*/
{
    int new_mode = *(int *) call_data;
    DataBrowserWidget top = client_data;

    if (top->dataBrowser.dataclass_mode == new_mode) return;
    top->dataBrowser.dataclass_mode = new_mode;
    _DataBrowser_refresh (top);
}   /*  End Function dataclass_mode_cbk  */

static void _DataBrowser_refresh (DataBrowserWidget w)
{
    KDisplayDataWindow window;
    KDisplayDataBlinkState bs;
    struct blinkstate *blinkstate = w->dataBrowser.current_blinkstate;

    XtVaSetValues (w->dataBrowser.browser_mode_menu,
		   XkwNsetChoice, browser_mode_choices[browser_mode],
		   NULL);
    if (!blinkstate->window) return;
    bs = blinkstate->generic_blinkstate;
    window = blinkstate->window;
    _DataBrowser_set_identity (blinkstate);
    XtSetSensitive (w->dataBrowser.destroy_btn,
		    (dispdata_get_first_blinkstate (window, NULL) ==
		     bs) ? False : True);
    _DataBrowser_active_change ( blinkstate,
				 dispdata_test_blinkstate_active (bs) );
    if (dataclasses[w->dataBrowser.dataclass_mode] ==
	blinkstate->array_list.dataclass)
    {
	XtSetMappedWhenManaged (w->dataBrowser.overlay.form, False);
	w->dataBrowser.num_entries = &blinkstate->array_list.num_entries;
	w->dataBrowser.show_details = _DataBrowser_array_show_details;
	w->dataBrowser.process_event = _DataBrowser_array_process_event;
	_DataBrowser_array_refresh_func (&blinkstate->array_list,
					 blinkstate->array_list.dataclass);
    }
    else if (dataclasses[w->dataBrowser.dataclass_mode] ==
	     blinkstate->overlay_list.dataclass)
    {
	XtSetMappedWhenManaged (w->dataBrowser.array.form, False);
	w->dataBrowser.num_entries = &blinkstate->overlay_list.num_entries;
	w->dataBrowser.show_details = _DataBrowser_overlay_show_details;
	w->dataBrowser.process_event = _DataBrowser_overlay_process_event;
	_DataBrowser_overlay_refresh_func (&blinkstate->overlay_list,
					   blinkstate->overlay_list.dataclass);
    }
}   /*  End Function refresh  */

static void _DataBrowser_set_image_display_bg (DataBrowserWidget w,flag active)
{
    if (active)
	XtVaSetValues (w->dataBrowser.image_display_widget,
		       XtVaTypedArg,
		           XtNbackground,
		           XtRString,
		           "red",
		           4,
		       NULL);
    else XtVaSetValues (w->dataBrowser.image_display_widget,
			XtNbackground, w->dataBrowser.image_display_bg,
			NULL);
}   /*  End Function set_image_display_bg  */

static void _DataBrowser_set_identity (struct blinkstate *blinkstate)
{
    char txt[STRING_LENGTH];

    if (!blinkstate->viewer) return;
    sprintf (txt, "Browser %u of %u for display window %u",
	     dispdata_get_blinkstate_index (blinkstate->generic_blinkstate) +1,
	     dispdata_get_num_blinkstates (blinkstate->window),
	     dispdata_get_window_index (blinkstate->window) + 1);
    XtVaSetValues ( (Widget) blinkstate->viewer, XtNtitle, txt, NULL );
}   /*  End Function set_identity  */

static void _DataBrowser_set_identities (KDisplayDataWindow window)
{
    KDisplayDataBlinkState bs;
    struct blinkstate *blinkstate;

    for ( bs = dispdata_get_first_blinkstate (window, (void **) &blinkstate);
	  bs;
	  bs = dispdata_get_next_blinkstate (bs, FALSE,
					     (void **) &blinkstate) )
	_DataBrowser_set_identity (blinkstate);
}   /*  End Function set_identities  */

static flag _DataBrowser_listcanvas_position_func (KPixCanvas canvas,
						   int x, int y,
						   unsigned int event_code,
						   void *e_info, void **f_info)
{
    KDisplayDataWindow window;
    KDisplayDataBlinkState bs;
    int db_event = -1;
    unsigned long mask = 0;
    DataBrowserWidget top = *f_info;
    struct blinkstate *blinkstate = top->dataBrowser.current_blinkstate;
    struct blinkstate *next;
    char *key;

    switch (event_code)
    {
      case K_CANVAS_EVENT_KEYSYM_PRESS:
	key = e_info;
	if (strncmp (key, "KP_", 3) == 0) key += 3;
	m_copy (&mask, key + strlen (key) + 1, sizeof mask);
	if (strcmp (key, "Left") == 0) db_event = EVENT_CODE_LEFT_ARROW;
	else if (strcmp (key, "Right") == 0) db_event = EVENT_CODE_RIGHT_ARROW;
	break;
      case K_CANVAS_EVENT_PLAIN_KEY_PRESS:
	key = e_info;
	db_event = *key;
	break;
      case K_CANVAS_EVENT_MODIFIED_KEY_PRESS:
	key = e_info;
	m_copy (&mask, key + strlen (key) + 1, sizeof mask);
	db_event = *key;
	break;
    }
    if (db_event < 0) return (FALSE);
    /*  Do special event processing that doesn't care about list contents  */
    bs = blinkstate->generic_blinkstate;
    window = dispdata_get_window_from_blinkstate (bs);
    switch (db_event)
    {
      case EVENT_CODE_LEFT_ARROW:
	if (browser_mode == BROWSER_MODE_ALL) return (TRUE);
	if (mask & K_CANVAS_EVENT_CONTROL_KEY_MASK)
	{
	    if (browser_mode != BROWSER_MODE_ONE) return (TRUE);
	    _DataBrowser_set_image_display_bg (blinkstate->owner, FALSE);
	    window = dispdata_get_prev_window (window, TRUE, NULL);
	    bs = dispdata_get_last_blinkstate (window, (void **) &next);
	    _DataBrowser_set_image_display_bg (next->owner, TRUE);
	}
	else bs = dispdata_get_prev_blinkstate (bs, TRUE, (void **) &next);
	if (mask & K_CANVAS_EVENT_SHIFTED_MASK)
	    dispdata_activate_blinkstate (bs);
	blinkstate->viewer = NULL;
	next->viewer = top;
	top->dataBrowser.current_blinkstate = next;
	_DataBrowser_refresh (top);
	return (TRUE);
	/*break;*/
      case EVENT_CODE_RIGHT_ARROW:
	if (browser_mode == BROWSER_MODE_ALL) return (TRUE);
	if (mask & K_CANVAS_EVENT_CONTROL_KEY_MASK)
	{
	    if (browser_mode != BROWSER_MODE_ONE) return (TRUE);
	    _DataBrowser_set_image_display_bg (blinkstate->owner, FALSE);
	    window = dispdata_get_next_window (window, TRUE, NULL);
	    bs = dispdata_get_first_blinkstate (window, (void **) &next);
	    _DataBrowser_set_image_display_bg (next->owner, TRUE);
	}
	else bs = dispdata_get_next_blinkstate (bs, TRUE, (void **) &next);
	if (mask & K_CANVAS_EVENT_SHIFTED_MASK)
	    dispdata_activate_blinkstate (bs);
	blinkstate->viewer = NULL;
	next->viewer = top;
	top->dataBrowser.current_blinkstate = next;
	_DataBrowser_refresh (top);
	return (TRUE);
	/*break;*/
      case 'b':
      case '>':
	bs = dispdata_get_active_blinkstate (window, NULL);
	bs = dispdata_get_next_blinkstate (bs, TRUE, NULL);
	dispdata_activate_blinkstate (bs);
	return (TRUE);
	/*break;*/
      case '<':
	bs = dispdata_get_active_blinkstate (window, NULL);
	bs = dispdata_get_prev_blinkstate (bs, TRUE, NULL);
	dispdata_activate_blinkstate (bs);
	return (TRUE);
	/*break;*/
      case 'B':
	dispdata_activate_blinkstate (bs);
	return (TRUE);
	/*break;*/
    }
    return (FALSE);
}   /*  End Function listcanvas_position_func  */

static flag _DataBrowser_list_position_func (KListWidget list,
					     unsigned int index,
					     unsigned int event_code,
					     void *e_info, void **f_info)
{
    int db_event = -1;
    unsigned long mask = 0;
    DataBrowserWidget top = *f_info;
    char *key;

    switch (event_code)
    {
      case K_CANVAS_EVENT_POINTER_MOVE:
      case K_CANVAS_EVENT_POINTER_ENTER:
	db_event = EVENT_CODE_MOUSE_MOVE;
	break;
      case K_CANVAS_EVENT_LEFT_MOUSE_RELEASE:
	db_event = EVENT_CODE_LEFT_MOUSE;
	break;
      case K_CANVAS_EVENT_MIDDLE_MOUSE_RELEASE:
	db_event = EVENT_CODE_MIDDLE_MOUSE;
	break;
      case K_CANVAS_EVENT_RIGHT_MOUSE_RELEASE:
	db_event = EVENT_CODE_RIGHT_MOUSE;
	break;
      case K_CANVAS_EVENT_PLAIN_KEY_PRESS:
	key = e_info;
	db_event = *key;
	break;
      case K_CANVAS_EVENT_MODIFIED_KEY_PRESS:
	key = e_info;
	m_copy (&mask, key + strlen (key) + 1, sizeof mask);
	db_event = *key;
	break;
    }
    if (db_event < 0) return (FALSE);
    switch (db_event)
    {
      case EVENT_CODE_MOUSE_MOVE:
	if (top->dataBrowser.active_mode)
	    _DataBrowser_select_entry (top, index);
	break;
      case EVENT_CODE_LEFT_MOUSE:
      case ' ':
	if (top->dataBrowser.active_mode)
	    top->dataBrowser.active_mode = FALSE;
	else if (index == top->dataBrowser.selected)
	    top->dataBrowser.active_mode = TRUE;
	_DataBrowser_select_entry (top, index);
	break;
      default:
	(*top->dataBrowser.process_event) (top, index, db_event, mask);
	break;
    }
    return (TRUE);
}   /*  End Function list_position_func  */

static void _DataBrowser_select_entry (DataBrowserWidget w, unsigned int index)
{
    if (index == w->dataBrowser.selected) return;
    w->dataBrowser.selected = index;
    listw_set_attributes (w->dataBrowser.listw, TRUE,
			  LISTW_ATT_REVERSE_VIDEO_INDEX, index,
			  LISTW_ATT_END);
    (w->dataBrowser.show_details) (w);
}   /*  End Function select_entry  */

static void _DataBrowser_show_source (KDisplayDataObject dataobject,Widget lbl)
{
    char *domain, *directory, *ptr;
    char line[STRING_LENGTH];
    char *label = line;

    dispdata_get_data_attributes
	(dataobject,
	 KDISPLAYDATA_DATA_ATT_DOMAIN, &domain,
	 KDISPLAYDATA_DATA_ATT_DIRECTORY, &directory,
	 KDISPLAYDATA_DATA_ATT_END);
    ptr = strstr (domain, "FILE");
    if (ptr) sprintf (line, "FB: %s Dir: %s", ptr + 4, directory);
    else if ( ( ptr = strstr (domain, "COMPOSITE:") ) != NULL )
	label = domain + 10;
    else sprintf (line, "Domain: %s", domain);
    XtVaSetValues (lbl, XtNlabel, label, NULL);
}   /*  End Function show_source  */

static void _DataBrowser_array_refresh_func (void *info,
					     KDisplayDataClass dataclass)
{
    KDisplayDataObject new_dataobject;
    flag image_is_cube = FALSE;
    int first_cube = -1;
    unsigned int contour_mode, count;
    DataBrowserWidget browser;
    Bool show_new_data_as_image = False;
    struct array_list *list = info;
    struct blinkstate *blinkstate = list->parent;

    browser = blinkstate->viewer;
    if (!browser) return;
    if (dataclasses[browser->dataBrowser.dataclass_mode] != dataclass) return;
    /*  Construct list  */
    list->strings = dispdata_regenerate_namelist (dataclass,
						  &list->namelist,
						  NUM_RESERVED_CHAR_ARRAY,
						  &list->num_entries,
						  NULL, NULL, NULL);
    list->blinkentries =
	dispdata_regenerate_blinkentry_list (list->class_blinkstate,
					     &list->num_entries);
    dispdata_get_dataclass_attributes
	(dataclass,
	 KDISPLAYDATA_CLASS_ATT_CURRENTLY_CREATING, &new_dataobject,
	 KDISPLAYDATA_CLASS_ATT_END);
    if (new_dataobject && (blinkstate->generic_blinkstate == round_robin_bs)
	&& show_new_data_as_image_tgl)
	XtVaGetValues (show_new_data_as_image_tgl,
		       XtNstate, &show_new_data_as_image,
		       NULL);
    if ( !list->main_image && (list->num_entries > 0) )
    {
	dispdata_set_blinkentry_attributes
	    (list->blinkentries[0],
	     KDISPLAYDATA_BLINKENTRY_ATT_IMAGE, TRUE,
	     KDISPLAYDATA_BLINKENTRY_ATT_END);
    }
    list->movie_control = NULL;
    list->main_profile = NULL;
    for (count = 0; count < list->num_entries; ++count)
    {
	KDisplayDataBlinkEntry be = list->blinkentries[count];
	KDisplayDataObject dataobject;
	flag replace;
	flag image, movie_master, movie_slave, profile_master, profile_slave;
	iarray pseudo, red, arr;
	struct array_blink_entry *blinkentry;

	dataobject = dispdata_get_dataobject_from_blinkentry (be);
	dispdata_get_data_attributes
	    (dataobject,
	     KDISPLAYDATA_DATA_ATT_REPLACE, &replace,
	     KDISPLAYDATA_DATA_ATT_PSEUDO_ARRAY, &pseudo,
	     KDISPLAYDATA_DATA_ATT_RED_ARRAY, &red,
	     KDISPLAYDATA_DATA_ATT_END);
	dispdata_get_blinkentry_attributes
	    (be,
	     KDISPLAYDATA_BLINKENTRY_ATT_IMAGE, &image,
	     KDISPLAYDATA_BLINKENTRY_ATT_CONTOUR_MODE, &contour_mode,
	     KDISPLAYDATA_BLINKENTRY_ATT_MOVIE_MASTER, &movie_master,
	     KDISPLAYDATA_BLINKENTRY_ATT_MOVIE_SLAVE, &movie_slave,
	     KDISPLAYDATA_BLINKENTRY_ATT_PROFILE_MASTER, &profile_master,
	     KDISPLAYDATA_BLINKENTRY_ATT_PROFILE_SLAVE, &profile_slave,
	     KDISPLAYDATA_BLINKENTRY_ATT_END);
	if (show_new_data_as_image)
	{
	    if (dataobject == new_dataobject)
	    {
		image = TRUE;
		dispdata_set_blinkentry_attributes
		    (be,
		     KDISPLAYDATA_BLINKENTRY_ATT_IMAGE, TRUE,
		     KDISPLAYDATA_BLINKENTRY_ATT_END);
	    }
	    else image = FALSE;
	}
	arr = pseudo ? pseudo : red;
	blinkentry = dispdata_get_blinkentry_info (be);
	blinkentry->index = count;
	if ( (iarray_num_dim (arr) > 2) && (first_cube == -1) )
	    first_cube = count;
	if (replace) list->strings[count][CHAR_INDEX_REPLACE] = '=';
	if (image)
	{
	    list->strings[count][CHAR_INDEX_ARRAY_IMAGE] = 'I';
	    list->main_image = blinkentry;
	    if (list->alt_image == blinkentry) list->alt_image = NULL;
	    if (iarray_num_dim (arr) > 2) image_is_cube = TRUE;
	}
	if (!list->alt_image && !image && pseudo) list->alt_image = blinkentry;
	if (list->alt_image == blinkentry)
	    list->strings[count][CHAR_INDEX_ARRAY_IMAGE] = 'A';
	switch (contour_mode)
	{
	  case KDISPLAYDATA_CONTOUR_NORMAL:
	    list->strings[count][CHAR_INDEX_ARRAY_CONTOUR] = 'c';
	    break;
	  case KDISPLAYDATA_CONTOUR_RENZOGRAM:
	    list->strings[count][CHAR_INDEX_ARRAY_CONTOUR] = 'r';
	    break;
	}
	if (!list->user_selected_movie_control)
	{   /*  Automatic movie master selection required  */
	    movie_master = (image && (iarray_num_dim (arr) > 2) && pseudo ) ?
		TRUE : FALSE;
	    dispdata_set_blinkentry_attributes
		(be,
		 KDISPLAYDATA_BLINKENTRY_ATT_MOVIE_MASTER, movie_master,
		 KDISPLAYDATA_BLINKENTRY_ATT_END);
	}
	if (movie_master)
	{
	    list->strings[count][CHAR_INDEX_ARRAY_MOVIE] = 'M';
	    list->movie_control = blinkentry;
	}
	else if (movie_slave)
	    list->strings[count][CHAR_INDEX_ARRAY_MOVIE] = 'm';
	if (!list->user_selected_main_profile)
	{   /*  Automatic profile master selection required  */
	    profile_master = (image && pseudo) ? TRUE : FALSE;
	    dispdata_set_blinkentry_attributes
		(be,
		 KDISPLAYDATA_BLINKENTRY_ATT_PROFILE_MASTER, profile_master,
		 KDISPLAYDATA_BLINKENTRY_ATT_END);
	}
	if (profile_master)
	{
	    list->strings[count][CHAR_INDEX_ARRAY_PROFILE] = 'P';
	    list->main_profile = blinkentry;
	}
	else if (profile_slave)
	    list->strings[count][CHAR_INDEX_ARRAY_PROFILE] = 'p';
    }
    if ( !list->movie_control && (first_cube >= 0) )
    {
	list->strings[first_cube][CHAR_INDEX_ARRAY_MOVIE] = 'M';
	dispdata_set_blinkentry_attributes
	    (list->blinkentries[first_cube],
	     KDISPLAYDATA_BLINKENTRY_ATT_MOVIE_MASTER, TRUE,
	     KDISPLAYDATA_BLINKENTRY_ATT_END);
	list->movie_control =
	    dispdata_get_blinkentry_info (list->blinkentries[first_cube]);
    }
    if ( !list->main_profile && (first_cube >= 0) )
    {
	flag profile_master;

	dispdata_set_blinkentry_attributes
	    (list->blinkentries[first_cube],
	     KDISPLAYDATA_BLINKENTRY_ATT_PROFILE_MASTER, TRUE,
	     KDISPLAYDATA_BLINKENTRY_ATT_END);
	dispdata_get_blinkentry_attributes
	    (list->blinkentries[first_cube],
	     KDISPLAYDATA_BLINKENTRY_ATT_PROFILE_MASTER, &profile_master,
	     KDISPLAYDATA_BLINKENTRY_ATT_END);
	if (profile_master)
	{
	    list->strings[first_cube][CHAR_INDEX_ARRAY_PROFILE] = 'P';
	    list->movie_control =
		dispdata_get_blinkentry_info (list->blinkentries[first_cube]);
	}
    }
    if (list->num_entries > 0)
    {
	if (browser->dataBrowser.selected >= list->num_entries)
	    browser->dataBrowser.selected = list->num_entries - 1;
	_DataBrowser_array_show_details (browser);
	XtSetMappedWhenManaged (browser->dataBrowser.array.form, True);
    }
    else
    {
	browser->dataBrowser.selected = 0;
	XtSetMappedWhenManaged (browser->dataBrowser.array.form, False);
    }
    listw_set_attributes (browser->dataBrowser.listw, TRUE,
			  LISTW_ATT_NUM_STRINGS, list->num_entries,
			  LISTW_ATT_STRINGS, list->strings,
			  LISTW_ATT_REVERSE_VIDEO_INDEX,
			      browser->dataBrowser.selected,
			  LISTW_ATT_END);
    _DataBrowser_set_movie_controls (list->parent);
}   /*  End Function array_refresh_func  */

static void *_DataBrowser_array_create_blink_entry (void *info,
						    KDisplayDataBlinkEntry be)
{
    Bool slave_movie_on_load = True;
    struct array_list *list = info;
    struct array_blink_entry *blinkentry;
    static char function_name[] = "DataBrowser::array_create_blink_entry";

    if ( ( blinkentry = m_calloc (sizeof *blinkentry) ) == NULL )
    {
	m_error_notify (function_name, "blink entry");
	return (NULL);
    }
    blinkentry->parent = list;
    blinkentry->generic_blinkentry = be;
    if (slave_movies_on_load_tgl)
	XtVaGetValues (slave_movies_on_load_tgl,
		       XtNstate, &slave_movie_on_load,
		       NULL);
    dispdata_set_blinkentry_attributes
	(blinkentry->generic_blinkentry,
	 KDISPLAYDATA_BLINKENTRY_ATT_MOVIE_SLAVE, slave_movie_on_load,
	 KDISPLAYDATA_BLINKENTRY_ATT_END);
    return (blinkentry);
}   /*  End Function array_create_blink_entry  */

static void _DataBrowser_array_destroy_blink_entry (void *info)
{
    struct array_blink_entry *blinkentry = info;

    if (blinkentry->parent->main_image == blinkentry)
	blinkentry->parent->main_image = NULL;
    if (blinkentry->parent->alt_image == blinkentry)
	blinkentry->parent->alt_image = NULL;
    if (blinkentry->parent->movie_control == blinkentry)
    {
	blinkentry->parent->movie_control = NULL;
	blinkentry->parent->user_selected_movie_control = FALSE;
    }
    if (blinkentry->parent->main_profile == blinkentry)
    {
	blinkentry->parent->main_profile = NULL;
	blinkentry->parent->user_selected_main_profile = FALSE;
    }
    m_free (blinkentry);
}   /*  End Function array_destroy_blink_entry  */

static void _DataBrowser_array_copy_blink_entry (void *dest_p, void *source_p)
{
    struct array_blink_entry *dest = dest_p;
    struct array_blink_entry *source = source_p;

    if (source->parent->main_image == source) dest->parent->main_image = dest;
    if (source->parent->alt_image == source) dest->parent->alt_image = dest;
}   /*  End Function array_copy_blink_entry  */

static void _DataBrowser_array_process_event (DataBrowserWidget w, int index,
					      int event_code,
					      unsigned long mask)
{
    KDisplayDataBlinkEntry be;
    KDisplayDataObject dataobject;
    flag changed_selected = FALSE;
    flag bool, slave;
    iarray pseudo, red, arr, old_pseudo;
    unsigned int contour_mode;
    struct blinkstate *blinkstate = w->dataBrowser.current_blinkstate;
    struct array_list *list = &blinkstate->array_list;
    struct array_blink_entry *blinkentry;
    char *string;

    if (list->num_entries < 1) return;
    if (mask & K_CANVAS_EVENT_CONTROL_KEY_MASK)
    {
	switch (event_code)
	{
	  case 'i':
	    if (list->main_image) index = list->main_image->index + 1;
	    else index = 0;
	    if (index >= list->num_entries) index = 0;
	    break;
	  case 'I':
	    if (list->main_image) index = list->main_image->index - 1;
	    else index = 0;
	    if (index < 0) index = list->num_entries - 1;
	    event_code = 'i';
	    break;
	  case 'c':
	    event_code = EVENT_CODE_CONTOUR_NORMAL;
	    break;
	  case 'r':
	    event_code = EVENT_CODE_CONTOUR_RENZOGRAM;
	    break;
	  default:
	    break;
	}
    }
    be = list->blinkentries[index];
    blinkentry = dispdata_get_blinkentry_info (be);
    dataobject = dispdata_get_dataobject_from_blinkentry (be);
    dispdata_get_data_attributes (dataobject,
				  KDISPLAYDATA_DATA_ATT_PSEUDO_ARRAY, &pseudo,
				  KDISPLAYDATA_DATA_ATT_RED_ARRAY, &red,
				  KDISPLAYDATA_DATA_ATT_END);
    arr = pseudo ? pseudo : red;
    if (list->main_image)
    {
	KDisplayDataBlinkEntry be = list->main_image->generic_blinkentry;

	dispdata_get_data_attributes
	    (dispdata_get_dataobject_from_blinkentry (be),
	     KDISPLAYDATA_DATA_ATT_PSEUDO_ARRAY, &old_pseudo,
	     KDISPLAYDATA_DATA_ATT_END);
    }
    else old_pseudo = NULL;
    switch (event_code)
    {
      case '=':
	dispdata_get_data_attributes (dataobject,
				      KDISPLAYDATA_DATA_ATT_REPLACE, &bool,
				      KDISPLAYDATA_DATA_ATT_END);
	bool = bool ? FALSE : TRUE;
	dispdata_set_data_attributes (dataobject,
				      KDISPLAYDATA_DATA_ATT_REPLACE, bool,
				      KDISPLAYDATA_DATA_ATT_END);
	return;
	/*break;*/
      case EVENT_CODE_MIDDLE_MOUSE:
      case 'i':
	if (list->strings[index][CHAR_INDEX_ARRAY_IMAGE] == 'I') return;
	if (old_pseudo)
	{
	    if (list->alt_image)
		list->strings[list->alt_image->index][CHAR_INDEX_ARRAY_IMAGE] =
		    ' ';
	    list->alt_image = list->main_image;
	    if (list->alt_image)
		list->strings[list->alt_image->index][CHAR_INDEX_ARRAY_IMAGE] =
		    'A';
	}
	else
	    list->strings[list->main_image->index][CHAR_INDEX_ARRAY_IMAGE]=' ';
	list->main_image = blinkentry;
	if (list->main_image == list->alt_image) list->alt_image = NULL;
	list->strings[index][CHAR_INDEX_ARRAY_IMAGE] = 'I';
	dispdata_set_blinkentry_attributes
	    (be,
	     KDISPLAYDATA_BLINKENTRY_ATT_IMAGE, TRUE,
	     KDISPLAYDATA_BLINKENTRY_ATT_END);
	if ( !list->user_selected_movie_control && (iarray_num_dim (arr) > 2) )
	{   /*  Automatic movie master selection required  */
	    if (list->movie_control)
	    {
		dispdata_get_blinkentry_attributes
		    (list->movie_control->generic_blinkentry,
		     KDISPLAYDATA_BLINKENTRY_ATT_MOVIE_SLAVE, &slave,
		     KDISPLAYDATA_BLINKENTRY_ATT_END);
		list->strings[list->movie_control->index]
		    [CHAR_INDEX_ARRAY_MOVIE] = slave ? 'm' : ' ';
	    }
	    dispdata_set_blinkentry_attributes
	    (be,
	     KDISPLAYDATA_BLINKENTRY_ATT_MOVIE_MASTER, TRUE,
	     KDISPLAYDATA_BLINKENTRY_ATT_END);
	    list->strings[index][CHAR_INDEX_ARRAY_MOVIE] = 'M';
	    list->movie_control = blinkentry;
	    _DataBrowser_set_movie_controls (list->parent);
	}
	if (!list->user_selected_main_profile && pseudo)
	{   /*  Automatic profile master selection required  */
	    if (list->main_profile)
	    {
		dispdata_get_blinkentry_attributes
		    (list->main_profile->generic_blinkentry,
		     KDISPLAYDATA_BLINKENTRY_ATT_PROFILE_SLAVE, &slave,
		     KDISPLAYDATA_BLINKENTRY_ATT_END);
		list->strings[list->main_profile->index]
		    [CHAR_INDEX_ARRAY_PROFILE] = slave ? 'p' : ' ';
	    }
	    dispdata_set_blinkentry_attributes
	    (be,
	     KDISPLAYDATA_BLINKENTRY_ATT_PROFILE_MASTER, TRUE,
	     KDISPLAYDATA_BLINKENTRY_ATT_END);
	    list->strings[index][CHAR_INDEX_ARRAY_PROFILE] = 'P';
	    list->main_profile = blinkentry;
	}
	changed_selected = TRUE;
	break;
      case 'a':
	if (!pseudo) return;
	if (list->strings[index][CHAR_INDEX_ARRAY_IMAGE] == 'I') return;
	if (list->strings[index][CHAR_INDEX_ARRAY_IMAGE] == 'A') return;
	if (list->alt_image)
	    list->strings[list->alt_image->index][CHAR_INDEX_ARRAY_IMAGE] =' ';
	list->alt_image = blinkentry;
	list->strings[index][CHAR_INDEX_ARRAY_IMAGE] = 'A';
	changed_selected = TRUE;
	break;
      case EVENT_CODE_RIGHT_MOUSE:
      case 'c':
      case 'r':
	if (!pseudo) return;
	dispdata_get_blinkentry_attributes
	    (be,
	     KDISPLAYDATA_BLINKENTRY_ATT_CONTOUR_MODE, &contour_mode,
	     KDISPLAYDATA_BLINKENTRY_ATT_END);
	switch (contour_mode)
	{
	  case KDISPLAYDATA_CONTOUR_OFF:
	    if (event_code == 'r')
	    {
		contour_mode = KDISPLAYDATA_CONTOUR_RENZOGRAM;
		list->strings[index][CHAR_INDEX_ARRAY_CONTOUR] = 'r';
	    }
	    else
	    {
		contour_mode = KDISPLAYDATA_CONTOUR_NORMAL;
		list->strings[index][CHAR_INDEX_ARRAY_CONTOUR] = 'c';
	    }
	    break;
	  default:
	    contour_mode = KDISPLAYDATA_CONTOUR_OFF;
	    list->strings[index][CHAR_INDEX_ARRAY_CONTOUR] = ' ';
	    break;
	}
	dispdata_set_blinkentry_attributes
	    (be,
	     KDISPLAYDATA_BLINKENTRY_ATT_CONTOUR_MODE, contour_mode,
	     KDISPLAYDATA_BLINKENTRY_ATT_END);
	break;
      case EVENT_CODE_CONTOUR_OFF:
	list->strings[index][CHAR_INDEX_ARRAY_CONTOUR] = ' ';
	dispdata_set_blinkentry_attributes
	    (be,
	     KDISPLAYDATA_BLINKENTRY_ATT_CONTOUR_MODE,
	     KDISPLAYDATA_CONTOUR_OFF,
	     KDISPLAYDATA_BLINKENTRY_ATT_END);
	break;
      case EVENT_CODE_CONTOUR_NORMAL:
	if (!pseudo) return;
	/* Set the levels so that later enabling causes contours to be shown */
	dispdata_get_data_attributes
	    (dataobject,
	     KDISPLAYDATA_DATA_ATT_CONTOUR_LEVELS, &string,
	     KDISPLAYDATA_DATA_ATT_END);
	if (!string)
	    dispdata_set_data_attributes
		(dataobject,
		 KDISPLAYDATA_DATA_ATT_CONTOUR_LEVELS, INIT_CONTOUR_STRING,
		 KDISPLAYDATA_DATA_ATT_END);
	/*  Now enable contours  */
	list->strings[index][CHAR_INDEX_ARRAY_CONTOUR] = 'c';
	dispdata_set_blinkentry_attributes
	    (be,
	     KDISPLAYDATA_BLINKENTRY_ATT_CONTOUR_MODE,
	     KDISPLAYDATA_CONTOUR_NORMAL,
	     KDISPLAYDATA_BLINKENTRY_ATT_END);
	break;
      case EVENT_CODE_CONTOUR_RENZOGRAM:
	if (!pseudo) return;
	/* Set the levels so that later enabling causes contours to be shown */
	dispdata_get_data_attributes
	    (dataobject,
	     KDISPLAYDATA_DATA_ATT_RENZOGRAM_LEVELS, &string,
	     KDISPLAYDATA_DATA_ATT_END);
	if (!string)
	    dispdata_set_data_attributes
		(dataobject,
		 KDISPLAYDATA_DATA_ATT_RENZOGRAM_LEVELS, INIT_RENZOGRAM_STRING,
		 KDISPLAYDATA_DATA_ATT_END);
	/*  Now enable contours  */
	list->strings[index][CHAR_INDEX_ARRAY_CONTOUR] = 'r';
	dispdata_set_blinkentry_attributes
	    (be,
	     KDISPLAYDATA_BLINKENTRY_ATT_CONTOUR_MODE,
	     KDISPLAYDATA_CONTOUR_RENZOGRAM,
	     KDISPLAYDATA_BLINKENTRY_ATT_END);
	break;
      case 'M':
	if (iarray_num_dim (arr) < 3) return;
	dispdata_get_blinkentry_attributes
	    (be,
	     KDISPLAYDATA_BLINKENTRY_ATT_MOVIE_MASTER, &bool,
	     KDISPLAYDATA_BLINKENTRY_ATT_MOVIE_SLAVE, &slave,
	     KDISPLAYDATA_BLINKENTRY_ATT_END);
	bool = bool ? FALSE : TRUE;
	list->user_selected_movie_control = bool;
	dispdata_set_blinkentry_attributes
	    (be,
	     KDISPLAYDATA_BLINKENTRY_ATT_MOVIE_MASTER, bool,
	     KDISPLAYDATA_BLINKENTRY_ATT_END);
	if (bool) list->strings[index][CHAR_INDEX_ARRAY_MOVIE] = 'M';
	else if (slave) list->strings[index][CHAR_INDEX_ARRAY_MOVIE] = 'm';
	else list->strings[index][CHAR_INDEX_ARRAY_MOVIE] = ' ';
	if (bool && (list->movie_control != blinkentry) && list->movie_control)
	{   /*  Update entry for previous master  */
	    dispdata_get_blinkentry_attributes
		(list->movie_control->generic_blinkentry,
		 KDISPLAYDATA_BLINKENTRY_ATT_MOVIE_SLAVE, &slave,
		 KDISPLAYDATA_BLINKENTRY_ATT_END);
	    list->strings[list->movie_control->index]
		[CHAR_INDEX_ARRAY_MOVIE] = slave ? 'm' : ' ';
	    changed_selected = TRUE;
	}
	list->movie_control = bool ? blinkentry : NULL;
	_DataBrowser_set_movie_controls (list->parent);
	break;
      case 'm':
	if (iarray_num_dim (arr) < 3) return;
	dispdata_get_blinkentry_attributes
	    (be,
	     KDISPLAYDATA_BLINKENTRY_ATT_MOVIE_SLAVE, &slave,
	     KDISPLAYDATA_BLINKENTRY_ATT_END);
	slave = slave ? FALSE : TRUE;
	dispdata_set_blinkentry_attributes
	    (be,
	     KDISPLAYDATA_BLINKENTRY_ATT_MOVIE_SLAVE, slave,
	     KDISPLAYDATA_BLINKENTRY_ATT_END);
	if (blinkentry == list->movie_control) return;
	list->strings[index][CHAR_INDEX_ARRAY_MOVIE] = slave ? 'm' : ' ';
	break;
      case EVENT_CODE_MOVIE_OFF:
	list->strings[index][CHAR_INDEX_ARRAY_MOVIE] = ' ';
	dispdata_get_blinkentry_attributes
	    (be,
	     KDISPLAYDATA_BLINKENTRY_ATT_MOVIE_MASTER, &bool,
	     KDISPLAYDATA_BLINKENTRY_ATT_END);
	dispdata_set_blinkentry_attributes
	    (be,
	     KDISPLAYDATA_BLINKENTRY_ATT_MOVIE_MASTER, FALSE,
	     KDISPLAYDATA_BLINKENTRY_ATT_MOVIE_SLAVE, FALSE,
	     KDISPLAYDATA_BLINKENTRY_ATT_END);
	if (bool) list->movie_control = NULL;
	list->user_selected_movie_control = FALSE;
	break;
      case EVENT_CODE_MOVIE_MASTER:
	list->strings[index][CHAR_INDEX_ARRAY_MOVIE] = 'M';
	dispdata_set_blinkentry_attributes
	    (be,
	     KDISPLAYDATA_BLINKENTRY_ATT_MOVIE_MASTER, TRUE,
	     KDISPLAYDATA_BLINKENTRY_ATT_END);
	if ( (list->movie_control != blinkentry) && list->movie_control )
	{   /*  Update entry for previous master  */
	    dispdata_get_blinkentry_attributes
		(list->movie_control->generic_blinkentry,
		 KDISPLAYDATA_BLINKENTRY_ATT_MOVIE_SLAVE, &slave,
		 KDISPLAYDATA_BLINKENTRY_ATT_END);
	    list->strings[list->movie_control->index]
		[CHAR_INDEX_ARRAY_MOVIE] = slave ? 'm' : ' ';
	    changed_selected = TRUE;
	}
	list->movie_control = blinkentry;
	list->user_selected_movie_control = TRUE;
	break;
      case EVENT_CODE_MOVIE_SLAVE:
	list->strings[index][CHAR_INDEX_ARRAY_MOVIE] = 'm';
	dispdata_set_blinkentry_attributes
	    (be,
	     KDISPLAYDATA_BLINKENTRY_ATT_MOVIE_MASTER, FALSE,
	     KDISPLAYDATA_BLINKENTRY_ATT_MOVIE_SLAVE, TRUE,
	     KDISPLAYDATA_BLINKENTRY_ATT_END);
	list->user_selected_movie_control = FALSE;
	break;
      case 'P':
	if (!pseudo) return;
	dispdata_get_blinkentry_attributes
	    (be,
	     KDISPLAYDATA_BLINKENTRY_ATT_PROFILE_MASTER, &bool,
	     KDISPLAYDATA_BLINKENTRY_ATT_PROFILE_SLAVE, &slave,
	     KDISPLAYDATA_BLINKENTRY_ATT_END);
	bool = bool ? FALSE : TRUE;
	list->user_selected_main_profile = bool;
	dispdata_set_blinkentry_attributes
	    (be,
	     KDISPLAYDATA_BLINKENTRY_ATT_PROFILE_MASTER, bool,
	     KDISPLAYDATA_BLINKENTRY_ATT_END);
	if (bool) list->strings[index][CHAR_INDEX_ARRAY_PROFILE] = 'P';
	else if (slave) list->strings[index][CHAR_INDEX_ARRAY_PROFILE] = 'p';
	else list->strings[index][CHAR_INDEX_ARRAY_PROFILE] = ' ';
	if (bool && (list->main_profile != blinkentry) && list->main_profile)
	{   /*  Update entry for previous master  */
	    dispdata_get_blinkentry_attributes
		(list->main_profile->generic_blinkentry,
		 KDISPLAYDATA_BLINKENTRY_ATT_PROFILE_SLAVE, &slave,
		 KDISPLAYDATA_BLINKENTRY_ATT_END);
	    list->strings[list->main_profile->index]
		[CHAR_INDEX_ARRAY_PROFILE] = slave ? 'p' : ' ';
	    changed_selected = TRUE;
	}
	list->main_profile = bool ? blinkentry : NULL;
	break;
      case 'p':
	if (!pseudo) return;
	dispdata_get_blinkentry_attributes
	    (be,
	     KDISPLAYDATA_BLINKENTRY_ATT_PROFILE_SLAVE, &slave,
	     KDISPLAYDATA_BLINKENTRY_ATT_END);
	slave = slave ? FALSE : TRUE;
	dispdata_set_blinkentry_attributes
	    (be,
	     KDISPLAYDATA_BLINKENTRY_ATT_PROFILE_SLAVE, slave,
	     KDISPLAYDATA_BLINKENTRY_ATT_END);
	if (blinkentry == list->main_profile) return;
	list->strings[index][CHAR_INDEX_ARRAY_PROFILE] = slave ? 'p' : ' ';
	break;
      case EVENT_CODE_PROFILE_OFF:
	list->strings[index][CHAR_INDEX_ARRAY_PROFILE] = ' ';
	dispdata_get_blinkentry_attributes
	    (be,
	     KDISPLAYDATA_BLINKENTRY_ATT_PROFILE_MASTER, &bool,
	     KDISPLAYDATA_BLINKENTRY_ATT_END);
	dispdata_set_blinkentry_attributes
	    (be,
	     KDISPLAYDATA_BLINKENTRY_ATT_PROFILE_MASTER, FALSE,
	     KDISPLAYDATA_BLINKENTRY_ATT_PROFILE_SLAVE, FALSE,
	     KDISPLAYDATA_BLINKENTRY_ATT_END);
	if (bool) list->main_profile = NULL;
	list->user_selected_main_profile = FALSE;
	break;
      case EVENT_CODE_PROFILE_MASTER:
	list->strings[index][CHAR_INDEX_ARRAY_PROFILE] = 'P';
	dispdata_set_blinkentry_attributes
	    (be,
	     KDISPLAYDATA_BLINKENTRY_ATT_PROFILE_MASTER, TRUE,
	     KDISPLAYDATA_BLINKENTRY_ATT_END);
	if ( (list->main_profile != blinkentry) && list->main_profile )
	{   /*  Update entry for previous master  */
	    dispdata_get_blinkentry_attributes
		(list->main_profile->generic_blinkentry,
		 KDISPLAYDATA_BLINKENTRY_ATT_PROFILE_SLAVE, &slave,
		 KDISPLAYDATA_BLINKENTRY_ATT_END);
	    list->strings[list->main_profile->index]
		[CHAR_INDEX_ARRAY_PROFILE] = slave ? 'p' : ' ';
	    changed_selected = TRUE;
	}
	list->main_profile = blinkentry;
	list->user_selected_main_profile = TRUE;
	break;
      case EVENT_CODE_PROFILE_SLAVE:
	list->strings[index][CHAR_INDEX_ARRAY_PROFILE] = 'p';
	dispdata_set_blinkentry_attributes
	    (be,
	     KDISPLAYDATA_BLINKENTRY_ATT_PROFILE_MASTER, FALSE,
	     KDISPLAYDATA_BLINKENTRY_ATT_PROFILE_SLAVE, TRUE,
	     KDISPLAYDATA_BLINKENTRY_ATT_END);
	list->user_selected_main_profile = FALSE;
	break;
    }
    listw_set_attributes (w->dataBrowser.listw, TRUE,
			  LISTW_ATT_NUM_STRINGS, list->num_entries,
			  LISTW_ATT_STRINGS, list->strings,
			  LISTW_ATT_END);
    if (index == w->dataBrowser.selected) changed_selected = TRUE;
    if (changed_selected) _DataBrowser_array_show_settings (w);
}   /*  End Function array_process_event  */

static void _DataBrowser_array_show_details (DataBrowserWidget w)
{
    KDisplayDataObject dataobject;
    KDisplayDataBlinkEntry be;
    iarray pseudo, red, green, blue, arr;
    char multiplier = 0;
    unsigned int count, num_dim, size;
    double min, max;
    struct array_form *form = &w->dataBrowser.array;
    struct array_list *list = &w->dataBrowser.current_blinkstate->array_list;
    char *object, *type;
    char line[STRING_LENGTH], txt[STRING_LENGTH], t1[STRING_LENGTH];
    extern char host_type_sizes[NUMTYPES];
    extern char *data_type_names[NUMTYPES];

    /*  Get settings and show them. Later set input sensitivity states  */
    be = list->blinkentries[w->dataBrowser.selected];
    dataobject = dispdata_get_dataobject_from_blinkentry (be);
    _DataBrowser_show_source (dataobject, form->source_lbl);
    dispdata_get_data_attributes
	(dataobject,
	 KDISPLAYDATA_DATA_ATT_OBJECT, &object,
	 KDISPLAYDATA_DATA_ATT_PSEUDO_ARRAY, &pseudo,
	 KDISPLAYDATA_DATA_ATT_RED_ARRAY, &red,
	 KDISPLAYDATA_DATA_ATT_GREEN_ARRAY, &green,
	 KDISPLAYDATA_DATA_ATT_BLUE_ARRAY, &blue,
	 KDISPLAYDATA_DATA_ATT_SCALED_MINIMUM, &min,
	 KDISPLAYDATA_DATA_ATT_SCALED_MAXIMUM, &max,
	 KDISPLAYDATA_DATA_ATT_END);
    arr = pseudo ? pseudo : red;
    if (object) sprintf (line, "Object: %s", object);
    else strcpy (line, "No OBJECT keyword");
    XtVaSetValues (form->objectNdate_lbl, XtNlabel, line, NULL);
    if (pseudo) sprintf ( line, "Format: unknown  Unit: %s",
			  iarray_value_name (pseudo) );
    else sprintf (line, "Format: unknown  Unit: intensity");
    XtVaSetValues (form->fu_lbl, XtNlabel, line, NULL);
    num_dim = iarray_num_dim (arr);
    if (pseudo) size = host_type_sizes[iarray_type (pseudo)];
    else size = host_type_sizes[iarray_type (red)] +
	     host_type_sizes[iarray_type (green)] +
	     host_type_sizes[iarray_type (blue)];
    for (count = 0, txt[0] = '\0'; count < num_dim; ++count)
    {
	size *= iarray_dim_length (arr, count);
	if (size > 4096) size = size >> 10, ++multiplier;
	if (count == 0)
	    sprintf ( t1, "%lu", iarray_dim_length (arr, num_dim - count -1) );
	else
	    sprintf ( t1, "*%lu", iarray_dim_length (arr, num_dim -count -1) );
	strcat (txt, t1);
    }
    if (pseudo) type = data_type_names[iarray_type (pseudo)];
    else type = "RGB";
    switch (multiplier)
    {
      case 0:
	sprintf (line, "Size: %u B  Dim: %s %s", size, txt, type);
	break;
      case 1:
	sprintf (line, "Size: %u KiB  Dim: %s %s", size, txt, type);
	break;
      case 2:
	sprintf (line, "Size: %u MiB  Dim: %s %s", size, txt, type);
	break;
      case 3:
	sprintf (line, "Size: %u GiB  Dim: %s %s", size, txt, type);
	break;
      case 4:
	sprintf (line, "Size: %u TiB  Dim: %s %s", size, txt, type);
	break;
    }
    XtVaSetValues (form->size_lbl, XtNlabel, line, NULL);
    if (pseudo) sprintf (line, "Min: %.10g  Max: %.10g", min, max);
    else line[0] = '\0';
    XtVaSetValues (form->minmax_lbl, XtNlabel, line, NULL);
    _DataBrowser_array_show_settings (w);
    /*  Set input sensitivity states  */
    XtSetSensitive (form->histogram_btn, pseudo ? True : False);
    XtSetSensitive (form->contour_mode_menu, pseudo ? True : False);
    XtSetSensitive (form->movie_mode_menu,
		    (iarray_num_dim (arr) < 3) ? False : True);
    XtSetSensitive (form->profile_mode_menu, pseudo ? True : False);
    XtSetSensitive (form->low_clip_txt, pseudo ? True : False);
    XtSetSensitive (form->high_clip_txt, pseudo ? True : False);
    XtSetSensitive (form->contour_levels_txt, pseudo ? True : False);
    XtSetSensitive (form->contour_neg_width_sld, pseudo ? True : False);
    XtSetSensitive (form->contour_pos_width_sld, pseudo ? True : False);
    XtSetSensitive (form->contour_neg_colour_txt, pseudo ? True : False);
    XtSetSensitive (form->contour_pos_colour_txt, pseudo ? True : False);
    XtSetSensitive (form->renzogram_levels_txt,
		    ( pseudo && (iarray_num_dim (arr) > 2) ) ? True : False);
    XtSetSensitive (form->renzogram_channels_txt,
		    ( pseudo && (iarray_num_dim (arr) > 2) ) ? True : False);
    XtSetSensitive (form->profile_dark_colour_txt, pseudo ? True : False);
    XtSetSensitive (form->profile_light_colour_txt, pseudo ? True : False);
}   /*  End Function array_show_details  */

static void _DataBrowser_array_show_settings (DataBrowserWidget w)
{
    KDisplayDataObject dataobject;
    KDisplayDataBlinkEntry be;
    iarray pseudo;
    flag movie_master, movie_slave, profile_master, profile_slave;
    unsigned int uival;
    struct array_form *form = &w->dataBrowser.array;
    struct array_list *list = &w->dataBrowser.current_blinkstate->array_list;
    struct array_blink_entry *blinkentry;
    char *ptr;

    be = list->blinkentries[w->dataBrowser.selected];
    dataobject = dispdata_get_dataobject_from_blinkentry (be);
    dispdata_get_data_attributes
	(dataobject,
	 KDISPLAYDATA_DATA_ATT_PSEUDO_ARRAY, &pseudo,
	 KDISPLAYDATA_DATA_ATT_END);
    blinkentry = dispdata_get_blinkentry_info (be);
    if (blinkentry == list->main_image)
	XtVaSetValues (form->image_mode_menu,
		       XkwNsetChoice, image_mode_choices[IMAGE_MODE_MAIN],
		       NULL);
    else if (blinkentry == list->alt_image)
	XtVaSetValues (form->image_mode_menu,
		       XkwNsetChoice, image_mode_choices[IMAGE_MODE_ALTERNATE],
		       NULL);
    else XtVaSetValues (form->image_mode_menu,
			XkwNsetChoice, image_mode_choices[IMAGE_MODE_OFF],
			NULL);
    dispdata_get_blinkentry_attributes
	(be,
	 KDISPLAYDATA_BLINKENTRY_ATT_CONTOUR_MODE, &uival,
	 KDISPLAYDATA_BLINKENTRY_ATT_MOVIE_MASTER, &movie_master,
	 KDISPLAYDATA_BLINKENTRY_ATT_MOVIE_SLAVE, &movie_slave,
	 KDISPLAYDATA_BLINKENTRY_ATT_PROFILE_MASTER, &profile_master,
	 KDISPLAYDATA_BLINKENTRY_ATT_PROFILE_SLAVE, &profile_slave,
	 KDISPLAYDATA_BLINKENTRY_ATT_END);
    switch (uival)
    {
      case KDISPLAYDATA_CONTOUR_OFF:
      default:
	ptr = contour_mode_choices[CONTOUR_MODE_OFF];
	break;
      case KDISPLAYDATA_CONTOUR_NORMAL:
	ptr = contour_mode_choices[CONTOUR_MODE_NORMAL];
	break;
      case KDISPLAYDATA_CONTOUR_RENZOGRAM:
	ptr = contour_mode_choices[CONTOUR_MODE_RENZOGRAM];
	break;
    }
    XtVaSetValues (form->contour_mode_menu, XkwNsetChoice, ptr, NULL);
    if (movie_master) ptr = movie_mode_choices[MOVIE_MODE_MASTER];
    else if (movie_slave) ptr = movie_mode_choices[MOVIE_MODE_SLAVE];
    else ptr = movie_mode_choices[MOVIE_MODE_OFF];
    XtVaSetValues (form->movie_mode_menu, XkwNsetChoice, ptr, NULL);
    if (profile_master) ptr = profile_mode_choices[PROFILE_MODE_MASTER];
    else if (profile_slave) ptr = profile_mode_choices[PROFILE_MODE_SLAVE];
    else ptr = profile_mode_choices[PROFILE_MODE_OFF];
    XtVaSetValues (form->profile_mode_menu, XkwNsetChoice, ptr, NULL);
    if (pseudo) _DataBrowser_array_settings_refresh_one (w, REFRESH_ARRAY_ALL);
    else
    {
	XtVaSetValues (form->low_clip_txt, XtNstring, "", NULL);
	XtVaSetValues (form->high_clip_txt, XtNstring, "", NULL);
	XtVaSetValues (form->contour_levels_txt, XtNstring, "", NULL);
	XtVaSetValues (form->contour_neg_colour_txt, XtNstring, "", NULL);
	XtVaSetValues (form->contour_pos_colour_txt, XtNstring, "", NULL);
	XtVaSetValues (form->renzogram_levels_txt, XtNstring, "", NULL);
	XtVaSetValues (form->renzogram_channels_txt, XtNstring, "", NULL);
	XtVaSetValues (form->profile_dark_colour_txt, XtNstring, "", NULL);
	XtVaSetValues (form->profile_light_colour_txt, XtNstring, "", NULL);
    }
}   /*  End Function array_show_settings  */

static void _DataBrowser_array_settings_refresh_one (DataBrowserWidget w,
						     unsigned int mask)
{
    KDisplayDataObject dataobject;
    KDisplayDataBlinkEntry be;
    iarray pseudo;
    int ival;
    double low, high, neg_width, pos_width;
    struct array_form *form = &w->dataBrowser.array;
    struct array_list *list = &w->dataBrowser.current_blinkstate->array_list;
    char *clevel, *neg_colour, *pos_colour, *renzogram_levels;
    char *renzogram_channels, *dark_colour, *light_colour;
    char txt[STRING_LENGTH];

    be = list->blinkentries[w->dataBrowser.selected];
    dataobject = dispdata_get_dataobject_from_blinkentry (be);
    dispdata_get_data_attributes
	(dataobject,
	 KDISPLAYDATA_DATA_ATT_PSEUDO_ARRAY, &pseudo,
	 KDISPLAYDATA_DATA_ATT_SCALED_LOWER_CLIP, &low,
	 KDISPLAYDATA_DATA_ATT_SCALED_UPPER_CLIP, &high,
	 KDISPLAYDATA_DATA_ATT_CONTOUR_LEVELS, &clevel,
	 KDISPLAYDATA_DATA_ATT_RENZOGRAM_LEVELS, &renzogram_levels,
	 KDISPLAYDATA_DATA_ATT_RENZOGRAM_CHANNELS, &renzogram_channels,
	 KDISPLAYDATA_DATA_ATT_CONTOUR_NEG_LINEWIDTH, &neg_width,
	 KDISPLAYDATA_DATA_ATT_CONTOUR_POS_LINEWIDTH, &pos_width,
	 KDISPLAYDATA_DATA_ATT_CONTOUR_NEG_COLOUR, &neg_colour,
	 KDISPLAYDATA_DATA_ATT_CONTOUR_POS_COLOUR, &pos_colour,
	 KDISPLAYDATA_DATA_ATT_PROFILE_DARK_COLOUR, &dark_colour,
	 KDISPLAYDATA_DATA_ATT_PROFILE_LIGHT_COLOUR, &light_colour,
	 KDISPLAYDATA_DATA_ATT_END);
    if (mask & REFRESH_ARRAY_RANGE)
    {
	sprintf (txt, "%.10g", low);
	XtVaSetValues (form->low_clip_txt, XtNstring, txt, NULL);
	sprintf (txt, "%.10g", high);
	XtVaSetValues (form->high_clip_txt, XtNstring, txt, NULL);
    }
    if (mask & REFRESH_ARRAY_CLEVELS)
    {
	if (!clevel) clevel = INIT_CONTOUR_STRING;
	XtVaSetValues (form->contour_levels_txt, XtNstring, clevel, NULL);
    }
    if (mask & REFRESH_ARRAY_NEG_WIDTH)
    {
	ival = neg_width * 10.0;
	XtVaSetValues (form->contour_neg_width_sld, XtNvalue, ival, NULL);
    }
    if (mask & REFRESH_ARRAY_POS_WIDTH)
    {
	ival = pos_width * 10.0;
	XtVaSetValues (form->contour_pos_width_sld, XtNvalue, ival, NULL);
    }
    if (mask & REFRESH_ARRAY_NEG_COLOUR)
	XtVaSetValues (form->contour_neg_colour_txt,XtNstring,neg_colour,NULL);
    if (mask & REFRESH_ARRAY_POS_COLOUR)
	XtVaSetValues (form->contour_pos_colour_txt,XtNstring,pos_colour,NULL);
    if (mask & REFRESH_ARRAY_RENZO_LEVELS)
    {
	if (!renzogram_levels) renzogram_levels = INIT_RENZOGRAM_STRING;
	if (iarray_num_dim (pseudo) < 3) renzogram_levels = "";
	XtVaSetValues (form->renzogram_levels_txt,
		       XtNstring, renzogram_levels,
		       NULL);
    }
    if (mask & REFRESH_ARRAY_RENZO_CHANNELS)
    {
	if (iarray_num_dim (pseudo) < 3) renzogram_channels = "";
	XtVaSetValues (form->renzogram_channels_txt,
		       XtNstring, renzogram_channels,
		       NULL);
    }
    if (mask & REFRESH_ARRAY_DARK_COLOUR)
	XtVaSetValues (form->profile_dark_colour_txt,
		       XtNstring, dark_colour,
		       NULL);
    if (mask & REFRESH_ARRAY_LIGHT_COLOUR)
	XtVaSetValues (form->profile_light_colour_txt,
		       XtNstring, light_colour,
		       NULL);
}   /*  End Function array_settings_refresh_one  */

static void _DataBrowser_array_settings_refresh_all
    (KDisplayDataObject dataobject, unsigned int mask)
{
    KDisplayDataWindow window;
    KDisplayDataBlinkState bs;
    struct blinkstate *blinkstate;

    /*  Run through all data browsers, updating the selected entry if
	appropriate  */
    for ( window = dispdata_get_first_window (NULL);
	  window;
	  window = dispdata_get_next_window (window, FALSE, NULL) )
	for ( bs = dispdata_get_first_blinkstate (window,
						  (void **) &blinkstate);
	      bs;
	      bs = dispdata_get_next_blinkstate (bs, FALSE,
						 (void **) &blinkstate) )
	{
	    DataBrowserWidget browser = blinkstate->viewer;

	    if (!browser) continue;
	    if (dataclasses[browser->dataBrowser.dataclass_mode] !=
		blinkstate->array_list.dataclass)
		continue;
	    _DataBrowser_array_settings_refresh_one (browser, mask);
	}
}   /*  End Function array_settings_refresh_all  */

static void _DataBrowser_histogram_btn_cbk (Widget w, XtPointer client_data,
					    XtPointer call_data)
/*  [SUMMARY] Callback for creating/showing Dataclip widget for a dataset.
*/
{
    KDisplayDataObject dataobject;
    KDisplayDataBlinkEntry blinkentry;
    iarray array;
    double min, max;
    DataBrowserWidget top = client_data;
    Widget dataclip;
    Widget TOP = (Widget) top;
    struct array_list *list = &top->dataBrowser.current_blinkstate->array_list;
    char *name;
    char title[STRING_LENGTH];

    blinkentry = list->blinkentries[top->dataBrowser.selected];
    dataobject = dispdata_get_dataobject_from_blinkentry (blinkentry);
    dataclip = dispdata_get_associated_data_object (dataobject, histogram_key);
    if (dataclip)
    {
	XtPopup (dataclip, XtGrabNone);
	return;
    }
    dispdata_get_data_attributes (dataobject,
				  KDISPLAYDATA_DATA_ATT_NAME, &name,
				  KDISPLAYDATA_DATA_ATT_PSEUDO_ARRAY, &array,
				  KDISPLAYDATA_DATA_ATT_RAW_MINIMUM, &min,
				  KDISPLAYDATA_DATA_ATT_RAW_MAXIMUM, &max,
				  KDISPLAYDATA_DATA_ATT_END);
    sprintf (title, "Intensity Zoom for %s", name);
    dataclip = XtVaCreatePopupShell ("izoomwinpopup", dataclipWidgetClass, TOP,
				     XtNtitle, title,
				     XkwNshowBlankControl, TRUE,
				     NULL);
    XtAddCallback (dataclip, XkwNscaledRegionCallback,
		   _DataBrowser_dataclip_cbk, dataobject);
    XkwDataclipNewArray (dataclip, array, min, max, TRUE);
    dispdata_associate_data_object (dataobject, histogram_key,
				    ( void (*) (void *) ) XtDestroyWidget,
				    dataclip);
    XtPopup (dataclip, XtGrabNone);
}   /*  End Function histogram_btn_cbk  */

static void _DataBrowser_array_new_data (void *info,
					 KDisplayDataObject dataobject)
{
    iarray array;
    double min, max;
    Bool show_new_data_as_image = False;
    Widget dataclip;
    char *name;
    char title[STRING_LENGTH];

    if (show_new_data_as_image_tgl)
	XtVaGetValues (show_new_data_as_image_tgl,
		       XtNstate, &show_new_data_as_image,
		       NULL);
    if (show_new_data_as_image)
	round_robin_bs = _DataBrowser_get_next_blinkstate (round_robin_bs);
    dataclip = dispdata_get_associated_data_object (dataobject, histogram_key);
    if (!dataclip) return;
    dispdata_get_data_attributes (dataobject,
				  KDISPLAYDATA_DATA_ATT_NAME, &name,
				  KDISPLAYDATA_DATA_ATT_PSEUDO_ARRAY, &array,
				  KDISPLAYDATA_DATA_ATT_RAW_MINIMUM, &min,
				  KDISPLAYDATA_DATA_ATT_RAW_MAXIMUM, &max,
				  KDISPLAYDATA_DATA_ATT_END);
    sprintf (title, "Intensity Zoom for %s", name);
    XtVaSetValues (dataclip, XtNtitle, title, NULL);
    XkwDataclipNewArray (dataclip, array, min, max, TRUE);
    XtRemoveAllCallbacks (dataclip, XkwNscaledRegionCallback);
    XtAddCallback (dataclip, XkwNscaledRegionCallback,
		   _DataBrowser_dataclip_cbk, dataobject);
}   /*  End Function array_new_data  */

static void _DataBrowser_array_clone_cbk (Widget w, XtPointer client_data,
					  XtPointer call_data)
/*  [SUMMARY] Callback for cloning a dataset.
*/
{
    KDisplayDataObject dataobject;
    KDisplayDataBlinkEntry blinkentry;
    DataBrowserWidget top = client_data;
    struct array_list *list = &top->dataBrowser.current_blinkstate->array_list;

    blinkentry = list->blinkentries[top->dataBrowser.selected];
    dataobject = dispdata_get_dataobject_from_blinkentry (blinkentry);
    dispdata_array_clone_data (dataobject);
}   /*  End Function array_clone_cbk  */

static void _DataBrowser_array_unload_cbk (Widget w, XtPointer client_data,
					   XtPointer call_data)
/*  [SUMMARY] Callback for unloading a dataset.
*/
{
    KDisplayDataObject dataobject;
    KDisplayDataBlinkEntry blinkentry;
    DataBrowserWidget top = client_data;
    struct array_list *list = &top->dataBrowser.current_blinkstate->array_list;

    blinkentry = list->blinkentries[top->dataBrowser.selected];
    dataobject = dispdata_get_dataobject_from_blinkentry (blinkentry);
    dispdata_destroy_dataobject (dataobject);
}   /*  End Function array_unload_cbk  */

static void _DataBrowser_dataclip_cbk (Widget w, XtPointer client_data,
				       XtPointer call_data)
/*  This is the region callback.
*/
{
    KDisplayDataClass dataclass;
    KDisplayDataObject dataobject = client_data;
    Bool change_all = False;
    DataclipRegions *regions = (DataclipRegions *) call_data;
    static char function_name[] = "DataBrowserWidget::dataclip_cbk";

    if (regions->num_regions != 1)
    {
	fprintf (stderr, "num_regions: %u is not 1\n", regions->num_regions);
	a_prog_bug (function_name);
    }
    if (change_all_ranges_tgl)
	XtVaGetValues (change_all_ranges_tgl, XtNstate, &change_all, NULL);
    if (!change_all)
    {
	dispdata_set_data_attributes
	    (dataobject,
	     KDISPLAYDATA_DATA_ATT_SCALED_LOWER_CLIP, regions->minima[0],
	     KDISPLAYDATA_DATA_ATT_SCALED_UPPER_CLIP, regions->maxima[0],
	     KDISPLAYDATA_DATA_ATT_END);
	_DataBrowser_array_settings_refresh_all (dataobject,
						 REFRESH_ARRAY_RANGE);
	return;
    }
    dataclass = dispdata_array_create_dataclass ();
    for ( dataobject = dispdata_get_first_dataobject (dataclass);
	  dataobject;
	  dataobject = dispdata_get_next_dataobject (dataobject) )
    {
	dispdata_set_data_attributes
	    (dataobject,
	     KDISPLAYDATA_DATA_ATT_SCALED_LOWER_CLIP, regions->minima[0],
	     KDISPLAYDATA_DATA_ATT_SCALED_UPPER_CLIP, regions->maxima[0],
	     KDISPLAYDATA_DATA_ATT_END);
	_DataBrowser_array_settings_refresh_all (dataobject,
						 REFRESH_ARRAY_RANGE);
    }
}   /*  End Function dataclip_cbk   */

static void _DataBrowser_image_mode_cbk (Widget w, XtPointer client_data,
					 XtPointer call_data)
{
    int keycode;
    DataBrowserWidget top = client_data;
    int choice = *(int *) call_data;
    struct array_form *form = &top->dataBrowser.array;
    struct array_list *list = &top->dataBrowser.current_blinkstate->array_list;

    switch (choice)
    {
      case IMAGE_MODE_MAIN:
	keycode = 'i';
	break;
      case IMAGE_MODE_ALTERNATE:
	keycode = 'a';
	break;
      default:
	keycode = 0;
	break;
    }
    _DataBrowser_array_process_event (top, top->dataBrowser.selected, keycode,
				      0);
    switch (list->strings[top->dataBrowser.selected][CHAR_INDEX_ARRAY_IMAGE])
    {
      case 'I':
	if (choice != IMAGE_MODE_MAIN)
	    XtVaSetValues (form->image_mode_menu,
			   XkwNsetChoice, image_mode_choices[IMAGE_MODE_MAIN],
			   NULL);
	break;
      case 'A':
	if (choice != IMAGE_MODE_ALTERNATE)
	    XtVaSetValues (form->image_mode_menu,
			   XkwNsetChoice,
			   image_mode_choices[IMAGE_MODE_ALTERNATE],
			   NULL);
	break;
      case ' ':
	if (choice != IMAGE_MODE_OFF)
	    XtVaSetValues (form->image_mode_menu,
			   XkwNsetChoice, image_mode_choices[IMAGE_MODE_OFF],
			   NULL);
	break;
    }
}   /*  End Function image_mode_cbk  */

static void _DataBrowser_contour_mode_cbk (Widget w, XtPointer client_data,
					   XtPointer call_data)
{
    int keycode;
    DataBrowserWidget top = client_data;
    int choice = *(int *) call_data;

    switch (choice)
    {
      case CONTOUR_MODE_OFF:
      default:
	keycode = EVENT_CODE_CONTOUR_OFF;
	break;
      case CONTOUR_MODE_NORMAL:
	keycode = EVENT_CODE_CONTOUR_NORMAL;
	break;
      case CONTOUR_MODE_RENZOGRAM:
	keycode = EVENT_CODE_CONTOUR_RENZOGRAM;
	break;
    }
    _DataBrowser_array_process_event (top, top->dataBrowser.selected, keycode,
				      0);
}   /*  End Function contour_mode_cbk  */

static void _DataBrowser_movie_mode_cbk (Widget w, XtPointer client_data,
					 XtPointer call_data)
{
    int keycode;
    DataBrowserWidget top = client_data;
    int choice = *(int *) call_data;

    switch (choice)
    {
      case MOVIE_MODE_OFF:
      default:
	keycode = EVENT_CODE_MOVIE_OFF;
	break;
      case MOVIE_MODE_MASTER:
	keycode = EVENT_CODE_MOVIE_MASTER;
	break;
      case MOVIE_MODE_SLAVE:
	keycode = EVENT_CODE_MOVIE_SLAVE;
	break;
    }
    _DataBrowser_array_process_event (top, top->dataBrowser.selected, keycode,
				      0);
}   /*  End Function movie_mode_cbk  */

static void _DataBrowser_profile_mode_cbk (Widget w, XtPointer client_data,
					   XtPointer call_data)
{
    int keycode;
    DataBrowserWidget top = client_data;
    int choice = *(int *) call_data;

    switch (choice)
    {
      case PROFILE_MODE_OFF:
      default:
	keycode = EVENT_CODE_PROFILE_OFF;
	break;
      case PROFILE_MODE_MASTER:
	keycode = EVENT_CODE_PROFILE_MASTER;
	break;
      case PROFILE_MODE_SLAVE:
	keycode = EVENT_CODE_PROFILE_SLAVE;
	break;
    }
    _DataBrowser_array_process_event (top, top->dataBrowser.selected, keycode,
				      0);
}   /*  End Function profile_mode_cbk  */

static void _DataBrowser_minmax_cbk (Widget w, XtPointer client_data,
				     XtPointer call_data)
{
    KDisplayDataClass dataclass;
    KDisplayDataObject dataobject;
    KDisplayDataBlinkEntry blinkentry;
    DataBrowserWidget top = client_data;
    double min, max, low, high, dval1, dval2;
    Bool change_all = False;
    String string;
    struct array_form *form = &top->dataBrowser.array;
    struct array_list *list = &top->dataBrowser.current_blinkstate->array_list;

    blinkentry = list->blinkentries[top->dataBrowser.selected];
    dataobject = dispdata_get_dataobject_from_blinkentry (blinkentry);
    dispdata_get_data_attributes(dataobject,
				 KDISPLAYDATA_DATA_ATT_SCALED_LOWER_CLIP,&low,
				 KDISPLAYDATA_DATA_ATT_SCALED_UPPER_CLIP,&high,
				 KDISPLAYDATA_DATA_ATT_SCALED_MINIMUM, &min,
				 KDISPLAYDATA_DATA_ATT_SCALED_MAXIMUM, &max,
				 KDISPLAYDATA_DATA_ATT_END);
    dval1 = min;
    XtVaGetValues (form->low_clip_txt, XtNstring, &string, NULL);
    sscanf (string, "%lg", &dval1);
    dval2 = max;
    XtVaGetValues (form->high_clip_txt, XtNstring, &string, NULL);
    sscanf (string, "%lg", &dval2);
    if (dval1 < dval2) low = dval1, high = dval2;
    if (change_all_ranges_tgl)
	XtVaGetValues (change_all_ranges_tgl, XtNstate, &change_all, NULL);
    if (!change_all)
    {
	dispdata_set_data_attributes
	    (dataobject,
	     KDISPLAYDATA_DATA_ATT_SCALED_LOWER_CLIP, low,
	     KDISPLAYDATA_DATA_ATT_SCALED_UPPER_CLIP, high,
	     KDISPLAYDATA_DATA_ATT_END);
	_DataBrowser_array_settings_refresh_all (dataobject,
						 REFRESH_ARRAY_RANGE);
	return;
    }
    dataclass = dispdata_array_create_dataclass ();
    for ( dataobject = dispdata_get_first_dataobject (dataclass);
	  dataobject;
	  dataobject = dispdata_get_next_dataobject (dataobject) )
    {
	dispdata_set_data_attributes
	    (dataobject,
	     KDISPLAYDATA_DATA_ATT_SCALED_LOWER_CLIP, low,
	     KDISPLAYDATA_DATA_ATT_SCALED_UPPER_CLIP, high,
	     KDISPLAYDATA_DATA_ATT_END);
	_DataBrowser_array_settings_refresh_all (dataobject,
						 REFRESH_ARRAY_RANGE);
    }
}   /*  End Function minmax_cbk  */

static void _DataBrowser_string_cbk (DataBrowserWidget top, String string,
				     unsigned int attribute, unsigned int mask)
{
    KDisplayDataObject dataobject;
    KDisplayDataBlinkEntry blinkentry;
    struct array_list *list = &top->dataBrowser.current_blinkstate->array_list;

    blinkentry = list->blinkentries[top->dataBrowser.selected];
    dataobject = dispdata_get_dataobject_from_blinkentry (blinkentry);
    dispdata_set_data_attributes (dataobject,
				  attribute, string,
				  KDISPLAYDATA_DATA_ATT_END);
    _DataBrowser_array_settings_refresh_all (dataobject, mask);
}   /*  End Function string_cbk  */

static void _DataBrowser_clevels_cbk (Widget w, XtPointer client_data,
				      XtPointer call_data)
{
    String string = call_data;

    if (string[0] == '\0') string = INIT_CONTOUR_STRING;
    _DataBrowser_string_cbk (client_data, string,
			     KDISPLAYDATA_DATA_ATT_CONTOUR_LEVELS,
			     REFRESH_ARRAY_CLEVELS);
}   /*  End Function clevels_cbk  */

static void _DataBrowser_neg_width_cbk (Widget w, XtPointer client_data,
					XtPointer call_data)
{
    KDisplayDataObject dataobject;
    KDisplayDataBlinkEntry blinkentry;
    DataBrowserWidget top = client_data;
    int ival;
    double neg_width;
    struct array_form *form = &top->dataBrowser.array;
    struct array_list *list = &top->dataBrowser.current_blinkstate->array_list;

    blinkentry = list->blinkentries[top->dataBrowser.selected];
    dataobject = dispdata_get_dataobject_from_blinkentry (blinkentry);
    XtVaGetValues (form->contour_neg_width_sld, XtNvalue, &ival, NULL);
    neg_width = ival * 1e-1;
    dispdata_set_data_attributes
	(dataobject,
	 KDISPLAYDATA_DATA_ATT_CONTOUR_NEG_LINEWIDTH, neg_width,
	 KDISPLAYDATA_DATA_ATT_END);
    _DataBrowser_array_settings_refresh_all (dataobject,
					     REFRESH_ARRAY_NEG_WIDTH);
}   /*  End Function neg_width_cbk  */

static void _DataBrowser_pos_width_cbk (Widget w, XtPointer client_data,
					XtPointer call_data)
{
    KDisplayDataObject dataobject;
    KDisplayDataBlinkEntry blinkentry;
    DataBrowserWidget top = client_data;
    int ival;
    double pos_width;
    struct array_form *form = &top->dataBrowser.array;
    struct array_list *list = &top->dataBrowser.current_blinkstate->array_list;

    blinkentry = list->blinkentries[top->dataBrowser.selected];
    dataobject = dispdata_get_dataobject_from_blinkentry (blinkentry);
    XtVaGetValues (form->contour_pos_width_sld, XtNvalue, &ival, NULL);
    pos_width = ival * 1e-1;
    dispdata_set_data_attributes
	(dataobject,
	 KDISPLAYDATA_DATA_ATT_CONTOUR_POS_LINEWIDTH, pos_width,
	 KDISPLAYDATA_DATA_ATT_END);
    _DataBrowser_array_settings_refresh_all (dataobject,
					     REFRESH_ARRAY_POS_WIDTH);
}   /*  End Function pos_width_cbk  */

static void _DataBrowser_neg_colour_cbk (Widget w, XtPointer client_data,
					 XtPointer call_data)
{
    _DataBrowser_string_cbk (client_data, call_data,
			     KDISPLAYDATA_DATA_ATT_CONTOUR_NEG_COLOUR,
			     REFRESH_ARRAY_NEG_COLOUR);
}   /*  End Function neg_colour_cbk  */

static void _DataBrowser_pos_colour_cbk (Widget w, XtPointer client_data,
					 XtPointer call_data)
{
    _DataBrowser_string_cbk (client_data, call_data,
			     KDISPLAYDATA_DATA_ATT_CONTOUR_POS_COLOUR,
			     REFRESH_ARRAY_POS_COLOUR);
}   /*  End Function pos_colour_cbk  */

static void _DataBrowser_renzo_levels_cbk (Widget w, XtPointer client_data,
					   XtPointer call_data)
{
    String string = call_data;

    if (string[0] == '\0') string = INIT_RENZOGRAM_STRING;
    _DataBrowser_string_cbk (client_data, string,
			     KDISPLAYDATA_DATA_ATT_RENZOGRAM_LEVELS,
			     REFRESH_ARRAY_RENZO_LEVELS);
}   /*  End Function renzo_levels_cbk  */

static void _DataBrowser_renzo_channels_cbk (Widget w, XtPointer client_data,
					     XtPointer call_data)
{
    _DataBrowser_string_cbk (client_data, call_data,
			     KDISPLAYDATA_DATA_ATT_RENZOGRAM_CHANNELS,
			     REFRESH_ARRAY_RENZO_CHANNELS);
}   /*  End Function renzo_channels_cbk  */

static void _DataBrowser_dark_colour_cbk (Widget w, XtPointer client_data,
					  XtPointer call_data)
{
    _DataBrowser_string_cbk (client_data, call_data,
			     KDISPLAYDATA_DATA_ATT_PROFILE_DARK_COLOUR,
			     REFRESH_ARRAY_DARK_COLOUR);
}   /*  End Function dark_colour_cbk  */

static void _DataBrowser_light_colour_cbk (Widget w, XtPointer client_data,
					    XtPointer call_data)
{
    _DataBrowser_string_cbk (client_data, call_data,
			     KDISPLAYDATA_DATA_ATT_PROFILE_LIGHT_COLOUR,
			     REFRESH_ARRAY_LIGHT_COLOUR);
}   /*  End Function light_colour_cbk  */

static void _DataBrowser_active_change (void *blinkstate_p, flag active)
{
    Pixel pixel;
    struct blinkstate *blinkstate = blinkstate_p;
    char *colour;

    if (active) _DataBrowser_set_movie_controls (blinkstate);
    if (!blinkstate->viewer) return;
    XtVaSetValues (blinkstate->viewer->dataBrowser.active_btn,
		   XtNlabel, active ? " Active " : "Inactive",
		   NULL);
    colour = active ? "red" : "black";
    XtVaSetValues (blinkstate->viewer->dataBrowser.list_widget,
		   XtVaTypedArg,
		     XtNborder,
		     XtRString,
		     colour,
		     strlen (colour) + 1,
		   NULL);
    if (active) XtVaGetValues (blinkstate->viewer->dataBrowser.list_widget,
			       XtNborder, &pixel,
			       NULL);
    else XtVaGetValues (blinkstate->viewer->dataBrowser.main_form,
			XtNbackground, &pixel,
			NULL);
    XtVaSetValues (blinkstate->viewer->dataBrowser.active_btn,
		   XtNbackground, pixel,
		   NULL);
}   /*  End Function active_change  */

static void _DataBrowser_set_movie_controls (struct blinkstate *blinkstate)
{
    unsigned int frame_number = 0;
    unsigned int num_frames = 0;

    if ( !dispdata_test_blinkstate_active (blinkstate->generic_blinkstate) )
	return;
    if (blinkstate->array_list.movie_control)
    {
	KDisplayDataWindow window;
	KDisplayDataObject dataobject;
	unsigned int direction;

	window = dispdata_get_window_from_blinkstate
	    (blinkstate->generic_blinkstate);
	dispdata_get_window_attributes
	    (window,
	     KDISPLAYDATA_WINDOW_ATT_SLICE_DIRECTION, &direction,
	     KDISPLAYDATA_WINDOW_ATT_END);
	dataobject = dispdata_get_dataobject_from_blinkentry
	    (blinkstate->array_list.movie_control->generic_blinkentry);
	dispdata_get_data_attributes
	    (dataobject,
	     KDISPLAYDATA_DATA_ATT_FRAME_NUMBER, &frame_number,
	     KDISPLAYDATA_DATA_ATT_NUM_FRAMES, &num_frames,
	     KDISPLAYDATA_DATA_ATT_DIRECTION_FOR_FRAME, direction,
	     KDISPLAYDATA_DATA_ATT_END);
    }
    XtVaSetValues (blinkstate->owner->dataBrowser.animatecontrol,
		   XkwNnumFrames, num_frames,
		   XkwNcurrentFrame, frame_number,
		   NULL);
}   /*  End Function set_movie_controls  */

static flag _DataBrowser_canvas_position_func (KWorldCanvas canvas,
					       double x, double y,
					       unsigned int event_code,
					       void *e_info,
					       void **f_info,
					       double x_lin, double y_lin,
					       int x_pix, int y_pix)
{
    KDisplayDataWindow window = *f_info;
    unsigned long mask;
    char *key = e_info;
    struct blinkstate *blinkstate;

    if (event_code != K_CANVAS_EVENT_MODIFIED_KEY_PRESS) return (FALSE);
    m_copy (&mask, key + strlen (key) + 1, sizeof mask);
    if (mask != K_CANVAS_EVENT_CONTROL_KEY_MASK) return (FALSE);
    dispdata_get_active_blinkstate (window, (void **) &blinkstate);
    if (!blinkstate->viewer) return (TRUE);
    if (dataclasses[blinkstate->viewer->dataBrowser.dataclass_mode] ==
	blinkstate->array_list.dataclass)
	_DataBrowser_array_process_event (blinkstate->viewer, 0, *key, mask);
    return (TRUE);
}   /*  End Function canvas_position_func  */

static void _DataBrowser_overlay_refresh_func (void *info,
					       KDisplayDataClass dataclass)
{
    unsigned int count;
    DataBrowserWidget browser;
    struct overlay_list *list = info;

    browser = list->parent->viewer;
    if (!browser) return;
    if (dataclasses[browser->dataBrowser.dataclass_mode] != dataclass) return;
    /*  Construct list  */
    list->strings = dispdata_regenerate_namelist (dataclass,
						  &list->namelist,
						  NUM_RESERVED_CHAR_OVERLAY,
						  &list->num_entries,
						  NULL, NULL, NULL);
    list->blinkentries =
	dispdata_regenerate_blinkentry_list (list->class_blinkstate,
					     &list->num_entries);
    for (count = 0; count < list->num_entries; ++count)
    {
	KDisplayDataBlinkEntry be = list->blinkentries[count];
	KDisplayDataObject dataobject;
	flag replace, active;
	struct overlay_blink_entry *blinkentry;

	dataobject = dispdata_get_dataobject_from_blinkentry (be);
	dispdata_get_data_attributes
	    (dataobject,
	     KDISPLAYDATA_DATA_ATT_REPLACE, &replace,
	     KDISPLAYDATA_DATA_ATT_END);
	dispdata_get_blinkentry_attributes
	    (be,
	     KDISPLAYDATA_BLINKENTRY_ATT_ACTIVE, &active,
	     KDISPLAYDATA_BLINKENTRY_ATT_END);
	blinkentry = dispdata_get_blinkentry_info (be);
	blinkentry->index = count;
	if (replace) list->strings[count][CHAR_INDEX_REPLACE] = '=';
	list->strings[count][CHAR_INDEX_OVERLAY_ACTIVE] = active ? 'a' : ' ';
    }
    if (list->num_entries > 0)
    {
	if (browser->dataBrowser.selected >= list->num_entries)
	    browser->dataBrowser.selected = list->num_entries - 1;
	_DataBrowser_overlay_show_details (browser);
	XtSetMappedWhenManaged (browser->dataBrowser.overlay.form, True);
    }
    else
    {
	browser->dataBrowser.selected = 0;
	XtSetMappedWhenManaged (browser->dataBrowser.overlay.form, False);
    }
    listw_set_attributes (browser->dataBrowser.listw, TRUE,
			  LISTW_ATT_NUM_STRINGS, list->num_entries,
			  LISTW_ATT_STRINGS, list->strings,
			  LISTW_ATT_REVERSE_VIDEO_INDEX,
			      browser->dataBrowser.selected,
			  LISTW_ATT_END);
}   /*  End Function overlay_refresh_func  */

static void *_DataBrowser_overlay_create_blink_entry(void *info,
						     KDisplayDataBlinkEntry be)
{
    Bool activate_on_load = True;
    struct overlay_list *list = info;
    struct overlay_blink_entry *blinkentry;
    static char function_name[] = "DataBrowser::overlay_create_blink_entry";

    if ( ( blinkentry = m_calloc (sizeof *blinkentry) ) == NULL )
    {
	m_error_notify (function_name, "blink entry");
	return (NULL);
    }
    if (activate_annotations_on_load_tgl)
	XtVaGetValues (activate_annotations_on_load_tgl,
		       XtNstate, &activate_on_load,
		       NULL);
    dispdata_set_blinkentry_attributes (be,
					KDISPLAYDATA_BLINKENTRY_ATT_ACTIVE,
					    activate_on_load,
					KDISPLAYDATA_BLINKENTRY_ATT_END);
    blinkentry->parent = list;
    blinkentry->generic_blinkentry = be;
    return (blinkentry);
}   /*  End Function overlay_create_blink_entry  */

static void _DataBrowser_overlay_destroy_blink_entry (void *info)
{
    struct overlay_blink_entry *blinkentry = info;

    m_free (blinkentry);
}   /*  End Function overlay_destroy_blink_entry  */

static void _DataBrowser_overlay_copy_blink_entry (void *dest_p,
						   void *source_p)
{
}   /*  End Function overlay_copy_blink_entry  */

static void _DataBrowser_overlay_process_event (DataBrowserWidget w, int index,
						int event_code,
						unsigned long mask)
{
    KDisplayDataBlinkEntry be;
    KDisplayDataObject dataobject;
    flag changed_selected = FALSE;
    flag active;
    flag bool;
    struct blinkstate *blinkstate = w->dataBrowser.current_blinkstate;
    struct overlay_list *list = &blinkstate->overlay_list;

    if (list->num_entries < 1) return;
    be = list->blinkentries[index];
    dataobject = dispdata_get_dataobject_from_blinkentry (be);
    switch (event_code)
    {
      case '=':
	dispdata_get_data_attributes (dataobject,
				      KDISPLAYDATA_DATA_ATT_REPLACE, &bool,
				      KDISPLAYDATA_DATA_ATT_END);
	bool = bool ? FALSE : TRUE;
	dispdata_set_data_attributes (dataobject,
				      KDISPLAYDATA_DATA_ATT_REPLACE, bool,
				      KDISPLAYDATA_DATA_ATT_END);
	return;
	/*break;*/
      case EVENT_CODE_MIDDLE_MOUSE:
      case 'a':
	dispdata_get_blinkentry_attributes
	    (be,
	     KDISPLAYDATA_BLINKENTRY_ATT_ACTIVE, &active,
	     KDISPLAYDATA_BLINKENTRY_ATT_END);
	active = active ? FALSE : TRUE;
	list->strings[index][CHAR_INDEX_OVERLAY_ACTIVE] = active ? 'a' : ' ';
	dispdata_set_blinkentry_attributes
	    (be,
	     KDISPLAYDATA_BLINKENTRY_ATT_ACTIVE, active,
	     KDISPLAYDATA_BLINKENTRY_ATT_END);
	break;
    }
    listw_set_attributes (w->dataBrowser.listw, TRUE,
			  LISTW_ATT_NUM_STRINGS, list->num_entries,
			  LISTW_ATT_STRINGS, list->strings,
			  LISTW_ATT_END);
    if (index == w->dataBrowser.selected) changed_selected = TRUE;
    if (changed_selected) _DataBrowser_overlay_show_settings (w);
}   /*  End Function overlay_process_event  */

static void _DataBrowser_overlay_show_details (DataBrowserWidget w)
{
    KDisplayDataObject dataobject;
    KDisplayDataBlinkEntry be;
    struct overlay_form *form = &w->dataBrowser.overlay;
    struct overlay_list *list=&w->dataBrowser.current_blinkstate->overlay_list;

    /*  Get settings and show them. Later set input sensitivity states  */
    be = list->blinkentries[w->dataBrowser.selected];
    dataobject = dispdata_get_dataobject_from_blinkentry (be);
    _DataBrowser_show_source (dataobject, form->source_lbl);
    _DataBrowser_overlay_show_settings (w);
}   /*  End Function overlay_show_details  */

static void _DataBrowser_overlay_show_settings (DataBrowserWidget w)
{
    KDisplayDataBlinkEntry be;
    flag active;
    struct overlay_form *form = &w->dataBrowser.overlay;
    struct overlay_list *list=&w->dataBrowser.current_blinkstate->overlay_list;

    be = list->blinkentries[w->dataBrowser.selected];
    dispdata_get_blinkentry_attributes
	(be,
	 KDISPLAYDATA_BLINKENTRY_ATT_ACTIVE, &active,
	 KDISPLAYDATA_BLINKENTRY_ATT_END);
    XtVaSetValues (form->active_tgl, XtNstate, active, NULL);
}   /*  End Function overlay_show_settings  */

static void _DataBrowser_overlay_unload_cbk (Widget w, XtPointer client_data,
					     XtPointer call_data)
/*  [SUMMARY] Callback for unloading a dataset.
*/
{
    KDisplayDataObject dataobject;
    KDisplayDataBlinkEntry blinkentry;
    DataBrowserWidget top = client_data;
    struct overlay_list *list =
	&top->dataBrowser.current_blinkstate->overlay_list;

    blinkentry = list->blinkentries[top->dataBrowser.selected];
    dataobject = dispdata_get_dataobject_from_blinkentry (blinkentry);
    dispdata_destroy_dataobject (dataobject);
}   /*  End Function overlay_unload_cbk  */

static void _DataBrowser_overlay_active_cbk (Widget w, XtPointer client_data,
					     XtPointer call_data)
/*  [SUMMARY] Callback for changing the active state of an overlay list.
*/
{
    DataBrowserWidget top = client_data;

    _DataBrowser_overlay_process_event (top, top->dataBrowser.selected, 'a',
					0);
}   /*  End Function overlay_active_cbk  */

static Widget _DataBrowser_create_advanced_control (Widget w)
/*  [SUMMARY] Create the advanced control window.
    <w> Any widget.
    [RETURNS] The shell widget.
*/
{
    Widget parent, form, upper_widget;
    static Widget popupshell = NULL;

    if (popupshell) return (popupshell);
    for ( parent = w; w; w = XtParent (w) ) parent = w;
    popupshell = XtVaCreatePopupShell ("advancedDataBrowserShell",
				       topLevelShellWidgetClass, parent,
				       XtNtitle, "Advanced Control Panel",
				       NULL);
    form = XtVaCreateManagedWidget ("form", formWidgetClass, popupshell,
				    XtNborderWidth, 0,
				    NULL);
    w = XtVaCreateManagedWidget ("closeButton", commandWidgetClass, form,
				 XtNlabel, "Close",
				 NULL);
    XtAddCallback (w, XtNcallback, xtmisc_popdown_cbk, popupshell);
    w = XtVaCreateManagedWidget ("clipModeMenu", exclusiveMenuWidgetClass,
				 form,
				 XtNmenuName, "clipModeMenu",
				 XkwNchoiceName, "Clip Mode",
				 XtNfromHoriz, w,
				 XtNtop, XtChainTop,
				 XtNbottom, XtChainTop,
				 XkwNnumItems,
				     KDISPLAYDATA_NUM_CLIP_MODE_CHOICES,
				 XkwNitemStrings, dispdata_clip_mode_choices,
				 NULL);
    upper_widget = w;
    XtAddCallback (w, XkwNselectCallback, _DataBrowser_clip_mode_cbk, NULL);
    w = XtVaCreateManagedWidget ("toggle", ktoggleWidgetClass, form,
				 XtNlabel, "Range changes affect all arrays",
				 XkwNcrosses, False,
				 XtNfromVert, upper_widget,
				 NULL);
    change_all_ranges_tgl = w;
    w = XtVaCreateManagedWidget ("toggle", ktoggleWidgetClass, form,
				 XtNlabel, "Make slaved movie on load",
				 XkwNcrosses, False,
				 XtNfromVert, w,
				 XtNstate, True,
				 NULL);
    slave_movies_on_load_tgl = w;
    w = XtVaCreateManagedWidget ("toggle", ktoggleWidgetClass, form,
				 XtNlabel, "Activate annotations on load",
				 XkwNcrosses, False,
				 XtNfromVert, w,
				 XtNstate, True,
				 NULL);
    activate_annotations_on_load_tgl = w;
    w = XtVaCreateManagedWidget ("toggle", ktoggleWidgetClass, form,
				 XtNlabel,
				     "Slave Profile H-range to image window",
				 XkwNcrosses, False,
				 XtNfromVert, w,
				 NULL);
    XtAddCallback (w, XtNcallback, _DataBrowser_slave_profile_h_range_cbk,
		   NULL);
    w = XtVaCreateManagedWidget ("showNewDataAsImageToggle",
				 ktoggleWidgetClass, form,
				 XtNlabel, "Show new data as Image",
				 XkwNcrosses, False,
				 XtNfromVert, w,
				 NULL);
    show_new_data_as_image_tgl = w;
    return (popupshell);
}   /*  End Function _DataBrowser_create_advanced_control  */

static void _DataBrowser_advanced_cbk (Widget w, XtPointer client_data,
				       XtPointer call_data)
/*  [SUMMARY] Callback for popping up (and creating) the advanced window.
*/
{
    XtPopup (_DataBrowser_create_advanced_control (client_data), XtGrabNone);
}   /*  End Function advanced_cbk  */

static void _DataBrowser_clip_mode_cbk (Widget w, XtPointer client_data,
					XtPointer call_data)
/*  This is the clip mode callback.
*/
{
    KDisplayDataClass dataclass = dispdata_array_create_dataclass ();
    KDisplayDataObject dataobject;
    int new_mode = *(int *) call_data;

    dispdata_set_dataclass_attributes (dataclass,
				       KDISPLAYDATA_CLASS_ATT_CLIP_MODE,
				           new_mode,
				       KDISPLAYDATA_CLASS_ATT_END);
    for ( dataobject = dispdata_get_first_dataobject (dataclass);
	  dataobject;
	  dataobject = dispdata_get_next_dataobject (dataobject) )
	_DataBrowser_array_settings_refresh_all (dataobject,
						 REFRESH_ARRAY_RANGE);
}   /*  End Function clip_mode_cbk  */

static void _DataBrowser_construct_cbk (Widget W, XtPointer client_data,
					XtPointer call_data)
/*  [SUMMARY] Callback for popping up (and creating) the construct window.
*/
{
    DataBrowserWidget top = client_data;
    Widget parent, w;

    for ( w = parent = (Widget) top; w; w = XtParent (w) ) parent = w;
    XkwMakeDataBrowserPopupAll (parent);
}   /*  End Function construct_cbk  */

static void _DataBrowser_slave_profile_h_range_cbk (Widget w,
						    XtPointer client_data,
						    XtPointer call_data)
/*  This is the profile H-range slave mode callback.
*/
{
    KDisplayDataClass dataclass = dispdata_array_create_dataclass ();
    uaddr bool = (uaddr) call_data;

    dispdata_set_dataclass_attributes
	(dataclass,
	 KDISPLAYDATA_CLASS_ATT_SLAVE_PROFILE_H_ZOOM, bool,
	 KDISPLAYDATA_CLASS_ATT_END);
}   /*  End Function slave_profile_h_range_cbk  */

static KDisplayDataBlinkState _DataBrowser_get_next_blinkstate
    (KDisplayDataBlinkState blinkstate)
/*  [SUMMARY] Get the next blinkstate, window list is circular.
    <blinkstate> The current blinkstate.
    [RETURNS] The next blinkstate.
*/
{
    KDisplayDataBlinkState next;
    KDisplayDataWindow win;

    next = dispdata_get_next_blinkstate (blinkstate, FALSE, NULL);
    if (next) return (next);
    win = dispdata_get_window_from_blinkstate (blinkstate);
    win = dispdata_get_next_window (win, TRUE, NULL);
    next = dispdata_get_first_blinkstate (win, NULL);
    return (next);
}   /*  End Function get_next_blinkstate  */


/*  Public functions follow  */

/*EXPERIMENTAL_FUNCTION*/
void XkwDataBrowserPopup (KDisplayDataBlinkState bs)
/*  [SUMMARY] Popup a browser for a blinkstate.
    <bs> The blinkstate.
    [RETURNS] Nothing.
*/
{
    struct blinkstate *blinkstate = dispdata_get_blinkstate_info (bs);

    if (!blinkstate->viewer)
    {
	if (browser_mode != BROWSER_MODE_ONE) return;
	_DataBrowser_set_image_display_bg
	    (current_one_browser->dataBrowser.current_blinkstate->owner,FALSE);
	current_one_browser->dataBrowser.current_blinkstate->viewer = NULL;
	current_one_browser->dataBrowser.current_blinkstate = blinkstate;
	blinkstate->viewer = current_one_browser;
	_DataBrowser_set_image_display_bg (blinkstate->owner, TRUE);
	_DataBrowser_refresh (current_one_browser);
    }
    XtPopup ( (Widget) blinkstate->viewer, XtGrabNone );
    blinkstate->viewer->dataBrowser.user_popped_up = TRUE;
}   /*  End Function XkwDataBrowserPopup  */

/*EXPERIMENTAL_FUNCTION*/
KDisplayDataBlinkEntry XkwDataBrowserGetAlternate (KDisplayDataBlinkState bs)
/*  [SUMMARY] Get alternate image for a blinkstate.
    <bs> The blinkstate.
    [RETURNS] The blinkentry. NULL is returned if there is no alternate image.
*/
{
    struct blinkstate *blinkstate = dispdata_get_blinkstate_info (bs);

    if (!blinkstate->array_list.alt_image) return (NULL);
    return (blinkstate->array_list.alt_image->generic_blinkentry);
}   /*  End Function XkwDataBrowserGetAlternate  */
