/*LINTLIBRARY*/
/*  Dataclip.c

    This code provides a data clip control widget for Xt.

    Copyright (C) 1994-2004  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This file contains all routines needed for a data clip control widget for
  Xt.


    Written by      Richard Gooch   23-OCT-1994

    Updated by      Richard Gooch   26-NOV-1994

    Updated by      Richard Gooch   25-DEC-1994: Added autoValueScale resource.

    Uupdated by     Richard Gooch   27-DEC-1994: Delay computing histogram
  until widget popped up.

    Updated by      Richard Gooch   6-JAN-1995: Changed from passing flag
  by value to by reference for intensityScaleCallback.

    Updated by      Richard Gooch   9-APR-1995: Added initialisation of
  conv_type  field in  win_scale  structure.

    Updated by      Richard Gooch   17-APR-1995: Fixed mix of XtRBool and
  sizeof Boolean in resource fields. Use Bool because it is of int size.

    Updated by      Richard Gooch   21-APR-1995: Added Full Range button.

    Updated by      Richard Gooch   30-JUL-1995: Added destroy callback for
  Intelligent Array.

    Updated by      Richard Gooch   28-DEC-1995: Added verbose resource.

    Updated by      Richard Gooch   22-FEB-1996: Allow dragging of end-points.

    Updated by      Richard Gooch   18-MAR-1996: Display clip limits.

    Updated by      Richard Gooch   4-MAY-1996: Switched to KtoggleWidget and
  added showBlankControl resource.

    Updated by      Richard Gooch   24-MAY-1996: Completed save functionality
  and set XtNresizable resource to TRUE for the save dialog widget so that
  longer names may be typed and viewed.

    Updated by      Richard Gooch   26-MAY-1996: Cleaned code to keep
  gcc -Wall -pedantic-errors happy.

    Updated by      Richard Gooch   4-JUN-1996: Switched to left-right
  bottom-top co-ordinate specification instead of min-max x and y.

    Updated by      Richard Gooch   15-JUN-1996: Added scaling information when
  writing integer types.

    Updated by      Richard Gooch   16-JUN-1996: Read input scaling information
  and combine with computed scaling when writing integer types.

    Updated by      Richard Gooch   20-JUN-1996: Fixed bug in <SetValues> where
  destroy callback was not being correctly cleared and set.

    Updated by      Richard Gooch   21-JUN-1996: Fixed bug in
  <iarr_destroy_callback> where destroy callback was not being cleared.

    Updated by      Richard Gooch   15-AUG-1996: Added autoPopdown resource.

    Updated by      Richard Gooch   2-OCT-1996: Added <get_colour> routine to
  make widget more robust when insufficient colours are available. Added
  cursor position label. Apply cube scaling attachments to all value displays.

    Updated by      Richard Gooch   13-OCT-1996: Added Zoom and Unzoom buttons.

    Updated by      Richard Gooch   12-NOV-1996: Fixed bug in <apply_cbk> so
  that when <<num_regions>> is zero, no region callbacks are called.

    Updated by      Richard Gooch   31-DEC-1996: Created <XkwDataclipNewArray>
  function.

    Updated by      Richard Gooch   13-JAN-1997: Fixed bug in <update_limits>:
  did not check if array is NULL.

    Updated by      Richard Gooch   30-JAN-1997: Created
  <XkwDataclipArrayChanged> function.

    Updated by      Richard Gooch   13-APR-1997: Also display cursor position
  when clicking or dragging left or right mouse buttons. Show values with
  preferred units.

    Updated by      Richard Gooch   15-APR-1997: Added axis labels for
  histogram display (no tick marks or scale).

    Updated by      Richard Gooch   1-JUN-1997: Move centre of region using
  horizontal position of middle button.

    Updated by      Richard Gooch   4-JUN-1997: Do not require cursor to be in
  region when moving centre if only one region is defined.

    Updated by      Richard Gooch   9-JUN-1997: Change region width using
  vertical position of middle button.

    Updated by      Richard Gooch   12-AUG-1997: Allow auto apply state to be
  set from resources.

    Updated by      Richard Gooch   30-SEP-1997: Added "Clip Outliers" button.

    Updated by      Richard Gooch   2-OCT-1997: Add right button first click if
  no regions are define: left boundary is set to data minimum.

    Updated by      Richard Gooch   31-OCT-1997: Added new <<reset_hscale>>
  parameter to <XkwDataclipNewArray>.

    Updated by      Richard Gooch   8-DEC-1997: Added range of clip buttons.

    Updated by      Richard Gooch   9-DEC-1997: Don't apply in outlier clip
  if auto apply is off.

    Updated by      Richard Gooch   12-DEC-1997: Removed most border chaining.

    Updated by      Richard Gooch   22-DEC-1997: Turned on "Auto Apply" by
  default.

    Updated by      Richard Gooch   25-MAR-1998: Fixed bug in position handler:
  did not trap zero range when moving middle mouse button.

    Updated by      Richard Gooch   17-MAY-1998: Trap illegal values.

    Updated by      Richard Gooch   19-MAY-1998: Set data range when writing
  scaled output.

    Updated by      Richard Gooch   14-JUL-1998: Fixed some problems with
  persistent invalid states.

    Updated by      Richard Gooch   12-AUG-1998: Now update cursor position on
  any canvas event.

    Updated by      Richard Gooch   17-SEP-1998: Took account of some CONST
  declarations in <ex> package.

    Updated by      Richard Gooch   24-OCT-1998: Added "DataclipWidget::Clip"
  event.

    Updated by      Richard Gooch   25-MAR-1999: Fixed bug in scaling
  information when writing float files.

    Updated by      Richard Gooch   9-APR-1999: Added "Zero Left" button.

    Updated by      Richard Gooch   15-AUG-1999: Made egcs happy.

    Updated by      Richard Gooch   17-JUN-2000: Added support for "join" and
  "steps" display modes.

    Updated by      Richard Gooch   19-JUN-2000: Removed support for "join" and
  "steps" display modes: doesn't really make much sense.

    Updated by      Richard Gooch   27-NOV-2000: Fixed bug in function
  <histcanvas_size_control_func>: use total pixel canvas width to determine
  array length.

    Updated by      Richard Gooch   10-MAR-2001: Added Y position label.

    Updated by      Richard Gooch   11-MAR-2001: Changed Y position label to
  show number of values in bin.

    Updated by      Richard Gooch   20-MAR-2001: Created Destroy method.

    Updated by      Richard Gooch   15-NOV-2001: Added 'o' and 'u' keybindings.
  Created experimental function <XkwDataclipSetRegions>.

    Updated by      Richard Gooch   16-NOV-2001: Split histogram into two
  arrays: a raw, higher resolution array, and a lower resolution display array.
  The raw array is used for percentile clipping but not display (to avoid
  "combing" effects.

    Updated by      Richard Gooch   19-NOV-2001: Added "subset" toggle which
  decimates the input array. This speeds up computation and may avoid outliers.

    Updated by      Richard Gooch   20-NOV-2001: Added 'i' key binding.

    Updated by      Richard Gooch   7-DEC-2002: Documentation improvement.

    Updated by      Richard Gooch   6-AUG-2003: Made use of new experimental
  function <c_unregister_callback2>.

    Updated by      Richard Gooch   3-OCT-2003: Switch from working mostly in
  raw values to scaled values. This makes coping with negative scales easier,
  since low scaled values should appear on the left.

    Last updated by Richard Gooch   2-MAY-2004: Changed bottom y limit from
  1.0 to 0.5 and removed 6 decade limit so that single counts can be seen.
  Enabled display of horizontal scale.


*/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <sys/wait.h>
#include <signal.h>
#include <X11/Xos.h>
#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/Label.h>
#include <X11/Xaw/Box.h>
#include <X11/Xaw/Dialog.h>
#ifndef X11
#  define X11
#endif
#include <karma.h>
#define NEW_WIN_SCALE
#include <k_event_codes.h>
#include <karma_canvas.h>
#include <karma_iarray.h>
#include <karma_c.h>
#include <karma_xtmisc.h>
#include <karma_ds.h>
#include <karma_ex.h>
#include <karma_m.h>
#include <karma_a.h>
#include <Xkw/DataclipP.h>
#include <Xkw/Canvas.h>
#include <Xkw/Ktoggle.h>
#include <Xkw/ExclusiveMenu.h>


#define MIN_RAW_HISTOGRAM_LENGTH  10000


/*  External functions  */


/*  Private functions  */
STATIC_FUNCTION (void _Dataclip_ClassInititialise, () );
STATIC_FUNCTION (void _Dataclip_Initialise, (Widget request, Widget new) );
STATIC_FUNCTION (void _Dataclip_Destroy, (Widget W) );
STATIC_FUNCTION (void _Dataclip_cnv_realise_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (Boolean _Dataclip_SetValues,
		 (Widget current, Widget request, Widget new) );
STATIC_FUNCTION (void immediate_apply_cbk, (Widget w, XtPointer client_data,
					    XtPointer call_data) );
STATIC_FUNCTION (void _Dataclip_subset_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void _Dataclip_update_clip_labels, (DataclipWidget w) );
STATIC_FUNCTION (void _Dataclip_compute_subset, (DataclipWidget w) );
STATIC_FUNCTION (void apply_cbk, (Widget w, XtPointer client_data,
				  XtPointer call_data) );
STATIC_FUNCTION (void histcanvas_size_control_func,
		 (KWorldCanvas canvas, int width, int height,
		  struct win_scale_type *win_scale,
		  void **info, flag *boundary_clear) );
STATIC_FUNCTION (void _Dataclip_canvas_refresh_func,
		 (KWorldCanvas canvas, int width, int height,
		  struct win_scale_type *win_scale, Kcolourmap cmap,
		  flag cmap_resize, void **info) );
STATIC_FUNCTION (flag _Dataclip_canvas_position_func,
		 (KWorldCanvas canvas, double x, double y,
		  unsigned int event_code, void *e_info, void **f_info,
		  double x_lin, double y_lin, int x_pix, int y_pix) );
STATIC_FUNCTION (void pop_cbk, (Widget w, XtPointer client_data,
				XtPointer call_data) );
STATIC_FUNCTION (void full_range_cbk, (Widget w, XtPointer client_data,
				       XtPointer call_data) );
STATIC_FUNCTION (void iarr_destroy_callback, (iarray arr,DataclipWidget top) );
STATIC_FUNCTION (void blank_data_cbk, (Widget w, XtPointer client_data,
				       XtPointer call_data) );
STATIC_FUNCTION (void type_menu_cbk, (Widget w, XtPointer client_data,
				      XtPointer call_data) );
STATIC_FUNCTION (void save_cbk, (Widget w, XtPointer client_data,
				 XtPointer call_data) );
STATIC_FUNCTION (unsigned long get_colour,
		 (KWorldCanvas canvas, CONST char *colourname) );
STATIC_FUNCTION (void zoom_cbk, (Widget w, XtPointer client_data,
				 XtPointer call_data) );
STATIC_FUNCTION (void unzoom_cbk, (Widget w, XtPointer client_data,
				   XtPointer call_data) );
STATIC_FUNCTION (void do_zoom, (DataclipWidget w, double factor, double mid) );
STATIC_FUNCTION (void update_limits, (DataclipWidget w, flag reset_hscale) );
STATIC_FUNCTION (void clip_cbk, (Widget w, XtPointer client_data,
				 XtPointer call_data) );
STATIC_FUNCTION (void clip_outliers, (DataclipWidget top, double proportion) );
STATIC_FUNCTION (flag _Dataclip_ds_event_func,
		 (void *object, multi_array *multi_desc,
		  CONST char *domain, CONST char *name) );
STATIC_FUNCTION (void _Dataclip_invalidate_data, (DataclipWidget w) );


#define offset(field) XtOffset(DataclipWidget, dataclip.field)

static XtResource DataclipResources[] = 
{
    {XkwNmaxDataRegions, XkwCMaxDataRegions, XtRInt, sizeof (int),
     offset (maxRegions), XtRImmediate, (XtPointer) 1},
    {XkwNscaledRegionCallback, XtCCallback, XtRCallback, sizeof (caddr_t),
     offset (scaledRegionCallback), XtRCallback, (caddr_t) NULL},
    {XkwNrawRegionCallback, XtCCallback, XtRCallback, sizeof (caddr_t),
     offset (rawRegionCallback), XtRCallback, (caddr_t) NULL},
    {XkwNshowBlankControl, XkwCShowBlankControl, XtRBool, sizeof (Bool),
     offset (showBlankControl), XtRImmediate, (XtPointer) False},
    {XkwNfixedOutputType, XkwCFixedOutputType, XtRCardinal, sizeof (Cardinal),
     offset (fixedOutputType), XtRImmediate, (XtPointer) NONE},
    {XkwNverbose, XkwCVerbose, XtRBool, sizeof (Bool),
     offset (verbose), XtRImmediate, False},
    {XkwNautoPopdown, XkwCAutoPopdown, XtRBool, sizeof (Bool),
     offset (autoPopdown), XtRImmediate, (XtPointer) FALSE},
};
#undef offset

DataclipClassRec dataclipClassRec = 
{
  {
      /* CoreClassPart */
    (WidgetClass)&topLevelShellClassRec,  /* superclass */
    "Dataclip",                    /* class_name */
    sizeof (DataclipRec),          /* widget_size */
    _Dataclip_ClassInititialise,   /* class_initialise */
    NULL,                          /* class_part_initialise */
    FALSE,                         /* class_init */
    (XtInitProc) _Dataclip_Initialise,  /* initialise */
    NULL,                          /* initialise_hook */
    XtInheritRealize,              /* realise */
    NULL,                          /* actions */
    0,                             /* num_actions */
    DataclipResources,             /* resources */
    XtNumber (DataclipResources),  /* num_resources */
    NULLQUARK,                     /* xrm_class */
    TRUE,                          /* compress_motion */
    TRUE,                          /* compress_exposure */
    TRUE,                          /* compress_enterleave */
    TRUE,                          /* visible_interest */
    _Dataclip_Destroy,             /* destroy */
    XtInheritResize,               /* resize */
    NULL,                          /* expose */
    (XtSetValuesFunc) _Dataclip_SetValues,   /* set_values */
    NULL,                          /* set_values_hook */
    XtInheritSetValuesAlmost,      /* set_values_almost */
    NULL,                          /* get_values_hook */
    NULL,                          /* accept_focus */
    XtVersion,                     /* version */
    NULL,                          /* callback_private */
    NULL,                          /* tm_translations */
    NULL,
    NULL,
    NULL,
  },
  {     /* CompositeClassPart */
      /* geometry_manager */	XtInheritGeometryManager,
      /* change_managed	  */	XtInheritChangeManaged,
      /* insert_child	  */	XtInheritInsertChild,
      /* delete_child	  */	XtInheritDeleteChild,
      /* extension	  */	NULL
  },
  { /* Shell */
      /* extension	  */	NULL
  },
  { /* WMShell */
      /* extension	  */	NULL
  },
  { /* VendorShell */
      /* extension	  */	NULL
  },
  { /* TopLevelShell */
      /* extension	  */	NULL
  },
  {  /* DataClipClassPart */
      {0}, {NULL}, 0      /* empty */
  }
};

WidgetClass dataclipWidgetClass = (WidgetClass) &dataclipClassRec;

static void _Dataclip_ClassInititialise ()
{
    unsigned int count, index;
    extern char *data_type_names[NUMTYPES];

    for (count = 0, index = 0; count < NUMTYPES; ++count)
    {
	if ( !ds_element_is_legal (count) ) continue;
	if ( !ds_element_is_atomic (count) ) continue;
	if ( ds_element_is_complex (count) ) continue;
	dataclipClassRec.dataclip_class.type_index_to_type[index] = count;
	dataclipClassRec.dataclip_class.type_names[index] =
	    data_type_names[count];
	++index;
    }
    dataclipClassRec.dataclip_class.num_types = index;
}   /*  End Function ClassInitialise  */

static void _Dataclip_Initialise (Widget Request, Widget New)
{
    unsigned int count;
    /*DataclipWidget request = (DataclipWidget) Request;*/
    DataclipWidget new = (DataclipWidget) New;
    Widget form;
    Widget close_btn, save_btn;
    Widget cnv;
    Widget w, upper_widget;
    Bool bool_val;
    static String clips[] = {"95%", "98%", "99%", "99.5%", "99.9%", "99.99%"};
    static char function_name[] = "DataclipWidget::Initialise";

    new->dataclip.pixcanvas = NULL;
    new->dataclip.worldcanvas = NULL;
    new->dataclip.iarr_destroy_callback = NULL;
    new->dataclip.last_was_left = FALSE;
    new->dataclip.left_pos = TOOBIG;
    new->dataclip.popped_up = FALSE;
    new->dataclip.blank_data = FALSE;
    new->dataclip.array = NULL;
    new->dataclip.reset_canvas_scale = TRUE;
    XtAddCallback (New, XtNpopupCallback, pop_cbk, (XtPointer) TRUE);
    XtAddCallback (New, XtNpopdownCallback, pop_cbk, (XtPointer) FALSE);
    if (new->dataclip.maxRegions < 1)
    {
	fprintf (stderr, "maxRegions: %u  must be greater than 0\n",
		 new->dataclip.maxRegions);
	a_prog_bug (function_name);
    }
    if ( ( new->dataclip.minima =
	   m_alloc (sizeof *new->dataclip.minima * new->dataclip.maxRegions) )
	 == NULL )
    {
	m_abort (function_name, "array of minima");
    }
    if ( ( new->dataclip.maxima =
	   m_alloc (sizeof *new->dataclip.maxima * new->dataclip.maxRegions) )
	 == NULL )
    {
	m_abort (function_name, "array of maxima");
    }
    new->dataclip.num_regions = 0;
    memset (&new->dataclip.raw_histogram, 0,
	    sizeof new->dataclip.raw_histogram);
    memset (&new->dataclip.display_histogram, 0,
	    sizeof new->dataclip.display_histogram);
    memset (&new->dataclip.subset, 0, sizeof new->dataclip.subset);
    form = XtVaCreateManagedWidget ("form", formWidgetClass, New,
				    XtNborderWidth, 0,
				    NULL);
    w = XtVaCreateManagedWidget ("closeButton", commandWidgetClass, form,
				 XtNlabel, "Close",
				 XtNheight, 20,
				 NULL);
    close_btn = w;
    XtAddCallback (close_btn, XtNcallback, xtmisc_popdown_cbk, New);
    if (new->dataclip.showBlankControl)
    {
	w = XtVaCreateManagedWidget ("blankToggle", ktoggleWidgetClass, form,
				     XtNlabel, "Blank",
				     XtNfromHoriz, w,
				     XtNheight, 20,
				     XtNstate, new->dataclip.blank_data,
				     XkwNcrosses, False,
				     NULL);
	XtAddCallback (w, XtNcallback, blank_data_cbk, (XtPointer) New);
    }
    w = XtVaCreateManagedWidget ("button", commandWidgetClass, form,
				 XtNlabel, "Zoom",
				 XtNfromHoriz, w,
				 XtNheight, 20,
				 NULL);
    XtAddCallback (w, XtNcallback, zoom_cbk, New);
    w = XtVaCreateManagedWidget ("button", commandWidgetClass, form,
				 XtNlabel, "Unzoom",
				 XtNfromHoriz, w,
				 XtNheight, 20,
				 NULL);
    XtAddCallback (w, XtNcallback, unzoom_cbk, New);
    w = XtVaCreateManagedWidget ("button", commandWidgetClass, form,
				 XtNlabel, "Full Range",
				 XtNfromHoriz, w,
				 /*XtNfromVert, close_btn,*/
				 XtNheight, 20,
				 NULL);
    XtAddCallback (w, XtNcallback, full_range_cbk, (XtPointer) New);
    w = XtVaCreateManagedWidget ("button", commandWidgetClass, form,
				 XtNlabel, "Apply",
				 XtNfromVert, close_btn,
				 XtNheight, 20,
				 NULL);
    XtAddCallback (w, XtNcallback, apply_cbk, New);
    upper_widget = w;
    if (new->dataclip.maxRegions == 1)
    {
	w = XtVaCreateManagedWidget ("autoApplyToggle",ktoggleWidgetClass,form,
				     XtNlabel, "Auto Apply",
				     XtNfromHoriz, w,
				     XtNfromVert, close_btn,
				     XtNheight, 20,
				     XkwNcrosses, False,
				     XtNstate, True,
				     NULL);
	XtAddCallback (w, XtNcallback, immediate_apply_cbk, (XtPointer) New);
	XtVaGetValues (w, XtNstate, &bool_val, NULL);
	new->dataclip.immediate_apply = bool_val;
    }
    w = XtVaCreateManagedWidget ("subsetToggle", ktoggleWidgetClass, form,
				 XtNlabel, "Subset",
				 XtNfromHoriz, w,
				 XtNfromVert, close_btn,
				 XtNheight, 20,
				 XkwNcrosses, False,
				 NULL);
    XtAddCallback (w, XtNcallback, _Dataclip_subset_cbk, (XtPointer) New);
    new->dataclip.subset_tgl = w;
    w = XtVaCreateManagedWidget ("button", commandWidgetClass, form,
				 XtNlabel, "Zero Left",
				 XtNfromHoriz, w,
				 XtNfromVert, close_btn,
				 XtNheight, 20,
				 NULL);
    XtAddCallback (w, XtNcallback, clip_cbk, (XtPointer) New);
    w = NULL;
    for (count = 0; count < sizeof clips / sizeof (String); ++count)
    {
	w = XtVaCreateManagedWidget ("button", commandWidgetClass, form,
				     XtNlabel, clips[count],
				     XtNfromVert, upper_widget,
				     XtNfromHoriz, w,
				     XtNheight, 20,
				     NULL);
	XtAddCallback (w, XtNcallback, clip_cbk, (XtPointer) New);
    }
    upper_widget = w;
    w = XtVaCreateManagedWidget ("label", labelWidgetClass, form,
				 XtNlabel, "Data Minimum:              ",
				 XtNborderWidth, 0,
				 XtNfromVert, upper_widget,
				 XtNheight, 20,
				 XtNresizable, True,
				 NULL);
    new->dataclip.min_label = w;
    w = XtVaCreateManagedWidget ("label", labelWidgetClass, form,
				 XtNlabel, "Data Maximum:              ",
				 XtNborderWidth, 0,
				 XtNfromVert, w,
				 XtNheight, 20,
				 XtNresizable, True,
				 NULL);
    new->dataclip.max_label = w;
    w = XtVaCreateManagedWidget ("label", labelWidgetClass, form,
				 XtNlabel, "Lower Clip:                      ",
				 XtNborderWidth, 0,
				 XtNfromVert, w,
				 XtNheight, 20,
				 XtNresizable, True,
				 NULL);
    new->dataclip.lower_label = w;
    w = XtVaCreateManagedWidget ("label", labelWidgetClass, form,
				 XtNlabel, "Upper Clip:                      ",
				 XtNborderWidth, 0,
				 XtNfromVert, w,
				 XtNheight, 20,
				 XtNresizable, True,
				 NULL);
    new->dataclip.upper_label = w;
    w = XtVaCreateManagedWidget ("label", labelWidgetClass, form,
				 XtNlabel, "Bin value:                       ",
				 XtNborderWidth, 0,
				 XtNfromVert, w,
				 XtNheight, 20,
				 XtNresizable, True,
				 NULL);
    new->dataclip.cursor_x_label = w;
    w = XtVaCreateManagedWidget ("label", labelWidgetClass, form,
				 XtNlabel, "Number of values in bin:         ",
				 XtNborderWidth, 0,
				 XtNfromVert, w,
				 XtNheight, 20,
				 XtNresizable, True,
				 NULL);
    new->dataclip.cursor_y_label = w;
    save_btn = XtVaCreateManagedWidget ("button", commandWidgetClass, form,
					XtNlabel, "Save",
					XtNfromVert, w,
					NULL);
    if (new->dataclip.fixedOutputType == NONE)
    {
	w = XtVaCreateManagedWidget("menuButton", exclusiveMenuWidgetClass,
				    form,
				    XtNmenuName, "typeMenu",
				    XkwNchoiceName, "Output Type",
				    XtNfromVert, w,
				    XtNfromHoriz, save_btn,
				    XkwNnumItems,
				    dataclipClassRec.dataclip_class.num_types,
				    XkwNitemStrings,
				    dataclipClassRec.dataclip_class.type_names,
				    NULL);
	XtAddCallback (w, XkwNselectCallback, type_menu_cbk, New);
	new->dataclip.output_type =
	    dataclipClassRec.dataclip_class.type_index_to_type[0];
    }
    else new->dataclip.output_type = new->dataclip.fixedOutputType;
    w = save_btn;
    w = XtVaCreateManagedWidget ("filename", dialogWidgetClass, form,
				 XtNlabel, "Filename: ",
				 XtNborderWidth, 0,
				 XtNfromVert, w,
				 XtNvalue, "scaled",
				 XtNresizable, TRUE,
				 NULL);
    new->dataclip.dialog = w;
    XtAddCallback (save_btn, XtNcallback, save_cbk, New);
    cnv = XtVaCreateManagedWidget ("pseudoColourCanvas", canvasWidgetClass,
				   form,
				   XtNfromVert, w,
				   XtNleft, XtChainLeft,
				   XtNright, XtChainRight,
				   XtNbottom, XtChainBottom,
				   XtNwidth, 360,
				   XtNheight, 280,
				   XkwNsilenceUnconsumed, True,
				   XtNresizable, True,
				   NULL);
    new->dataclip.cnv = cnv;
    XtAddCallback (cnv, XkwNrealiseCallback, _Dataclip_cnv_realise_cbk, New);
    ds_event_register_func (_Dataclip_ds_event_func, New);
}   /*  End Function Initialise  */

static void _Dataclip_Destroy (Widget W)
{
    DataclipWidget w = (DataclipWidget) W;

    if (w->dataclip.minima) m_free (w->dataclip.minima);
    if (w->dataclip.maxima) m_free (w->dataclip.maxima);
    if (w->dataclip.raw_histogram.array)
	m_free (w->dataclip.raw_histogram.array);
    if (w->dataclip.display_histogram.array)
	m_free (w->dataclip.display_histogram.array);
    if (w->dataclip.subset.array) m_free (w->dataclip.subset.array);
    c_unregister_callback2 (w->dataclip.iarr_destroy_callback);
}   /*  End Function Destroy  */

static void _Dataclip_cnv_realise_cbk (Widget w, XtPointer client_data,
				       XtPointer call_data)
/*  This is the canvas realise callback.
*/
{
    KPixCanvas pixcanvas = (KPixCanvas) call_data;
    KWorldCanvas wc;
    GC gc;
    XGCValues gcvalues;
    struct win_scale_type win_scale;
    DataclipWidget top = (DataclipWidget) client_data;
    char *hlabel;
    /*static char function_name[] = "DataclipWidget::cnv_realise_cbk";*/

    top->dataclip.pixcanvas = pixcanvas;
    /*  Get the Graphics Context of the pixel canvas  */
    gc = kwin_get_gc_x (pixcanvas);
    if (XGetGCValues (XtDisplay (w), gc, GCForeground | GCBackground,
		      &gcvalues) == 0)
    {
	fprintf (stderr, "Error getting GC values\n");
	exit (RV_UNDEF_ERROR);
    }
    /*  Create the offset control world canvas  */
    canvas_init_win_scale (&win_scale, K_WIN_SCALE_MAGIC_NUMBER);
    /*  No histogram data yet  */
    win_scale.min_sat_pixel = gcvalues.background;
    win_scale.max_sat_pixel = gcvalues.foreground;
    win_scale.conv_type = CONV_CtoR_REAL;
    if ( ( wc = canvas_create (pixcanvas, NULL, &win_scale) ) == NULL )
    {
	exit (RV_UNDEF_ERROR);
    }
    top->dataclip.worldcanvas = wc;
    canvas_use_log_scale (wc, FALSE, TRUE);
    canvas_register_size_control_func (wc, histcanvas_size_control_func,
				       (void *) top);
    canvas_register_refresh_func (wc, _Dataclip_canvas_refresh_func,
				  (void *) top);
    canvas_register_position_event_func (wc, _Dataclip_canvas_position_func,
					 (void *) top);
    if (top->dataclip.array == NULL) hlabel = "Data Value";
    else hlabel = iarray_value_name (top->dataclip.array);
    canvas_set_dressing (wc,
			 KCD_ATT_DISPLAY, TRUE,
			 KCD_ATT_BOTTOM_TICKS, TRUE,
			 KCD_ATT_HLABEL, hlabel,
			 KCD_ATT_VLABEL, "Log(Incidence)",
			 KCD_ATT_SHOW_LABELS, TRUE,
			 KCD_ATT_END);
}   /*  End Function cnv_realise_cbk   */

static Boolean _Dataclip_SetValues (Widget Current, Widget Request, Widget New)
{
    DataclipWidget current = (DataclipWidget) Current;
    /*DataclipWidget request = (DataclipWidget) Request;*/
    DataclipWidget new = (DataclipWidget) New;
    static char function_name[] = "DataclipWidget::SetValues";

    if (new->dataclip.maxRegions != current->dataclip.maxRegions)
    {
	fprintf (stderr, "Cannot change number of regions\n");
	a_prog_bug (function_name);
    }
    return False;
}   /*  End Function SetValues  */

static void immediate_apply_cbk (Widget w, XtPointer client_data,
				 XtPointer call_data)
{
    flag bool = (uaddr) call_data;
    DataclipWidget top = (DataclipWidget) client_data;

    top->dataclip.immediate_apply = bool;
}   /*  End Function immediate_apply_cbk  */

static void apply_cbk (Widget w, XtPointer client_data, XtPointer call_data)
{
    unsigned int count;
    double scale, offset;
    DataclipWidget top = (DataclipWidget) client_data;
    DataclipRegions regions;

    if (top->dataclip.array == NULL) return;
    if (top->dataclip.num_regions < 1) return;
    regions.blank_data_outside_regions = top->dataclip.blank_data;
    regions.num_regions = top->dataclip.num_regions;
    if ( ( regions.minima = m_dup (top->dataclip.minima,
				   sizeof *top->dataclip.minima *
				   regions.num_regions) )
	== NULL ) return;
    if ( ( regions.maxima = m_dup (top->dataclip.maxima,
				   sizeof *top->dataclip.maxima *
				   regions.num_regions ) )
	== NULL )
    {
	m_free (regions.minima);
	return;
    }
    _Dataclip_update_clip_labels (top);
    XtCallCallbacks ( (Widget) top, XkwNscaledRegionCallback,
		      (XtPointer) &regions );
    iarray_get_data_scaling (top->dataclip.array, &scale, &offset);
    for (count = 0; count < regions.num_regions; ++count)
    {
	double min, max, tmp;

	min = (top->dataclip.minima[count] - offset) / scale;
	max = (top->dataclip.maxima[count] - offset) / scale;
	if (max < min) swap (min, max, tmp);
	regions.minima[count] = min;
	regions.maxima[count] = max;
    }
    XtCallCallbacks ( (Widget) top, XkwNrawRegionCallback, &regions );
    m_free (regions.minima);
    m_free (regions.maxima);
}   /*  End Function apply_cbk  */

static void _Dataclip_subset_cbk (Widget w, XtPointer client_data,
				  XtPointer call_data)
{
    DataclipWidget top = (DataclipWidget) client_data;
    flag subset = (uaddr) call_data;

    if (top->dataclip.array == NULL) return;
    top->dataclip.raw_histogram.length = 0;
    top->dataclip.display_histogram.length = 0;
    update_limits (top, FALSE);
    kwin_refresh_if_visible (top->dataclip.pixcanvas, TRUE);
    if (subset) full_range_cbk (NULL, client_data, NULL);
}   /*  End Function subset_cbk  */

static void _Dataclip_update_clip_labels (DataclipWidget w)
{
    char txt[STRING_LENGTH], val[STRING_LENGTH];

    iarray_format_value (w->dataclip.array, val, w->dataclip.minima[0],
			 1.0, 0.0);
    sprintf (txt, "Lower Clip: %s", val + 7);
    XtVaSetValues (w->dataclip.lower_label, XtNlabel, txt, NULL);
    iarray_format_value (w->dataclip.array, val,
			 w->dataclip.maxima[w->dataclip.num_regions - 1],
			 1.0, 0.0);
    sprintf (txt, "Upper Clip: %s", val + 7);
    XtVaSetValues (w->dataclip.upper_label, XtNlabel, txt, NULL);
}   /*  End Function update_clip_labels  */

static void _Dataclip_compute_subset (DataclipWidget w)
{
    unsigned int num_dim, count, factor, elem_type;
    unsigned long array_length;
    float min = TOOBIG;
    float max = -TOOBIG;
    double scale, offset;
    char *data;
    static char function_name[] = "DataclipWidget::compute_subset";

    num_dim = iarray_num_dim (w->dataclip.array);
    for (count = 0, array_length = 1; count < num_dim; ++count)
	array_length *= iarray_dim_length (w->dataclip.array, count);
    if (array_length < 1000000) factor = 11;
    else if (array_length < 10000000) factor = 101;
    else factor = 1009;
    array_length /= factor;
    if (w->dataclip.subset.buf_len < array_length)
    {
	if (w->dataclip.subset.array) m_free (w->dataclip.subset.array);
	if ( ( w->dataclip.subset.array =
	       m_alloc (sizeof *w->dataclip.subset.array * array_length) )
	     == NULL )
	    m_abort (function_name, "subset array");
	w->dataclip.subset.buf_len = array_length;
    }
    /*  Naughty: direct access of iarray members  */
    data = w->dataclip.array->data;
    elem_type = iarray_type (w->dataclip.array);
    factor *= ds_get_packet_size (w->dataclip.array->arr_desc->packet);
    for (count = 0; count < array_length; ++count, data += factor)
    {
	float val;
	double value[2];

	ds_get_element (data, elem_type, value, NULL);
	val = value[0];
	w->dataclip.subset.array[count] = val;
	if ( DOUBLE_IS_VALID (value[0]) )
	{
	    if (val < min) min = val;
	    if (val > max) max = val;
	}
    }
    w->dataclip.subset.length = array_length;
    w->dataclip.subset.min_raw = min;
    w->dataclip.subset.max_raw = max;
    iarray_get_data_scaling (w->dataclip.array, &scale, &offset);
    w->dataclip.subset.min_scaled = min * scale + offset;
    w->dataclip.subset.max_scaled = max * scale + offset;
    if (scale < 0.0)
    {
	double tmp;

	swap (w->dataclip.subset.min_scaled,w->dataclip.subset.max_scaled,tmp);
    }
    w->dataclip.subset.valid = TRUE;
}   /*  End Function compute_subset  */

static void histcanvas_size_control_func (KWorldCanvas canvas,
					  int width, int height,
					  struct win_scale_type *win_scale,
					  void **info, flag *boundary_clear)
/*  [SUMMARY] Size control callback.
    [PURPOSE] This routine will modify the window scaling information for a
    world canvas. While this routine is running, colourmap resize events are
    ignored. Hence this routine may safely cause the associated Kcolourmap
    object to be resized.
    <canvas> The world canvas object.
    <width> The width of the canvas in pixels after removing the space consumed
    by the dressing.
    <height> The height of the canvas in pixels after removing the space
    consumed by the dressing.
    <win_scale> The window scaling information. This may be modified.
    <info> A pointer to the arbitrary canvas information pointer.
    <boundary_clear> If the value TRUE is written here then the
    [<canvas_resize>] routine will attempt to clear only the boundary between
    the pixel canvas and the world canvas. If the value FALSE is written here
    or nothing is written here, the [<canvas_resize>] routine will clear the
    entire pixel canvas as appropriate.
    [RETURNS] Nothing.
*/
{
    flag ok;
    unsigned int count, raw_length, display_length, factor;
    unsigned long hpeak;
    double scale, offset, min_rawx, max_rawx, tmp;
    DataclipWidget w = (DataclipWidget) *info;
    Bool use_subset;
    static char function_name[]="DataclipWidget::histcanvas_size_control_func";

    win_scale->x_offset = 0;
    win_scale->y_offset = 0;
    win_scale->x_pixels = width;
    win_scale->y_pixels = height;
    if (w->dataclip.array == NULL) return;
    XtVaGetValues (w->dataclip.subset_tgl, XtNstate, &use_subset, NULL);
    if (use_subset && !w->dataclip.subset.valid) _Dataclip_compute_subset (w);
    /*  Use real width, otherwise histogram will be computed multiple times  */
    kwin_get_size (canvas_get_pixcanvas (canvas), &width, NULL);
    display_length = width;
    /*  Compute over-sampling factor  */
    for (factor = 1; width * factor < MIN_RAW_HISTOGRAM_LENGTH; ++factor);
    raw_length = width * factor;
    if (w->dataclip.verbose)
	fprintf (stderr, "factor: %u  raw_length: %u\n", factor, raw_length);
    iarray_get_data_scaling (w->dataclip.array, &scale, &offset);
    if (w->dataclip.reset_canvas_scale)
    {
	if (use_subset)
	{
	    min_rawx = w->dataclip.subset.min_raw;
	    max_rawx = w->dataclip.subset.max_raw;
	    win_scale->left_x = w->dataclip.subset.min_scaled;
	    win_scale->right_x = w->dataclip.subset.max_scaled;
	}
	else
	{
	    min_rawx = w->dataclip.bigset.min_raw;
	    max_rawx = w->dataclip.bigset.max_raw;
	    win_scale->left_x = w->dataclip.bigset.min_scaled;
	    win_scale->right_x = w->dataclip.bigset.max_scaled;
	}
	w->dataclip.reset_canvas_scale = FALSE;
    }
    else
    {
	min_rawx = (win_scale->left_x - offset) / scale;
	max_rawx = (win_scale->right_x - offset) / scale;
	if (max_rawx < min_rawx) swap (min_rawx, max_rawx, tmp);
    }
    if (display_length == w->dataclip.display_histogram.length) return;
    /*  Ensure buffers are large enough  */
    if (raw_length > w->dataclip.raw_histogram.buf_length)
    {
	/*  Increase histogram array size  */
	if (w->dataclip.raw_histogram.array)
	{
	    m_free (w->dataclip.raw_histogram.array);
	    w->dataclip.raw_histogram.buf_length = 0;
	}
	if ( ( w->dataclip.raw_histogram.array =
	       m_alloc (sizeof *w->dataclip.raw_histogram.array * raw_length) )
	     == NULL )
	    m_abort (function_name, "raw histogram array");
	w->dataclip.raw_histogram.buf_length = raw_length;
    }
    w->dataclip.raw_histogram.length = raw_length;
    m_clear (w->dataclip.raw_histogram.array,
	     sizeof *w->dataclip.raw_histogram.array * raw_length);
    if (display_length > w->dataclip.display_histogram.buf_length)
    {
	/*  Increase histogram array size  */
	if (w->dataclip.display_histogram.array)
	{
	    m_free (w->dataclip.display_histogram.array);
	    w->dataclip.display_histogram.buf_length = 0;
	}
	if ( ( w->dataclip.display_histogram.array =
	       m_alloc (sizeof *w->dataclip.display_histogram.array *
			display_length) )
	     == NULL )
	    m_abort (function_name, "display histogram array");
	w->dataclip.display_histogram.buf_length = display_length;
    }
    w->dataclip.display_histogram.length = display_length;
    hpeak = 0;
    w->dataclip.raw_histogram.index_of_peak = 0;
    if (use_subset)
	ok =ds_find_single_histogram((CONST char *) w->dataclip.subset.array,
				     K_FLOAT, win_scale->conv_type,
				     w->dataclip.subset.length,
				     NULL, sizeof (float), min_rawx, max_rawx,
				     raw_length,
				     w->dataclip.raw_histogram.array, &hpeak,
				     &w->dataclip.raw_histogram.index_of_peak);
    else
	ok =iarray_compute_histogram(w->dataclip.array, win_scale->conv_type,
				     min_rawx, max_rawx,
				     raw_length,
				     w->dataclip.raw_histogram.array, &hpeak,
				     &w->dataclip.raw_histogram.index_of_peak);
    if (!ok)
    {
	fprintf (stderr, "Error computing histogram\n");
	a_prog_bug (function_name);
    }
    if (w->dataclip.verbose) fprintf (stderr, "histogram peak: %lu\n", hpeak);
    if (scale < 0.0)
    {   /*  Flip histogram left-to-right  */
	unsigned long tmp;

	for (count = 0; count < raw_length / 2; ++count)
	{
	    swap (w->dataclip.raw_histogram.array[count],
		  w->dataclip.raw_histogram.array[raw_length - 1 - count],
		  tmp);
	}
	tmp = w->dataclip.raw_histogram.index_of_peak;
	w->dataclip.raw_histogram.index_of_peak = raw_length - 1 - tmp;
    }
    for (count = 0; count < display_length; ++count)
    {
	unsigned int subcount;
	unsigned long sum;

	for (subcount = 0, sum = 0; subcount < factor; ++subcount)
	    sum += w->dataclip.raw_histogram.array[count * factor + subcount];
	w->dataclip.display_histogram.array[count] = sum;
    }
    hpeak = w->dataclip.raw_histogram.index_of_peak / factor;
    w->dataclip.display_histogram.index_of_peak = hpeak;
    /*  Have histogram data: setup logarithmic scale  */
    hpeak = w->dataclip.display_histogram.array[hpeak];
    win_scale->top_y = hpeak;
#if 0  /*  Try not to limit the number of decades  */
    if (hpeak > 1)
    {
	double frac;

	frac = log10 (win_scale->top_y);
	if (frac > 6.0) frac = 6.0;
	else if (frac < 1.0) frac = 1.0;
	win_scale->bottom_y = win_scale->top_y / pow (10.0, frac);
	if (win_scale->bottom_y < 0.5) win_scale->bottom_y = 0.5;
    }
    else
#endif
	win_scale->bottom_y = 0.5;
}   /*  End Function histcanvas_size_control_func  */

static void _Dataclip_canvas_refresh_func (KWorldCanvas canvas,
					   int width, int height,
					   struct win_scale_type *win_scale,
					   Kcolourmap cmap, flag cmap_resize,
					   void **info)
/*  This routine is a refresh event consumer for a world canvas.
    The canvas is given by  canvas  .
    The width of the canvas in pixels is given by  width  .
    The height of the canvas in pixels is given by  height  .
    The window scaling information is pointed to by  win_scale  .
    The colourmap associated with the canvas is given by  cmap  .
    If the refresh function was called as a result of a colourmap resize
    the value of  cmap_resize  will be TRUE.
    The arbitrary canvas information pointer is pointed to by  info  .
    The routine returns nothing.
*/
{
    KPixCanvas pixcanvas;
    int count;
    unsigned long pixel_value;
    DataclipWidget w = (DataclipWidget) *info;
    double x, y, x0, x1, left_x, right_x, bottom_y, top_y;
    /*static char function_name[] = "_Dataclip_canvas_refresh_func";*/

    if (w->dataclip.array == NULL) return;
    canvas_get_attributes (canvas,
			   CANVAS_ATT_LEFT_X, &left_x,
			   CANVAS_ATT_RIGHT_X, &right_x,
			   CANVAS_ATT_BOTTOM_Y, &bottom_y,
			   CANVAS_ATT_TOP_Y, &top_y,
			   CANVAS_ATT_MAX_SAT_PIXEL, &pixel_value,
			   CANVAS_ATT_END);
    canvas_draw_lines_p (canvas, (double *) NULL,
			 w->dataclip.display_histogram.array,
			 w->dataclip.display_histogram.length,
			 pixel_value);
    pixel_value = get_colour (canvas, "yellow4");
    if ( DOUBLE_IS_VALID (w->dataclip.left_pos) )
    {
	x0 = w->dataclip.left_pos;
	canvas_draw_line_p (canvas, x0, bottom_y, x0, top_y, pixel_value);
    }
    if (w->dataclip.num_regions < 1) return;
    pixcanvas = canvas_get_pixcanvas (canvas);
    for (count = 0; count < w->dataclip.num_regions; ++count)
    {
	x0 = w->dataclip.minima[count];
	x1 = w->dataclip.maxima[count];
	/*  Draw the two vertical lines  */
	canvas_draw_line_p (canvas, x0, bottom_y, x0, top_y, pixel_value);
	canvas_draw_line_p (canvas, x1, bottom_y, x1, top_y, pixel_value);
	/*  Draw the diagonal lines  */
	canvas_draw_line_p (canvas, x0, bottom_y, x1, top_y, pixel_value);
	canvas_draw_line_p (canvas, x1, bottom_y, x0, top_y, pixel_value);
	/*  Draw the "width" ellipse  */
	x = (x0 + x1) * 0.5;
	y = log (top_y / bottom_y) * (x1 - x0) / (right_x - left_x);
	y = bottom_y * exp (y);
	canvas_convert_from_canvas_coords (canvas, FALSE, FALSE, 1, &x, &y,
					   &x, &y);
	kwin_draw_ellipse (pixcanvas, x, y, 5.0, 5.0, pixel_value);
    }
}   /*  End Function _Dataclip_canvas_refresh_func  */

static flag _Dataclip_canvas_position_func (KWorldCanvas canvas,
					    double x, double y,
					    unsigned int event_code,
					    void *e_info, void **f_info,
					    double x_lin, double y_lin,
					    int x_pix, int y_pix)
/*  [SUMMARY] Position event callback.
    [PURPOSE] This routine is a position event consumer for a world canvas.
    <canvas> The canvas on which the event occurred.
    <x> The horizontal world co-ordinate of the event.
    <y> The vertical world co-ordinate of the event.
    <event_code> The arbitrary event code.
    <e_info> A pointer to arbitrary event information.
    <f_info> A pointer to an arbitrary function information pointer.
    <x_lin> The horizontal linear world co-ordinate prior to the transform
    function being called.
    <y_lin> The vertical linear world co-ordinate prior to the transform
    function being called.
    <x_pix> The horizontal pixel co-ordinate.
    <y_pix> The vertical pixel co-ordinate.
    [NOTE] The world co-ordinate values will have been transformed by the
    registered transform function (see [<canvas_register_transform_func>]).
    [RETURNS] TRUE if the event was consumed, else FALSE indicating that
    the event is still to be processed.
*/
{
    flag apply = FALSE;
    unsigned int region, index;
    double range, x0, x1, left_x, right_x, bottom_y, top_y;
    DataclipWidget w = (DataclipWidget) *f_info;
    CONST char *key;
    char txt[STRING_LENGTH], val[STRING_LENGTH];

    if (w->dataclip.array == NULL) return (FALSE);
    canvas_get_attributes (canvas,
			   CANVAS_ATT_LEFT_X, &left_x,
			   CANVAS_ATT_RIGHT_X, &right_x,
			   CANVAS_ATT_BOTTOM_Y, &bottom_y,
			   CANVAS_ATT_TOP_Y, &top_y,
			   CANVAS_ATT_END);
    iarray_format_value (w->dataclip.array, val, x, 1.0, 0.0);
    sprintf (txt, "Bin value: %s", val + 7);
    XtVaSetValues (w->dataclip.cursor_x_label,
		   XtNlabel, txt,
		   NULL);
    index = rint ( (x - left_x) / (right_x - left_x) *
		   (double) (w->dataclip.display_histogram.length - 1) );
    sprintf (txt, "Number of values in bin: %g",
	     w->dataclip.display_histogram.array[index]);
    XtVaSetValues (w->dataclip.cursor_y_label, XtNlabel, txt, NULL);
    switch (event_code)
    {
      case K_CANVAS_EVENT_POINTER_MOVE:
	return (TRUE);
	/*break;*/
      case K_CANVAS_EVENT_LEFT_MOUSE_CLICK:
      case K_CANVAS_EVENT_LEFT_MOUSE_DRAG:
	if (w->dataclip.last_was_left)
	{
	    XBell (XtDisplay (w), 100);
	    return (TRUE);
	}
	if ( (w->dataclip.maxRegions == 1) &&
	     (w->dataclip.num_regions == 1) )
	{
	    if (x >= w->dataclip.maxima[0])
	    {
		XBell (XtDisplay (w), 100);
		return (TRUE);
	    }
	    w->dataclip.minima[0] = x;
	    if (w->dataclip.immediate_apply) apply = TRUE;
	}
	else if (w->dataclip.num_regions >= w->dataclip.maxRegions)
	{
	    XBell (XtDisplay (w), 100);
	    return (TRUE);
	}
	else
	{
	    w->dataclip.left_pos = x;
	    w->dataclip.last_was_left = TRUE;
	}
	break;
      case K_CANVAS_EVENT_MIDDLE_MOUSE_CLICK:
      case K_CANVAS_EVENT_MIDDLE_MOUSE_DRAG:
	/*  Search for region  */
	for (region = 0; region < w->dataclip.num_regions; ++region)
	{
	    if ( (x >= w->dataclip.minima[region]) &&
		 (x <= w->dataclip.maxima[region]) ) break;
	}
	if (w->dataclip.num_regions == 1) region = 0;
	if (region >= w->dataclip.num_regions)
	{
	    XBell (XtDisplay (w), 100);
	    return (TRUE);
	}
	range = log (y / bottom_y) / log (top_y / bottom_y);
	range = (right_x - left_x) * range;
	x0 = x - range * 0.5;
	x1 = x + range * 0.5;
	if ( (x0 < left_x) || (x1 > right_x) || (x0 >= x1) )
	{
	    XBell (XtDisplay (w), 100);
	    return (TRUE);
	}
	w->dataclip.minima[region] = x0;
	w->dataclip.maxima[region] = x1;
	if (w->dataclip.immediate_apply) apply = TRUE;
	break;
      case K_CANVAS_EVENT_RIGHT_MOUSE_CLICK:
      case K_CANVAS_EVENT_RIGHT_MOUSE_DRAG:
	if ( (w->dataclip.maxRegions == 1) &&
	     (w->dataclip.num_regions == 1) )
	{
	    if (x <= w->dataclip.minima[0])
	    {
		XBell (XtDisplay (w), 100);
		return (TRUE);
	    }
	    w->dataclip.maxima[0] = x;
	}
	else if (!w->dataclip.last_was_left)
	{
	    if (w->dataclip.num_regions > 0)
	    {
		XBell (XtDisplay (w), 100);
		return (TRUE);
	    }
	    w->dataclip.minima[0] = w->dataclip.curr_min;
	    w->dataclip.maxima[0] = x;
	    w->dataclip.num_regions = 1;
	}
	else if (w->dataclip.num_regions >= w->dataclip.maxRegions)
	{
	    XBell (XtDisplay (w), 100);
	    return (TRUE);
	}
	else if (x <= w->dataclip.left_pos)
	{
	    XBell (XtDisplay (w), 100);
	    return (TRUE);
	}
	else
	{
	    w->dataclip.minima[w->dataclip.num_regions] = w->dataclip.left_pos;
	    w->dataclip.maxima[w->dataclip.num_regions++] = x;
	    w->dataclip.last_was_left = FALSE;
	    w->dataclip.left_pos = TOOBIG;
	}
	if (w->dataclip.immediate_apply) apply = TRUE;
	break;
      case K_CANVAS_EVENT_PLAIN_KEY_PRESS:
	key = e_info;
	switch (*key)
	{
	  case 'i':
	    do_zoom (w, 0.5, x_lin);
	    canvas_get_attributes (canvas,
				   CANVAS_ATT_BOTTOM_Y, &bottom_y,
				   CANVAS_ATT_TOP_Y, &top_y,
				   CANVAS_ATT_END);
	    canvas_warp_pointer (canvas, x_lin, (bottom_y + top_y) * 0.5, TRUE,
				 TRUE);
	    return (TRUE);
	    /*break;*/
	  case 'o':
	    do_zoom (w, 2.0, TOOBIG);
	    return (TRUE);
	    /*break;*/
	  case 'u':
	    unzoom_cbk (NULL, w, NULL);
	    return (TRUE);
	    /*break;*/
	}
	break;
      default:
	return (FALSE);
	/*break;*/
    }
    canvas_resize (canvas, NULL, TRUE);
    if (apply)
    {
	XSync (XtDisplay (w), False);
	apply_cbk ( (Widget) NULL, (XtPointer) w, NULL );
    }
    return (TRUE);
}   /*  End Function _Dataclip_canvas_position_func  */

static void pop_cbk (w, client_data, call_data)
/*  This is the generic popup/popdown callback.
*/
Widget w;
XtPointer client_data;
XtPointer call_data;
{
    DataclipWidget top = (DataclipWidget) w;

    top->dataclip.popped_up = (iaddr) client_data;
}   /*  End Function pop_cbk   */

static void full_range_cbk (Widget w, XtPointer client_data,
			    XtPointer call_data)
/*  This is the generic popup/popdown callback.
*/
{
    DataclipWidget top = (DataclipWidget) client_data;

    top->dataclip.last_was_left = FALSE;
    top->dataclip.left_pos = TOOBIG;
    if (top->dataclip.array == NULL) return;
    top->dataclip.num_regions = 1;
    top->dataclip.minima[0] = top->dataclip.curr_min;
    top->dataclip.maxima[0] = top->dataclip.curr_max;
    unzoom_cbk (NULL, top, NULL);
    XSync (XtDisplay ( (Widget) client_data ), False);
    apply_cbk ( (Widget) NULL, (XtPointer) top, NULL );
}   /*  End Function full_range_cbk   */

static void iarr_destroy_callback (iarray arr, DataclipWidget top)
/*  [PURPOSE] This routine will register the destruction of the Intelligent
    Array.
    <arr> The Intelligent Array.
    <top> The Dataclip widget.
    [RETURNS] Nothing.
*/
{
    top->dataclip.array = NULL;
    top->dataclip.iarr_destroy_callback = NULL;
    if (top->dataclip.autoPopdown) XtPopdown ( (Widget) top );
}   /*  End Function iarr_destroy_callback  */

static void blank_data_cbk (Widget w, XtPointer client_data,
				 XtPointer call_data)
{
    flag bool = (uaddr) call_data;
    DataclipWidget top = (DataclipWidget) client_data;

    top->dataclip.blank_data = bool;
    apply_cbk ( (Widget) NULL, (XtPointer) top, NULL );
}   /*  End Function blank_data_cbk  */

static void type_menu_cbk (Widget w, XtPointer client_data,XtPointer call_data)
{
    unsigned int type;
    unsigned int index = *(int *) call_data;
    DataclipWidget top = (DataclipWidget) client_data;

    type = dataclipClassRec.dataclip_class.type_index_to_type[index];
    top->dataclip.output_type = type;
}   /*  End Function type_menu_cbk  */

static void save_cbk (Widget w, XtPointer client_data, XtPointer call_data)
{
    iarray new, old;
    flag rescale = TRUE;
    unsigned int count, num_dim;
    Widget dialog;
    DataclipWidget top = (DataclipWidget) client_data;
    double inp_scale, inp_offset, min, max, scale, offset, low, high, tmp;
    char *arrayfile_name;
    unsigned long *dim_lengths;
    CONST char **dim_names;
    static char function_name[] = "DataclipWidget::save_cbk";

    dialog = top->dataclip.dialog;
    if (top->dataclip.num_regions < 1)
    {
	XBell (XtDisplay (w), 100);
	return;
    }
    if (top->dataclip.num_regions > 1)
    {
	fprintf (stderr, "Multiple regions not supported yet\n");
	return;
    }
    arrayfile_name = XawDialogGetValueString (dialog);
    if (arrayfile_name == NULL) arrayfile_name = "scaled";
    if (strlen (arrayfile_name) < 1) arrayfile_name = "scaled";
    old = top->dataclip.array;
    num_dim = iarray_num_dim (old);
    if ( ( dim_names = m_alloc (num_dim * sizeof *dim_names) ) == NULL )
    {
	m_abort (function_name, "array of dimension names pointers");
    }
    if ( ( dim_lengths = m_alloc (num_dim * sizeof *dim_lengths) ) == NULL )
    {
	m_abort (function_name, "array of dimension lengths");
    }
    for (count = 0; count < num_dim; ++count)
    {
	dim_names[count] = iarray_dim_name (old, count);
	dim_lengths[count] = iarray_dim_length (old, count);
    }
    if ( ( new = iarray_create (top->dataclip.output_type, num_dim, dim_names,
				dim_lengths, iarray_value_name (old), old) )
	 == NULL )
    {
	m_error_notify (function_name, "output Intelligent Array");
	m_free (dim_names);
	m_free (dim_lengths);
	return;
    }
    m_free (dim_names);
    m_free (dim_lengths);
    for (count = 0; count < num_dim; ++count)
    {
	iarray_get_world_coords (old, count, &min, &max);
	iarray_set_world_coords (new, count, min, max);
    }
    iarray_get_data_scaling (old, &inp_scale, &inp_offset);
    iarray_set_data_range (new, top->dataclip.minima[0],
			   top->dataclip.maxima[0]);
    /*  Determine desired new minimum and maximum range  */
    switch (top->dataclip.output_type)
    {
      case K_BYTE:
	min = -127.0;
	max = 127.0;
	break;
      case K_UBYTE:
	min = 0.0;
	max = 255.0;
	break;
      case K_SHORT:
	min = -32767.0;
	max = 32767.0;
	break;
      case K_USHORT:
	min = 0.0;
	max = 65535.0;
	break;
      case K_INT:
	min = -2147483647.0;
	max = 2147483647.0;
	break;
      case K_UINT:
	min = 0.0;
	max = 4294967296.0;
	break;
      default:
	rescale = FALSE;
	break;
    }
    low  = (top->dataclip.minima[0] - inp_offset) / inp_scale;
    high = (top->dataclip.maxima[0] - inp_offset) / inp_scale;
    if (high < low) swap (low, high, tmp);
    if (rescale)
    {
	/*  Must fit the defined range of data into a range  */
	/*  TODO: it would be nice to generate output with +ve scale  */
	scale = (max - min) / (high - low);
	offset = min - scale * low;
	iarray_set_data_scaling (new, inp_scale / scale,
				 inp_offset - offset * inp_scale / scale);
    }
    else
    {
	scale = inp_scale;
	offset = inp_offset;
	iarray_set_data_scaling (new, 1.0, 0.0);
    }
    iarray_clip_scale_and_offset (new, old, scale, offset, low, high,
				  top->dataclip.blank_data);
    if ( !iarray_write (new, arrayfile_name) )
    {
	iarray_dealloc (new);
	return;
    }
    iarray_dealloc (new);
    XBell (XtDisplay (w), -90);
}   /*  End Function save_cbk  */

static unsigned long get_colour (KWorldCanvas canvas, CONST char *colourname)
/*  [SUMMARY] Get colour.
    <canvas> The world canvas.
    <colourname> The name of the colour.
    [RETURNS] The specified colour if available, else a default colour.
*/
{
    KPixCanvas pixcanvas;
    unsigned long pixel_value;

    pixcanvas = canvas_get_pixcanvas (canvas);
    if ( kwin_get_colour (pixcanvas, colourname, &pixel_value,
			  NULL, NULL, NULL) )
    {
	return (pixel_value);
    }
    if ( kwin_get_colour (pixcanvas, "white", &pixel_value, NULL, NULL, NULL) )
    {
	fprintf (stderr,
		 "Error allocating colour: \"%s\" defaulting to white\n",
		 colourname);
	return (pixel_value);
    }
    fprintf (stderr, "Error allocating colour: \"%s\" defaulting to 0\n",
	     colourname);
    return (0);
}   /*  End Function get_colour  */

static void zoom_cbk (Widget w, XtPointer client_data, XtPointer call_data)
/*  [SUMMARY] Zoom callback.
    <w> The widget.
    <client_data> Client data pointer.
    <call_data> Call-specific data pointer.
    [RETURNS] Nothing.
*/
{
    double min, max, left, right;
    DataclipWidget top = (DataclipWidget) client_data;

    if (top->dataclip.array == NULL) return;
    if ( DOUBLE_IS_BLANK (top->dataclip.curr_min) ) return;
    if (top->dataclip.num_regions < 1) return;
    canvas_get_attributes (top->dataclip.worldcanvas,
			   CANVAS_ATT_LEFT_X, &left,
			   CANVAS_ATT_RIGHT_X, &right,
			   CANVAS_ATT_END);
    min = top->dataclip.minima[0];
    max = top->dataclip.maxima[top->dataclip.num_regions - 1];
    if ( (min == left) && (max == right) ) return;
    canvas_set_attributes (top->dataclip.worldcanvas,
			   CANVAS_ATT_LEFT_X, min,
			   CANVAS_ATT_RIGHT_X, max,
			   CANVAS_ATT_END);
    /*  Force histogram to be recomputed  */
    top->dataclip.raw_histogram.length = 0;
    top->dataclip.display_histogram.length = 0;
    kwin_refresh_if_visible (top->dataclip.pixcanvas, TRUE);
}   /*  End Function zoom_cbk   */

static void unzoom_cbk (Widget w, XtPointer client_data, XtPointer call_data)
/*  [SUMMARY] Unzoom callback.
    <w> The widget.
    <client_data> Client data pointer.
    <call_data> Call-specific data pointer.
    [RETURNS] Nothing.
*/
{
    double min, max, left, right;
    DataclipWidget top = (DataclipWidget) client_data;

    if (top->dataclip.array == NULL) return;
    if ( DOUBLE_IS_BLANK (top->dataclip.curr_min) ) return;
    canvas_get_attributes (top->dataclip.worldcanvas,
			   CANVAS_ATT_LEFT_X, &left,
			   CANVAS_ATT_RIGHT_X, &right,
			   CANVAS_ATT_END);
    min = top->dataclip.curr_min;
    max = top->dataclip.curr_max;
    if ( (min == left) && (max == right) )
    {
	/*  Refresh in case called from <full_range_cbk>  */
	kwin_refresh_if_visible (top->dataclip.pixcanvas, TRUE);
	return;
    }
    canvas_set_attributes (top->dataclip.worldcanvas,
			   CANVAS_ATT_LEFT_X, min,
			   CANVAS_ATT_RIGHT_X, max,
			   CANVAS_ATT_END);
    /*  Force histogram to be recomputed  */
    top->dataclip.raw_histogram.length = 0;
    top->dataclip.display_histogram.length = 0;
    kwin_refresh_if_visible (top->dataclip.pixcanvas, TRUE);
}   /*  End Function unzoom_cbk   */

static void do_zoom (DataclipWidget w, double factor, double mid)
/*  [SUMMARY] Zoom the histogram display by a specified factor.
    <w> The Dataclip widget.
    <factor> The factor to zoom by. Values greater than one zoom out.
    <mid> The midpoint to zoom in on, in linear co-ordinates. If this is
    TOOBIG, the midpoint of the current range is used.
    [RETURNS] Nothing.
*/
{
    double old_left, old_right, new_left, new_right, range;

    if (w->dataclip.array == NULL) return;
    if ( DOUBLE_IS_BLANK (w->dataclip.curr_min) ) return;
    canvas_get_attributes (w->dataclip.worldcanvas,
			   CANVAS_ATT_LEFT_X, &old_left,
			   CANVAS_ATT_RIGHT_X, &old_right,
			   CANVAS_ATT_END);
    if ( DOUBLE_IS_BLANK (mid) ) mid = (old_left + old_right) * 0.5;
    range = old_right - old_left;
    new_left = mid - range * factor * 0.5;
    new_right = mid + range * factor * 0.5;
    if (new_left < w->dataclip.curr_min) new_left = w->dataclip.curr_min;
    if (new_right > w->dataclip.curr_max) new_right = w->dataclip.curr_max;
    if ( (new_left == old_left) && (new_right == old_right) ) return;
    canvas_set_attributes (w->dataclip.worldcanvas,
			   CANVAS_ATT_LEFT_X, new_left,
			   CANVAS_ATT_RIGHT_X, new_right,
			   CANVAS_ATT_END);
    /*  Force histogram to be recomputed  */
    w->dataclip.raw_histogram.length = 0;
    w->dataclip.display_histogram.length = 0;
    kwin_refresh_if_visible (w->dataclip.pixcanvas, TRUE);
}   /*  End Function do_zoom   */

static void update_limits (DataclipWidget w, flag reset_hscale)
/*  [SUMMARY] Update the displayed data range.
    <w> The Dataclip widget.
    <reset_hscale> If TRUE the canvas horizontal scale is reset.
    [RETURNS] Nothing.
*/
{
    Bool use_subset;
    char txt[STRING_LENGTH], val[STRING_LENGTH];

    if (w->dataclip.array == NULL) return;
    XtVaGetValues (w->dataclip.subset_tgl, XtNstate, &use_subset, NULL);
    if (use_subset && !w->dataclip.subset.valid) _Dataclip_compute_subset (w);
    if ( DOUBLE_IS_BLANK (w->dataclip.bigset.min_raw) )
    {   /*  Recompute the minimum and maximum  */
	double min, max;

	if ( !iarray_min_max (w->dataclip.array, CONV1_REAL, &min, &max) )
	{
	    fprintf (stderr, "Error getting image range\n");
	    return;
	}
	if (min >= max)
	{
	    w->dataclip.array = NULL;
	    return;
	}
	w->dataclip.bigset.min_raw = min;
	w->dataclip.bigset.max_raw = max;
    }
    if ( DOUBLE_IS_BLANK (w->dataclip.bigset.min_scaled) )
    {
	double scale, offset, tmp;

	iarray_get_data_scaling (w->dataclip.array, &scale, &offset);
	w->dataclip.bigset.min_scaled = w->dataclip.bigset.min_raw * scale
	    + offset;
	w->dataclip.bigset.max_scaled = w->dataclip.bigset.max_raw * scale
	    + offset;
	if (scale < 0.0)
	    swap (w->dataclip.bigset.min_scaled, w->dataclip.bigset.max_scaled,
		  tmp);
    }
    if (use_subset)
    {
	w->dataclip.curr_min = w->dataclip.subset.min_scaled;
	w->dataclip.curr_max = w->dataclip.subset.max_scaled;
    }
    else
    {
	w->dataclip.curr_min = w->dataclip.bigset.min_scaled;
	w->dataclip.curr_max = w->dataclip.bigset.max_scaled;
    }
    iarray_format_value (w->dataclip.array, val, w->dataclip.curr_min,
			 1.0, 0.0);
    sprintf (txt, "Data Min: %s", val + 7);
    XtVaSetValues (w->dataclip.min_label, XtNlabel, txt, NULL);
    iarray_format_value (w->dataclip.array, val, w->dataclip.curr_max,
			 1.0, 0.0);
    sprintf (txt, "Data Max: %s", val + 7);
    XtVaSetValues (w->dataclip.max_label, XtNlabel, txt, NULL);
    w->dataclip.reset_canvas_scale = reset_hscale;
}   /*  End Function update_limits  */

static void clip_cbk (Widget w, XtPointer client_data, XtPointer call_data)
/*  [SUMMARY] Clip callback.
    <w> The widget.
    <client_data> Client data pointer.
    <call_data> Call-specific data pointer.
    [RETURNS] Nothing.
*/
{
    double proportion;
    DataclipWidget top = (DataclipWidget) client_data;
    String label;
    CONST char *p;
    char txt[STRING_LENGTH];

    if (top->dataclip.array == NULL) return;
    /*  Find how much to clip  */
    XtVaGetValues (w, XtNlabel, &label, NULL);
    if (strcmp (label, "Zero Left") == 0)
    {
	if (top->dataclip.curr_min > 0.0) return;
	top->dataclip.minima[0] = 0.0;
	if (top->dataclip.num_regions < 1)
	{
	    top->dataclip.maxima[0] = top->dataclip.curr_max;
	    top->dataclip.num_regions = 1;
	}
	kwin_refresh_if_visible (top->dataclip.pixcanvas, TRUE);
	if (top->dataclip.immediate_apply)
	{
	    XSync (XtDisplay (top), False);
	    apply_cbk ( (Widget) NULL, (XtPointer) top, NULL );
	}
	return;
    }
    strcpy (txt, label);
    txt[strlen (txt) - 1] = '\0';
    proportion = ex_float (txt, &p) / 100.0;
    clip_outliers (top, proportion);
}   /*  End Function clip_cbk   */

static void clip_outliers (DataclipWidget top, double proportion)
/*  [SUMMARY] Clip outliers.
    <top> The Dataclip widget.
    <proportion> The proportion of data to clip to.
    [RETURNS] Nothing.
*/
{
    unsigned int length;
    unsigned long index;
    double total, sum, left, right, factor;

    if ( DOUBLE_IS_BLANK (top->dataclip.curr_min) ) return;
    length = top->dataclip.raw_histogram.length;
    if (length < 2) return;
    canvas_get_attributes (top->dataclip.worldcanvas,
			   CANVAS_ATT_LEFT_X, &left,
			   CANVAS_ATT_RIGHT_X, &right,
			   CANVAS_ATT_END);
    factor = (right - left) / (length - 1);
    /*  Find cutoff points which emcompass most of the data values. Search from
	the peak, as that is optimal for distributions with a long tail, which
	is why the user is using this option in the first place  */
    /*  First the lower cutoff  */
    for (index = top->dataclip.raw_histogram.index_of_peak, total = 0.0;
	 index > 0;
	 --index)
	total += top->dataclip.raw_histogram.array[index];
    total *= proportion;
    for (index = top->dataclip.raw_histogram.index_of_peak, sum = 0.0;
	 (index > 0) && (sum < total);
	 --index)
	sum += top->dataclip.raw_histogram.array[index];
    top->dataclip.minima[0] = left + index * factor;
    /*  Next the upper cutoff  */
    for (index = top->dataclip.raw_histogram.index_of_peak, total = 0.0;
	 index < length;
	 ++index)
    {
	total += top->dataclip.raw_histogram.array[index];
    }
    total *= proportion;
    for (index = top->dataclip.raw_histogram.index_of_peak, sum = 0.0;
	 (index < length) && (sum < total);
	 ++index)
	sum += top->dataclip.raw_histogram.array[index];
    top->dataclip.maxima[0] = left + index * factor;
    top->dataclip.num_regions = 1;
    kwin_refresh_if_visible (top->dataclip.pixcanvas, TRUE);
    if (top->dataclip.immediate_apply)
    {
	XSync (XtDisplay (top), False);
	apply_cbk ( (Widget) NULL, (XtPointer) top, NULL );
    }
}   /*  End Function clip_outliers   */

static flag _Dataclip_ds_event_func (void *object, multi_array *multi_desc,
				      CONST char *domain, CONST char *name)
/*  [SUMMARY] Process an event.
    <object> The object information pointer.
    <multi_desc> The multi_array descriptor.
    <domain> The domain from where the event was generated.
    <name> The name of the event in its domain.
    [RETURNS] TRUE if further callbacks should not be called, else FALSE.
*/
{
    double min, max;
    DataclipWidget top = (DataclipWidget) object;
    packet_desc *pack_desc;
    char *packet, *d_title;
    double value[2];

    if (strcmp (name, "DataclipWidget::Clip") != 0) return (FALSE);
    if (top->dataclip.array == NULL) return (FALSE);
    if (multi_desc->num_arrays != 1) return (FALSE);
    pack_desc = multi_desc->headers[0];
    packet = multi_desc->data[0];
    if ( ( d_title = ds_get_unique_named_string (pack_desc, packet,
						 "title") ) == NULL )
	return (FALSE);
    if (strcmp (top->wm.title, d_title) != 0)
    {
	m_free (d_title);
	return (FALSE);
    }
    m_free (d_title);
    if ( ds_get_unique_named_value (pack_desc, packet, "hist_clip", NULL,
				    value) )
    {
	clip_outliers (top, value[0]);
	return (TRUE);
    }
    if ( !ds_get_unique_named_value (pack_desc, packet, "minimum", NULL,
				     value) ) return (FALSE);
    min = value[0];
    if ( !ds_get_unique_named_value (pack_desc, packet, "maximum", NULL,
				     value) ) return (FALSE);
    max = value[0];
    if (min < top->dataclip.curr_min) min = top->dataclip.curr_min;
    if (max > top->dataclip.curr_max) max = top->dataclip.curr_max;
    top->dataclip.num_regions = 1;
    top->dataclip.minima[0] = min;
    top->dataclip.maxima[0] = max;
    if (top->dataclip.worldcanvas != NULL)
	canvas_resize (top->dataclip.worldcanvas, NULL, TRUE);
    apply_cbk ( (Widget) NULL, (XtPointer) top, NULL );
    return (TRUE);
}   /*  End Function _Dataclip_ds_event_func  */

static void _Dataclip_invalidate_data (DataclipWidget w)
/*  [SUMMARY] Invalidate current data.
    <w> The Dataclip widget.
    [RETURNS] Nothing.
*/
{
    w->dataclip.bigset.min_raw = TOOBIG;
    w->dataclip.bigset.max_raw = -TOOBIG;
    w->dataclip.bigset.min_scaled = TOOBIG;
    w->dataclip.bigset.max_scaled = -TOOBIG;
    w->dataclip.curr_min = TOOBIG;
    w->dataclip.curr_max = -TOOBIG;
    w->dataclip.raw_histogram.length = 0;
    w->dataclip.display_histogram.length = 0;
    w->dataclip.subset.valid = FALSE;
}   /*  End Function _Dataclip_invalidate_data  */


/*  Public functions follow  */

/*EXPERIMENTAL_FUNCTION*/
void XkwDataclipNewArray (Widget W, iarray array, double min, double max,
			  flag reset_hscale)
/*  [SUMMARY] Register new array.
    <W> The Dataclip widget.
    <array> The new array. This may be NULL.
    <min> The minimum (unscaled) data value in the array.
    <max> The maximum (unscaled) data value in the array. If this is less than
    <<min>> then the minimum and maximum values are computed.
    <reset_hscale> If TRUE the horizontal scale is reset.
    [RETURNS] Nothing.
*/
{
    DataclipWidget w = (DataclipWidget) W;
    static char function_name[] = "XkwDataclipNewArray";

    FLAG_VERIFY (reset_hscale);
    /*  Clean up any old array  */
    c_unregister_callback2 (w->dataclip.iarr_destroy_callback);
    w->dataclip.iarr_destroy_callback = NULL;
    w->dataclip.last_was_left = FALSE;
    w->dataclip.left_pos = TOOBIG;
    _Dataclip_invalidate_data (w);
    w->dataclip.array = array;
    if (max >= min)
    {
	w->dataclip.bigset.min_raw = min;
	w->dataclip.bigset.max_raw = max;
    }
    update_limits (w, reset_hscale);
    if (array)
    {
	/*  Process new array  */
	w->dataclip.iarr_destroy_callback = 
	    iarray_register_destroy_func(w->dataclip.array,
					 ( flag (*) () ) iarr_destroy_callback,
					 w);
	if (w->dataclip.worldcanvas)
	{
	    canvas_set_dressing (w->dataclip.worldcanvas,
				 KCD_ATT_HLABEL, iarray_value_name (array),
				 KCD_ATT_END);
	}
    }
    /*  Force computation and display of histogram if new data available,
	else clear canvas  */
    if (w->dataclip.popped_up)
	kwin_refresh_if_visible (w->dataclip.pixcanvas, TRUE);
}   /*  End Function XkwDataclipNewArray  */

/*EXPERIMENTAL_FUNCTION*/
void XkwDataclipArrayChanged (Widget W, double min, double max)
/*  [SUMMARY] Register a change in the array.
    <W> The Dataclip widget.
    <min> The minimum (unscaled) data value in the array.
    <max> The maximum (unscaled) data value in the array. If this is less than
    <<min>> then the minimum and maximum values are computed.
    [RETURNS] Nothing.
*/
{
    DataclipWidget w = (DataclipWidget) W;
    /*static char function_name[] = "XkwDataclipArrayChanged";*/

    _Dataclip_invalidate_data (w);
    if (max >= min)
    {
	w->dataclip.bigset.min_raw = min;
	w->dataclip.bigset.max_raw = max;
    }
    update_limits (w, TRUE);
    /*  Force computation and display of histogram if new data available,
	else clear canvas  */
    if (w->dataclip.popped_up)
	kwin_refresh_if_visible (w->dataclip.pixcanvas, TRUE);
}   /*  End Function XkwDataclipArrayChanged  */

/*EXPERIMENTAL_FUNCTION*/
void XkwDataclipSetRegions (Widget W, CONST DataclipRegions *regions)
/*  [SUMMARY] Set the regions.
    <W> The Dataclip widget.
    <regions> The regions. These are unscaled (raw) values.
    [RETURNS] Nothing.
*/
{
    double scale, offset, min, max, tmp;
    DataclipWidget w = (DataclipWidget) W;
    static char function_name[] = "XkwDataclipSetRegions";

    if (w->dataclip.array == NULL) return;
    if (regions->num_regions != 1)
    {
	fprintf (stderr, "%s: illegal number of regions: %u\n",
		 function_name, regions->num_regions);
	return;
    }
    w->dataclip.num_regions = 1;
    iarray_get_data_scaling (w->dataclip.array, &scale, &offset);
    min = regions->minima[0] * scale + offset;
    max = regions->maxima[0] * scale + offset;
    if (max < min) swap (min, max, tmp);
    w->dataclip.minima[0] = (min < w->dataclip.curr_min) ?
	w->dataclip.curr_min : min;
    w->dataclip.maxima[0] = (max > w->dataclip.curr_max) ?
	w->dataclip.curr_max : max;
    _Dataclip_update_clip_labels (w);
    if (w->dataclip.popped_up)
	kwin_refresh_if_visible (w->dataclip.pixcanvas, TRUE);
}   /*  End Function XkwDataclipSetRegions  */
