/*LINTLIBRARY*/
/*  DressingControl.c

    This code provides a dressing control widget for Xt.

    Copyright (C) 1997-2004  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This file contains all routines needed for a dressing control widget for
  Xt.


    Written by      Richard Gooch   15-FEB-1997

    Updated by      Richard Gooch   17-FEB-1997: Added colour dialogs.

    Updated by      Richard Gooch   19-FEB-1997: Added background colour dialog
  and "Screen Colours" and "Paper Colours" buttons.

    Updated by      Richard Gooch   20-FEB-1997: Added "Change Colourmap"
  toggle.

    Updated by      Richard Gooch   16-MAR-1997: Fixed bug in <apply_cbk> where
  canvas array pointer was not checked.

    Updated by      Richard Gooch   22-FEB-1998: Created function
  <XkwDressingControlAddCanvas>.

    Updated by      Richard Gooch   5-MAY-2000: Added grid colour control.

    Updated by      Richard Gooch   6-MAY-2000: Replaced several callbacks with
  <toggle_cbk>. Added internal ticks control.

    Updated by      Richard Gooch   11-JUL-2000: Added #include <karma_kcmap.h>
  and <karma_kwin.h>

    Updated by      Richard Gooch   12-JUL-2000: Removed unnecessary #includes.

    Updated by      Richard Gooch   20-MAR-2001: Created Destroy method.

    Updated by      Richard Gooch   6-JUN-2001: Set KCD_ATT_DISPLAY resource in
  <XkwDressingControlAddCanvas>.

    Updated by      Richard Gooch   5-NOV-2001: Added font scale slider.

    Updated by      Richard Gooch   31-JAN-2002: Switch to use new experimental
  function <kwin_get_colour_harder>. Added "Show Colourbar" toggle.

    Updated by      Richard Gooch   6-FEB-2002: Only create "Show Colourbar"
  toggle if grandparent is an ImageDisplay widget.

    Updated by      Richard Gooch   7-JAN-2004: Renamed some toggles and
  added "Show Scale" toggle.

    Updated by      Richard Gooch   11-JAN-2004: Minor cosmetic changes.

    Last updated by Richard Gooch   1-APR-2004: Changed to SimpleTextEntry
  widget class.


*/
#include <stdio.h>
#include <math.h>
#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Label.h>
#include <X11/Xaw/Form.h>
#include <karma.h>
#include <k_event_codes.h>
#include <karma_viewimg.h>
#include <karma_xtmisc.h>
#include <karma_kcmap.h>
#include <karma_kwin.h>
#include <karma_m.h>
#include <karma_a.h>
#include <Xkw/DressingControlP.h>
#include <Xkw/Ktoggle.h>
#include <Xkw/ExclusiveMenu.h>
#include <Xkw/SimpleSlider.h>
#include <Xkw/ImageDisplay.h>
#include <Xkw/SimpleTextEntry.h>


struct canvas_list
{
    KWorldCanvas canvas;
    struct canvas_list *next;
};


/*  Private functions  */
STATIC_FUNCTION (void _DressingControl_Initialise,
		 (Widget request, Widget new) );
STATIC_FUNCTION (void _DressingControl_Destroy, (Widget W) );
STATIC_FUNCTION (void _DressingControl_apply_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void _DressingControl_set_bool,
		 (DressingControlWidget top, unsigned int att, flag val) );
STATIC_FUNCTION (void _DressingControl_set_double,
		 (DressingControlWidget top, unsigned int att, double val) );
STATIC_FUNCTION (void _DressingControl_auto_refresh_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void _DressingControl_toggle_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void _DressingControl_screen_colours_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void _DressingControl_paper_colours_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void _DressingControl_set_cmap_bool,
		 (DressingControlWidget top, unsigned int att, flag val) );
STATIC_FUNCTION (void _DressingControl_slider_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void _DressingControl_colourbar_toggle_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );


#define offset(field) XtOffset(DressingControlWidget, dressingControl.field)

static XtResource DressingControlResources[] = 
{
    {"empty", "Empty", XtRInt, sizeof (int),
     offset (dummy), XtRImmediate, False},
};

#undef offset


DressingControlClassRec dressingControlClassRec = 
{
  {
      /* CoreClassPart */
    (WidgetClass) &topLevelShellClassRec,    /* superclass */
    "DressingControl",                       /* class_name */
    sizeof (DressingControlRec),             /* widget_size */
    NULL,                                    /* class_initialise */
    NULL,                                    /* class_part_initialise */
    FALSE,                                   /* class_init */
    (XtInitProc) _DressingControl_Initialise,/* initialise */
    NULL,                                    /* initialise_hook */
    XtInheritRealize,                        /* realise */
    NULL,                                    /* actions */
    0,                                       /* num_actions */
    DressingControlResources,                /* resources */
    XtNumber (DressingControlResources),     /* num_resources */
    NULLQUARK,                               /* xrm_class */
    TRUE,                                    /* compress_motion */
    TRUE,                                    /* compress_exposure */
    TRUE,                                    /* compress_enterleave */
    TRUE,                                    /* visible_interest */
    _DressingControl_Destroy,                /* destroy */
    XtInheritResize,                         /* resize */
    NULL,                                    /* expose */
    NULL,                                    /* set_values */
    NULL,                                    /* set_values_hook */
    XtInheritSetValuesAlmost,                /* set_values_almost */
    NULL,                                    /* get_values_hook */
    NULL,                                    /* accept_focus */
    XtVersion,                               /* version */
    NULL,                                    /* callback_private */
    NULL,                                    /* tm_translations */
    NULL,
    NULL,
    NULL,
  },
  {     /* CompositeClassPart */
    /* geometry_manager	  */	XtInheritGeometryManager,
    /* change_managed	  */	XtInheritChangeManaged,
    /* insert_child	  */	XtInheritInsertChild,
    /* delete_child	  */	XtInheritDeleteChild,
    /* extension	  */	NULL
  },{ /* Shell */
    /* extension	  */	NULL
  },{ /* WMShell */
    /* extension	  */	NULL
  },{ /* VendorShell */
    /* extension	  */	NULL
  },{ /* TopLevelShell */
    /* extension	  */	NULL
  },
  {  /* DressingControlClassPart */
    0 /* empty */
  }
};

WidgetClass dressingControlWidgetClass = (WidgetClass)&dressingControlClassRec;

static void _DressingControl_Initialise (Widget Request, Widget New)
{
    /*DressingControlWidget request = (DressingControlWidget) Request;*/
    DressingControlWidget new = (DressingControlWidget) New;
    Widget form, w, upper_widget;
    /*static char function_name[] = "DressingControlWidget::Initialise";*/

    new->dressingControl.auto_refresh = TRUE;
    new->dressingControl.first_canvas = NULL;
    form = XtVaCreateManagedWidget ("form", formWidgetClass, New,
				    XtNborderWidth, 0,
				    NULL);
    w = XtVaCreateManagedWidget ("closeButton", commandWidgetClass, form,
				 XtNlabel, "Close",
				 XtNtop, XtChainTop,
				 XtNbottom, XtChainTop,
				 XtNleft, XtChainLeft,
				 XtNheight, 20,
				 NULL);
    XtAddCallback (w, XtNcallback, xtmisc_popdown_cbk, New);
    w = XtVaCreateManagedWidget ("displayAxisLabelsToggle",
				 ktoggleWidgetClass, form,
				 XtNlabel, "Enable",
				 XtNfromHoriz, w,
				 XkwNcrosses, False,
				 NULL);
    XtAddCallback (w, XtNcallback, _DressingControl_toggle_cbk,
		   (XtPointer) KCD_ATT_DISPLAY);
    w = XtVaCreateManagedWidget ("apply", commandWidgetClass, form,
				 XtNlabel, "Apply",
				 XtNfromHoriz, w,
				 NULL);
    XtAddCallback (w, XtNcallback, _DressingControl_apply_cbk, New);
    w = XtVaCreateManagedWidget ("autoRefreshToggle", ktoggleWidgetClass, form,
				 XtNlabel, "Auto Refresh",
				 XtNstate, True,
				 XkwNcrosses, False,
				 XtNfromHoriz, w,
				 NULL);
    upper_widget = w;
    XtAddCallback (w, XtNcallback, _DressingControl_auto_refresh_cbk, New);
    w = XtVaCreateManagedWidget ("showLabelsToggle", ktoggleWidgetClass, form,
				 XtNlabel, "Show Labels",
				 XtNfromVert, upper_widget,
				 XkwNcrosses, False,
				 XtNstate, True,
				 NULL);
    XtAddCallback (w, XtNcallback, _DressingControl_toggle_cbk,
		   (XtPointer) KCD_ATT_SHOW_LABELS);
    w = XtVaCreateManagedWidget ("showLabelsToggle", ktoggleWidgetClass, form,
				 XtNlabel, "Show Scale",
				 XtNfromVert, upper_widget,
				 XtNfromHoriz, w,
				 XkwNcrosses, False,
				 XtNstate, True,
				 NULL);
    upper_widget = w;
    XtAddCallback (w, XtNcallback, _DressingControl_toggle_cbk,
		   (XtPointer) KCD_ATT_SHOW_SCALE);
    w = XtVaCreateManagedWidget ("topTicksToggle", ktoggleWidgetClass, form,
				 XtNlabel, "Top Tick Marks",
				 XtNfromVert, upper_widget,
				 XkwNcrosses, False,
				 XtNstate, True,
				 NULL);
    XtAddCallback (w, XtNcallback, _DressingControl_toggle_cbk,
		   (XtPointer) KCD_ATT_TOP_TICKS);
    w = XtVaCreateManagedWidget ("bottomTicksToggle", ktoggleWidgetClass, form,
				 XtNlabel, "Bottom Tick Marks",
				 XtNfromVert, upper_widget,
				 XtNfromHoriz, w,
				 XkwNcrosses, False,
				 XtNstate, True,
				 NULL);
    upper_widget = w;
    XtAddCallback (w, XtNcallback, _DressingControl_toggle_cbk,
		   (XtPointer) KCD_ATT_BOTTOM_TICKS);
    w = XtVaCreateManagedWidget ("leftTicksToggle", ktoggleWidgetClass, form,
				 XtNlabel, "Left Tick Marks",
				 XtNfromVert, upper_widget,
				 XkwNcrosses, False,
				 XtNstate, True,
				 NULL);
    XtAddCallback (w, XtNcallback, _DressingControl_toggle_cbk,
		   (XtPointer) KCD_ATT_LEFT_TICKS);
    w = XtVaCreateManagedWidget ("rightTicksToggle", ktoggleWidgetClass, form,
				 XtNlabel, "Right Tick Marks",
				 XtNfromVert, upper_widget,
				 XtNfromHoriz, w,
				 XkwNcrosses, False,
				 XtNstate, True,
				 NULL);
    upper_widget = w;
    XtAddCallback (w, XtNcallback, _DressingControl_toggle_cbk,
		   (XtPointer) KCD_ATT_RIGHT_TICKS);
    w = XtVaCreateManagedWidget ("internalTicksLabelsToggle",
				 ktoggleWidgetClass, form,
				 XtNlabel, "Internal Ticks",
				 XtNfromVert, upper_widget,
				 XkwNcrosses, False,
				 XtNstate, True,
				 NULL);
    XtAddCallback (w, XtNcallback, _DressingControl_toggle_cbk,
		   (XtPointer) KCD_ATT_INTERNAL_TICKS);
    w = XtVaCreateManagedWidget ("gridToggle", ktoggleWidgetClass, form,
				 XtNlabel, "Grid Lines",
				 XtNfromVert, upper_widget,
				 XtNfromHoriz, w,
				 XkwNcrosses, False,
				 XtNstate, FALSE,
				 NULL);
    upper_widget = w;
    XtAddCallback (w, XtNcallback, _DressingControl_toggle_cbk,
		   (XtPointer) KCD_ATT_GRID);
    w = XtVaCreateManagedWidget ("button", commandWidgetClass, form,
				 XtNlabel, "Screen Colours",
				 XtNfromVert, upper_widget,
				 NULL);
    XtAddCallback (w, XtNcallback, _DressingControl_screen_colours_cbk, New);
    w = XtVaCreateManagedWidget ("button", commandWidgetClass, form,
				 XtNlabel, "Paper Colours",
				 XtNfromVert, upper_widget,
				 XtNfromHoriz, w,
				 NULL);
    upper_widget = w;
    XtAddCallback (w, XtNcallback, _DressingControl_paper_colours_cbk, New);
    w = XtVaCreateManagedWidget ("cmapToggle", ktoggleWidgetClass, form,
				 XtNlabel, "Change Colourmap",
				 XtNfromVert, upper_widget,
				 XkwNcrosses, False,
				 XtNstate, TRUE,
				 NULL);
    new->dressingControl.change_cmap_tgl = w;
    if ( XtIsImageDisplay ( XtParent ( XtParent (New) ) ) )
    {   /*  Disgusting hack  */
	w = XtVaCreateManagedWidget ("colourbarToggle", ktoggleWidgetClass,
				     form,
				     XtNlabel, "Show Colourbar",
				     XtNfromVert, upper_widget,
				     XtNfromHoriz, w,
				     XkwNcrosses, False,
				     NULL);
	XtAddCallback (w, XtNcallback, _DressingControl_colourbar_toggle_cbk,
		       New);
    }
    upper_widget = w;
    w = XtVaCreateManagedWidget ("slider", simpleSliderWidgetClass, form,
				 XtNfromVert, upper_widget,
				 XtNlabel, "Font scale",
				 XtNwidth, 220,
				 XkwNminimum, 5,
				 XkwNmaximum, 50,
				 XkwNmodifier, 1,
				 XtNvalue, 10,
				 XkwNscaledValuePtr,
				 &new->dressingControl.font_scale,
				 XkwNscaledFormat, "%.1f",
				 NULL);
    upper_widget = w;
    XtAddCallback (w, XkwNvalueChangeCallback, _DressingControl_slider_cbk,
		   (XtPointer) KCD_ATT_FONT_SCALE);
    XkwSimpleSliderSetScale (w, 1e-1, 0.0, FALSE, TRUE);
    w = XtVaCreateManagedWidget ("divider", widgetClass, form,
				 XtNfromVert, upper_widget,
				 XtNleft, XtChainLeft,
				 XtNright, XtChainRight,
				 XtNwidth, 330,
				 XtNheight, 1,
				 XtVaTypedArg, XtNbackground,
				 XtRString, "red", 4,
				 NULL);
    upper_widget = w;
    w = XtVaCreateManagedWidget ("label", labelWidgetClass, form,
				 XtNlabel, "            Colours:",
				 XtNborderWidth, 0,
				 XtNfromVert, upper_widget,
				 NULL);
    upper_widget = w;
    w = XtVaCreateManagedWidget ("label", labelWidgetClass, form,
				 XtNlabel, "Background:",
				 XtNborderWidth, 0,
				 XtNfromVert, upper_widget,
				 NULL);
    w = XtVaCreateManagedWidget ("backgroundColour",
				 simpleTextEntryWidgetClass, form,
				 XtNfromVert, upper_widget,
				 XtNfromHoriz, w,
				 XtNstring, "black",
				 XtNwidth, 220,
				 XtNeditType, XawtextEdit,
				 NULL);
    upper_widget = w;
    new->dressingControl.background_colour_txt = w;
    XtAddCallback (w, XkwNnewLineCallback, _DressingControl_apply_cbk, New);
    w = XtVaCreateManagedWidget ("label", labelWidgetClass, form,
				 XtNlabel, "Label:     ",
				 XtNborderWidth, 0,
				 XtNfromVert, upper_widget,
				 NULL);
    w = XtVaCreateManagedWidget ("labelColour", simpleTextEntryWidgetClass,
				 form,
				 XtNfromVert, upper_widget,
				 XtNfromHoriz, w,
				 XtNstring, "green",
				 XtNwidth, 220,
				 XtNeditType, XawtextEdit,
				 NULL);
    upper_widget = w;
    new->dressingControl.label_colour_txt = w;
    XtAddCallback (w, XkwNnewLineCallback, _DressingControl_apply_cbk, New);
    w = XtVaCreateManagedWidget ("label", labelWidgetClass, form,
				 XtNlabel, "Grid:      ",
				 XtNborderWidth, 0,
				 XtNfromVert, upper_widget,
				 NULL);
    w = XtVaCreateManagedWidget ("gridColour", simpleTextEntryWidgetClass,
				 form,
				 XtNfromVert, upper_widget,
				 XtNfromHoriz, w,
				 XtNstring, "green",
				 XtNwidth, 220,
				 XtNeditType, XawtextEdit,
				 NULL);
    new->dressingControl.grid_colour_txt = w;
    XtAddCallback (w, XkwNnewLineCallback, _DressingControl_apply_cbk, New);
}   /*  End Function Initialise  */

static void _DressingControl_Destroy (Widget W)
{
    DressingControlWidget w = (DressingControlWidget) W;
    struct canvas_list *curr, *next;

    for (curr = w->dressingControl.first_canvas; curr; curr = next)
    {
	next = curr->next;
	m_free (curr);
    }
}   /*  End Function Destroy  */

static void _DressingControl_apply_cbk (Widget w, XtPointer client_data,
					XtPointer call_data)
/*  This is the apply callback.
*/
{
    KPixCanvas pixcanvas;
    DressingControlWidget top = (DressingControlWidget) client_data;
    unsigned long background_pixel;
    struct canvas_list *curr;
    String background_colour, label_colour, grid_colour;
    static char function_name[] = "DressingControlWidget::apply_cbk";

    XtVaGetValues (top->dressingControl.background_colour_txt,
		   XtNstring, &background_colour,
		   NULL);
    XtVaGetValues (top->dressingControl.label_colour_txt,
		   XtNstring, &label_colour,
		   NULL);
    XtVaGetValues (top->dressingControl.grid_colour_txt,
		   XtNstring, &grid_colour,
		   NULL);
    for (curr = top->dressingControl.first_canvas; curr; curr = curr->next)
    {
	canvas_set_dressing (curr->canvas,
			     KCD_ATT_TITLE_COLOUR, label_colour,
			     KCD_ATT_GRID_COLOUR, grid_colour,
			     KCD_ATT_SCALE_COLOUR, label_colour,
			     KCD_ATT_LABEL_COLOUR, label_colour,
			     KCD_ATT_END);
	pixcanvas = canvas_get_pixcanvas (curr->canvas);
	if ( !kwin_get_colour_harder (pixcanvas, background_colour,
				      &background_pixel) )
	{
	    fprintf (stderr, "%s: cannot allocate colour: \"%s\"\n",
		     function_name, background_colour);
	    fprintf (stderr, "Using pixel value 0 instead\n");
	    background_pixel = 0;
	}
	kwin_set_attributes (pixcanvas,
			     KWIN_ATT_BACKGROUND_PIXEL, background_pixel,
			     KWIN_ATT_END);
	kwin_refresh_if_visible (pixcanvas, TRUE);
    }
}   /*  End Function apply_cbk   */

static void _DressingControl_set_bool (DressingControlWidget top,
				       unsigned int att, flag val)
{
    struct canvas_list *curr;

    for (curr = top->dressingControl.first_canvas; curr; curr = curr->next)
    {
	canvas_set_dressing (curr->canvas,
			     att, val,
			     KCD_ATT_END);
    }
    if (top->dressingControl.auto_refresh)
	_DressingControl_apply_cbk (NULL, top, NULL);
}   /*  End Function set_bool   */

static void _DressingControl_set_double (DressingControlWidget top,
					 unsigned int att, double val)
{
    struct canvas_list *curr;

    for (curr = top->dressingControl.first_canvas; curr; curr = curr->next)
    {
	canvas_set_dressing (curr->canvas,
			     att, val,
			     KCD_ATT_END);
    }
    if (top->dressingControl.auto_refresh)
	_DressingControl_apply_cbk (NULL, top, NULL);
}   /*  End Function set_double   */

static void _DressingControl_auto_refresh_cbk (Widget w, XtPointer client_data,
					       XtPointer call_data)
/*  This is the auto refresh callback.
*/
{
    DressingControlWidget top = (DressingControlWidget) client_data;

    top->dressingControl.auto_refresh = (uaddr) call_data;
}   /*  End Function auto_refresh_cbk   */

static void _DressingControl_toggle_cbk (Widget w, XtPointer client_data,
					 XtPointer call_data)
/*  This is the generic toggle callback.
*/
{
    static char function_name[] = "DressingControl::toggle_cbk";

    while (w)
    {
	if ( XtIsDressingControl (w) ) break;
	w = XtParent (w);
    }
    if (w == NULL)
    {
	fprintf (stderr, "No parent DressingControl widget found!\n");
	a_prog_bug (function_name);
    }
    _DressingControl_set_bool ( (DressingControlWidget) w, (uaddr) client_data,
				(uaddr) call_data );
}   /*  End Function toggle_cbk   */

static void _DressingControl_screen_colours_cbk (Widget w,
						 XtPointer client_data,
						 XtPointer call_data)
/*  This is the screen colours callback.
*/
{
    DressingControlWidget top = (DressingControlWidget) client_data;

    _DressingControl_set_cmap_bool (top, KCMAP_ATT_REVERSE, FALSE);
    XtVaSetValues (top->dressingControl.background_colour_txt,
		   XtNstring, "black",
		   NULL);
    XtVaSetValues (top->dressingControl.label_colour_txt,
		   XtNstring, "green",
		   NULL);
    XtVaSetValues (top->dressingControl.grid_colour_txt,
		   XtNstring, "green",
		   NULL);
    _DressingControl_apply_cbk (NULL, top, NULL);
}   /*  End Function screen_colours_cbk   */

static void _DressingControl_paper_colours_cbk (Widget w,
						XtPointer client_data,
						XtPointer call_data)
/*  This is the paper colours callback.
*/
{
    DressingControlWidget top = (DressingControlWidget) client_data;

    _DressingControl_set_cmap_bool (top, KCMAP_ATT_REVERSE, TRUE);
    XtVaSetValues (top->dressingControl.background_colour_txt,
		   XtNstring, "white",
		   NULL);
    XtVaSetValues (top->dressingControl.label_colour_txt,
		   XtNstring, "black",
		   NULL);
    XtVaSetValues (top->dressingControl.grid_colour_txt,
		   XtNstring, "black",
		   NULL);
    _DressingControl_apply_cbk (NULL, top, NULL);
}   /*  End Function paper_colours_cbk   */

static void _DressingControl_set_cmap_bool (DressingControlWidget top,
					    unsigned int att, flag val)
{
    Kcolourmap cmap;
    Bool bool;
    struct canvas_list *curr;

    XtVaGetValues (top->dressingControl.change_cmap_tgl,
		   XtNstate, &bool,
		   NULL);
    if (!bool) return;
    for (curr = top->dressingControl.first_canvas; curr; curr = curr->next)
    {
	if ( ( cmap = canvas_get_cmap (curr->canvas) ) == NULL ) continue;
	kcmap_set_attributes (cmap,
			      att, val,
			      KCMAP_ATT_END);
    }
}   /*  End Function set_cmap_bool   */

static void _DressingControl_slider_cbk (Widget W, XtPointer client_data,
					 XtPointer call_data)
/*  This is the display labels callback.
*/
{
    DressingControlWidget w;
    static char function_name[] = "DressingControl::slider_cbk";

    while (W)
    {
	if ( XtIsDressingControl (W) ) break;
	W = XtParent (W);
    }
    if (W == NULL)
    {
	fprintf (stderr, "No parent DressingControl widget found!\n");
	a_prog_bug (function_name);
    }
    w = (DressingControlWidget) W;
    _DressingControl_set_double (w, (uaddr) client_data,
				 w->dressingControl.font_scale);
}   /*  End Function slider_cbk  */

static void _DressingControl_colourbar_toggle_cbk (Widget w,
						   XtPointer client_data,
						   XtPointer call_data)
/*  This is the colourbar toggle callback.
*/
{
    flag bool = (uaddr) call_data;
    DressingControlWidget top = (DressingControlWidget) client_data;
    struct canvas_list *curr;
    /*static char function_name[] = "DressingControl::colourbar_toggle_cbk";*/

    for (curr = top->dressingControl.first_canvas; curr; curr = curr->next)
	viewimg_set_canvas_attributes (curr->canvas,
				       VIEWIMG_ATT_SHOW_COLOURBAR, bool,
				       VIEWIMG_ATT_END);
    if (top->dressingControl.auto_refresh)
	_DressingControl_apply_cbk (NULL, top, NULL);
}   /*  End Function colourbar_toggle_cbk   */


/*  Public functions follow  */

void XkwDressingControlAddCanvas (Widget W, KWorldCanvas canvas)
/*  [SUMMARY] Add a KWorldCanvas to a DressingControl widget canvas list.
    <W> The widget.
    <canvas> The canvas to add to the list. The widget will then manipulate
    this canvas. If this is NULL the routine does nothing.
    [RETURNS] Nothing.
*/
{
    DressingControlWidget w;
    Widget tgl;
    Bool state;
    struct canvas_list *new;
    static char function_name[] = "XkwDressingControlAddCanvas";

    if ( !XtIsDressingControl (W) )
    {
	fprintf (stderr, "Invalid widget passed\n");
	a_prog_bug (function_name);
    }
    w = (DressingControlWidget) W;
    if (canvas == NULL) return;
    if ( ( new = m_alloc (sizeof *new) ) == NULL )
	m_abort (function_name, "canvas_list structure");
    new->canvas = canvas;
    new->next = w->dressingControl.first_canvas;
    w->dressingControl.first_canvas = new;
    tgl = XtNameToWidget (W, "form.displayAxisLabelsToggle");
    XtVaGetValues (tgl, XtNstate, &state, NULL);
    canvas_set_dressing (canvas,
			 KCD_ATT_DISPLAY, state,
			 KCD_ATT_END);
}   /*  End Function XkwDressingControlAddCanvas  */
