/*LINTLIBRARY*/
/*  ImageDisplay.c

    This code provides an image display widget for Xt.

    Copyright (C) 1994-2004  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*

    This file contains all routines needed for an image display widget for Xt.


    Written by      Richard Gooch   18-DEC-1994

    Updated by      Richard Gooch   25-DEC-1994

    Updated by      Richard Gooch   27-DEC-1994: Trapped NULL visibleCanvas
  resource in SetValues.

    Updated by      Richard Gooch   3-JAN-1995: Removed *ColourCanvas resources
  Use XtNameToWidget instead.

    Updated by      Richard Gooch   6-JAN-1995: Took account of change in
  DataclipWidget from passing flag by value to by reference for
  intensityScaleCallback.

    Updated by      Richard Gooch   12-JAN-1995: Created cmapSize resource.

    Updated by      Richard Gooch   20-JAN-1995: No longer set visible canvas
  upon canvas refresh since not all refreshes are due to CanvasWidget mapping.

    Updated by      Richard Gooch   30-JAN-1995: Added stereo support.

    Updated by      Richard Gooch   17-APR-1995: Fixed mix of XtRBool and
  sizeof Boolean in resource fields. Use Bool because it is of int size.

    Updated by      Richard Gooch   10-JUL-1995: Added showAnimateButton
  resource.

    Updated by      Richard Gooch   13-JUL-1995: Made use of
  VIEWIMG_ATT_ALLOW_TRUNCATION attribute.

    Updated by      Richard Gooch   6-AUG-1995: Don't assume PseudoColour
  visual is always available.

    Updated by      Richard Gooch   29-AUG-1995: Set saturation pixels to auto.

    Updated by      Richard Gooch   1-SEP-1995: Moved to <kcmap_va_create>.

    Updated by      Richard Gooch   4-SEP-1995: Ensure one of the mono canvases
  is mapped when managed.

    Updated by      Richard Gooch   6-SEP-1995: Added "Export" menu. Supported
  Sun Rasterfile format.

    Updated by      Richard Gooch   7-SEP-1995: Update imageName member with
  dynamically allocated private copy of image name.

    Updated by      Richard Gooch   8-SEP-1995: Added experimental fullscreen
  resource.

    Updated by      Richard Gooch   23-SEP-1995: Took account of stereo modes.

    Uupdated by     Richard Gooch   11-OCT-1995: Forced use of same Kcolourmap
  for mono and stereo pseudocolour world canvases.

    Updated by      Richard Gooch   13-OCT-1995: Prevented unmap/map when
  switching active canvas from left<->right within same visual type.

    Updated by      Richard Gooch   28-DEC-1995: Added verbose resource.

    Updated by      Richard Gooch   21-FEB-1996: Created
  <XkwImageDisplayRefresh> routine.

    Updated by      Richard Gooch   26-FEB-1996: Commented out registering of
  crosshair event consumers for stereo canvases, since the events are not
  properly dealt with yet.

    Updated by      Richard Gooch   17-MAR-1996: Made use of <xtmisc_popup_cbk>

    Updated by      Richard Gooch   21-APR-1996: Made use of
  XkwNsimpleColourbar resource for Cmapwinpopup widget.

    Updated by      Richard Gooch   29-APR-1996: Added control for DirectColour
  colourmaps.

    Updated by      Richard Gooch   3-MAY-1996: Switched to KtoggleWidget.

    Updated by      Richard Gooch   11-MAY-1996: Added raise button and keyed
  right mouse button press to lower fullscreen canvas.

    Updated by      Richard Gooch   12-MAY-1996: Made use of ZoomPolicy widget.

    Updated by      Richard Gooch   16-MAY-1996: Removed <iscale_cbk>.

    Updated by      Richard Gooch   19-MAY-1996: Changed from using window
  scale structure to using <canvas_set_attributes>.

    Updated by      Richard Gooch   26-MAY-1996: Cleaned code to keep
  gcc -Wall -pedantic-errors happy.

    Updated by      Richard Gooch   4-JUN-1996: Switched to left-right
  bottom-top co-ordinate specification instead of min-max x and y. Made use of
  colourmap for DirectColour visual (if needed) when exporting Sun Rasterfiles.
  Added export of PPM files.

    Updated by      Richard Gooch   7-JUN-1996: Register image name with
  PostScript widget when new name is given, no longer register name inside
  <postscript_cbk>.

    Updated by      Richard Gooch   9-JUN-1996: Added magnifier canvases.

    Updated by      Richard Gooch   14-JUN-1996: Added XkwNnumTrackLabels
  resource.

    Updated by      Richard Gooch   28-JUN-1996: Made use of XkwNautoIncrement
  resource for PostScript widget.

    Updated by      Richard Gooch   21-JUL-1996: Removed unnecessary world
  canvas refresh function.

    Updated by      Richard Gooch   6-AUG-1996: Added export of entire dataset
  in FITS format.

    Updated by      Richard Gooch   7-AUG-1996: Added export of entire dataset
  in Miriad Image format.

    Updated by      Richard Gooch   8-AUG-1996: Added export of entire dataset
  in GIPSY format.

    Updated by      Richard Gooch   15-AUG-1996: Made use of improved
  dialogpopup widget when saving foreign data.

    Updated by      Richard Gooch   15-SEP-1996: Made use of new <xv_*>
  package.

    Updated by      Richard Gooch   28-SEP-1996: Added export of entire dataset
  in Karma format.

    Updated by      Richard Gooch   1-OCT-1996: Move export functionality into
  ExportMenu widget.

    Updated by      Richard Gooch   21-OCT-1996: Update magnifierPseudoCanvas
  in <region_cbk>.

    Updated by      Richard Gooch   22-OCT-1996: Added statistics reporting.

    Updated by      Richard Gooch   6-NOV-1996: Copied XkwNforceNewCmap
  resource for magnifier canvases to be the same as their corresponding main
  canvases.

    Updated by      Richard Gooch   10-NOV-1996: Made use of new attributes for
  ViewableImage objects to override the canvas intensity range.

    Updated by      Richard Gooch   30-NOV-1996: Added header display.

    Updated by      Richard Gooch   30-DEC-1996: Created cmapMaster resource.

    Updated by      Richard Gooch   20-JAN-1997: Created integerXZoom and
  integerYZoom resources.

    Updated by      Richard Gooch   24-JAN-1997: Created view2Datasets
  resource.

    Updated by      Richard Gooch   29-JAN-1997: Created numDatasets resource.

    Updated by      Richard Gooch   3-FEB-1997: Removed autoIntensityScale,
  integerXZoom and integerYZoom resources.

    Updated by      Richard Gooch   10-FEB-1997: Made use of new
  VIEWIMG_ATT_AUTO_WCS attribute.

    Updated by      Richard Gooch   15-FEB-1997: Made use of new
  DressingControl widget.

    Updated by      Richard Gooch   17-FEB-1997: Enable tick marks and labels
  (dressing itself not enabled).

    Updated by      Richard Gooch   18-FEB-1997: Tolerate reduced colourmap
  size for DirectColour canvas.

    Updated by      Richard Gooch   19-FEB-1997: Set blank pixel to auto.

    Updated by      Richard Gooch   15-MAR-1997: Moved Zoom Policy to Zoom
  menu.

    Updated by      Richard Gooch   16-MAR-1997: Fixed bug with setting canvas
  array introduced in previous change. Created allRealisedCallback resource.

    Updated by      Richard Gooch   26-MAR-1997: Moved View2Datasets widget
  creation after Zoom Menu creation so that View2Datasets widget can manipulate
  sensitivities of Zoom Menu selections.

    Updated by      Richard Gooch   20-APR-1997: Made use of XkwNloadFiles
  resource for Filewinpopup widget.

    Updated by      Richard Gooch   18-MAY-1997: Create software colourmap to
  emulate PseudoColour canvas if only TrueColour is available.

    Updated by      Richard Gooch   4-JUN-1997: Enable statistics reporting for
  any visual.

    Updated by      Richard Gooch   5-JUN-1997: Added zoom in and zoom out
  entries in zoom menu.

    Updated by      Richard Gooch   18-JUN-1997: Changed to use new function
  <kcmap_create_scmap>.

    Updated by      Richard Gooch   2-JUL-1997: Pass Kcolourmap for
  DirectColour canvas to world canvas create.

    Updated by      Richard Gooch   18-JUL-1997: Switched to OverlayMenu rather
  than directly manipulating the DressingControl widget.

    Updated by      Richard Gooch   12-AUG-1997: Initialise <contour> package.

    Updated by      Richard Gooch   5-DEC-1997: Created fastMagnifier resource.

    Updated by      Richard Gooch   6-DEC-1997: Moved fastMagnifier resource to
  SlaveImageDisplayPopup widget.

    Updated by      Richard Gooch   7-DEC-1997: Colourmap buttons and intensity
  scaling menu items into a new intensity menu. Moved magnifier and panner into
  zoom menu.

    Updated by      Richard Gooch   8-DEC-1997: Took account of creation of new
  IntensityPolicy widget.

    Updated by      Richard Gooch   12-DEC-1997: Created multiCanvas resource.

    Updated by      Richard Gooch   18-DEC-1997: Created fastPanner resource.

    Updated by      Richard Gooch   13-FEB-1998: No longer set transient
  resource for View2Datasets. New visibility check in <XkwImageDisplayRefresh>.

    Updated by      Richard Gooch   22-FEB-1998: Switched to <Xkw*AddCanvas>
  functions.

    Updated by      Richard Gooch   25-FEB-1998: Fixed bug in <intensity_cbk>
  when selecting DirectColour colourmap. Since code was getting too complicated
  changed to string-based scheme.

    Updated by      Richard Gooch   22-JUN-1998: Fixed bug in function
  <get_pc_colourmap> where depth and visual was not provided.

    Updated by      Richard Gooch   26-AUG-1998: Added call to function
  <kcmap_prepare_for_mastery> in function <get_pc_colourmap>.

    Updated by      Richard Gooch   31-AUG-1998: Created canvasList resource.

    Updated by      Richard Gooch   5-SEP-1998: Fixed bug in <intensity_cbk>:
  dataset names were not scanned.

    Updated by      Richard Gooch   8-SEP-1998: Made use of new function
  <xc_get_location2>.

    Updated by      Richard Gooch   9-JUN-1999: Added image edit control.

    Updated by      Richard Gooch   15-AUG-1999: Took account of change to
  <XkwIntensityPolicyAddCanvas> function.

    Updated by      Richard Gooch   26-AUG-1999: Set default filename tester
  for Filepopup widget.

    Updated by      Richard Gooch   9-SEP-1999: Added updateTrackLabels
  resource.

    Updated by      Richard Gooch   12-SEP-1999: Added <intensity_range_cbk>.

    Updated by      Richard Gooch   29-SEP-1999: Took account of interface
  change to <viewimg_zoom>.

    Updated by      Richard Gooch   15-OCT-1999: Took account of resource name
  change for MultiCanvas widget.

    Updated by      Richard Gooch   12-MAY-2000: Created <set_irange> and
  update all canvases in <region_cbk> (TrueColour and DirectColour canvases in
  magnifier and panner were not updated).

    Updated by      Richard Gooch   19-MAY-2000: Cosmetic changes.

    Updated by      Richard Gooch   9-JUN-2000: Removed unused variable.

    Updated by      Richard Gooch   14-JUN-2000: Added values reporting.

    Updated by      Richard Gooch   26-JUN-2000: Do not update track labels on
  pointer enter events.

    Updated by      Richard Gooch   11-JUL-2000: Added #include <karma_kcmap.h>

    Updated by      Richard Gooch   13-JUL-2000: Created Destroy method.

    Updated by      Richard Gooch   20-MAR-2001: Fixed bug in Destroy method:
  should not destroy canvases because this will be done by automatic Canvas
  widget destruction. Made use of new <kcmap_register_destroy_func> and no
  longer destroy display handle in Destroy method.

    Updated by      Richard Gooch   21-MAR-2001: Deallocate array of intensity
  choice strings after menu is created.

    Updated by      Richard Gooch   28-MAY-2001: Moved View2Datasets widget
  creation after Export Menu creation so that View2Datasets widget can set
  save PPM movie function.

    Updated by      Richard Gooch   15-NOV-2001: Added calls to new function
  <XkwDataclipSetRegions> in <intensity_range_cbk>.

    Updated by      Richard Gooch   27-NOV-2001: Created new function
  <setup_wc> and registered <viewimg_statistics_geom_func> and
  <viewimg_values_geom_func> with editor context. Created resource
  newEditorContextCallback.

    Updated by      Richard Gooch   22-JAN-2002: Increase desired number of
  colours in <get_dc_colourmap> from 255 to 256.

    Updated by      Richard Gooch   24-NOV-2002: Created function
  <XkwImageDisplayGetWindow>.

    Updated by      Richard Gooch   5-DEC-2002: Renamed resource for view
  dataset control. Created DataBrowser widget if needed.

    Updated by      Richard Gooch   9-DEC-2002: Registered binding list.

    Updated by      Richard Gooch   22-DEC-2002: Stupid compiler warning fix.

    Updated by      Richard Gooch   26-DEC-2002: Added "New" and "Destroy"
  buttons.

    Updated by      Richard Gooch   24-DEC-2002: Made use of new support for
  families of file browsers.

    Updated by      Richard Gooch   28-JAN-2003: No longer use the resource
  mappedWhenManaged and instead set the visible canvas in <canvas_realise_cbk>.

    Updated by      Richard Gooch   12-FEB-2003: Search for appropriate parent
  widgets in <new_cbk>.

    Updated by      Richard Gooch   14-FEB-2003: Added XkwNallowedMask resource
  and added "Independent Colourmap (24 bit)" option to "New" menu.

    Updated by      Richard Gooch   19-FEB-2003: Added Renzogram colourmap.

    Updated by      Richard Gooch   13-JUN-2003: Made use of new experimental
  function <dispdata_get_num_windows>.

    Updated by      Richard Gooch   14-JUN-2003: Split realised callback.

    Updated by      Richard Gooch   13-AUG-2003: Removed Renzogram Colourmap
  option for non-<kvis> to prevent crashes when user selects what they should
  not.

    Updated by      Richard Gooch   3-OCT-2003: Took account of interface
  change for Dataclip widget.

    Updated by      Richard Gooch   16-OCT-2003: Create function
  <XkwImageDisplayCheckObsolete>.

    Updated by      Richard Gooch   29-MAR-2004: Removed showAnimateButton
  resource (obsolete).

    Updated by      Richard Gooch   1-MAY-2004: Moved realise callbacks to
  after slave Renzogram colourmap is created.

    Updated by      Richard Gooch   6-MAY-2004: Published XkwNrenzoPseudoCmap
  and XkwNrenzoTrueCmap resources.

    Updated by      Richard Gooch   18-JUN-2004: Took account of change to
  <xtmisc_ext_call> API.

    Updated by      Richard Gooch   8-JUL-2004: Made use of
  KCMAP_ATT_STEAL_COLOURS_FROM for PC Renzogram colourmap. Added
  XkwNpseudoXColourMap resource.

    Updated by      Richard Gooch   11-JUL-2004: Created function
  <XkwImageDisplayCreateWindow>.

    Last updated by Richard Gooch   13-JUL-2004: Changed numTrackLabels type
  from Cardinal to int, since former doesn't work with default resources. No
  longer use <kcmap_register_destroy_func> and instead use
  <<REGISTER_FREE_FUNC>> so as to fix bug where display handle was destroyed
  before Renzogram colourmap was destroyed (for window with independent cmap).
  


*/
#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>
#include <math.h>
#include <unistd.h>
#ifndef X11
#  define X11
#endif
#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>
#include <X11/Xaw/XawInit.h>
#include <X11/Xaw/Label.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Box.h>
#include <karma.h>
#define NEW_WIN_SCALE
#include <k_event_codes.h>
#include <karma_dispdata.h>
#include <karma_viewimg.h>
#include <karma_contour.h>
#include <karma_xtmisc.h>
#include <karma_kcmap.h>
#include <karma_dsxfr.h>
#include <karma_kwin.h>
#include <karma_conn.h>
#include <karma_xv.h>
#include <karma_xc.h>
#include <karma_st.h>
#include <karma_ds.h>
#include <karma_m.h>
#include <karma_a.h>
#include <karma_r.h>
#include <karma_c.h>
#include <Xkw/ImageDisplayP.h>
#include <Xkw/MultiCanvas.h>
#include <Xkw/Filepopup.h>
#include <Xkw/Filewin.h>
#include <Xkw/Cmapwinpopup.h>
#include <Xkw/Dialogpopup.h>
#include <Xkw/Postscript.h>
#include <Xkw/Dataclip.h>
#include <Xkw/ChoiceMenu.h>
#include <Xkw/IntensityPolicy.h>
#include <Xkw/ZoomPolicy.h>
#include <Xkw/OverlayMenu.h>
#include <Xkw/ExportMenu.h>
#include <Xkw/View2Datasets.h>
#include <Xkw/ViewDatasets.h>
#include <Xkw/MagnifierPopup.h>
#include <Xkw/PannerPopup.h>
#include <Xkw/ImageEditorControl.h>
#include <Xkw/DataBrowser.h>


#define DEFAULT_COLOURMAP_NAME "Greyscale1"
#define RENZOGRAM_FRACTION 1/5

#define REGISTER_FREE_FUNC(w,func,object) \
    c_register_callback (&(w)->imageDisplay.destroy_list, \
			 ( flag (*) () ) (func), (object), \
			 NULL, FALSE, NULL, FALSE, FALSE);


STATIC_FUNCTION (void _ImageDisplay_Initialise, (Widget request, Widget new) );
STATIC_FUNCTION (void _ImageDisplay_Destroy, (Widget W) );
STATIC_FUNCTION (Boolean _ImageDisplay_SetValues, () );
STATIC_FUNCTION (Widget _ImageDisplay_wc_to_cnv,
		 (ImageDisplayWidget top, KWorldCanvas wc) );
STATIC_FUNCTION (void quit_cbk, (Widget w, XtPointer client_data,
				 XtPointer call_data) );
STATIC_FUNCTION (void _ImageDisplay_canvas_realise_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void setup_wc, (ImageDisplayWidget top,KWorldCanvas canvas) );
STATIC_FUNCTION (void handle_wc,
		 (ImageDisplayWidget top, KWorldCanvas canvas,
		  KWorldCanvas **ptr, flag slave) );
STATIC_FUNCTION (void zoom_cbk, (Widget w, XtPointer client_data,
				 XtPointer call_data) );
STATIC_FUNCTION (void region_cbk, (Widget w, XtPointer client_data,
				   XtPointer call_data) );
STATIC_FUNCTION (flag _ImageDisplay_set_irange,
		 (KWorldCanvas vc, KWorldCanvas *canvases,
		  double minimum, double maximum) );
STATIC_FUNCTION (void _ImageDisplay_intensity_range_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (flag _ImageDisplay_canvas_event_handler,
		 (KWorldCanvas canvas, double x, double y,
		  unsigned int event_code, void *e_info,
		  void **f_info, double x_lin, double y_lin) );
STATIC_FUNCTION (flag _ImageDisplay_vimage_event_func,
		 (ViewableImage vimage, double x, double y, void *value,
		  unsigned int event_code, void *e_info, void **f_info,
		  double x_lin, double y_lin, unsigned int value_type,
		  double x_im, double y_im, int x_pix, int y_pix) );
STATIC_FUNCTION (void zoom, (ImageDisplayWidget w, double zoom) );
STATIC_FUNCTION (void unzoom, (ImageDisplayWidget w) );
STATIC_FUNCTION (void _ImageDisplay_intensity_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void raise_cbk, (Widget w, XtPointer client_data,
				  XtPointer call_data) );
STATIC_FUNCTION (Kcolourmap _ImageDisplay_get_pc_colourmap,
		 (ImageDisplayWidget w, Visual *visual,
		  KPixCanvas canvas, Colormap xcmap, int depth) );
STATIC_FUNCTION (Kcolourmap _ImageDisplay_create_pc_colourmap,
		 (ImageDisplayWidget w, Visual *visual, KPixCanvas canvas,
		  Colormap xcmap, int depth, unsigned int num_ccells,
		  Widget *colourmapwinpopup, Kcolourmap donor) );
STATIC_FUNCTION (Kcolourmap get_dc_colourmap,
		 (ImageDisplayWidget w, Visual *visual, Colormap xcmap,
		  int depth, KPixCanvas pixcanvas) );
STATIC_FUNCTION (void handle_canvas,
		 (ImageDisplayWidget w, Widget canvas) );
STATIC_FUNCTION (void _ImageDisplay_show_dispdata_canvas,
		 (void *info, KWorldCanvas canvas) );
STATIC_FUNCTION (void _ImageDisplay_new_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void _ImageDisplay_kcmap_modify_func,
		 (Kcolourmap source, void **info) );
STATIC_FUNCTION (void _ImageDisplay_renumber_func, (ImageDisplayWidget w) );


static char *def_image_name = "fred";

#define XkwNpseudoCmap        "pseudoCmap"
#define XkwNallowedMask       "allowedMask"
#define XkwCColourmap         "Colourmap"
#define XkwCMask              "Mask"

#define offset(field) XtOffsetOf(ImageDisplayRec, imageDisplay.field)

static XtResource resources[] =
{
    {XkwNpseudoColourCanvas, XkwCWorldCanvas, XtRPointer, sizeof (XtPointer),
     offset (pseudoCanvas), XtRImmediate, NULL},
    {XkwNdirectColourCanvas, XkwCWorldCanvas, XtRPointer, sizeof (XtPointer),
     offset (directCanvas), XtRImmediate, NULL},
    {XkwNtrueColourCanvas, XkwCWorldCanvas, XtRPointer, sizeof (XtPointer),
     offset (trueCanvas), XtRImmediate, NULL},
    {XkwNpseudoColourLeftCanvas, XkwCWorldCanvas, XtRPointer,
     sizeof (XtPointer), offset (pseudoCanvasLeft), XtRImmediate, NULL},
    {XkwNpseudoColourRightCanvas, XkwCWorldCanvas, XtRPointer,
     sizeof (XtPointer), offset (pseudoCanvasRight), XtRImmediate, NULL},
    {XkwNdirectColourLeftCanvas, XkwCWorldCanvas, XtRPointer,
     sizeof (XtPointer), offset (directCanvasLeft), XtRImmediate, NULL},
    {XkwNdirectColourRightCanvas, XkwCWorldCanvas, XtRPointer,
     sizeof (XtPointer), offset (directCanvasRight), XtRImmediate, NULL},
    {XkwNtrueColourLeftCanvas, XkwCWorldCanvas, XtRPointer,
     sizeof (XtPointer), offset (trueCanvasLeft), XtRImmediate, NULL},
    {XkwNtrueColourRightCanvas, XkwCWorldCanvas, XtRPointer,
     sizeof (XtPointer), offset (trueCanvasRight), XtRImmediate, NULL},
    {XkwNvisibleCanvas, XkwCWorldCanvas, XtRPointer, sizeof (XtPointer),
     offset (visibleCanvas), XtRImmediate, NULL},
    {XkwNcanvasList, XkwCWorldCanvas, XtRPointer, sizeof (XtPointer),
     offset (canvasList), XtRImmediate, NULL},
    {XkwNmagnifier, XtCWidget, XtRWidget, sizeof (Widget),
     offset (magnifier), XtRImmediate, NULL},
    {XkwNimageName, XkwCImageName, XtRString, sizeof (String),
     offset (imageName), XtRImmediate, NULL},
    {XkwNcreateFilepopup, XkwCCreateFilepopup, XtRBool, sizeof (Bool),
     offset (createFilepopup), XtRImmediate, (XtPointer) True},
    {XkwNenableAnimation, XkwCEnableAnimation, XtRBool, sizeof (Bool),
     offset (enableAnimation), XtRImmediate, (XtPointer) False},
    {XkwNshowQuitButton, XkwCShowQuitButton, XtRBool, sizeof (Bool),
     offset (showQuitButton), XtRImmediate, (XtPointer) True},
    {XkwNcmapSize, XkwCCmapSize, XtRInt, sizeof (int),
     offset (cmapSize), XtRImmediate, (XtPointer) 200},
    {XkwNcmapMaster, XkwCCmapMaster, XtRString, sizeof (String),
     offset (cmapMaster), XtRImmediate, NULL},
    {XkwNfullscreen, XkwCFullscreen, XtRBool, sizeof (Bool),
     offset (fullscreen), XtRImmediate, (XtPointer) False},
    {XkwNnumTrackLabels, XkwCNumTrackLabels, XtRInt, sizeof (int),
     offset (numTrackLabels), XtRImmediate, (XtPointer) 0},
    {XkwNviewDatasets, XkwCViewDatasets, XtRBool, sizeof (Bool),
     offset (viewDatasets), XtRImmediate, False},
    {XkwNnumDatasets, XkwCNumDatasets, XtRInt, sizeof (int),
     offset (numDatasets), XtRImmediate, (XtPointer) 1},
    {XkwNmainRealisedCallback, XtCCallback, XtRCallback, sizeof (caddr_t),
     offset (mainRealisedCallback), XtRCallback, (caddr_t) NULL},
    {XkwNafterMainRealisedCallback, XtCCallback, XtRCallback, sizeof (caddr_t),
     offset (afterMainRealisedCallback), XtRCallback, (caddr_t) NULL},
    {XkwNdatasetNames, XkwCDatasetNames, XtRPointer, sizeof (XtPointer),
     offset (datasetNames), XtRImmediate, NULL},
    {XkwNnewVisibleCanvasCallback, XtCCallback, XtRCallback, sizeof (caddr_t),
     offset (newVisibleCanvasCallback), XtRCallback, (caddr_t) NULL},
    {XkwNpanner, XtCWidget, XtRWidget, sizeof (Widget),
     offset (panner), XtRImmediate, NULL},
    {XkwNfastPanner, XkwCFastPanner, XtRBool, sizeof (Bool),
     offset (fastPanner), XtRImmediate, (XtPointer) True},
    {XkwNupdateTrackLabels, XkwCUpdateTrackLabels, XtRBool, sizeof (Bool),
     offset (updateTrackLabels), XtRImmediate, False},
    {XkwNnewEditorContextCallback, XtCCallback, XtRCallback, sizeof (caddr_t),
     offset (newEditorContextCallback), XtRCallback, (caddr_t) NULL},
    {XkwNmultiCanvas, XtCWidget, XtRWidget, sizeof (Widget),
     offset (multiCanvas), XtRImmediate, NULL},
    {XkwNpseudoCmap, XkwCColourmap, XtRPointer, sizeof (XtPointer),
     offset (pseudoCmap), XtRImmediate, NULL},
    {XkwNrenzoPseudoCmap, XkwCColourmap, XtRPointer, sizeof (XtPointer),
     offset (renzoPseudoCmap), XtRImmediate, NULL},
    {XkwNrenzoTrueCmap, XkwCColourmap, XtRPointer, sizeof (XtPointer),
     offset (renzoTrueCmap), XtRImmediate, NULL},
    {XkwNallowedMask, XkwCMask, XtRInt, sizeof (int),
     offset (allowedMask), XtRImmediate, (XtPointer) ~0},
    {XkwNpseudoXColourMap, XtCColormap, XtRColormap, sizeof (Colormap),
     offset (pseudoXColourMap), XtRImmediate, None},
    {XkwNverbose, XkwCVerbose, XtRBool, sizeof (Bool),
     offset (verbose), XtRImmediate, False},
};
#undef offset


#define ZOOMCODE_MAGNIFIER        0
#define ZOOMCODE_PANNER           1
#define ZOOMCODE_ZOOMIN2          2
#define ZOOMCODE_ZOOMIN4          3
#define ZOOMCODE_ZOOMOUT2         4
#define ZOOMCODE_ZOOMOUT4         5
#define ZOOMCODE_UNZOOM           6
#define ZOOMCODE_POLICY           7
#define NUM_ZOOM_CHOICES          8

static char *zoom_choices[NUM_ZOOM_CHOICES] =
{
    "Magnifier", "Panner", "Zoom in 2x", "Zoom in 4x", "Zoom out 2x",
    "Zoom out 4x", "Unzoom", "Zoom Policy",
};


#define CMAP_STRING_PC               "PseudoColour (8 bit)"
#define CMAP_STRING_PC_EMULATED      "PseudoColour (emulated)"
#define CMAP_STRING_DC               "DirectColour (24 bit)"
#define CMAP_STRING_RENZO            "Renzogram Colourmap"
#define INTENSITY_POLICY_STRING      "Intensity Policy"
#define MIN_ITENSITY_MENU_ITEMS      4


/*  Keep these consistent with include/Xkw/ImageDisplay.h  */
#define NUM_NEW_WINDOW_CHOICES       3

static char *new_window_choices[NUM_NEW_WINDOW_CHOICES] =
{
    "Slave Colourmap", "Independent Colourmap",
    "Independent Colourmap (24 bit)",
};


ImageDisplayClassRec imageDisplayClassRec =
{
    {
	/* core fields  */
	(WidgetClass) &formClassRec,    /*  superclass             */
	"ImageDisplay",                 /*  class_name             */
	sizeof (ImageDisplayRec),       /*  widget_size            */
	NULL,                           /*  class_initialise       */
	NULL,                           /*  class_part_initialise  */
	FALSE,                          /*  class_inited           */
	(XtInitProc) _ImageDisplay_Initialise, /*  initialise             */
	NULL,                           /*  initialise_hook        */
	XtInheritRealize,               /*  realise                */
	NULL,                           /*  actions                */
	0,                              /*  num_actions            */
	resources,                      /*  resources              */
	XtNumber (resources),           /*  num_resources          */
	NULLQUARK,                      /*  xrm_class              */
	TRUE,                           /*  compress_motion        */
	TRUE,                           /*  compress_exposure      */
	TRUE,                           /*  compress_enterleave    */
	TRUE,                           /*  visible_interest       */
	_ImageDisplay_Destroy,          /*  destroy                */
	XtInheritResize,                /*  resize                 */
	NULL,                           /*  expose                 */
	(XtSetValuesFunc) _ImageDisplay_SetValues, /*  set_values  */
	NULL,                           /*  set_values_hook        */
	XtInheritSetValuesAlmost,       /*  set_values_almost      */
	NULL,                           /*  get_values_hook        */
	NULL,                           /*  accept_focus           */
	XtVersion,                      /*  version                */
	NULL,                           /*  callback_private       */
	NULL,                           /*  tm_table               */
	XtInheritQueryGeometry,         /*  query_geometry         */
	XtInheritDisplayAccelerator,    /*  display_accelerator    */
	NULL                            /*  extension              */
    },
    {     /* CompositeClassPart */
	XtInheritGeometryManager        /* geometry_manager        */,
	XtInheritChangeManaged          /* change_managed          */,
	XtInheritInsertChild            /* insert_child            */,
	XtInheritDeleteChild            /* delete_child            */,
	NULL                            /* extension               */
    },
    { /* Constraint */
	NULL                            /* subresourses       */,
	0                               /* subresource_count  */,
	sizeof (ImageDisplayConstraintsRec) /* constraint_size    */,
	NULL                            /* initialise         */,
	NULL                            /* destroy            */,
	NULL                            /* set_values         */,
	NULL                            /* extension          */ 
    },
    { /* Form */
	XtInheritLayout                 /* layout             */
    },
    {
	/*  imageDisplay fields */
	0                               /*  ignore                 */
    }
};

WidgetClass imageDisplayWidgetClass = (WidgetClass) &imageDisplayClassRec;

static KPixCanvasBinding main_bindings[] =
{
  {"s",             "show statistics"},
  {"v",             "show scaled values"},
  {"V",             "show raw values"},
  {"h",             "show header"},
  {"H",             "show header with history"},
  {"l",             "print track co-ordinates to terminal window"},
  {NULL,           NULL}
};


static void _ImageDisplay_Initialise (Widget Request, Widget New)
{
    flag pc_cmap = FALSE;
    flag dc_cmap = FALSE;
    /*ImageDisplayWidget request = (ImageDisplayWidget) Request;*/
    ImageDisplayWidget new = (ImageDisplayWidget) New;
    int canvas_types, count, vertical_distance, num_intensity_choices;
    int canvas_width = 512;
    int canvas_height = 512;
    int num_new_window_choices = NUM_NEW_WINDOW_CHOICES;
    Bool use_data_browser;
    Widget filewin, files_btn, app_box, m_cnv, w;
    Widget intensity_btn = NULL;
    Widget quit_btn, menu_btn, raise_btn;
    Widget izoomwinpopup;
    Widget animatepopup = NULL;  /*  Initialised to keep compiler happy  */
    Widget viewcontrolpopup = NULL;
    Widget shell = NULL;         /*  Initialised to keep compiler happy  */
    Widget pseudo_cnvs, direct_cnvs, true_cnvs;
    Widget left_widget, top_widget;
    Screen *screen;
    Visual *pseudocolour_visual, *truecolour_visual, *directcolour_visual;
    CONST char **datasetNames;
    char txt[STRING_LENGTH], title[STRING_LENGTH];
    static char function_name[] = "ImageDisplayWidget::Initialise";

    if (new->imageDisplay.imageName == NULL)
	new->imageDisplay.imageName = def_image_name;
    new->imageDisplay.override_shell = NULL;
    new->imageDisplay.visibleCanvas = NULL;
    new->imageDisplay.direct_cmap = NULL;
    new->imageDisplay.cmapwinpopup_psuedo = NULL;
    new->imageDisplay.cmapwinpopup_direct = NULL;
    new->imageDisplay.cmapwinpopup_renzogram = NULL;
    new->imageDisplay.num_unrealised = 0;
    new->imageDisplay.dispdata_window = NULL;
    new->imageDisplay.destroy_list = NULL;
    screen = XtScreen (New);
    datasetNames = (CONST char **) new->imageDisplay.datasetNames;
    if (new->imageDisplay.fullscreen)
    {
	canvas_width = WidthOfScreen (screen) - 20;
	canvas_height = HeightOfScreen (screen) - 20;
    }
    else
    {
	if (HeightOfScreen (screen) < 600)
	{
	    canvas_width = 350;
	    canvas_height = 350;
	}
    }
    /*  Create as many popups as possible first. Note that the colourmap
	control popup(s) cannot be created yet since they require a visual
	structure which cannot be predicted until their respective canvas
	widgets are realised. It is possible to check which visuals are
	available for each visual type and then setup the colourmap popup
	button/menu
	*/
    if (new->imageDisplay.createFilepopup)
    {
	char *domain = NULL;
	char *title;

	if (new->imageDisplay.numDatasets < 1)
	{
	    domain = "ARRAY:FILE%d";
	    title = "Array File Selector %d";
	}
	else title = "Array File Selector";
	filewin = XtVaCreatePopupShell ("filewinPopup", filepopupWidgetClass,
					New,
					XkwNdomainName, domain,
					XtNtitle, title,
					XkwNautoPopdown, True,
					XkwNloadFiles, True,
					XkwNfilenameTester,
					XkwFilewinStandardFileTester_nD,
					NULL);
    }
    else filewin = NULL;
    w = XtVaCreatePopupShell ("intensityPolicyPopup",
			      intensityPolicyWidgetClass, New,
			      NULL);
    new->imageDisplay.intensity_policy_popup = w;
    for (count = 0; count < new->imageDisplay.numDatasets; ++count)
    {
	if (new->imageDisplay.numDatasets < 2)
	{
	    strcpy (txt, "izoomwinpopup");
	    strcpy (title, "Intensity Zoom");
	}
	else
	{
	    sprintf (txt, "izoomwinpopup%d", count);
	    if (datasetNames == NULL)
		sprintf (title, "Intensity Zoom for Dataset %d", count + 1);
	    else sprintf (title, "Intensity Zoom for %s", datasetNames[count]);
	}
	izoomwinpopup = XtVaCreatePopupShell (txt, dataclipWidgetClass, New,
					      XtNtitle, title,
					      XkwNshowBlankControl, TRUE,
					      NULL);
	if (new->imageDisplay.numDatasets == 1)
	{
	    XtAddCallback (izoomwinpopup, XkwNrawRegionCallback, region_cbk,
			   New);
	    XtAddCallback (new->imageDisplay.intensity_policy_popup,
			   XkwNintensityRangeCallback,
			   _ImageDisplay_intensity_range_cbk, New);
	}
    }
    if (new->imageDisplay.enableAnimation)
    {
	animatepopup = XtVaCreatePopupShell ("animatepopup",
					     animateControlWidgetClass, New,
					     XtNtitle, "Animation Control",
					     NULL);
	w = XtNameToWidget (animatepopup, "form.antiFlickerToggle");
	XtAddCallback (w, XtNcallback, XkwImageDisplaySetSmoothCallback, New);

    }
    w = XtVaCreatePopupShell ("zoomPolicyPopup", zoomPolicyWidgetClass, New,
			      NULL);
    new->imageDisplay.zoom_policy_popup = w;
    w = XtVaCreatePopupShell ("editorPopup", imageEditorControlWidgetClass,
			      (Widget) new,
			      XtNtitle, "Image Editor",
			      NULL);
    new->imageDisplay.image_editor_popup = w;
    /*  Now create the widgets for the main window  */
    if (new->imageDisplay.createFilepopup)
    {
	files_btn = XtVaCreateManagedWidget ("filesButton", commandWidgetClass,
					     New,
					     XtNlabel, "Files",
					     NULL);
	if (new->imageDisplay.numDatasets < 1)
	    XtAddCallback (files_btn, XtNcallback, XkwFilepopupPopupCallback,
			   filewin);
	else
	    XtAddCallback (files_btn, XtNcallback, xtmisc_popup_cbk, filewin);
    }
    else files_btn = NULL;
    /*  A bit of buggerising is required to setup the colourmap popup button or
	menu  */
    xv_get_visuals (screen, &pseudocolour_visual, &truecolour_visual,
		    &directcolour_visual);
    if (pseudocolour_visual || truecolour_visual) pc_cmap = TRUE;
    if (directcolour_visual) dc_cmap = TRUE;
    if (!pc_cmap && !dc_cmap && (new->imageDisplay.numDatasets < 1) )
	left_widget = files_btn;
    else
    {
	/*  Construct array of intensity choices including Dataclip widgets  */
	char **intensity_choices, **ptr;

	/*  Construct an array big enough  */
	if ( ( intensity_choices =
	       m_alloc ( (MIN_ITENSITY_MENU_ITEMS +
			  new->imageDisplay.numDatasets) *
			 sizeof *intensity_choices ) ) == NULL )
	    m_abort (function_name, "array of intensity choices");
	num_intensity_choices = 0;
	ptr = intensity_choices;
	if (pc_cmap)
	{
	    if (!new->imageDisplay.pseudoCmap)
	    {
		char *string = CMAP_STRING_PC_EMULATED;

		if ( pseudocolour_visual &&
		     (new->imageDisplay.allowedMask &
		      XkwCanvasTypePseudoColour) ) string = CMAP_STRING_PC;
		*ptr++ = st_dup (string);
		++num_intensity_choices;
		if ( new->imageDisplay.viewDatasets &&
		     (new->imageDisplay.numDatasets < 1) )
		{
		    *ptr++ = st_dup (CMAP_STRING_RENZO);
		    ++num_intensity_choices;
		}
	    }
	    *ptr++ = st_dup (INTENSITY_POLICY_STRING);
	    ++num_intensity_choices;
	}
	if (dc_cmap)
	{
	    *ptr++ = st_dup (CMAP_STRING_DC);
	    ++num_intensity_choices;
	}
	for (count = 0; count < new->imageDisplay.numDatasets; ++count)
	{
	    if (new->imageDisplay.numDatasets < 2)
		strcpy (txt, "IScale for Dataset");
	    else if (datasetNames == NULL)
		sprintf (txt, "IScale for Dataset %d", count + 1);
	    else sprintf (txt, "IScale for %s", datasetNames[count]);
	    if ( ( *ptr++ = st_dup (txt) ) == NULL )
		m_abort (function_name, "intensity choice string");
	    ++num_intensity_choices;
	}
	intensity_btn = XtVaCreateManagedWidget
	    ("intensityMenuButton", choiceMenuWidgetClass, New,
	     XtNlabel, "Intensity",
	     XkwNmenuTitle, "Intensity Menu",
	     XtNfromHoriz, files_btn,
	     XtNmenuName, "intensityMenu",
	     XkwNnumItems, num_intensity_choices,
	     XkwNitemStrings, intensity_choices,
	     NULL);
	XtAddCallback (intensity_btn, XkwNselectCallback,
		       _ImageDisplay_intensity_cbk, New);
	left_widget = intensity_btn;
	/*  Now free the strings  */
	for (count = 0; count < num_intensity_choices; ++count)
	    m_free (intensity_choices[count]);
	m_free (intensity_choices);
    }
    menu_btn = XtVaCreateManagedWidget ("zoomMenuButton",
					choiceMenuWidgetClass, New,
					XtNlabel, "Zoom",
					XkwNmenuTitle, "Zoom Menu",
					XtNfromHoriz, left_widget,
					XtNmenuName, "zoomMenu",
					XkwNnumItems, NUM_ZOOM_CHOICES,
					XkwNitemStrings, zoom_choices,
					NULL);
    left_widget = menu_btn;
    XtAddCallback (menu_btn, XkwNselectCallback, zoom_cbk, New);
    use_data_browser = ( new->imageDisplay.viewDatasets &&
			 (new->imageDisplay.numDatasets == 0) ) ? TRUE : FALSE;
    w = XtVaCreateManagedWidget ("overlayMenu", overlayMenuWidgetClass, New,
				 XtNfromHoriz, left_widget,
				 XkwNuseDataBrowser, use_data_browser,
				 NULL);
    new->imageDisplay.overlay_menu = w;
    left_widget = w;
    menu_btn = XtVaCreateManagedWidget ("exportMenu", exportMenuWidgetClass,
					New,
					XtNfromHoriz, left_widget,
					NULL);
    new->imageDisplay.export_menu = menu_btn;
    left_widget = menu_btn;
    /*  The View2Datasets widget must be created after the export menu  */
    if (new->imageDisplay.viewDatasets)
    {
	if (new->imageDisplay.numDatasets == 2)
	    viewcontrolpopup = XtVaCreatePopupShell ("twinviewpopup",
						     view2DatasetsWidgetClass,
						     New,
						     XtNtitle, "View Control",
						     NULL);
	else
	    viewcontrolpopup = XtVaCreatePopupShell ("viewcontrolpopup",
						     viewDatasetsWidgetClass,
						     New,
						     NULL);
    }
    if (new->imageDisplay.viewDatasets)
    {
	w = XtVaCreateManagedWidget ("button", commandWidgetClass, New,
				     XtNlabel, "View",
				     XtNfromHoriz, left_widget,
				     NULL);
	left_widget = w;
	XtAddCallback (w, XtNcallback, xtmisc_popup_cbk, viewcontrolpopup);
    }
    if (new->imageDisplay.fullscreen)
    {
	/*  Need fullscreen canvas: create the raise button  */
	raise_btn = XtVaCreateManagedWidget ("button", commandWidgetClass, New,
					     XtNlabel, "Raise",
					     XtNfromHoriz, left_widget,
					     NULL);
	left_widget = raise_btn;
	XtAddCallback (raise_btn, XtNcallback, raise_cbk, new);
    }
    w = XtVaCreateManagedWidget ("button", commandWidgetClass, New,
				 XtNlabel, "Edit",
				 XtNfromHoriz, left_widget,
				 NULL);
    XtAddCallback (w, XtNcallback, xtmisc_popup_cbk,
		   new->imageDisplay.image_editor_popup);
    left_widget = w;
    if (use_data_browser)
    {
	if (!truecolour_visual) --num_new_window_choices;
	w = XtVaCreateManagedWidget ("newMenuButton", choiceMenuWidgetClass,
				     New,
				     XtNlabel, "New",
				     XkwNmenuTitle, "New Window Menu",
				     XtNfromHoriz, left_widget,
				     XtNmenuName, "newWindowMenu",
				     XkwNnumItems, num_new_window_choices,
				     XkwNitemStrings, new_window_choices,
				     NULL);
	left_widget = w;
	XtAddCallback (w, XkwNselectCallback, _ImageDisplay_new_cbk, new);
    }
    if (new->imageDisplay.showQuitButton)
    {
	quit_btn = XtVaCreateManagedWidget ("quit", commandWidgetClass, New,
					    XtNlabel, "Quit",
					    XtNfromHoriz, left_widget,
					    NULL);
	XtAddCallback (quit_btn, XtNcallback, quit_cbk, NULL);
	left_widget = quit_btn;
    }
    else if (use_data_browser)
    {
	w = XtVaCreateManagedWidget ("destroyButton", commandWidgetClass, New,
				     XtNlabel, "Destroy",
				     XtNfromHoriz, left_widget,
				     NULL);
	XtAddCallback ( w, XtNcallback, xtmisc_destroy_cbk, XtParent (New) );
	left_widget = w;
    }
    app_box = XtVaCreateManagedWidget ("applicationBox", boxWidgetClass, New,
				       XtNfromVert, menu_btn,
				       XtNwidth, canvas_width,
				       XtNheight, 0,
				       XtNborderWidth, 0,
				       XtNleft, XtChainLeft,
				       XtNright, XtChainRight,
				       XtNorientation, XtorientVertical,
				       XtNhorizDistance, 0,
				       XtNvertDistance, 0,
				       NULL);
    top_widget = app_box;
    vertical_distance = 0;
    for (count = 0; count < new->imageDisplay.numTrackLabels; ++count)
    {
	sprintf (txt, "trackLabel%d", count);
	w = XtVaCreateManagedWidget (txt, labelWidgetClass, New,
				     XtNlabel, "Track Output",
				     XtNwidth, canvas_width,
				     XtNfromVert, top_widget,
				     XtNvertDistance, vertical_distance,
				     NULL);
	top_widget = w;
	vertical_distance = new->form.default_spacing;
    }
    /*  Create the canvases  */
    canvas_types = XkwCanvasTypePseudoColour | XkwCanvasTypeDirectColour;
    canvas_types |= XkwCanvasTypeTrueColour | XkwCanvasTypeStereo;
    if (new->imageDisplay.fullscreen)
    {
	/*  Create the override shell  */
	shell = XtVaCreatePopupShell ("fullscreenPopup",
				      overrideShellWidgetClass, New,
				      XtNx, 0,
				      XtNy, 0,
				      XtNwidth, WidthOfScreen (screen),
				      XtNheight, HeightOfScreen (screen),
				      XtNborderWidth, 0,
				      NULL);
	new->imageDisplay.override_shell = shell;
	m_cnv = XtVaCreateManagedWidget ("multiCanvas",
					 multiCanvasWidgetClass, shell,
					 XkwNrequestMask, canvas_types,
					 XtNwidth, WidthOfScreen (screen),
					 XtNheight, HeightOfScreen (screen),
					 XkwNallowedMask,
					     new->imageDisplay.allowedMask,
					 XkwNpseudoXColourMap,
					    new->imageDisplay.pseudoXColourMap,
					 NULL);
    }
    else
    {
	m_cnv = XtVaCreateManagedWidget ("multiCanvas", multiCanvasWidgetClass,
					 New,
					 XtNfromVert, top_widget,
					 XtNwidth, canvas_width,
					 XtNheight, canvas_height,
					 XkwNrequestMask, canvas_types,
					 XkwNallowedMask,
					     new->imageDisplay.allowedMask,
					 XkwNpseudoXColourMap,
					    new->imageDisplay.pseudoXColourMap,
					 NULL);
    }
    new->imageDisplay.multiCanvas = m_cnv;
    /*  Setup for mono canvases  */
    pseudo_cnvs = XtNameToWidget (m_cnv, "pseudoColourCanvas");
    direct_cnvs = XtNameToWidget (m_cnv, "directColourCanvas");
    true_cnvs = XtNameToWidget (m_cnv, "trueColourCanvas");
    handle_canvas (new, pseudo_cnvs);
    handle_canvas (new, direct_cnvs);
    handle_canvas (new, true_cnvs);
    /*  Setup for stereo canvases  */
    pseudo_cnvs = XtNameToWidget (m_cnv, "pseudoColourStereoCanvas");
    direct_cnvs = XtNameToWidget (m_cnv, "directColourStereoCanvas");
    true_cnvs = XtNameToWidget (m_cnv, "trueColourStereoCanvas");
    handle_canvas (new, pseudo_cnvs);
    handle_canvas (new, direct_cnvs);
    handle_canvas (new, true_cnvs);
    /*  Create magnifier window  */
    w= XtVaCreatePopupShell ("magnifierShell",
			     magnifierPopupWidgetClass, New,
			     XkwNenablePanning, TRUE,
			     NULL);
    new->imageDisplay.magnifier = w;
    /*  Create panner window  */
    w = XtVaCreatePopupShell ("pannerShell", pannerPopupWidgetClass, New,
			      NULL);
    new->imageDisplay.panner = w;
    /*  Realise if needed  */
    xtmisc_ext_call (New, "WidgetInitialise", 0, NULL, FALSE, Request, New);
    if (new->imageDisplay.fullscreen)
    {
	XtRealizeWidget (shell);
	XtPopup (shell, XtGrabNone);
    }
}   /*  End Function Initialise  */

static void _ImageDisplay_Destroy (Widget W)
/*  Destroy method. This is called after Xt destroy callbacks.
*/
{
    ImageDisplayWidget w = (ImageDisplayWidget) W;

    if ( w->imageDisplay.imageName &&
	 (w->imageDisplay.imageName != def_image_name) )
	m_free (w->imageDisplay.imageName);
    c_call_callbacks (w->imageDisplay.destroy_list, NULL);
    c_destroy_list (w->imageDisplay.destroy_list);
}   /*  End Function Destroy  */

static Boolean _ImageDisplay_SetValues (Widget Current, Widget Request,
					Widget New)
{
    KWorldCanvas old_wc, new_wc;
    ImageDisplayWidget current = (ImageDisplayWidget) Current;
    /*ImageDisplayWidget request = (ImageDisplayWidget) Request;*/
    ImageDisplayWidget new = (ImageDisplayWidget) New;
    Widget multi_canvas;
    Widget old_cnv = NULL;      /*  Initialised to keep compiler happy  */
    Widget new_cnv = NULL;      /*  Initialised to keep compiler happy  */
    Widget w;
    static char function_name[] = "ImageDisplayWidget::SetValues";

    if (new->imageDisplay.cmapSize != current->imageDisplay.cmapSize)
    {
	if ( !kcmap_change (new->imageDisplay.pseudoCmap, NULL,
			    (unsigned int) new->imageDisplay.cmapSize, TRUE) )
	{
	    fprintf (stderr, "Error resizing colourmap to: %d cells\n",
		     new->imageDisplay.cmapSize);
	    new->imageDisplay.cmapSize = current->imageDisplay.cmapSize;
	}
    }
    if (new->imageDisplay.imageName != current->imageDisplay.imageName)
    {
	if ( current->imageDisplay.imageName &&
	     (current->imageDisplay.imageName != def_image_name) )
	{
	    /*  There was an old name and it's not the default: free it  */
	    m_free (current->imageDisplay.imageName);
	}
	if (new->imageDisplay.imageName == NULL)
	    new->imageDisplay.imageName = def_image_name;
	else
	{
	    if ( ( new->imageDisplay.imageName =
		  st_dup (new->imageDisplay.imageName) ) == NULL )
		m_abort (function_name, "image name");
	}
	w = XtNameToWidget (new->imageDisplay.export_menu, "savePopup");
	XtVaSetValues (w,
		       XkwNdefaultName, new->imageDisplay.imageName,
		       NULL);
	w = XtNameToWidget (new->imageDisplay.export_menu,
			    "postscriptwinpopup");
	XkwPostscriptRegisterImageAndName (w, NULL,
					   new->imageDisplay.imageName);
	
    }
    /*  Deal with requests to change active canvas  */
    old_wc = current->imageDisplay.visibleCanvas;
    new_wc = new->imageDisplay.visibleCanvas;
    multi_canvas = new->imageDisplay.multiCanvas;
    if (new_wc != old_wc)
    {
	if (new_wc == NULL)
	{
	    fprintf (stderr, "NULL visibleCanvas resource!\n");
	    a_prog_bug (function_name);
	}
	if (old_wc == NULL) old_cnv = NULL;
	else old_cnv = _ImageDisplay_wc_to_cnv (new, old_wc);
	new_cnv = _ImageDisplay_wc_to_cnv (new, new_wc);
	if (old_cnv == new_cnv) return False;
	if (old_cnv) XtUnmapWidget (old_cnv);
	XtMapWidget (new_cnv);
	XtVaSetValues (new->imageDisplay.export_menu,
		       XkwNworldCanvas, new_wc,
		       NULL);
	XtCallCallbacks ( (Widget) new, XkwNnewVisibleCanvasCallback, new_wc );
	return False;
    }
    return True;
}   /*  End Function SetValues  */

static Widget _ImageDisplay_wc_to_cnv (ImageDisplayWidget top, KWorldCanvas wc)
{
    Widget multi_canvas = top->imageDisplay.multiCanvas;
    static char function_name[] = "ImageDisplayWidget::wc_to_cnv";

    if (wc == top->imageDisplay.pseudoCanvas)
	return XtNameToWidget (multi_canvas, "pseudoColourCanvas");
    if (wc == top->imageDisplay.directCanvas)
	return XtNameToWidget (multi_canvas, "directColourCanvas");
    if (wc == top->imageDisplay.trueCanvas)
	return XtNameToWidget (multi_canvas, "trueColourCanvas");
    if (wc == top->imageDisplay.pseudoCanvasLeft)
	return XtNameToWidget (multi_canvas, "pseudoColourStereoCanvas");
    if (wc == top->imageDisplay.pseudoCanvasRight)
	return XtNameToWidget (multi_canvas, "pseudoColourStereoCanvas");
    if (wc == top->imageDisplay.directCanvasLeft)
	return XtNameToWidget (multi_canvas, "directColourStereoCanvas");
    if (wc == top->imageDisplay.directCanvasRight)
	return XtNameToWidget (multi_canvas, "directColourStereoCanvas");
    if (wc == top->imageDisplay.trueCanvasLeft)
	return XtNameToWidget (multi_canvas, "trueColourStereoCanvas");
    if (wc == top->imageDisplay.trueCanvasRight)
	return XtNameToWidget (multi_canvas, "trueColourStereoCanvas");
    fprintf (stderr, "Visible canvas: %p unknown!\n", wc);
    a_prog_bug (function_name);
    return (NULL);
}   /*  End Function wc_to_cnv  */

static void quit_cbk (Widget w, XtPointer client_data, XtPointer call_data)
/*  This is the quit button callback.
*/
{
    exit (RV_OK);
}   /*  End Function quit_cbk   */

static void _ImageDisplay_canvas_realise_cbk (Widget w, XtPointer client_data,
					      XtPointer call_data)
/*  [SUMMARY] Canvas realise callback.
    [PURPOSE] This routine is called when a Canvas widget is realised. A
    Kcolourmap object is created if required, then the KWorldCanvas is created
    and initialised for the [<viewimg>] package.
    <w> The Canvas widget that has just been realised.
    <client_data> The ImageDisplay widget.
    <call_data> The pixel canvas(es).
    [RETURNS] Nothing.
*/
{
    Kcolourmap kcmap = NULL;        /*  Initialised to keep compiler happy  */
    flag verbose;
    KPixCanvas pixcanvas;
    KWorldCanvas wc_pseudo, wc_direct, wc_true;
    KPixCanvas *stereopixcanvases = NULL;/*Initialised to keep compiler happy*/
    KPixCanvasEditorContext ctx;
    int stereo_mode, depth;
    unsigned int visual_type;
    unsigned long background_pixel;
    ImageDisplayWidget top = client_data;
    Colormap xcmap;
    Widget multicanvas;
    Widget pseudo_cnvs, direct_cnvs, true_cnvs;
    KWorldCanvas *worldcanvas = NULL;
    KWorldCanvas *leftworldcanvas = NULL;
    KWorldCanvas *rightworldcanvas = NULL;
    KWorldCanvas canvases[10], *ptr;
    Visual *visual;
    struct win_scale_type win_scale;
    static char function_name[] = "ImageDisplayWidget::canvas_realise_cbk";

    verbose = top->imageDisplay.verbose;
    multicanvas = top->imageDisplay.multiCanvas;
    XtVaGetValues (w,
		   XkwNstereoMode, &stereo_mode,
		   NULL);
    if (stereo_mode == XkwSTEREO_MODE_MONO)
    {
	pseudo_cnvs = XtNameToWidget (multicanvas, "pseudoColourCanvas");
	direct_cnvs = XtNameToWidget (multicanvas, "directColourCanvas");
	true_cnvs = XtNameToWidget (multicanvas, "trueColourCanvas");
	pixcanvas = (KPixCanvas) call_data;
    }
    else
    {
	pseudo_cnvs = XtNameToWidget (multicanvas, "pseudoColourStereoCanvas");
	direct_cnvs = XtNameToWidget (multicanvas, "directColourStereoCanvas");
	true_cnvs = XtNameToWidget (multicanvas, "trueColourStereoCanvas");
	stereopixcanvases = (KPixCanvas *) call_data;
	pixcanvas = stereopixcanvases[0];
    }
    kwin_get_attributes (pixcanvas,
			 KWIN_ATT_VISUAL, &visual_type,
			 KWIN_ATT_END);
    XtVaGetValues (w,
		   XtNcolormap, &xcmap,
		   XtNdepth, &depth,
		   XtNbackground, &background_pixel,
		   XtNvisual, &visual,
		   NULL);
    if (verbose) fprintf ( stderr, "%s: visual: %p visualID: %#lx\n",
			   function_name, (void *) visual,
			   XVisualIDFromVisual (visual) );
    /*  First initialise win_scale since all data is wiped here  */
    canvas_init_win_scale (&win_scale, K_WIN_SCALE_MAGIC_NUMBER);
    win_scale.blank_pixel = background_pixel;
    /*  Visual-specific work  */
    if (w == pseudo_cnvs)
    {
	if (visual_type != KWIN_VISUAL_PSEUDOCOLOUR)
	{
	    fprintf (stderr, "pseudo_canvas not PseudoColour visual!\n");
	    a_prog_bug (function_name);
	}
	kcmap = _ImageDisplay_get_pc_colourmap (top, visual, pixcanvas, xcmap,
						depth);
	worldcanvas = &top->imageDisplay.pseudoCanvas;
	leftworldcanvas = &top->imageDisplay.pseudoCanvasLeft;
	rightworldcanvas = &top->imageDisplay.pseudoCanvasRight;
    }
    else if (w == direct_cnvs)
    {
	if (visual_type != KWIN_VISUAL_DIRECTCOLOUR)
	{
	    fprintf (stderr, "direct_canvas not DirectColour visual!\n");
	    a_prog_bug (function_name);
	}
	/*  Giving the world canvas a colourmap is a handy way of having the
	    min. and max. sat pixels kept up to date  */
	kcmap = get_dc_colourmap (top, visual, xcmap, depth, pixcanvas);
	worldcanvas = &top->imageDisplay.directCanvas;
	leftworldcanvas = &top->imageDisplay.directCanvasLeft;
	rightworldcanvas = &top->imageDisplay.directCanvasRight;
    }
    else if (w == true_cnvs)
    {
	if (visual_type != KWIN_VISUAL_TRUECOLOUR)
	{
	    fprintf (stderr, "true_canvas not TrueColour visual!\n");
	    a_prog_bug (function_name);
	}
	kcmap = (pseudo_cnvs ==
		 NULL) ? _ImageDisplay_get_pc_colourmap (top, visual,
							 pixcanvas, xcmap,
							 depth) : NULL;
	worldcanvas = &top->imageDisplay.trueCanvas;
	leftworldcanvas = &top->imageDisplay.trueCanvasLeft;
	rightworldcanvas = &top->imageDisplay.trueCanvasRight;
    }
    else
    {
	fprintf (stderr, "Bad canvas passed: %p\n", (void *) w);
	a_prog_bug (function_name);
    }
    /*  Create world canvas(es) and initialise for higher level packages  */
    /*  Create the main world canvas (with dummy world min and max)  */
    win_scale.conv_type = CONV1_REAL;
    if (stereo_mode == XkwSTEREO_MODE_MONO)
    {   /*  Monoscopic mode  */
	if ( ( *worldcanvas = canvas_create (pixcanvas, kcmap, &win_scale) )
	     == NULL )
	    exit (RV_UNDEF_ERROR);
	setup_wc (top, *worldcanvas);
	ctx = viewimg_create_drag_and_zoom_interface (*worldcanvas);
	viewimg_register_position_event_func (*worldcanvas,
					      viewimg_statistics_position_func,
					      NULL);
	kwin_editor_register_event_func (ctx, viewimg_statistics_geom_func,
					 *worldcanvas);
	kwin_editor_register_event_func (ctx, viewimg_values_geom_func,
					 *worldcanvas);
	viewimg_register_position_event_func (*worldcanvas,
					      viewimg_values_position_func,
					      NULL);
	viewimg_register_position_event_func (*worldcanvas,
					      viewimg_header_position_func,
					      NULL);
	viewimg_register_position_event_func (*worldcanvas,
					      _ImageDisplay_vimage_event_func,
					      top);
	contour_init (*worldcanvas,
		      CONTOUR_CANVAS_ATT_COLOURNAME, "green",
		      CONTOUR_CANVAS_ATT_END);
	XtCallCallbacks ( (Widget) top, XkwNnewEditorContextCallback,
			  (XtPointer) ctx );
	kwin_register_bindings (pixcanvas, main_bindings, FALSE);
	/*HACK*/
	if ( r_getenv ("LABELS_UNDER_OVERLAYS") )
	{
	    static flag first_time = TRUE;

	    if (first_time)
	    {
		first_time = FALSE;
		fprintf (stderr, "Labels will be drawn under overlays\n");
	    }
	    canvas_sequence_dressing_refresh (*worldcanvas);
	}
    }
    else
    {   /*  Stereoscopic mode  */
	if ( ( *leftworldcanvas = canvas_create (stereopixcanvases[0], kcmap,
						 &win_scale) ) == NULL )
	    exit (RV_UNDEF_ERROR);
	setup_wc (top, *leftworldcanvas);
	if ( ( *rightworldcanvas = canvas_create (stereopixcanvases[1], kcmap,
						  &win_scale) ) == NULL )
	    exit (RV_UNDEF_ERROR);
	setup_wc (top, *rightworldcanvas);
	contour_init (*leftworldcanvas,
		      CONTOUR_CANVAS_ATT_COLOURNAME, "green",
		      CONTOUR_CANVAS_ATT_END);
	contour_init (*rightworldcanvas,
		      CONTOUR_CANVAS_ATT_COLOURNAME, "green",
		      CONTOUR_CANVAS_ATT_END);
    }
    if (top->imageDisplay.num_unrealised == 0)
    {
	fprintf (stderr, "%s: spurious realise\n", function_name);
	return;
    }
    if (--top->imageDisplay.num_unrealised > 0) return;
    /*  All canvases are now realised  */
    ptr = canvases;
    /*  PseudoColour canvases  */
    handle_wc (top, top->imageDisplay.pseudoCanvas, &ptr, TRUE);
    handle_wc (top, top->imageDisplay.pseudoCanvasLeft, &ptr, FALSE);
    handle_wc (top, top->imageDisplay.pseudoCanvasRight, &ptr, FALSE);
    /*  DirectColour canvases  */
    handle_wc (top, top->imageDisplay.directCanvas, &ptr, TRUE);
    handle_wc (top, top->imageDisplay.directCanvasLeft, &ptr, FALSE);
    handle_wc (top, top->imageDisplay.directCanvasRight, &ptr, FALSE);
    /*  TrueColour canvases  */
    handle_wc (top, top->imageDisplay.trueCanvas, &ptr, TRUE);
    handle_wc (top, top->imageDisplay.trueCanvasLeft, &ptr, FALSE);
    handle_wc (top, top->imageDisplay.trueCanvasRight, &ptr, FALSE);
    *ptr++ = NULL;
    top->imageDisplay.canvasList = m_alloc ( sizeof *ptr * (ptr - canvases) );
    if (top->imageDisplay.canvasList == NULL)
	m_abort (function_name, "canvas list");
    REGISTER_FREE_FUNC (top, m_free, top->imageDisplay.canvasList);
    m_copy (top->imageDisplay.canvasList, canvases,
	    sizeof *ptr * (ptr - canvases) );
    if (top->imageDisplay.renzoPseudoCmap && top->imageDisplay.trueCanvas)
    {   /*Make TrueColour Renzogram colourmap a slave of the PseudoColour one*/
	unsigned int num_ccells;
	unsigned long red_mask, green_mask, blue_mask;
	unsigned short colours[256 * 3];

	kwin_get_attributes (canvas_get_pixcanvas
			     (top->imageDisplay.trueCanvas),
			     KWIN_ATT_PIX_RED_MASK, &red_mask,
			     KWIN_ATT_PIX_GREEN_MASK, &green_mask,
			     KWIN_ATT_PIX_BLUE_MASK, &blue_mask,
			     KWIN_ATT_END);
	num_ccells = kcmap_read_rgb_values (top->imageDisplay.renzoPseudoCmap,
					    256, colours, colours + 1,
					    colours + 2, 3);
	if ( ( kcmap = kcmap_create_scmap (DEFAULT_COLOURMAP_NAME,
					   num_ccells, TRUE,
					   red_mask, green_mask, blue_mask) )
	     == NULL )
	{
	    fprintf (stderr,
		     "Error creating software Karma PseudoColour cmap\n");
	    a_prog_bug (function_name);
	}
	REGISTER_FREE_FUNC (top, kcmap_destroy, kcmap);
	top->imageDisplay.renzoTrueCmap = kcmap;
	kcmap_write_rgb_values (kcmap, num_ccells,
				colours, colours + 1, colours + 2, 3);
	kcmap_register_modify_func (top->imageDisplay.renzoPseudoCmap,
				    _ImageDisplay_kcmap_modify_func, kcmap);
    }
    XtCallCallbacks ( (Widget) top, XkwNmainRealisedCallback,
		      (XtPointer) canvases );
    XtCallCallbacks ( (Widget) top, XkwNafterMainRealisedCallback,
		      (XtPointer) canvases );
    if (top->imageDisplay.viewDatasets && (top->imageDisplay.numDatasets < 1) )
    {   /*  Now that Renzogram colourmaps are done, can create browser  */
	static flag ft = TRUE;

	XtVaCreatePopupShell (ft ? "firstDataBrowserShell" :"dataBrowserShell",
			      dataBrowserWidgetClass, (Widget) top,
			      NULL);
	ft = FALSE;
    }
    if (!top->imageDisplay.visibleCanvas)
    {
	KWorldCanvas wc = top->imageDisplay.pseudoCanvas;

	if (!wc) wc = top->imageDisplay.trueCanvas;
	if (!wc) wc = top->imageDisplay.directCanvas;
	XtVaSetValues ( (Widget) top,
			XkwNvisibleCanvas, wc,
			NULL);
    }
    /*  Now that magnifier and panner are realised, add their canvases too  */
    XtVaGetValues (top->imageDisplay.magnifier,
		   XkwNpseudoColourCanvas, &wc_pseudo,
		   XkwNdirectColourCanvas, &wc_direct,
		   XkwNtrueColourCanvas, &wc_true,
		   NULL);
    XkwIntensityPolicyAddCanvas (top->imageDisplay.intensity_policy_popup,
				 wc_pseudo, TRUE);
    XkwIntensityPolicyAddCanvas (top->imageDisplay.intensity_policy_popup,
				 wc_direct, TRUE);
    XkwIntensityPolicyAddCanvas (top->imageDisplay.intensity_policy_popup,
				 wc_true, TRUE);
    XtVaGetValues (top->imageDisplay.panner,
		   XkwNpseudoColourCanvas, &wc_pseudo,
		   XkwNdirectColourCanvas, &wc_direct,
		   XkwNtrueColourCanvas, &wc_true,
		   NULL);
    XkwIntensityPolicyAddCanvas (top->imageDisplay.intensity_policy_popup,
				 wc_pseudo, FALSE);
    XkwIntensityPolicyAddCanvas (top->imageDisplay.intensity_policy_popup,
				 wc_direct, FALSE);
    XkwIntensityPolicyAddCanvas (top->imageDisplay.intensity_policy_popup,
				 wc_true, FALSE);
    xtmisc_ext_call (client_data, "WidgetRealise", 0, NULL, FALSE,client_data);
}   /*  End Function canvas_realise_cbk   */

static void setup_wc (ImageDisplayWidget top, KWorldCanvas canvas)
{
    Bool integer_x_zoom, integer_y_zoom, auto_iscale, fix_aspect;

    XtVaGetValues (top->imageDisplay.zoom_policy_popup,
		   XkwNintegerXZoom, &integer_x_zoom,
		   XkwNintegerYZoom, &integer_y_zoom,
		   XkwNfixAspect, &fix_aspect,
		   NULL);
    XtVaGetValues (top->imageDisplay.intensity_policy_popup,
		   XkwNautoIntensityScale, &auto_iscale,
		   NULL);
    canvas_set_attributes (canvas,
			   CANVAS_ATT_AUTO_MIN_SAT, TRUE,
			   CANVAS_ATT_AUTO_MAX_SAT, TRUE,
			   CANVAS_ATT_AUTO_BLANK, TRUE,
			   CANVAS_ATT_END);
    canvas_set_dressing (canvas,
			 KCD_ATT_SHOW_LABELS, TRUE,
			 KCD_ATT_TOP_TICKS, TRUE,
			 KCD_ATT_BOTTOM_TICKS, TRUE,
			 KCD_ATT_LEFT_TICKS, TRUE,
			 KCD_ATT_RIGHT_TICKS, TRUE,
			 KCD_ATT_END);
    canvas_register_position_event_func (canvas,
					 _ImageDisplay_canvas_event_handler,
					 top);
    viewimg_init (canvas);
    viewimg_set_canvas_attributes (canvas,
				   VIEWIMG_ATT_MAINTAIN_ASPECT, TRUE,
				   VIEWIMG_ATT_ALLOW_TRUNCATION, TRUE,
				   VIEWIMG_ATT_INT_X, integer_x_zoom,
				   VIEWIMG_ATT_INT_Y, integer_y_zoom,
				   VIEWIMG_ATT_AUTO_V, auto_iscale,
				   VIEWIMG_ATT_MAINTAIN_ASPECT, fix_aspect,
				   VIEWIMG_ATT_AUTO_WCS, TRUE,
				   VIEWIMG_ATT_FAST_PANNER,
				   top->imageDisplay.fastPanner,
				   VIEWIMG_ATT_END);
}   /*  End Function setup_wc  */

static void handle_wc (ImageDisplayWidget top, KWorldCanvas canvas,
		       KWorldCanvas **ptr, flag slave)
{
    if (canvas == NULL) return;
    **ptr = canvas;
    ++*ptr;
    canvas_set_attributes (canvas,
			   CANVAS_ATT_AUTO_MIN_SAT, TRUE,
			   CANVAS_ATT_AUTO_MAX_SAT, TRUE,
			   CANVAS_ATT_AUTO_BLANK, TRUE,
			   CANVAS_ATT_END);
    canvas_set_dressing (canvas,
			 KCD_ATT_SHOW_LABELS, TRUE,
			 KCD_ATT_TOP_TICKS, TRUE,
			 KCD_ATT_BOTTOM_TICKS, TRUE,
			 KCD_ATT_LEFT_TICKS, TRUE,
			 KCD_ATT_RIGHT_TICKS, TRUE,
			 KCD_ATT_END);
    XkwIntensityPolicyAddCanvas (top->imageDisplay.intensity_policy_popup,
				 canvas, TRUE);
    XkwZoomPolicyAddCanvas (top->imageDisplay.zoom_policy_popup, canvas);
    XkwOverlayMenuAddCanvas (top->imageDisplay.overlay_menu, canvas);
    XkwImageEditorControlAddCanvas (top->imageDisplay.image_editor_popup,
				    canvas);
    if (!slave) return;
    XkwMagnifierPopupAddCanvas (top->imageDisplay.magnifier, canvas);
    XkwPannerPopupAddCanvas (top->imageDisplay.panner, canvas);
}   /*  End Function handle_wc  */

static void zoom_cbk (w, client_data, call_data)
/*  This is the zoom menu callback.
*/
Widget w;
XtPointer client_data;
XtPointer call_data;
{
    unsigned int zoomcode = *(int *) call_data;
    ImageDisplayWidget top = (ImageDisplayWidget) client_data;
    static char function_name[] = "ImageDisplayWidget::zoom_cbk";

    switch (zoomcode)
    {
      case ZOOMCODE_MAGNIFIER:
	XtPopup (top->imageDisplay.magnifier, XtGrabNone);
	break;
      case ZOOMCODE_PANNER:
	XtPopup (top->imageDisplay.panner, XtGrabNone);
	break;
      case ZOOMCODE_ZOOMIN2:
	zoom (top, 2.0);
	break;
      case ZOOMCODE_ZOOMIN4:
	zoom (top, 4.0);
	break;
      case ZOOMCODE_ZOOMOUT2:
	zoom (top, 0.5);
	break;
      case ZOOMCODE_ZOOMOUT4:
	zoom (top, 0.25);
	break;
      case ZOOMCODE_UNZOOM:
	unzoom (top);
	break;
      case ZOOMCODE_POLICY:
	XtPopup (top->imageDisplay.zoom_policy_popup, XtGrabNone);
	break;
      default:
	fprintf (stderr, "Illegal zoom code: %u\n", zoomcode);
	a_prog_bug (function_name);
	break;
    }
}   /*  End Function zoom_cbk   */

static void region_cbk (Widget w, XtPointer client_data, XtPointer call_data)
/*  This is the region callback.
*/
{
    DataclipRegions *regions = (DataclipRegions *) call_data;
    KWorldCanvas *canvases;
    ImageDisplayWidget top = (ImageDisplayWidget) client_data;
    static char function_name[] = "ImageDisplayWidget::region_cbk";

    if (regions->num_regions != 1)
    {
	fprintf (stderr, "num_regions: %u is not 1\n", regions->num_regions);
	a_prog_bug (function_name);
    }
    if (top->imageDisplay.visibleCanvas == NULL) return;
    XtVaGetValues (top->imageDisplay.magnifier, XkwNcanvasList,&canvases,NULL);
    _ImageDisplay_set_irange (NULL, canvases,
			      regions->minima[0], regions->maxima[0]);
    XtVaGetValues (top->imageDisplay.panner, XkwNcanvasList, &canvases, NULL);
    _ImageDisplay_set_irange (NULL, canvases,
			      regions->minima[0], regions->maxima[0]);
    if ( _ImageDisplay_set_irange (top->imageDisplay.visibleCanvas,
				   top->imageDisplay.canvasList,
				   regions->minima[0], regions->maxima[0]) )
    {
	XkwImageDisplayRefresh ( (Widget) top, TRUE );
	XkwSlaveImageDisplayPopupRefresh ( (Widget)top->imageDisplay.magnifier,
					   TRUE );
	XkwSlaveImageDisplayPopupRefresh ( (Widget) top->imageDisplay.panner,
					   TRUE );
    }
}   /*  End Function region_cbk   */

static flag _ImageDisplay_set_irange (KWorldCanvas vc, KWorldCanvas *canvases,
				      double minimum, double maximum)
/*  [SUMMARY] Set the intensity range for a set of world canvases.
    <vc> The visible KWorldCanvas object.
    <canvases> The array of KWorldCanvas objects to update. This must be NULL
    terminated.
    <minimum> The minimum of the range.
    <maximum> The maximum of the range.
    [RETURNS] TRUE if a refresh should be performed, else FALSE.
*/
{
    flag refresh = FALSE;

    for (; *canvases; ++canvases)
    {
	flag truecolour;
	ViewableImage vimage;
	double min, max;

	canvas_set_attributes (*canvases,
			       CANVAS_ATT_VALUE_MIN, minimum,
			       CANVAS_ATT_VALUE_MAX, maximum,
			       CANVAS_ATT_END);
	if ( ( vimage = viewimg_get_active (*canvases) ) == NULL ) continue;
	viewimg_get_attributes (vimage,
				VIEWIMG_VATT_TRUECOLOUR, &truecolour,
				VIEWIMG_VATT_END);
	if (truecolour) continue;
	viewimg_get_attributes (vimage,
				VIEWIMG_VATT_VALUE_MIN, &min,
				VIEWIMG_VATT_VALUE_MAX, &max,
				VIEWIMG_VATT_END);
	min = (min < TOOBIG) ? minimum : TOOBIG;
	max = (max < TOOBIG) ? maximum : TOOBIG;
	viewimg_set_attributes (vimage,
				VIEWIMG_VATT_VALUE_MIN, min,
				VIEWIMG_VATT_VALUE_MAX, max,
				VIEWIMG_VATT_END);
	if (*canvases == vc) refresh = TRUE;
    }
    return (refresh);
}   /*  End Function set_irange  */

static void _ImageDisplay_intensity_range_cbk (Widget w, XtPointer client_data,
					       XtPointer call_data)
{
    DataclipRegions region;
    ImageDisplayWidget top = (ImageDisplayWidget) client_data;
    Widget izoom;
    double *range = call_data;

    region.num_regions = 1;
    region.minima = range;
    region.maxima = range + 1;
    region_cbk (NULL, client_data, &region);
    izoom = XtNameToWidget ( (Widget) top, "izoomwinpopup" );
    XkwDataclipSetRegions (izoom, &region);
}   /*  End Function intensity_range_cbk  */

static flag _ImageDisplay_canvas_event_handler (KWorldCanvas canvas,
						double x, double y,
						unsigned int event_code,
						void *e_info, void **f_info,
						double x_lin, double y_lin)
/*  [SUMMARY] Position event callback.
    [PURPOSE] This routine is a position event consumer for a world canvas.
    <canvas> The canvas on which the event occurred.
    <x> The horizontal world co-ordinate of the event.
    <y> The vertical world co-ordinate of the event.
    [NOTE] These values will have been transformed by the registered
    transform function (see [<canvas_register_transform_func>]).
    <event_code> The arbitrary event code.
    <e_info> A pointer to arbitrary event information.
    <f_info> A pointer to an arbitrary function information pointer.
    <x_lin> The horizontal linear world co-ordinate prior to the transform
    function being called.
    <y_lin> The vertical linear world co-ordinate prior to the transform
    function being called.
    [RETURNS] TRUE if the event was consumed, else FALSE indicating that
    the event is still to be processed.
*/
{
    ImageDisplayWidget w = (ImageDisplayWidget) *f_info;
    /*static char function_name[]="ImageDisplayWidget::canvas_event_handler";*/

    if ( (event_code == K_CANVAS_EVENT_RIGHT_MOUSE_CLICK) &&
	 w->imageDisplay.override_shell )
    {
	/*  Lower fullscreen canvas  */
	XLowerWindow ( XtDisplay (w),
		       XtWindow (w->imageDisplay.override_shell) );
	return (TRUE);
    }
    return (FALSE);
}   /*  End Function canvas_event_handler  */

static flag _ImageDisplay_vimage_event_func (ViewableImage vimage,
					     double x, double y,
					     void *value,
					     unsigned int event_code,
					     void *e_info, void **f_info,
					     double x_lin, double y_lin,
					     unsigned int value_type,
					     double x_im, double y_im,
					     int x_pix, int y_pix)
/*  [SUMMARY] Position event callback.
    [PURPOSE] This routine is a position event consumer for a world canvas
    which has a number of ViewableImage objects associated with it.
    <viewimg> The active viewable image.
    <x> The horizontal world co-ordinate of the event.
    <y> The vertical world co-ordinate of the event.
    These values will have been transformed by the registered transform
    function (see [<canvas_register_transform_func>]) for the associated
    world canvas.
    <value> A pointer to the data value in the viewable image corresponding
    to the event co-ordinates.
    <event_code> The arbitrary event code.
    <e_info> The arbitrary event information.
    <f_info> The arbitrary function information pointer.
    <x_lin> The linear horizontal world co-ordinate (the co-ordinate prior
    to the transform function being called).
    <y_lin> The linear vertical world co-ordinate (the co-ordinate prior
    to the transform function being called).
    <value_type> The type of the data value. This may be K_DCOMPLEX or
    K_UB_RGB.
    <x_im> The horizontal image pixel co-ordinate. 0.0 is the pixel centre.
    <y_im> The vertical image pixel co-ordinate. 0.0 is the pixel centre.
    <x_pix> The horizontal screen pixel co-ordinate.
    <y_pix> The vertical screen pixel co-ordinate.
    [RETURNS] TRUE if the event was consumed, else FALSE indicating that
    the event is still to be processed.
*/
{
    ImageDisplayWidget w = (ImageDisplayWidget) *f_info;
    Widget W = (Widget) *f_info;
    Widget first_track_label, second_track_label, third_track_label;
    char *string = e_info;
    char pix_string[STRING_LENGTH];
    char world_string[STRING_LENGTH], extra_string[STRING_LENGTH];
    /*static char function_name[] = "ImageDisplay::vimage_event_func";*/

    if (event_code == K_CANVAS_EVENT_POINTER_ENTER) return (FALSE);
    if (!w->imageDisplay.updateTrackLabels) return (FALSE);
    viewimg_track_compute (vimage, value, value_type, x, y, x_im, y_im,
			   pix_string, world_string, extra_string);
    if ( (event_code == K_CANVAS_EVENT_PLAIN_KEY_PRESS) &&
	 (strcmp (string, "l") == 0) )
	puts (world_string);
    first_track_label = XtNameToWidget (W, "trackLabel0");
    if (first_track_label == NULL) return (FALSE);
    XtVaSetValues (first_track_label, XtNlabel, pix_string, NULL);
    second_track_label = XtNameToWidget (W, "trackLabel1");
    if (second_track_label == NULL) return (FALSE);
    XtVaSetValues (second_track_label, XtNlabel, world_string, NULL);
    third_track_label = XtNameToWidget (W, "trackLabel2");
    if (third_track_label == NULL) return (FALSE);
    XtVaSetValues (third_track_label, XtNlabel, extra_string, NULL);
    return (FALSE);
}   /*  End Function vimage_event_func  */

static void zoom (ImageDisplayWidget w, double zoom)
/*  This routine will zoom the canvas.
    The routine returns nothing.
*/
{
    KWorldCanvas visible_worldcanvas = w->imageDisplay.visibleCanvas;
    ViewableImage vimage;
    /*static char function_name[] = "zoom";*/

    /*  Zoom  */
    vimage = viewimg_get_active (visible_worldcanvas);
    if (vimage == NULL) return;
    viewimg_set_canvas_attributes (visible_worldcanvas,
				   VIEWIMG_ATT_AUTO_X, FALSE,
				   VIEWIMG_ATT_AUTO_Y, FALSE,
				   VIEWIMG_ATT_END);
    /*  Redraw the canvas  */
    viewimg_zoom (vimage, 0.5, 0.5, TRUE, zoom);
}   /*  End Function zoom  */

static void unzoom (ImageDisplayWidget w)
/*  This routine will unzoom the canvas.
    The routine returns nothing.
*/
{
    KWorldCanvas visible_worldcanvas = w->imageDisplay.visibleCanvas;
    /*static char function_name[] = "unzoom";*/

    /*  Unzoom  */
    viewimg_set_canvas_attributes (visible_worldcanvas,
				   VIEWIMG_ATT_AUTO_X, TRUE,
				   VIEWIMG_ATT_AUTO_Y, TRUE,
				   VIEWIMG_ATT_END);
    /*  Redraw the canvas  */
    if ( !kwin_resize (canvas_get_pixcanvas (visible_worldcanvas),
		       TRUE, 0, 0, 0, 0) )
    {
	fprintf (stderr, "Error refreshing window\n");
    }
}   /*  End Function unzoom  */

static void _ImageDisplay_intensity_cbk (Widget w, XtPointer client_data,
					 XtPointer call_data)
/*  This is the colourmap menu callback.
*/
{
    int len, count;
    unsigned int icode = *(int *) call_data;
    ImageDisplayWidget top = (ImageDisplayWidget) client_data;
    Widget popup = NULL;  /*  Initialised to keep compiler happy  */
    char *choice;
    char **choices;
    char **dataset_names = (char **) top->imageDisplay.datasetNames;
    char txt[STRING_LENGTH];
    static char *dataclip_name = "IScale for Dataset ";
    static char function_name[] = "ImageDisplayWidget::intensity_cbk";

    len = strlen (dataclip_name);
    XtVaGetValues (w, XkwNitemStrings, &choices, NULL);
    choice = choices[icode];
    if (strcmp (choice, CMAP_STRING_PC) == 0)
	popup = top->imageDisplay.cmapwinpopup_psuedo;
    else if (strcmp (choice, CMAP_STRING_PC_EMULATED) == 0)
	popup = top->imageDisplay.cmapwinpopup_psuedo;
    else if (strcmp (choice, INTENSITY_POLICY_STRING) == 0)
	popup = top->imageDisplay.intensity_policy_popup;
    else if (strcmp (choice, CMAP_STRING_DC) == 0)
	popup = top->imageDisplay.cmapwinpopup_direct;
    else if (strcmp (choice, CMAP_STRING_RENZO) == 0)
	popup = top->imageDisplay.cmapwinpopup_renzogram;
    else if ( (strncmp (choice, dataclip_name, len - 1) == 0) &&
	      (choice[len - 1] == '\0') )
	popup = XtNameToWidget ( (Widget) top, "izoomwinpopup" );
    else if (strncmp (choice, dataclip_name, len) == 0)
    {
	sprintf (txt, "izoomwinpopup%d", atoi (choice + len) - 1);
	popup = XtNameToWidget ( (Widget) top, txt );
    }
    else if (top->imageDisplay.datasetNames)
    {
	for (count = 0; count < top->imageDisplay.numDatasets; ++count)
	{
	    sprintf (txt, "IScale for %s", dataset_names[count]);
	    if (strcmp (choice, txt) == 0) break;
	}
	if (count >= top->imageDisplay.numDatasets)
	{
	    fprintf (stderr, "Unknown choice string: \"%s\"\n", choice);
	    a_prog_bug (function_name);
	}
	sprintf (txt, "izoomwinpopup%d", count);
	popup = XtNameToWidget ( (Widget) top, txt );
    }
    else
    {
	fprintf (stderr, "Unknown choice string: \"%s\"\n", choice);
	a_prog_bug (function_name);
    }
    if (popup == NULL)
    {
	fprintf (stderr, "No widget for choice: \"%s\"\n", choice);
	a_prog_bug (function_name);
    }
    XtPopup (popup, XtGrabNone);
}   /*  End Function intensity_cbk   */

static void raise_cbk (Widget w, XtPointer client_data, XtPointer call_data)
/*  This is the raise button callback.
*/
{
    ImageDisplayWidget top = (ImageDisplayWidget) client_data;

    XRaiseWindow ( XtDisplay (w),XtWindow (top->imageDisplay.override_shell) );
}   /*  End Function raise_cbk   */

static Kcolourmap _ImageDisplay_get_pc_colourmap (ImageDisplayWidget w,
						  Visual *visual,
						  KPixCanvas canvas,
						  Colormap xcmap, int depth)
/*  [SUMMARY] Get or create colourmap.
    <w> The ImageDisplay widget.
    <visual> The visual information.
    <canvas> The pixel canvas.
    <xcmap> The XColormap ID.
    <depth> The depth of the colourmap.
    [RETURN] The Kcolourmap object.
*/
{
    Kcolourmap kcmap;
    flag verbose;
    flag slave_cmap = FALSE;
    int port;
    unsigned int num_ccells, num_renzo_ccells, visual_type;
    Widget cmapwinpopup;
    char *slave_port_str = NULL;  /*  Initialised to keep compiler happy  */
    char slave_address[STRING_LENGTH];
    static flag first_time = TRUE;
    static char function_name[] = "ImageDisplayWidget::get_pc_colourmap";

    verbose = w->imageDisplay.verbose;
    if (w->imageDisplay.pseudoCmap) return (w->imageDisplay.pseudoCmap);
    kwin_get_attributes (canvas,
			 KWIN_ATT_VISUAL, &visual_type,
			 KWIN_ATT_END);
    if ( w->imageDisplay.viewDatasets && (w->imageDisplay.numDatasets < 1) )
    {   /*  Need to reserve some colourcells for Renzogram colourmap  */
	num_renzo_ccells = w->imageDisplay.cmapSize * RENZOGRAM_FRACTION;
	num_ccells = w->imageDisplay.cmapSize;
    }
    else
    {
	num_renzo_ccells = 0;
	num_ccells = w->imageDisplay.cmapSize;
    }
    if ( w->imageDisplay.cmapMaster && (*w->imageDisplay.cmapMaster != '\0') )
    {
	num_ccells = 2;
	slave_cmap = TRUE;
	strcpy (slave_address, w->imageDisplay.cmapMaster);
	if ( ( slave_port_str = strchr (slave_address, ':') ) == NULL )
	{
	    fprintf (stderr, "%s: bad slave colourmap address given: \"%s\"\n",
		     function_name, slave_address);
	    exit (RV_BAD_PARAM);
	}
	*slave_port_str++ = '\0';
    }
    kcmap = _ImageDisplay_create_pc_colourmap (w, visual, canvas, xcmap, depth,
					       num_ccells, &cmapwinpopup,NULL);
    w->imageDisplay.pseudoCmap = kcmap;
    w->imageDisplay.cmapwinpopup_psuedo = cmapwinpopup;
    if (first_time)
    {
	first_time = FALSE;
	kcmap_prepare_for_mastery (kcmap);
    }
    if (slave_cmap)
    {
	kcmap_prepare_for_slavery (kcmap);
	/*  Compute port number  */
	if ( isdigit (*slave_port_str) )
	    port = atoi (slave_port_str);
	else
	    port = r_get_def_port ( slave_port_str,
				    XDisplayString ( XtDisplay (w) ) );
	if (port < 0)
	{
	    fprintf (stderr, "Illegal port number: %d\n", port);
	    a_prog_bug (function_name);
	}
	if ( conn_attempt_connection (slave_address, port,
				      "colourmap_indices") ||
	     conn_attempt_connection (slave_address, port,
				      "full_colourmap") )
	{
	    fprintf (stderr,
		     "%s: made connection to colourmap master at: %s\n",
		     function_name, w->imageDisplay.cmapMaster);
	}
	else
	{
	    fprintf (stderr,
		     "%s: error connecting to colourmap master at: %s\n",
		     function_name, w->imageDisplay.cmapMaster);
	}
    }
    if (num_renzo_ccells < 2) return (kcmap);
    kcmap = _ImageDisplay_create_pc_colourmap
	(w, visual, canvas, xcmap, depth, num_renzo_ccells, &cmapwinpopup,
	 slave_cmap ? NULL : w->imageDisplay.pseudoCmap);
    if (visual_type == KWIN_VISUAL_PSEUDOCOLOUR)
	w->imageDisplay.renzoPseudoCmap = kcmap;
    else w->imageDisplay.renzoTrueCmap = kcmap;
    w->imageDisplay.cmapwinpopup_renzogram = cmapwinpopup;
    if (visual_type == KWIN_VISUAL_PSEUDOCOLOUR)
    {
	Kdisplay dpy_handle1;

	/*  Register destroy functions for display handle(s), now that destroy
	    functions for Kcolourmap objects are registered  */
	dpy_handle1 = kcmap_get_dpy_handle (w->imageDisplay.pseudoCmap);
	REGISTER_FREE_FUNC (w, xc_free_handle, dpy_handle1);
	if (w->imageDisplay.renzoPseudoCmap)
	{
	    Kdisplay dpy_handle2;

	    dpy_handle2 =
		kcmap_get_dpy_handle (w->imageDisplay.renzoPseudoCmap);
	    if (dpy_handle2 != dpy_handle1)
		REGISTER_FREE_FUNC (w, xc_free_handle, dpy_handle2);
	}
    }
    return (w->imageDisplay.pseudoCmap);
}   /*  End Function get_pc_colourmap  */

static Kcolourmap _ImageDisplay_create_pc_colourmap (ImageDisplayWidget w,
						     Visual *visual,
						     KPixCanvas canvas,
						     Colormap xcmap, int depth,
						     unsigned int num_ccells,
						     Widget *colourmapwinpopup,
						     Kcolourmap donor)
/*  [SUMMARY] Create colourmap.
    <w> The ImageDisplay widget.
    <visual> The visual information.
    <canvas> The pixel canvas.
    <xcmap> The XColormap ID.
    <depth> The depth of the colourmap.
    <donor> The Kcolourmap which will donate colourcells.
    [RETURN] The Kcolourmap object.
*/
{
    Kcolourmap kcmap;
    flag verbose;
    unsigned int visual_type;
    unsigned long *pixel_values;
    static char function_name[] = "ImageDisplayWidget::create_pc_colourmap";

    verbose = w->imageDisplay.verbose;
    kwin_get_attributes (canvas,
			 KWIN_ATT_VISUAL, &visual_type,
			 KWIN_ATT_END);
    if (visual_type == KWIN_VISUAL_PSEUDOCOLOUR)
    {
	Kdisplay dpy_handle = NULL;

	if (donor) dpy_handle = kcmap_get_dpy_handle (donor);
	else
	{
	    if ( ( dpy_handle = xc_get_dpy_handle (XtDisplay (w), xcmap) )
		 == NULL )
	    {
		fprintf (stderr, "Error getting display handle\n");
		a_prog_bug (function_name);
	    }
	}
	if ( ( kcmap = kcmap_va_create (DEFAULT_COLOURMAP_NAME,
					num_ccells, TRUE,
					dpy_handle,
					xc_alloc_colours, xc_free_colours,
					xc_store_colours, xc_get_location2,
					KCMAP_ATT_STEAL_COLOURS_FROM, donor,
					KCMAP_ATT_END) ) == NULL )
	{
	    fprintf (stderr,
		     "Error creating Karma PseudoColour cmap for Xcmap: %p\n",
		     (void *) xcmap);
	    a_prog_bug (function_name);
	}
    }
    else
    {
	unsigned long red_mask, green_mask, blue_mask;

	kwin_get_attributes (canvas,
			     KWIN_ATT_PIX_RED_MASK, &red_mask,
			     KWIN_ATT_PIX_GREEN_MASK, &green_mask,
			     KWIN_ATT_PIX_BLUE_MASK, &blue_mask,
			     KWIN_ATT_END);
	if ( ( kcmap = kcmap_create_scmap (DEFAULT_COLOURMAP_NAME,
					   num_ccells, TRUE,
					   red_mask, green_mask, blue_mask) )
	     == NULL )
	{
	    fprintf (stderr,
		     "Error creating software Karma PseudoColour cmap\n");
	    a_prog_bug (function_name);
	}
    }
    if (w->imageDisplay.verbose)
	fprintf (stderr,
		 "%s: created Karma colourmap: %p from X cmap: 0x%lx\n",
		 function_name, (void *) kcmap, xcmap);
    REGISTER_FREE_FUNC (w, kcmap_destroy, kcmap);
    num_ccells = kcmap_get_pixels (kcmap, &pixel_values);
    if (verbose) fprintf (stderr, "%s: num colours for PseudoColour: %u\n",
			  function_name, num_ccells);
    *colourmapwinpopup = XtVaCreatePopupShell
	("pseudoCmapwinpopup", cmapwinpopupWidgetClass, (Widget) w,
	 XtNcolormap, xcmap,
	 XtNdepth, depth,
	 XtNvisual, visual,
	 XkwNcolourbarVisual, visual,
	 XkwNkarmaColourmap, kcmap,
	 XkwNsimpleColourbar, True,
	 NULL);
    return (kcmap);
}   /*  End Function create_pc_colourmap  */

static Kcolourmap get_dc_colourmap (ImageDisplayWidget w, Visual *visual,
				    Colormap xcmap, int depth,
				    KPixCanvas pixcanvas)
/*  [SUMMARY] Get or create colourmap.
    <w> The ImageDisplay widget.
    <visual> The visual information.
    <xcmap> The XColormap ID.
    <depth> The depth of the colourmap.
    [RETURN] The Kcolourmap object.
*/
{
    Kcolourmap kcmap;
    Kdisplay dpy_handle;
    flag verbose;
    unsigned int num_ccells;
    Widget colourmapwinpopup;
    unsigned long *pixel_values;
    Display *dpy;
    static char function_name[] = "ImageDisplayWidget::get_dc_colourmap";

    dpy = XtDisplay (w);
    verbose = w->imageDisplay.verbose;
    if (w->imageDisplay.direct_cmap) return (w->imageDisplay.direct_cmap);
    if ( ( dpy_handle = xc_get_dpy_handle (dpy, xcmap) ) == NULL )
    {
	fprintf (stderr, "Error getting display handle\n");
	a_prog_bug (function_name);
    }
    if ( ( kcmap = kcmap_va_create (DEFAULT_COLOURMAP_NAME, 256, TRUE,
				    dpy_handle,
				    xc_alloc_colours, xc_free_colours,
				    xc_store_colours, xc_get_location2,
				    KCMAP_ATT_DIRECT_VISUAL, TRUE,
				    KCMAP_ATT_END) ) == NULL )
    {
	fprintf (stderr,
		 "Error creating Karma DirectColour colourmap for Xcmap: %p\n",
		 (void *) xcmap);
	a_prog_bug (function_name);
    }
    REGISTER_FREE_FUNC (w, kcmap_destroy, kcmap);        /*  Must be first   */
    REGISTER_FREE_FUNC (w, xc_free_handle, dpy_handle);  /*  Must be second  */
    w->imageDisplay.direct_cmap = kcmap;
    num_ccells = kcmap_get_pixels (kcmap, &pixel_values);
    if (verbose)
	fprintf (stderr, "%s: num colours for DirectColour: %u\n",
		 function_name, num_ccells);
    colourmapwinpopup = XtVaCreatePopupShell
	("directCmapwinpopup", cmapwinpopupWidgetClass, (Widget) w,
	 XkwNcolourbarVisual, visual,
	 XkwNkarmaColourmap, kcmap,
	 XkwNsimpleColourbar, True,
	 NULL);
    w->imageDisplay.cmapwinpopup_direct = colourmapwinpopup;
    return (kcmap);
}   /*  End Function get_dc_colourmap  */

static void handle_canvas (ImageDisplayWidget w, Widget canvas)
/*  [SUMMARY] Handle a canvas widget.
    <w> The ImageDisplay widget.
    <canvas> The Canvas widget.
    [RETURNS] Nothing.
*/
{
    if (canvas == NULL) return;
    XtVaSetValues (canvas,
		   XkwNsilenceUnconsumed, True,
		   NULL);
    XtAddCallback (canvas, XkwNrealiseCallback,
		   _ImageDisplay_canvas_realise_cbk, (XtPointer) w);
    ++w->imageDisplay.num_unrealised;
}   /*  End Function handle_canvas  */

static void _ImageDisplay_show_dispdata_canvas (void *info,KWorldCanvas canvas)
{
    ImageDisplayWidget w = info;

    if (w->imageDisplay.visibleCanvas == canvas)
	canvas_resize (canvas, NULL, FALSE);
    else XtVaSetValues ( (Widget) w, XkwNvisibleCanvas, canvas, NULL );
}   /*  End Function show_dispdata_canvas  */

static void _ImageDisplay_new_cbk (Widget w, XtPointer client_data,
				   XtPointer call_data)
/*  [SUMMARY] Callback to create a new ImageDisplay window.
*/
{
    unsigned int code = *(int *) call_data;
    Widget shell;

    shell = XkwImageDisplayCreateWindow (client_data, code);
    XtPopup (shell, XtGrabNone);
}   /*  End Function new_cbk   */

static void _ImageDisplay_kcmap_modify_func (Kcolourmap source, void **info)
/*  [SUMMARY] This routine registers a change in the colourmap values.
    <source> The colourmap.
    <info> A pointer to the arbitrary colourmap information pointer.
    [RETURNS] Nothing.
*/
{
    Kcolourmap dest = *info;
    unsigned int num_ccells;
    unsigned short colours[256 * 3];

    num_ccells = kcmap_read_rgb_values (source,
					256, colours, colours + 1, colours + 2,
					3);
    kcmap_write_rgb_values (dest, num_ccells,
			    colours, colours + 1, colours + 2, 3);
}   /*  End Function kcmap_modify_func  */

static void _ImageDisplay_renumber_func (ImageDisplayWidget w)
{
    Widget W = (Widget) w;
    char title[STRING_LENGTH];

    sprintf (title, "Display Window %u",
	     dispdata_get_window_index (w->imageDisplay.dispdata_window) + 1);
    XtVaSetValues (XtParent (W), XtNtitle, title, NULL);
}   /*  End Function renumber_func  */


/*  Public functions follow  */

/*EXPERIMENTAL_FUNCTION*/
void XkwImageDisplayRefresh (Widget W, flag clear)
/*  [SUMMARY] Refresh the active display canvas for an ImageDisplay widget.
    [PURPOSE] This routine will refresh the active display canvas for the
    ImageDisplay widget. If the active canvas is a stereo canvas both the left
    and right canvases are refreshed.
    <W> The widget.
    <clear> If TRUE, the canvas is cleared prior to refresh.
    [RETURNS] Nothing.
*/
{
    KWorldCanvas vc;
    flag visible;
    ImageDisplayWidget w;
    static char function_name[] = "XkwImageDisplayRefresh";

    if ( !XtIsImageDisplay (W) )
    {
	fprintf (stderr, "Invalid widget passed\n");
	a_prog_bug (function_name);
    }
    w = (ImageDisplayWidget) W;
    if ( (vc = w->imageDisplay.visibleCanvas) == NULL ) return;
    kwin_get_attributes (canvas_get_pixcanvas (vc),
			 KWIN_ATT_VISIBLE, &visible,
			 KWIN_ATT_END);
    if (!visible) return;
    if ( (vc == w->imageDisplay.pseudoCanvasLeft) ||
	 (vc == w->imageDisplay.pseudoCanvasRight) )
    {
	canvas_resize (w->imageDisplay.pseudoCanvasLeft, NULL, clear);
	canvas_resize (w->imageDisplay.pseudoCanvasRight, NULL, clear);
    }
    else if ( (vc == w->imageDisplay.directCanvasLeft) ||
	      (vc == w->imageDisplay.directCanvasRight) )
    {
	canvas_resize (w->imageDisplay.directCanvasLeft, NULL, clear);
	canvas_resize (w->imageDisplay.directCanvasRight, NULL, clear);
    }
    else if ( (vc == w->imageDisplay.trueCanvasLeft) ||
	      (vc == w->imageDisplay.trueCanvasRight) )
    {
	canvas_resize (w->imageDisplay.trueCanvasLeft, NULL, clear);
	canvas_resize (w->imageDisplay.trueCanvasRight, NULL, clear);
    }
    else canvas_resize (vc, NULL, clear);
}   /*  End Function XkwImageDisplayRefresh  */

/*EXPERIMENTAL_FUNCTION*/
KDisplayDataWindow XkwImageDisplayGetWindow (Widget W)
/*  [SUMMARY] Get/create the KDisplayDataWindow for an ImageDisplay widget.
    <W> The widget.
    [NOTE] The widget must be realised before calling this function.
    [RETURNS] The KDisplayDataWindow object.
*/
{
    KDisplayDataCanvas pc, true, direct, rgb;
    ImageDisplayWidget w;
    static char function_name[] = "XkwImageDisplayGetWindow";

    if ( !XtIsImageDisplay (W) )
    {
	fprintf (stderr, "Invalid widget passed\n");
	a_prog_bug (function_name);
    }
    w = (ImageDisplayWidget) W;
    if (w->imageDisplay.dispdata_window)
	return (w->imageDisplay.dispdata_window);
    /*  Need to create the KDisplayDataWindow  */
    pc.main = w->imageDisplay.pseudoCanvas;
    true.main = w->imageDisplay.trueCanvas;
    direct.main = w->imageDisplay.directCanvas;
    if (!pc.main && !true.main && !direct.main)
    {
	fputs ("No main canvases found (widget not realised yet?)\n", stderr);
	a_prog_bug (function_name);
    }
    XtVaGetValues (w->imageDisplay.magnifier,
		   XkwNpseudoColourCanvas, &pc.magnifier,
		   XkwNtrueColourCanvas, &true.magnifier,
		   XkwNdirectColourCanvas, &direct.magnifier,
		   NULL);
    if (!pc.magnifier && !true.magnifier && !direct.magnifier)
    {
	fputs ("No magnifier canvases found (widget not realised yet?)\n",
	       stderr);
	a_prog_bug (function_name);
    }
    XtVaGetValues (w->imageDisplay.panner,
		   XkwNpseudoColourCanvas, &pc.panner,
		   XkwNtrueColourCanvas, &true.panner,
		   XkwNdirectColourCanvas, &direct.panner,
		   NULL);
    if (!pc.panner && !true.panner && !direct.panner)
    {
	fputs ("No panner canvases found (widget not realised yet?)\n",stderr);
	a_prog_bug (function_name);
    }
    pc.renzogram_cmap = w->imageDisplay.renzoPseudoCmap;
    true.renzogram_cmap = w->imageDisplay.renzoTrueCmap;
    direct.renzogram_cmap = NULL;
    rgb = true.main ? true : direct;
    w->imageDisplay.dispdata_window =
	dispdata_create_window (pc, rgb, _ImageDisplay_show_dispdata_canvas,
				w);
    dispdata_set_window_attributes
	(w->imageDisplay.dispdata_window,
	 KDISPLAYDATA_WINDOW_ATT_SLICE_DIRECTION,
	   KDISPLAYDATA_SLICE_DIRECTION_XY,
	 KDISPLAYDATA_WINDOW_ATT_SET_ISCALE, TRUE,
	 KDISPLAYDATA_WINDOW_ATT_END);
    if (dispdata_get_window_index (w->imageDisplay.dispdata_window) > 0)
    {
	dispdata_register_window_renumber_func(w->imageDisplay.dispdata_window,
					       _ImageDisplay_renumber_func,
					       w);
	_ImageDisplay_renumber_func (w);
    }
    return (w->imageDisplay.dispdata_window);
}   /*  End Function XkwImageDisplayGetWindow  */

/*EXPERIMENTAL_FUNCTION*/
void XkwImageDisplaySetSmoothCallback (Widget w, XtPointer client_data,
				       XtPointer call_data)
/*  [SUMMARY] Ktoggle callback for double buffering for an ImageDisplay widget.
    <w> The widget the event occurred on.
    <client_data> Client data. This must point to the ImageDisplay widget.
    <call_data> Call data. This must point to the toggle state data.
    [NOTE] The ImageDisplay widget must be realised before calling this
    function.
    [RETURNS] Nothing.
*/
{
    ImageDisplayWidget top = client_data;
    flag smooth = (uaddr) call_data;
    static char function_name[] = "XkwImageDisplaySetSmoothCallback";

    if ( !XtIsImageDisplay ( (Widget) top ) )
    {
	fprintf (stderr, "Invalid widget passed\n");
	a_prog_bug (function_name);
    }
    if (top->imageDisplay.pseudoCanvas)
	kwin_set_attributes (canvas_get_pixcanvas
			     (top->imageDisplay.pseudoCanvas),
			     KWIN_ATT_DOUBLE_BUFFER, smooth,
			     KWIN_ATT_END);
    if (top->imageDisplay.trueCanvas)
	kwin_set_attributes (canvas_get_pixcanvas
			     (top->imageDisplay.trueCanvas),
			     KWIN_ATT_DOUBLE_BUFFER, smooth,
			     KWIN_ATT_END);
    if (top->imageDisplay.directCanvas)
	kwin_set_attributes (canvas_get_pixcanvas
			     (top->imageDisplay.directCanvas),
			     KWIN_ATT_DOUBLE_BUFFER, smooth,
			     KWIN_ATT_END);
    if (smooth) XkwImageDisplayRefresh ( (Widget) top, TRUE );
}   /*  End Function XkwImageDisplaySetSmoothCallback  */

void XkwImageDisplayCheckObsolete (int argc, char **argv)
{
    flag use_obsolete = FALSE;
    extern char module_name[];

    for (--argc, ++argv; argc > 0; --argc, ++argv)
        if (strcmp (*argv, "-obsolete") == 0) use_obsolete = TRUE;
    if (!use_obsolete)
    {
	fprintf (stderr,
		 "<%s> is obsolete and will be removed in the next release.\n",
		 module_name);
	fprintf (stderr, "Please use <kvis> instead.\n");
	fprintf (stderr,
		 "If you are desperate to use <%s>, use the  -obsolete  option.\n",
		 module_name);
	fprintf (stderr,
		 "Please contact rgooch@atnf.csiro.au if there is a critical feature in\n");
	fprintf (stderr, "<%s> which is not in <kvis>\n", module_name);
	exit (RV_OK);
    }
}   /*  End Function XkwImageDisplayCheckObsolete  */

Widget XkwImageDisplayCreateWindow (Widget w, unsigned int window_type)
/*  [SUMMARY] Create a new ImageDisplay window.
    <w> The parent ImageDisplay widget.
    <window_type> The type of the window.
    [RETURNS] The new shell widget.
*/
{
    Kcolourmap pc_kcmap = NULL;
    Kcolourmap renzo_pc_kcmap = NULL;
    Kcolourmap renzo_true_kcmap = NULL;
    int allowed_mask = ~0;
    Colormap xcmap;
    Colormap pc_xcmap = None;
    ImageDisplayWidget old = (ImageDisplayWidget) w;
    Widget parent = w;
    Widget shell, new, tmp;
    Screen *screen = XtScreen (w);
    char name[STRING_LENGTH];
    static char function_name[] = "XkwImageDisplayCreateWindow";

    switch (window_type)
    {
      case XkwNEW_WINDOW_SLAVE_CMAP:
	/*  Make parent for new window same as parent for old window  */
	for ( parent = XtParent (parent); parent; parent = XtParent (parent) )
	    if ( XtIsImageDisplay (parent) ) break;
	if (!parent) parent = w;
	XtVaGetValues (parent,
		       XkwNpseudoCmap, &pc_kcmap,
		       XkwNrenzoPseudoCmap, &renzo_pc_kcmap,
		       XkwNrenzoTrueCmap, &renzo_true_kcmap,
		       XtNcolormap, &xcmap,
		       XkwNallowedMask, &allowed_mask,
		       NULL );
	if ( ( tmp = XtNameToWidget (old->imageDisplay.multiCanvas,
				     "pseudoColourCanvas") ) != NULL )
	    XtVaGetValues (tmp, XtNcolormap, &pc_xcmap, NULL);
	break;
      case XkwNEW_WINDOW_INDEPENDENT_CMAP:
	/*  Make parent for new window the main application shell  */
	for ( ; XtParent (parent); parent = XtParent (parent) );
	xcmap = xtmisc_init_get_pc_colourmap (screen,
					      old->imageDisplay.cmapSize, 200,
					      FALSE, NULL);
	break;
      case XkwNEW_WINDOW_24_ONLY:
	allowed_mask &= ~XkwCanvasTypePseudoColour;
	/*  Make parent for new window the main application shell  */
	for ( ; XtParent (parent); parent = XtParent (parent) );
	XtVaGetValues (parent,
		       XtNcolormap, &xcmap,
		       NULL );
	break;
      default:
	fprintf (stderr, "Invalid window_type: %u\n", window_type);
	a_prog_bug (function_name);
	break;
    }
    sprintf ( name, "imageDisplayShell%u", dispdata_get_num_windows () );
    shell = XtVaCreatePopupShell (name, topLevelShellWidgetClass, parent,
				  XtNcolormap, xcmap,
				  NULL);
    new = XtVaCreateManagedWidget ("topForm", imageDisplayWidgetClass, shell,
				   XtNborderWidth, 0,
				   XkwNviewDatasets, True,
				   XkwNnumDatasets, 0,
				   XkwNshowQuitButton, False,
				   XkwNcreateFilepopup, False,
				   XkwNpseudoCmap, pc_kcmap,
				   XkwNrenzoPseudoCmap, renzo_pc_kcmap,
				   XkwNrenzoTrueCmap, renzo_true_kcmap,
				   XkwNallowedMask, allowed_mask,
				   XkwNpseudoXColourMap, pc_xcmap,
				   NULL);
    return (shell);
}   /*  End Function XkwImageDisplayCreateWindow   */
