/*LINTLIBRARY*/
/*  MomentGenerator.c

    This code provides a moment generator control widget for Xt.

    Copyright (C) 1996-2004  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This file contains all routines needed for a moment generator control
    widget for Xt.


    Written by      Richard Gooch   11-SEP-1996

    Updated by      Richard Gooch   14-SEP-1996

    Updated by      Richard Gooch   19-SEP-1996: Changed to special function
  call interface for setting new array.

    Updated by      Richard Gooch   26-SEP-1996: Changed to implement the new
  FITS-style co-ordinate handling, where dimension co-ordinates range from 0
  to length - 1.

    Updated by      Richard Gooch   30-SEP-1996: Fixed bug in initialisation
  where cube_arr and cube_ap were not initialised.

    Updated by      Richard Gooch   14-OCT-1996: Copy "OBSRA" and "OBSDEC" from
  cube to moment maps.

    Updated by      Richard Gooch   14-NOV-1996: Removed upper clip level.

    Updated by      Richard Gooch   15-JAN-1997: Switched to
  <wcs_astro_transform3>.

    Updated by      Richard Gooch   15-FEB-1997: No longer set 0th moment map
  to TOOBIG if below clip value, set to 0 instead.

    Updated by      Richard Gooch   22-JUN-1997: Switched to function
  <iarray_get_associated_wcs_astro> instead. Added warning when chromatic
  aberration becomes significant. Added start and end sliders (after v1.6.13
  beta release).

    Updated by      Richard Gooch   17-MAY-1998: Fixed bug where "BUNIT" was
  copied from cube for 1st moment: should be frequency/velocity name. Added
  traps for illegal data.

    Updated by      Richard Gooch   26-MAY-1998: Compute min&max and store in
  header.

    Updated by      Richard Gooch   17-SEP-1998: Took account of some CONST
  declarations in <ex> package.

    Updated by      Richard Gooch   12-AUG-1999: Support Z axis which is a sky
  axis.

    Updated by      Richard Gooch   11-JUN-2001: Copy over "CROTAn" keywords.

    Updated by      Richard Gooch   6-DEC-2001: Took account of interface
  change to <wcs_astro_transform*>.

    Updated by      Richard Gooch   6-AUG-2003: Made use of new experimental
  function <c_unregister_callback2>.

    Updated by      Richard Gooch   6-OCT-2003: Bug fix in function
  <XkwMomentGeneratorNewArray>: <<cube_arr>> was not set to NULL when a NULL
  <<array>> passed in.

    Updated by      Richard Gooch   15-MAY-2004: Made use of new experimental
  function <wcs_astro_has_chromatic_aberration>.

    Updated by      Richard Gooch   20-JUN-2004: Moved moment computing code to
  <iarray_moment_compute>.

    Last updated by Richard Gooch   22-JUN-2004: Moved more code to
  source/lib/karma/packages/iarray/moment.c


*/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <sys/wait.h>
#include <signal.h>
#include <X11/Xos.h>
#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/Label.h>
#include <X11/Xaw/Box.h>
#include <X11/Xaw/Dialog.h>
#ifndef X11
#  define X11
#endif
#include <karma.h>
#include <karma_iarray.h>
#include <karma_xtmisc.h>
#include <karma_ex.h>
#include <karma_c.h>
#include <Xkw/MomentGeneratorP.h>
#include <Xkw/Ktoggle.h>
#include <Xkw/ExclusiveMenu.h>
#include <Xkw/SimpleSlider.h>


/*  Private functions  */

STATIC_FUNCTION (void _MomentGenerator_Initialise,
		 (Widget request, Widget new) );
STATIC_FUNCTION (void _MomentGenerator_iarr_destroy_callback,
		 (iarray arr, MomentGeneratorWidget top) );
STATIC_FUNCTION (void _MomentGenerator_apply_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );


#define offset(field) XtOffset(MomentGeneratorWidget, momentGenerator.field)
#define XkwRDouble "Double"

static XtResource MomentGeneratorResources[] = 
{
    {XkwNverbose, XkwCVerbose, XtRBool, sizeof (Bool),
     offset (verbose), XtRImmediate, FALSE},
    {XkwNmom0Array, XkwCArray, XtRPointer, sizeof (XtPointer),
     offset (mom0Array), XtRImmediate, NULL},
    {XkwNmom1Array, XkwCArray, XtRPointer, sizeof (XtPointer),
     offset (mom1Array), XtRImmediate, NULL},
    {XkwNmomentCallback, XtCCallback, XtRCallback, sizeof (caddr_t),
     offset (momentCallback), XtRCallback, (caddr_t) NULL},
};

#undef offset

MomentGeneratorClassRec momentGeneratorClassRec = 
{
  {
      /* CoreClassPart */
    (WidgetClass)&topLevelShellClassRec,          /* superclass */
    "MomentGenerator",                            /* class_name */
    sizeof (MomentGeneratorRec),                  /* widget_size */
    NULL,                                         /* class_initialise */
    NULL,                                         /* class_part_initialise */
    FALSE,                                        /* class_init */
    (XtInitProc) _MomentGenerator_Initialise,     /* initialise */
    NULL,                                         /* initialise_hook */
    XtInheritRealize,                             /* realise */
    NULL,                                         /* actions */
    0,                                            /* num_actions */
    MomentGeneratorResources,                     /* resources */
    XtNumber (MomentGeneratorResources),          /* num_resources */
    NULLQUARK,                                    /* xrm_class */
    TRUE,                                         /* compress_motion */
    TRUE,                                         /* compress_exposure */
    TRUE,                                         /* compress_enterleave */
    TRUE,                                         /* visible_interest */
    NULL,                                         /* destroy */
    XtInheritResize,                              /* resize */
    NULL,                                         /* expose */
    NULL,                                         /* set_values */
    NULL,                                         /* set_values_hook */
    XtInheritSetValuesAlmost,                     /* set_values_almost */
    NULL,                                         /* get_values_hook */
    NULL,                                         /* accept_focus */
    XtVersion,                                    /* version */
    NULL,                                         /* callback_private */
    NULL,                                         /* tm_translations */
    NULL,
    NULL,
    NULL,
  },
  {     /* CompositeClassPart */
      /* geometry_manager */	XtInheritGeometryManager,
      /* change_managed	  */	XtInheritChangeManaged,
      /* insert_child	  */	XtInheritInsertChild,
      /* delete_child	  */	XtInheritDeleteChild,
      /* extension	  */	NULL
  },
  { /* Shell */
      /* extension	  */	NULL
  },
  { /* WMShell */
      /* extension	  */	NULL
  },
  { /* VendorShell */
      /* extension	  */	NULL
  },
  { /* TopLevelShell */
      /* extension	  */	NULL
  },
  {  /* MomentGeneratorClassPart */
      0                   /* empty */
  }
};

WidgetClass momentGeneratorWidgetClass = (WidgetClass)&momentGeneratorClassRec;


static void _MomentGenerator_Initialise (Widget Request, Widget New)
{
    /*MomentGeneratorWidget request = (MomentGeneratorWidget) Request;*/
    MomentGeneratorWidget new = (MomentGeneratorWidget) New;
    Widget form, close_btn, w;
    /*static char function_name[] = "MomentGeneratorWidget::Initialise";*/

    new->momentGenerator.cube_arr = NULL;
    new->momentGenerator.iarr_destroy_callback = NULL;
    new->momentGenerator.mom1_algorithm=IARRAY_MOMENT1_ALGORITHM_WEIGHTED_MEAN;
    form = XtVaCreateManagedWidget ("form", formWidgetClass, New,
				    XtNborderWidth, 0,
				    NULL);
    w = XtVaCreateManagedWidget ("closeButton", commandWidgetClass, form,
				 XtNlabel, "Close",
				 XtNtop, XtChainTop,
				 XtNbottom, XtChainTop,
				 XtNleft, XtChainLeft,
				 XtNheight, 20,
				 NULL);
    close_btn = w;
    XtAddCallback (close_btn, XtNcallback, xtmisc_popdown_cbk, New);
    w = XtVaCreateManagedWidget ("button", commandWidgetClass, form,
				   XtNlabel, "Apply Parameters",
				   XtNfromHoriz, w,
				   NULL);
    XtAddCallback (w, XtNcallback, _MomentGenerator_apply_cbk, New);
    w = XtVaCreateManagedWidget ("menuButton", exclusiveMenuWidgetClass, form,
				 XtNmenuName, "mom1AlgorithmMenu",
				 XtNfromVert, w,
				 XkwNchoiceName, "1st Moment Algorithm",
				 XkwNnumItems, IARRAY_NUM_MOMENT_CHOICES,
				 XkwNitemStrings,
				     iarray_moment_algorithm_alternatives,
				 XkwNvaluePtr,
				 &new->momentGenerator.mom1_algorithm,
				 NULL);
    w = XtVaCreateManagedWidget ("minLabel", labelWidgetClass, form,
				 XtNfromVert, w,
				 XtNborderWidth, 0,
				 XtNlabel, "Cube min:",
				 XtNresizable, TRUE,
				 NULL);
    new->momentGenerator.cube_min_label = w;
    w = XtVaCreateManagedWidget ("maxLabel", labelWidgetClass, form,
				 XtNfromVert, w,
				 XtNborderWidth, 0,
				 XtNlabel, "Cube max:",
				 XtNresizable, TRUE,
				 NULL);
    new->momentGenerator.cube_max_label = w;
    w = XtVaCreateManagedWidget ("minLabel", labelWidgetClass, form,
				 XtNfromVert, w,
				 XtNborderWidth, 0,
				 XtNlabel, "0th moment min:",
				 XtNresizable, TRUE,
				 NULL);
    new->momentGenerator.sum_min_label = w;
    w = XtVaCreateManagedWidget ("maxLabel", labelWidgetClass, form,
				 XtNfromVert, w,
				 XtNborderWidth, 0,
				 XtNlabel, "0th moment max:",
				 XtNresizable, TRUE,
				 NULL);
    new->momentGenerator.sum_max_label = w;
    w = XtVaCreateManagedWidget ("lowerClipLevel", dialogWidgetClass, form,
				 XtNlabel, "Lower Clip Level",
				 XtNfromVert, w,
				 XtNvalue, "5e-3",
				 NULL);
    new->momentGenerator.lower_clip_dlg = w;
    w = XtVaCreateManagedWidget ("sumClipLevel", dialogWidgetClass, form,
				 XtNlabel, "Sum Clip Level",
				 XtNfromVert,
				 new->momentGenerator.sum_max_label,
				 XtNfromHoriz, w,
				 XtNvalue, "5e-2",
				 NULL);
    new->momentGenerator.sum_clip_dlg = w;
    w = XtVaCreateManagedWidget ("startSlider", simpleSliderWidgetClass, form,
				 XtNlabel, "Start Channel",
				 XtNfromVert, w,
				 XtNwidth, 310,
				 XkwNmodifier, 1,
				 XkwNshowRange, TRUE,
				 NULL);
    new->momentGenerator.start_sld = w;
    w = XtVaCreateManagedWidget ("endSlider", simpleSliderWidgetClass, form,
				 XtNlabel, "End Channel",
				 XtNfromVert, w,
				 XtNwidth, 310,
				 XkwNmodifier, 1,
				 XkwNshowRange, TRUE,
				 NULL);
    new->momentGenerator.end_sld = w;
}   /*  End Function Initialise  */

static void _MomentGenerator_iarr_destroy_callback (iarray arr,
						    MomentGeneratorWidget top)
/*  [PURPOSE] This routine will register the destruction of the Intelligent
    Array.
    <arr> The Intelligent Array.
    <top> The MomentGenerator widget.
    [RETURNS] Nothing.
*/
{
    top->momentGenerator.cube_arr = NULL;
    top->momentGenerator.iarr_destroy_callback = NULL;
    iarray_dealloc (top->momentGenerator.mom0Array);
    iarray_dealloc (top->momentGenerator.mom1Array);
    top->momentGenerator.mom0Array = NULL;
    top->momentGenerator.mom1Array = NULL;
}   /*  End Function iarr_destroy_callback  */

static void _MomentGenerator_apply_cbk (Widget w, XtPointer client_data,
					XtPointer call_data)
/*  [SUMMARY] Moment computation callback.
    <w> The widget the event occurred on.
    <client_data> Client data. The MomentGeneratorWidget.
    <call_data> Call data.
    [RETURNS] Nothing.
*/
{
    int start_channel, end_channel;
    float lower_clip, sum_clip;
    MomentGeneratorWidget top = (MomentGeneratorWidget) client_data;
    float mom0_min, mom0_max;
    char *clip_str;
    CONST char *p;
    char txt[STRING_LENGTH];
    /*static char function_name[] = "MomentGeneratorWidget::apply_cbk";*/

    if (top->momentGenerator.cube_arr == NULL) return;
    XtVaGetValues (top->momentGenerator.lower_clip_dlg,
		   XtNvalue, &clip_str,
		   NULL);
    lower_clip = ex_float (clip_str, &p);
    XtVaGetValues (top->momentGenerator.sum_clip_dlg,
		   XtNvalue, &clip_str,
		   NULL);
    sum_clip = ex_float (clip_str, &p);
    XtVaGetValues (top->momentGenerator.start_sld,
		   XtNvalue, &start_channel,
		   NULL);
    XtVaGetValues (top->momentGenerator.end_sld,
		   XtNvalue, &end_channel,
		   NULL);
    fprintf (stderr, "Computing moment maps...");
    iarray_moment_compute (top->momentGenerator.mom0Array,
			   top->momentGenerator.mom1Array,
			   top->momentGenerator.cube_arr,
			   lower_clip, sum_clip,
			   top->momentGenerator.mom1_algorithm,
			   start_channel, end_channel + 1,
			   &mom0_min, &mom0_max);
    fprintf (stderr, "done\n");
    sprintf (txt, "0th moment min: %e", mom0_min);
    XtVaSetValues (top->momentGenerator.sum_min_label,
		   XtNlabel, txt,
		   NULL);
    sprintf (txt, "0th moment max: %e", mom0_max);
    XtVaSetValues (top->momentGenerator.sum_max_label,
		   XtNlabel, txt,
		   NULL);
    XtCallCallbacks ( (Widget) top, XkwNmomentCallback, NULL );
}   /*  End Function apply_cbk  */


/*  Public functions follow  */

/*EXPERIMENTAL_FUNCTION*/
void XkwMomentGeneratorNewArray (Widget W, iarray array, double min,double max)
/*  [SUMMARY] Register new array.
    <W> The MomentGenerator widget.
    <array> The new array. This may be NULL.
    <min> The minimum data value in the array.
    <max> The maximum data value in the array. If this is less than <<min>>
    then the minimum and maximum values are computed.
    [RETURNS] Nothing.
*/
{
    iarray mom0_arr, mom1_arr;
    unsigned int zlen;
    MomentGeneratorWidget w = (MomentGeneratorWidget) W;
    char txt[STRING_LENGTH];
    /*static char function_name[] = "XkwMomentGeneratorNewArray";*/

    /*  Clean up any old arrays  */
    w->momentGenerator.cube_arr = NULL;
    c_unregister_callback2 (w->momentGenerator.iarr_destroy_callback);
    w->momentGenerator.iarr_destroy_callback = NULL;
    iarray_dealloc2 (w->momentGenerator.mom0Array);
    w->momentGenerator.mom0Array = NULL;
    iarray_dealloc2 (w->momentGenerator.mom1Array);
    w->momentGenerator.mom1Array = NULL;
    /*  Process new cube  */
    if (array == NULL) return;
    if ( !iarray_moment_create (&mom0_arr, &mom1_arr, array) ) return;
    w->momentGenerator.cube_arr = array;
    if (max < min) iarray_min_max (array, CONV_CtoR_REAL, &min, &max);
    sprintf (txt, "Cube min: %e", min);
    XtVaSetValues (w->momentGenerator.cube_min_label,
		   XtNlabel, txt,
		   NULL);
    sprintf (txt, "Cube max: %e", max);
    XtVaSetValues (w->momentGenerator.cube_max_label,
		   XtNlabel, txt,
		   NULL);
    w->momentGenerator.mom0Array = mom0_arr;
    w->momentGenerator.mom1Array = mom1_arr;
    w->momentGenerator.iarr_destroy_callback = 
	iarray_register_destroy_func
	(array, ( flag (*) () ) _MomentGenerator_iarr_destroy_callback, w);
    zlen = iarray_dim_length (array, 0);
    XtVaSetValues (w->momentGenerator.start_sld,
		   XtNvalue, 0,
		   XkwNmaximum, zlen - 1,
		   NULL);
    XtVaSetValues (w->momentGenerator.end_sld,
		   XtNvalue, zlen - 1,
		   XkwNmaximum, zlen - 1,
		   NULL);
}   /*  End Function XkwMomentGeneratorNewArray  */
