/*LINTLIBRARY*/
/*  TracePopup.c

    This code provides a trace display widget for Xt.

    Copyright (C) 1996-2004  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This file contains all routines needed for a trace display widget for Xt.


    Written by      Richard Gooch   15-SEP-1996

    Updated by      Richard Gooch   16-SEP-1996

    Updated by      Richard Gooch   19-SEP-1996: Changed to special function
  call interface for setting new array.

    Updated by      Richard Gooch   9-OCT-1996: Store trace details and use
  refresh function. Added print button.

    Updated by      Richard Gooch   5-NOV-1996: Added track display.

    Updated by      Richard Gooch   6-NOV-1996: Added horizontal pixel
  co-ordinate in track display and put both track dislays on their own line.

    Updated by      Richard Gooch   11-NOV-1996: Only refresh canvas if visible

    Updated by      Richard Gooch   30-DEC-1996: Possible bugfix in
  <XkwTracePopupNewArray> when NULL array passed.

    Updated by      Richard Gooch   27-JAN-1997: Changed interface to
  <XkwTracePopupNewArray> function.

    Updated by      Richard Gooch   28-JAN-1997: Changed back to previous
  interface to <XkwTracePopupNewArray> function: now construct restriction
  information from co-ordinate indices.

    Updated by      Richard Gooch   31-JAN-1997: Added drag&zoom functionality.

    Updated by      Richard Gooch   1-FEB-1997: Made use of
  <wcs_astro_get_associated> function.

    Updated by      Richard Gooch   11-FEB-1997: Added "Unzoom" and "Smooth"
  controls.

    Updated by      Richard Gooch   21-FEB-1997: Added "Auto Zoom" toggle.

    Updated by      Richard Gooch   23-FEB-1997: Added "Axis Labels" button.

    Updated by      Richard Gooch   24-FEB-1997: Added simple clipping.

    Updated by      Richard Gooch   5-MAR-1997: Fixed bug when Y scale range
  was set to 0.

    Updated by      Richard Gooch   24-MAR-1997: Fixed bug drawing zero line
  when autoscaling is enabled: line was drawn before scale was changed.

    Updated by      Richard Gooch   27-MAR-1997: Worked around bug in GCC 2.7.2
  on an alpha.

    Updated by      Richard Gooch   10-APR-1997: Copy trace from array to
  internal buffer to deal with problems when data array changes (i.e. scrolling
  buffer).

    Updated by      Richard Gooch   13-APR-1997: Took account of change to
  <ds_format_unit> function.

    Updated by      Richard Gooch   22-APR-1997: Fixed bug in
  <XkwTracePopupNewArray>: trace buffer was being set to NULL.

    Updated by      Richard Gooch   4-JUN-1997: Supported non-float arrays.

    Updated by      Richard Gooch   7-JUN-1997: Add profile save.

    Updated by      Richard Gooch   19-JUN-1997: Switched to <ds_format_value>.

    Updated by      Richard Gooch   26-JUL-1997: Attempt to preserve horizontal
  range if user zooms in and then a new profile is given with
  <XkwTracePopupShowTrace>.

    Updated by      Richard Gooch   31-JUL-1997: Made use of new editor context
  in <kwin> package for zoom box. Created <XkwTracePopupShowTraces> function.

    Updated by      Richard Gooch   1-AUG-1997: Created function
  <XkwTracePopupShowBoxedTrace>.

    Updated by      Richard Gooch   4-AUG-1997: Added rescaling of "JY/PIX" and
  "JY/BEAM" to "JY" when appropriate.

    Updated by      Richard Gooch   6-AUG-1997: Added test to flux rescaling
  to ensure the two axes are sky co-ordinates (don't want one being frequency).
  Use canvas X & Y labels for track display and save format rather than getting
  from iarray (to capture flux rescaling).

    Updated by      Richard Gooch   12-AUG-1997: Fixed bug in Initialise:
  array was not initialised.

    Updated by      Richard Gooch   16-AUG-1997: Forced trace colour to be
  black if drawing to PostScript.

    Updated by      Richard Gooch   20-AUG-1997: Made use of new functions for
  testing if axes are longitude or latitude.

    Updated by      Richard Gooch   16-SEP-1997: More robust initialisation of
  trace_dim. Apply data scaling everywhere.

    Updated by      Richard Gooch   25-NOV-1997: Fixed bug in <geom_func> to
  ensure sensible boundaries passed to <canvas_set_attributes>.

    Updated by      Richard Gooch   17-DEC-1997: Improved auto-zooming.

    Updated by      Richard Gooch   22-FEB-1998: Switched to
  <XkwDressingControlAddCanvas>.

    Updated by      Richard Gooch   27-FEB-1998: Switched to overlay menu.

    Updated by      Richard Gooch   17-MAY-1998: Trap illegal values.

    Updated by      Richard Gooch   12-AUG-1998: Now update cursor position on
  any canvas event.

    Updated by      Richard Gooch   14-OCT-1998: Handle negative data scaling.

    Updated by      Richard Gooch   13-MAY-2000: Show multi-colour trace even
  for PostScript output.

    Updated by      Richard Gooch   14-JUN-2000: Do not show zero line if it
  falls outside canvas boundary.

    Updated by      Richard Gooch   15-JUN-2000: Implemented 'i' and 'o' zoom
  key bindings. Moved profile plotting into new experimental <profile>
  package. Added support for "join" and "steps" profile display modes.

    Updated by      Richard Gooch   20-JUN-2000: Took account of rename of
  <profile> package to <graph1d>. Created function <XkwTracePopupInvalidate>.

    Updated by      Richard Gooch   11-JUL-2000: Added #include <karma_kcmap.h>

    Updated by      Richard Gooch   22-SEP-2000: Took account of interface
  change in <ds_format_value>.

    Updated by      Richard Gooch   8-MAR-2001: Make calls to function
  <canvas_use_astro_transform> when canvas is realised and when a new array is
  provided instead of during refresh. This ensures projection information is
  up-to-date before the precompute callbacks for the canvas are called.

    Updated by      Richard Gooch   9-MAR-2001: Added calls to <canvas_specify>
  to clear axis labels when new array is provided or old one is destroyed.

    Updated by      Richard Gooch   10-MAR-2001: Fixed memory leak: function
  <iarr_destroy_callback> was dropping trace buffer without freeing. Fixed
  multiple traces so that output is not blank if at least one input is not
  blank.

    Updated by      Richard Gooch   11-MAR-2001: Set canvas restrictions based
  on midpoint of traces rather than first trace. Added filename, position and
  number of traces to saved ASCII profile output.

    Updated by      Richard Gooch   13-MAR-2001: Always report number of
  profiles used to produce output profile in ASCII output.

    Updated by      Richard Gooch   20-MAR-2001: Created Destroy method.

    Updated by      Richard Gooch   31-MAY-2001: Changed from <kwin_resize> to
  <kwin_clear> in <XkwTracePopupNewArray>.

    Updated by      Richard Gooch   16-AUG-2001: Made use of function
  <canvas_warp_pointer>.

    Updated by      Richard Gooch   20-AUG-2001: Fixed bugs in <do_zoom> when
  canvas boundaries exceeded. Only warp pointer for zoom in and zoom out.

    Updated by      Richard Gooch   4-SEP-2001: Discard events with key presses
  in <geom_func>. Created <_TracePopup_size_control_func> to smooth data and
  determine min&max, so that canvas range can be safely set.

    Updated by      Richard Gooch   10-NOV-2001: Fixed bug introduced in
  previous change which had inconsistent calculation of <<y_arr>> start.

    Updated by      Richard Gooch   4-DEC-2001: Made use of new functions
  <canvas_zoom_handle_mouse> and <canvas_zoom_handle_keys>.

    Updated by      Richard Gooch   6-DEC-2001: Took account of interface
  change to <wcs_astro_transform*>.

    Updated by      Richard Gooch   22-DEC-2001: Bug fix in function
  <canvas_realise_cbk>: was using uninitialised background pixel value.

    Updated by      Richard Gooch   23-DEC-2001: Bug fix in function
  <canvas_realise_cbk>: check for negative (undefined) trace_dim.

    Updated by      Richard Gooch   3-JAN-2002: Bug fix in function
  <refresh_func>: trap for zero-length arrays, add <<recompute_size>> flag.

    Updated by      Richard Gooch   12-FEB-2002: Created <XkwTracePopupRefresh>

    Updated by      Richard Gooch   4-JUL-2002: Moved beam correction code to
  new <iarray_get_beam_correction> function.

    Updated by      Richard Gooch   7-DEC-2002: Do nothing in function
  <XkwTracePopupNewArray> if array is the same.

    Updated by      Richard Gooch   25-DEC-2002: Made use of new KViewableGraph
  class.

    Updated by      Richard Gooch   26-DEC-2002: Fixed bug in <save_cbk>: was
  not using restriction information when converting co-ordinates.

    Updated by      Richard Gooch   27-DEC-2002: Made use of new
  KOrthogonalProfile class.

    Updated by      Richard Gooch   28-DEC-2002: Realise widget in
  <XkwTracePopupNewArray> to ensure canvas is available.

    Updated by      Richard Gooch   20-JAN-2003: Made use of new functions
  <graph1d_test_if_zoomed> and <graph1d_unzoom>.

    Updated by      Richard Gooch   22-JAN-2003: Switched to renamed
  <graph1d_test_if_zoomed_horizontally>.

    Updated by      Richard Gooch   6-FEB-2003: Set draw mode in
  <canvas_realise_cbk>.

    Updated by      Richard Gooch   3-MAR-2003: Fall back to function
  <canvas_track_compute> if <graph1d_orthogonal_track_compute> fails.

    Updated by      Richard Gooch   28-MAY-2003: Replaced smooth toggle with
  menu.

    Updated by      Richard Gooch   6-AUG-2003: Made use of new experimental
  function <c_unregister_callback2>.

    Updated by      Richard Gooch   30-DEC-2003: Renamed "Save Profile" button
  to "Save Data".

    Updated by      Richard Gooch   16-MAY-2004: Took account of API changes to
  <graph1d> functions.

    Updated by      Richard Gooch   17-MAY-2004: Renamed "Auto Zoom" to
  "Auto V Zoom".

    Last updated by Richard Gooch   15-AUG-2004: Blank track labels when
  pointer leaves window.


*/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <sys/wait.h>
#include <signal.h>
#include <errno.h>
#include <X11/Xos.h>
#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/Label.h>
#include <X11/Xaw/Box.h>
#include <X11/Xaw/Dialog.h>
#ifndef X11
#  define X11
#endif
#include <karma.h>
#include <k_event_codes.h>
#include <karma_graph1d.h>
#include <karma_iarray.h>
#include <karma_xtmisc.h>
#include <karma_kcmap.h>
#include <karma_wcs.h>
#include <karma_ds.h>
#include <karma_ex.h>
#include <karma_ic.h>
#include <karma_ch.h>
#include <karma_m.h>
#include <karma_a.h>
#include <karma_c.h>
#include <Xkw/TracePopupP.h>
#include <Xkw/Canvas.h>
#include <Xkw/Postscript.h>
#include <Xkw/Ktoggle.h>
#include <Xkw/OverlayMenu.h>
#include <Xkw/Dialogpopup.h>
#include <Xkw/ExclusiveMenu.h>


/*  Private functions  */
STATIC_FUNCTION (void _TracePopup_Initialise,
		 (Widget request, Widget new) );
STATIC_FUNCTION (void _TracePopup_Destroy, (Widget W) );
STATIC_FUNCTION (void _TracePopup_iarr_destroy_callback,
		 (iarray arr, TracePopupWidget top) );
STATIC_FUNCTION (void _TracePopup_canvas_realise_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (flag _TracePopup_position_func,
		 (KWorldCanvas canvas, double x, double y,
		  unsigned int event_code, void *e_info,
		  void **f_info, double x_lin, double y_lin,
		  int x_pix, int y_pix) );
STATIC_FUNCTION (void _TracePopup_unzoom_cbk, (Widget w, XtPointer client_data,
					       XtPointer call_data) );
STATIC_FUNCTION (void _TracePopup_smooth_mode_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void _TracePopup_auto_zoom_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void _TracePopup_save_cbk, (Widget w, XtPointer client_data,
					     XtPointer call_data) );
STATIC_FUNCTION (void _TracePopup_draw_mode_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );


#define offset(field) XtOffset(TracePopupWidget, tracePopup.field)
#define XkwRDouble "Double"

static XtResource TracePopupResources[] = 
{
    {XkwNverbose, XkwCVerbose, XtRBool, sizeof (Bool),
     offset (verbose), XtRImmediate, FALSE},
    {XkwNcanvasVisual, XtCVisual, XtRVisual, sizeof (Visual *),
     offset (canvasVisual), XtRImmediate, CopyFromParent},
    {XkwNworldCanvas, XkwCWorldCanvas, XtRPointer, sizeof (XtPointer),
     offset (worldCanvas), XtRImmediate, NULL},
    {XkwNrealiseCallback, XtCCallback, XtRCallback, sizeof (caddr_t),
     offset (realiseCallback), XtRCallback, (caddr_t) NULL},
    {XkwNkarmaColourmap, XkwCKarmaColourmap, XtRPointer, sizeof (XtPointer),
     offset (karmaColourmap), XtRImmediate, (caddr_t) NULL},
};
#undef offset

TracePopupClassRec tracePopupClassRec = 
{
  {
      /* CoreClassPart */
    (WidgetClass) &topLevelShellClassRec,     /* superclass */
    "TracePopup",                             /* class_name */
    sizeof (TracePopupRec),                   /* widget_size */
    NULL,                                     /* class_initialise */
    NULL,                                     /* class_part_initialise */
    FALSE,                                    /* class_init */
    (XtInitProc) _TracePopup_Initialise,      /* initialise */
    NULL,                                     /* initialise_hook */
    XtInheritRealize,                         /* realise */
    NULL,                                     /* actions */
    0,                                        /* num_actions */
    TracePopupResources,                      /* resources */
    XtNumber (TracePopupResources),           /* num_resources */
    NULLQUARK,                                /* xrm_class */
    TRUE,                                     /* compress_motion */
    TRUE,                                     /* compress_exposure */
    TRUE,                                     /* compress_enterleave */
    TRUE,                                     /* visible_interest */
    _TracePopup_Destroy,                      /* destroy */
    XtInheritResize,                          /* resize */
    NULL,                                     /* expose */
    NULL,                                     /* set_values */
    NULL,                                     /* set_values_hook */
    XtInheritSetValuesAlmost,                 /* set_values_almost */
    NULL,                                     /* get_values_hook */
    NULL,                                     /* accept_focus */
    XtVersion,                                /* version */
    NULL,                                     /* callback_private */
    NULL,                                     /* tm_translations */
    NULL,
    NULL,
    NULL,
  },
  {     /* CompositeClassPart */
      /* geometry_manager */	XtInheritGeometryManager,
      /* change_managed	  */	XtInheritChangeManaged,
      /* insert_child	  */	XtInheritInsertChild,
      /* delete_child	  */	XtInheritDeleteChild,
      /* extension	  */	NULL
  },
  { /* Shell */
      /* extension	  */	NULL
  },
  { /* WMShell */
      /* extension	  */	NULL
  },
  { /* VendorShell */
      /* extension	  */	NULL
  },
  { /* TopLevelShell */
      /* extension	  */	NULL
  },
  {  /* TracePopupClassPart */
      0                   /* empty */
  }
};

WidgetClass tracePopupWidgetClass = (WidgetClass) &tracePopupClassRec;


static void _TracePopup_Initialise (Widget Request, Widget New)
{
    /*TracePopupWidget request = (TracePopupWidget) Request;*/
    TracePopupWidget new = (TracePopupWidget) New;
    Widget form, w, print_btn, cnv, upper_widget;
    Widget save_dialog_popup;
    /*static char function_name[] = "TracePopupWidget::Initialise";*/

    new->tracePopup.worldCanvas = NULL;
    new->tracePopup.pixcanvas = NULL;
    new->tracePopup.array = NULL;
    new->tracePopup.iarr_destroy_callback = NULL;
    new->tracePopup.do_unzoom = FALSE;
    new->tracePopup.profile = NULL;
    xtmisc_set_icon (New, ic_write_kplot_icon);
    w = XtVaCreatePopupShell ("savePopup", dialogpopupWidgetClass, New,
			      XtNlabel, "Filename         ",
			      XkwNextension, "",
			      NULL);
    save_dialog_popup = w;
    XtAddCallback (w, XtNcallback, _TracePopup_save_cbk, New);
    form = XtVaCreateManagedWidget ("form", formWidgetClass, New,
				    XtNborderWidth, 0,
				    NULL);
    w = XtVaCreateManagedWidget ("closeButton", commandWidgetClass, form,
				 XtNlabel, "Close",
				 XtNtop, XtChainTop,
				 XtNbottom, XtChainTop,
				 XtNleft, XtChainLeft,
				 XtNheight, 20,
				 NULL);
    XtAddCallback (w, XtNcallback, xtmisc_popdown_cbk, New);
    w = XtVaCreateManagedWidget ("toggle", ktoggleWidgetClass, form,
				 XtNlabel, "Auto V Zoom",
				 XtNtop, XtChainTop,
				 XtNbottom, XtChainTop,
				 XtNfromHoriz, w,
				 XtNheight, 20,
				 XkwNcrosses, False,
				 NULL);
    XtAddCallback (w, XtNcallback, _TracePopup_auto_zoom_cbk, New);
    w = XtVaCreateManagedWidget ("button", commandWidgetClass, form,
				 XtNlabel, "Unzoom",
				 XtNtop, XtChainTop,
				 XtNbottom, XtChainTop,
				 XtNfromHoriz, w,
				 XtNheight, 20,
				 NULL);
    XtAddCallback (w, XtNcallback, _TracePopup_unzoom_cbk, New);
    w = XtVaCreateManagedWidget ("traceSmoothMenu", exclusiveMenuWidgetClass,
				 form,
				 XtNmenuName, "traceSmoothMenu",
				 XkwNchoiceName, "Smooth",
				 XtNtop, XtChainTop,
				 XtNbottom, XtChainTop,
				 XtNfromHoriz, w,
				 XkwNnumItems, GRAPH1D_NUM_SMOOTH_MODES,
				 XkwNitemStrings, graph1d_smooth_mode_choices,
				 NULL);
    XtAddCallback (w, XkwNselectCallback, _TracePopup_smooth_mode_cbk, New);
    upper_widget = w;
    w = XtVaCreateManagedWidget ("button", commandWidgetClass, form,
				 XtNlabel, "Print",
				 XtNtop, XtChainTop,
				 XtNbottom, XtChainTop,
				 XtNfromVert, upper_widget,
				 XtNheight, 20,
				 NULL);
    print_btn = w;
    w = XtVaCreateManagedWidget ("overlayMenu", overlayMenuWidgetClass, form,
				 XtNtop, XtChainTop,
				 XtNbottom, XtChainTop,
				 XtNfromVert, upper_widget,
				 XtNfromHoriz, w,
				 XtNheight, 20,
				 NULL);
    new->tracePopup.overlay_menu = w;
    w = XtVaCreateManagedWidget ("button", commandWidgetClass, form,
				 XtNlabel, "Save Data",
				 XtNtop, XtChainTop,
				 XtNbottom, XtChainTop,
				 XtNfromVert, upper_widget,
				 XtNfromHoriz, w,
				 XtNheight, 20,
				 NULL);
    XtAddCallback (w, XtNcallback, xtmisc_popup_cbk, save_dialog_popup);
    w = XtVaCreateManagedWidget ("traceStyleMenu", exclusiveMenuWidgetClass,
				 form,
				 XtNmenuName, "traceStyleMenu",
				 XkwNchoiceName, "Style",
				 XtNtop, XtChainTop,
				 XtNbottom, XtChainTop,
				 XtNfromVert, upper_widget,
				 XtNfromHoriz, w,
				 XkwNnumItems, GRAPH1D_NUM_DRAW_STYLES,
				 XkwNitemStrings, graph1d_draw_style_choices,
				 XkwNvaluePtr, &new->tracePopup.draw_mode,
				 NULL);
    XtAddCallback (w, XkwNselectCallback, _TracePopup_draw_mode_cbk, New);
    w = XtVaCreateManagedWidget ("trackLabel0", labelWidgetClass, form,
				 XtNlabel,
				     "Track Output                                         ",
				 XtNtop, XtChainTop,
				 XtNbottom, XtChainTop,
				 XtNfromVert, w,
				 XtNheight, 20,
				 NULL);
    new->tracePopup.track_label0 = w;
    w = XtVaCreateManagedWidget ("trackLabel1", labelWidgetClass, form,
				 XtNlabel,
				     "Track Output                                         ",
				 XtNtop, XtChainTop,
				 XtNbottom, XtChainTop,
				 XtNfromVert, w,
				 XtNheight, 20,
				 NULL);
    new->tracePopup.track_label1 = w;
    w = XtVaCreateManagedWidget ("pseudoColourCanvas", canvasWidgetClass, form,
				 XtNfromVert, w,
				 XtNleft, XtChainLeft,
				 XtNright, XtChainRight,
				 XtNwidth, 432,
				 XtNheight, 324,
				 XtNvisual, new->tracePopup.canvasVisual,
				 XkwNsilenceUnconsumed, True,
				 XtNresizable, True,
				 XtNtop, XtChainTop,
				 XtNbottom, XtChainBottom,
				 NULL);
    cnv = w;
    XtAddCallback (cnv, XkwNrealiseCallback, _TracePopup_canvas_realise_cbk,
		   New);
    w = XtVaCreatePopupShell ("postscriptwinpopup", postscriptWidgetClass, New,
			      XtNtitle, "Postscript Window",
			      XkwNautoIncrement, TRUE,
			      NULL);
    new->tracePopup.pswinpopup = w;
    XtAddCallback (print_btn, XtNcallback, xtmisc_popup_cbk, w);
}   /*  End Function Initialise  */

static void _TracePopup_Destroy (Widget W)
{
    TracePopupWidget w = (TracePopupWidget) W;

    c_unregister_callback2 (w->tracePopup.iarr_destroy_callback);
}   /*  End Function Destroy  */

static void _TracePopup_iarr_destroy_callback (iarray arr,
					       TracePopupWidget top)
/*  [PURPOSE] This routine will register the destruction of the Intelligent
    Array.
    <arr> The Intelligent Array.
    <top> The TracePopup widget.
    [RETURNS] Nothing.
*/
{
    top->tracePopup.array = NULL;
    top->tracePopup.profile = NULL;
    top->tracePopup.iarr_destroy_callback = NULL;
    canvas_use_astro_transform (top->tracePopup.worldCanvas, NULL);
    canvas_specify (top->tracePopup.worldCanvas, NULL, NULL,
		    0, (CONST char **) NULL, NULL);
}   /*  End Function iarr_destroy_callback  */

static void _TracePopup_canvas_realise_cbk (Widget w, XtPointer client_data,
					    XtPointer call_data)
/*  [SUMMARY] Canvas realise callback.
    [PURPOSE] This routine is called when a Canvas widget is realised. A
    Kcolourmap object is created if required, then the KWorldCanvas is created.
    <w> The Canvas widget that has just been realised.
    <client_data> The TracePopup widget.
    <call_data> The pixel canvas(es).
    [RETURNS] Nothing.
*/
{
    KPixCanvas pixcanvas = (KPixCanvas) call_data;
    TracePopupWidget top = (TracePopupWidget) client_data;
    /*static char function_name[] = "TracePopupWidget::canvas_realise_cbk";*/

    XkwPostscriptRegisterImageAndName (top->tracePopup.pswinpopup, pixcanvas,
				       NULL);
    top->tracePopup.pixcanvas = pixcanvas;
    top->tracePopup.worldCanvas = canvas_create(pixcanvas,
						top->tracePopup.karmaColourmap,
						NULL);
    canvas_set_dressing (top->tracePopup.worldCanvas,
			 KCD_ATT_SHOW_LABELS, TRUE,
			 KCD_ATT_TOP_TICKS, TRUE,
			 KCD_ATT_BOTTOM_TICKS, TRUE,
			 KCD_ATT_LEFT_TICKS, TRUE,
			 KCD_ATT_RIGHT_TICKS, TRUE,
			 KCD_ATT_END);
    canvas_register_position_event_func (top->tracePopup.worldCanvas,
					 _TracePopup_position_func, top);
    XtCallCallbacks ( (Widget) top, XkwNrealiseCallback,
		      (XtPointer) top->tracePopup.worldCanvas );
    XkwOverlayMenuAddCanvas (top->tracePopup.overlay_menu,
			     top->tracePopup.worldCanvas);
    graph1d_init (top->tracePopup.worldCanvas);
    graph1d_set_canvas_attributes (top->tracePopup.worldCanvas,
				   GRAPH1D_CANVAS_ATT_DRAW_STYLE,
				       top->tracePopup.draw_mode,
				   GRAPH1D_CANVAS_ATT_END);
}   /*  End Function canvas_realise_cbk   */

static flag _TracePopup_position_func (KWorldCanvas canvas, double x, double y,
				       unsigned int event_code, void *e_info,
				       void **f_info,
				       double x_lin, double y_lin,
				       int x_pix, int y_pix)
/*  [SUMMARY] Position event callback.
    [PURPOSE] This routine is a position event consumer for a world canvas.
    <canvas> The canvas on which the event occurred.
    <x> The horizontal world co-ordinate of the event.
    <y> The vertical world co-ordinate of the event.
    <event_code> The arbitrary event code.
    <e_info> A pointer to arbitrary event information.
    <f_info> A pointer to an arbitrary function information pointer.
    <x_lin> The horizontal linear world co-ordinate prior to the transform
    function being called.
    <y_lin> The vertical linear world co-ordinate prior to the transform
    function being called.
    <x_pix> The horizontal pixel co-ordinate.
    <y_pix> The vertical pixel co-ordinate.
    [NOTE] The world co-ordinate values will have been transformed by the
    registered transform function (see [<canvas_register_transform_func>]).
    [RETURNS] TRUE if the event was consumed, else FALSE indicating that
    the event is still to be processed.
*/
{
    TracePopupWidget w = (TracePopupWidget) *f_info;
    char x_string[STRING_LENGTH], y_string[STRING_LENGTH];

    if (event_code == K_CANVAS_EVENT_POINTER_LEAVE)
    {
	x_string[0] = '\0';
	y_string[0] = '\0';
    }
    else if ( !graph1d_orthogonal_track_compute (canvas, x, y, x_lin,
						 x_string, y_string) )
	canvas_track_compute (canvas, x, y, x_lin, y_lin, x_string, y_string);
    XtVaSetValues (w->tracePopup.track_label0, XtNlabel, x_string, NULL);
    XtVaSetValues (w->tracePopup.track_label1, XtNlabel, y_string, NULL);
    return (FALSE);
}   /*  End Function _TracePopup_position_func  */

static void _TracePopup_unzoom_cbk (Widget w, XtPointer client_data,
				    XtPointer call_data)
{
    TracePopupWidget top = (TracePopupWidget) client_data;

    kwin_process_position_event (top->tracePopup.pixcanvas, 0, 0,
				 TRUE, K_CANVAS_EVENT_PLAIN_KEY_PRESS, "u");
}   /*  End Function unzoom_cbk   */

static void _TracePopup_smooth_mode_cbk (Widget w, XtPointer client_data,
					 XtPointer call_data)
{
    int mode = *(int *) call_data;
    TracePopupWidget top = (TracePopupWidget) client_data;

    graph1d_set_canvas_attributes (top->tracePopup.worldCanvas,
				   GRAPH1D_CANVAS_ATT_SMOOTH_MODE, mode,
				   GRAPH1D_CANVAS_ATT_END);
    kwin_refresh_if_visible (top->tracePopup.pixcanvas, TRUE);
}   /*  End Function smooth_mode_cbk   */

static void _TracePopup_auto_zoom_cbk (Widget w, XtPointer client_data,
				       XtPointer call_data)
{
    flag bool = (uaddr) call_data;
    TracePopupWidget top = (TracePopupWidget) client_data;

    graph1d_set_canvas_attributes (top->tracePopup.worldCanvas,
				   GRAPH1D_CANVAS_ATT_AUTO_Y, bool,
				   GRAPH1D_CANVAS_ATT_END);
    if (bool) kwin_refresh_if_visible (top->tracePopup.pixcanvas, TRUE);
}   /*  End Function auto_zoom_cbk   */

static void _TracePopup_save_cbk (Widget w, XtPointer client_data,
				  XtPointer call_data)
{
    TracePopupWidget top = (TracePopupWidget) client_data;
    String fname = (String) call_data;

    graph1d_orthogonal_save (top->tracePopup.worldCanvas, fname);
}   /*  End Function save_cbk  */

static void _TracePopup_draw_mode_cbk (Widget w, XtPointer client_data,
				       XtPointer call_data)
{
    int mode = *(int *) call_data;
    TracePopupWidget top = (TracePopupWidget) client_data;

    graph1d_set_canvas_attributes (top->tracePopup.worldCanvas,
				   GRAPH1D_CANVAS_ATT_DRAW_STYLE, mode,
				   GRAPH1D_CANVAS_ATT_END);
    kwin_refresh_if_visible (top->tracePopup.pixcanvas, TRUE);
}   /*  End Function draw_mode_cbk  */


/*  Public functions follow  */

/*EXPERIMENTAL_FUNCTION*/
void XkwTracePopupNewArray (Widget W, iarray array, double min, double max)
/*  [SUMMARY] Register new array.
    <W> The TracePopup widget.
    <array> The new array. This may be NULL.
    <min> The minimum (unscaled) data value in the array.
    <max> The maximum (unscaled) data value in the array. If this is less than
    <<min>> then the minimum and maximum values are computed.
    [RETURNS] Nothing.
*/
{
    KwcsAstro ap;
    double scale, offset, tmp;
    TracePopupWidget w = (TracePopupWidget) W;
    /*static char function_name[] = "XkwTracePopupNewArray";*/

    /*  Clean up any old profile  */
    if (w->tracePopup.array == array) return;
    w->tracePopup.do_unzoom = TRUE;
    w->tracePopup.array = array;
    if (w->tracePopup.profile)
    {
	graph1d_orthogonal_destroy (w->tracePopup.profile);
	c_unregister_callback (w->tracePopup.iarr_destroy_callback);
	w->tracePopup.iarr_destroy_callback = NULL;
	w->tracePopup.profile = NULL;
    }
    XtRealizeWidget (W);
    canvas_use_astro_transform (w->tracePopup.worldCanvas, NULL);
    canvas_specify (w->tracePopup.worldCanvas, NULL, NULL,
		    0, (CONST char **) NULL, NULL);
    kwin_clear (w->tracePopup.pixcanvas, 0, 0, -1, -1);
    /*  Process new array  */
    if (array == NULL) return;
    if (max < min) iarray_min_max (array, CONV_CtoR_REAL, &min, &max);
    w->tracePopup.profile =
	graph1d_orthogonal_create (w->tracePopup.worldCanvas, array, NULL);
    if (w->tracePopup.profile == NULL) return;
    w->tracePopup.iarr_destroy_callback = 
	iarray_register_destroy_func
	(array,
	 ( flag (*) () ) _TracePopup_iarr_destroy_callback,
	 w);
    ap = iarray_get_associated_wcs_astro (array);
    canvas_use_astro_transform (w->tracePopup.worldCanvas, ap);
    iarray_get_data_scaling (array, &scale, &offset);
    min = min * scale + offset;
    max = max * scale + offset;
    if (min > max) swap (min, max, tmp);
    canvas_set_attributes (w->tracePopup.worldCanvas,
			   CANVAS_ATT_BOTTOM_Y, min,
			   CANVAS_ATT_TOP_Y, max,
			   CANVAS_ATT_END);
    graph1d_set_canvas_attributes (w->tracePopup.worldCanvas,
				   GRAPH1D_CANVAS_ATT_BOTTOM_Y, min,
				   GRAPH1D_CANVAS_ATT_TOP_Y, max,
				   GRAPH1D_CANVAS_ATT_END);
}   /*  End Function XkwTracePopupNewArray  */

/*EXPERIMENTAL_FUNCTION*/
void XkwTracePopupShowTrace (Widget W, CONST unsigned int *dim_indices,
			     CONST uaddr *coord_indices)
/*  [SUMMARY] Show a trace.
    <W> The widget.
    <dim_indices> The dimension indicies. The length of this array must be one
    less than the number of dimensions in the array.
    <coord_indices> The coordinate indices of the trace. The length of this
    array must be one less than the number of dimensions in the array.
    [RETURNS] Nothing.
*/
{
    XkwTracePopupShowTraces (W, dim_indices, TRUE, 1.0, 1, coord_indices,NULL);
}   /*  End Function XkwTracePopupShowTrace  */

/*EXPERIMENTAL_FUNCTION*/
void XkwTracePopupShowTraces (Widget W, CONST unsigned int *dim_indices,
			      flag sum, float global_weight,
			      unsigned int num_traces,
			      CONST uaddr *coord_indices, CONST float *weights)
/*  [SUMMARY] Show a weighted, summed trace.
    <W> The widget.
    <dim_indices> The dimension indicies. The length of this array must be one
    less than the number of dimensions in the array.
    <sum> If TRUE, the data are summed rather than averaged.
    <global_weight> This weight value is applied to each trace as it is
    combined.
    <num_traces> The number of traces to combine.
    <coord_indices> The co-ordinate indices for the traces. This is grouped
    into <<num_traces>> contiguous blocks of co-ordinate traces. Each block
    must contain <<num_dim>> - 1 co-ordinate indices.
    <weights> The weights to apply to each trace as it is combined. If this is
    NULL no weights are applied.
    [RETURNS] Nothing.
*/
{
    TracePopupWidget w = (TracePopupWidget) W;
    /*static char function_name[] = "XkwTracePopupShowTraces";*/

    graph1d_orthogonal_show_traces (w->tracePopup.profile, dim_indices, FALSE,
				    sum, global_weight, num_traces,
				    coord_indices, 0, weights);
    graph1d_orthogonal_configure_canvas (w->tracePopup.profile);
    if (!graph1d_test_if_zoomed_horizontally (w->tracePopup.worldCanvas) ||
	w->tracePopup.do_unzoom)
    {
	graph1d_unzoom (w->tracePopup.worldCanvas, TRUE, FALSE, FALSE);
	w->tracePopup.do_unzoom = FALSE;
    }
    kwin_refresh_if_visible (w->tracePopup.pixcanvas, TRUE);
}   /*  End Function XkwTracePopupShowTraces  */

/*EXPERIMENTAL_FUNCTION*/
void XkwTracePopupShowBoxedTrace (Widget W, CONST unsigned int *dim_indices,
				  flag sum,
				  CONST uaddr *start_indices,
				  CONST uaddr *stop_indices)
/*  [SUMMARY] Show a boxed trace.
    <W> The widget.
    <dim_indices> The dimension indicies. The length of this array must be one
    less than the number of dimensions in the array.
    <sum> If TRUE the traces are summed, else they are averaged.
    <start_indices> The starting coordinate indices of the box. The length of
    this array must be one less than the number of dimensions in the array.
    <stop_indices> The stopping coordinate indices of the box. The length of
    this array must be one less than the number of dimensions in the array.
    [RETURNS] Nothing.
*/
{
    TracePopupWidget w = (TracePopupWidget) W;
    /*static char function_name[] = "XkwTracePopupShowBoxedTrace";*/

    graph1d_orthogonal_show_boxed_trace (w->tracePopup.profile, dim_indices,
					 sum, start_indices, stop_indices, 0);
    graph1d_orthogonal_configure_canvas (w->tracePopup.profile);
    if (!graph1d_test_if_zoomed_horizontally (w->tracePopup.worldCanvas) ||
	w->tracePopup.do_unzoom)
    {
	graph1d_unzoom (w->tracePopup.worldCanvas, TRUE, FALSE, FALSE);
	w->tracePopup.do_unzoom = FALSE;
    }
    kwin_refresh_if_visible (w->tracePopup.pixcanvas, TRUE);
}   /*  End Function XkwTracePopupShowBoxedTrace  */

/*EXPERIMENTAL_FUNCTION*/
void XkwTracePopupInvalidate (Widget W)
/*  [SUMMARY] Invalidate trace data.
    <W> The widget.
    [RETURNS] Nothing.
*/
{
    char *xlabel;
    TracePopupWidget w = (TracePopupWidget) W;

    if (!w->tracePopup.profile) return;
    graph1d_orthogonal_invalidate (w->tracePopup.profile);
    canvas_get_specification (w->tracePopup.worldCanvas, &xlabel, NULL, NULL,
			      NULL, NULL);
    canvas_specify (w->tracePopup.worldCanvas, xlabel, "NO VALID DATA",
		    0, NULL, NULL);
    kwin_refresh_if_visible (w->tracePopup.pixcanvas, TRUE);
}   /*  End Function XkwTracePopupInvalidate  */

void XkwTracePopupRefresh (Widget W, flag clear)
/*  [PURPOSE] This routine will refresh the display canvas for the TracePopup
    widget.
    <W> The widget.
    <clear> If TRUE, the canvas is cleared prior to refresh.
    [RETURNS] Nothing.
*/
{
    TracePopupWidget w;
    static char function_name[] = "XkwTracePopupRefresh";

    if ( !XtIsTracePopup (W) )
    {
	fprintf (stderr, "Invalid widget passed\n");
	a_prog_bug (function_name);
    }
    w = (TracePopupWidget) W;
    kwin_refresh_if_visible (w->tracePopup.pixcanvas, clear);
}   /*  End Function XkwTracePopupRefresh  */
